/*
===========================================================================

Doom 3 BFG Edition GPL Source Code
Copyright (C) 2022 Robert Beckebans

This file is part of the Doom 3 BFG Edition GPL Source Code ("Doom 3 BFG Edition Source Code").

Doom 3 BFG Edition Source Code is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Doom 3 BFG Edition Source Code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Doom 3 BFG Edition Source Code.  If not, see <http://www.gnu.org/licenses/>.

In addition, the Doom 3 BFG Edition Source Code is also subject to certain additional terms. You should have received a copy of these additional terms immediately following the terms and conditions of the GNU General Public License which accompanied the Doom 3 BFG Edition Source Code.  If not, please request a copy in writing from id Software at the address below.

If you have questions concerning this license or the applicable additional terms, you may contact in writing id Software LLC, c/o ZeniMax Media Inc., Suite 120, Rockville, Maryland 20850 USA.

===========================================================================
*/

typedef struct
{
	const char*	quakeName;
	const char*	doomName;
} quakeToDoom_t;

static quakeToDoom_t textureConvertNames[] =
{
	// id1.wad
	{"textures/#04awater1", "textures/id1/04awater1"},
	{"textures/#04mwat1", "textures/id1/04mwat1"},
	{"textures/#04mwat2", "textures/id1/04mwat2"},
	{"textures/#04water1", "textures/id1/04water1"},
	{"textures/#04water2", "textures/id1/04water2"},
	{"textures/#lava1", "textures/id1/lava1"},
	{"textures/#slime", "textures/id1/slime"},
	{"textures/#slime0", "textures/id1/slime0"},
	{"textures/#slime1", "textures/id1/slime1"},
	{"textures/#teleport", "textures/id1/teleport"},
	{"textures/#water0", "textures/id1/water0"},
	{"textures/#water1", "textures/id1/water1"},
	{"textures/#water2", "textures/id1/water2"},
	{"textures/0_box_side", "textures/id1/0_box_side"},
	{"textures/0_box_top", "textures/id1/0_box_top"},
	{"textures/0_med100", "textures/id1/0_med100"},
	{"textures/0_med25", "textures/id1/0_med25"},
	{"textures/0_med25s", "textures/id1/0_med25s"},
	{"textures/0basebtn", "textures/id1/0basebtn"},
	{"textures/0butn", "textures/id1/0butn"},
	{"textures/0butnn", "textures/id1/0butnn"},
	{"textures/0button", "textures/id1/0button"},
	{"textures/0floorsw", "textures/id1/0floorsw"},
	{"textures/0light01", "textures/id1/0light01"},
	{"textures/0mtlsw", "textures/id1/0mtlsw"},
	{"textures/0planet", "textures/id1/0planet"},
	{"textures/0shoot", "textures/id1/0shoot"},
	{"textures/0slip", "textures/id1/0slip"},
	{"textures/0slipbot", "textures/id1/0slipbot"},
	{"textures/0sliptop", "textures/id1/0sliptop"},
	{"textures/1_box_side", "textures/id1/1_box_side"},
	{"textures/1_box_top", "textures/id1/1_box_top"},
	{"textures/1_med100", "textures/id1/1_med100"},
	{"textures/1_med25", "textures/id1/1_med25"},
	{"textures/1_med25s", "textures/id1/1_med25s"},
	{"textures/1basebtn", "textures/id1/1basebtn"},
	{"textures/1butn", "textures/id1/1butn"},
	{"textures/1butnn", "textures/id1/1butnn"},
	{"textures/1button", "textures/id1/1button"},
	{"textures/1floorsw", "textures/id1/1floorsw"},
	{"textures/1light01", "textures/id1/1light01"},
	{"textures/1mtlsw", "textures/id1/1mtlsw"},
	{"textures/1planet", "textures/id1/1planet"},
	{"textures/1shoot", "textures/id1/1shoot"},
	{"textures/1slip", "textures/id1/1slip"},
	{"textures/2_med100", "textures/id1/2_med100"},
	{"textures/2_med25", "textures/id1/2_med25"},
	{"textures/2butn", "textures/id1/2butn"},
	{"textures/2butnn", "textures/id1/2butnn"},
	{"textures/2button", "textures/id1/2button"},
	{"textures/2floorsw", "textures/id1/2floorsw"},
	{"textures/2light01", "textures/id1/2light01"},
	{"textures/2mtlsw", "textures/id1/2mtlsw"},
	{"textures/2planet", "textures/id1/2planet"},
	{"textures/2shoot", "textures/id1/2shoot"},
	{"textures/2slip", "textures/id1/2slip"},
	{"textures/3_med100", "textures/id1/3_med100"},
	{"textures/3_med25", "textures/id1/3_med25"},
	{"textures/3butn", "textures/id1/3butn"},
	{"textures/3butnn", "textures/id1/3butnn"},
	{"textures/3button", "textures/id1/3button"},
	{"textures/3floorsw", "textures/id1/3floorsw"},
	{"textures/3mtlsw", "textures/id1/3mtlsw"},
	{"textures/3planet", "textures/id1/3planet"},
	{"textures/3shoot", "textures/id1/3shoot"},
	{"textures/3slip", "textures/id1/3slip"},
	{"textures/4slip", "textures/id1/4slip"},
	{"textures/5slip", "textures/id1/5slip"},
	{"textures/6slip", "textures/id1/6slip"},
	{"textures/abasebtn", "textures/id1/abasebtn"},
	{"textures/abutn", "textures/id1/abutn"},
	{"textures/abutnn", "textures/id1/abutnn"},
	{"textures/abutton", "textures/id1/abutton"},
	{"textures/afloorsw", "textures/id1/afloorsw"},
	{"textures/amtlsw", "textures/id1/amtlsw"},
	{"textures/ashoot", "textures/id1/ashoot"},
	{"textures/adoor01_2", "textures/id1/adoor01_2"},
	{"textures/adoor02_2", "textures/id1/adoor02_2"},
	{"textures/adoor03_2", "textures/id1/adoor03_2"},
	{"textures/adoor03_3", "textures/id1/adoor03_3"},
	{"textures/adoor03_4", "textures/id1/adoor03_4"},
	{"textures/adoor03_5", "textures/id1/adoor03_5"},
	{"textures/adoor03_6", "textures/id1/adoor03_6"},
	{"textures/adoor09_1", "textures/id1/adoor09_1"},
	{"textures/adoor09_2", "textures/id1/adoor09_2"},
	{"textures/afloor1_3", "textures/id1/afloor1_3"},
	{"textures/afloor1_4", "textures/id1/afloor1_4"},
	{"textures/afloor1_8", "textures/id1/afloor1_8"},
	{"textures/afloor3_1", "textures/id1/afloor3_1"},
	{"textures/altar1_1", "textures/id1/altar1_1"},
	{"textures/altar1_3", "textures/id1/altar1_3"},
	{"textures/altar1_4", "textures/id1/altar1_4"},
	{"textures/altar1_6", "textures/id1/altar1_6"},
	{"textures/altar1_7", "textures/id1/altar1_7"},
	{"textures/altar1_8", "textures/id1/altar1_8"},
	{"textures/altarb_1", "textures/id1/altarb_1"},
	{"textures/altarb_2", "textures/id1/altarb_2"},
	{"textures/altarc_1", "textures/id1/altarc_1"},
	{"textures/arch7", "textures/id1/arch7"},
	{"textures/arrow_m", "textures/id1/arrow_m"},
	{"textures/az1_6", "textures/id1/az1_6"},
	{"textures/azfloor1_1", "textures/id1/azfloor1_1"},
	{"textures/azswitch3", "textures/id1/azswitch3"},
	{"textures/azwall1_5", "textures/id1/azwall1_5"},
	{"textures/azwall3_1", "textures/id1/azwall3_1"},
	{"textures/azwall3_2", "textures/id1/azwall3_2"},
	{"textures/basebutn3", "textures/id1/basebutn3"},
	{"textures/batt0sid", "textures/id1/batt0sid"},
	{"textures/batt0top", "textures/id1/batt0top"},
	{"textures/batt1sid", "textures/id1/batt1sid"},
	{"textures/batt1top", "textures/id1/batt1top"},
	{"textures/black", "textures/id1/black"},
	{"textures/bodiesa2_1", "textures/id1/bodiesa2_1"},
	{"textures/bodiesa2_4", "textures/id1/bodiesa2_4"},
	{"textures/bodiesa3_1", "textures/id1/bodiesa3_1"},
	{"textures/bodiesa3_2", "textures/id1/bodiesa3_2"},
	{"textures/bodiesa3_3", "textures/id1/bodiesa3_3"},
	{"textures/bricka2_1", "textures/id1/bricka2_1"},
	{"textures/bricka2_2", "textures/id1/bricka2_2"},
	{"textures/bricka2_4", "textures/id1/bricka2_4"},
	{"textures/bricka2_6", "textures/id1/bricka2_6"},
	{"textures/carch02", "textures/id1/carch02"},
	{"textures/carch03", "textures/id1/carch03"},
	{"textures/carch04_1", "textures/id1/carch04_1"},
	{"textures/carch04_2", "textures/id1/carch04_2"},
	{"textures/ceil1_1", "textures/id1/ceil1_1"},
	{"textures/ceiling1_3", "textures/id1/ceiling1_3"},
	{"textures/ceiling4", "textures/id1/ceiling4"},
	{"textures/ceiling5", "textures/id1/ceiling5"},
	{"textures/church1_2", "textures/id1/church1_2"},
	{"textures/church7", "textures/id1/church7"},
	{"textures/city1_4", "textures/id1/city1_4"},
	{"textures/city1_7", "textures/id1/city1_7"},
	{"textures/city2_1", "textures/id1/city2_1"},
	{"textures/city2_2", "textures/id1/city2_2"},
	{"textures/city2_3", "textures/id1/city2_3"},
	{"textures/city2_5", "textures/id1/city2_5"},
	{"textures/city2_6", "textures/id1/city2_6"},
	{"textures/city2_7", "textures/id1/city2_7"},
	{"textures/city2_8", "textures/id1/city2_8"},
	{"textures/city3_2", "textures/id1/city3_2"},
	{"textures/city3_4", "textures/id1/city3_4"},
	{"textures/city4_1", "textures/id1/city4_1"},
	{"textures/city4_2", "textures/id1/city4_2"},
	{"textures/city4_5", "textures/id1/city4_5"},
	{"textures/city4_6", "textures/id1/city4_6"},
	{"textures/city4_7", "textures/id1/city4_7"},
	{"textures/city4_8", "textures/id1/city4_8"},
	{"textures/city5_1", "textures/id1/city5_1"},
	{"textures/city5_2", "textures/id1/city5_2"},
	{"textures/city5_3", "textures/id1/city5_3"},
	{"textures/city5_4", "textures/id1/city5_4"},
	{"textures/city5_6", "textures/id1/city5_6"},
	{"textures/city5_7", "textures/id1/city5_7"},
	{"textures/city5_8", "textures/id1/city5_8"},
	{"textures/city6_3", "textures/id1/city6_3"},
	{"textures/city6_4", "textures/id1/city6_4"},
	{"textures/city6_7", "textures/id1/city6_7"},
	{"textures/city6_8", "textures/id1/city6_8"},
	{"textures/city8_2", "textures/id1/city8_2"},
	{"textures/citya1_1", "textures/id1/citya1_1"},
	{"textures/clip", "textures/common/clip"},
	{"textures/column01_3", "textures/id1/column01_3"},
	{"textures/column01_4", "textures/id1/column01_4"},
	{"textures/column1_2", "textures/id1/column1_2"},
	{"textures/column1_4", "textures/id1/column1_4"},
	{"textures/column1_5", "textures/id1/column1_5"},
	{"textures/comp1_1", "textures/id1/comp1_1"},
	{"textures/comp1_2", "textures/id1/comp1_2"},
	{"textures/comp1_3", "textures/id1/comp1_3"},
	{"textures/comp1_4", "textures/id1/comp1_4"},
	{"textures/comp1_5", "textures/id1/comp1_5"},
	{"textures/comp1_6", "textures/id1/comp1_6"},
	{"textures/comp1_7", "textures/id1/comp1_7"},
	{"textures/comp1_8", "textures/id1/comp1_8"},
	{"textures/cop1_1", "textures/id1/cop1_1"},
	{"textures/cop1_2", "textures/id1/cop1_2"},
	{"textures/cop1_3", "textures/id1/cop1_3"},
	{"textures/cop1_4", "textures/id1/cop1_4"},
	{"textures/cop1_5", "textures/id1/cop1_5"},
	{"textures/cop1_6", "textures/id1/cop1_6"},
	{"textures/cop1_7", "textures/id1/cop1_7"},
	{"textures/cop1_8", "textures/id1/cop1_8"},
	{"textures/cop2_1", "textures/id1/cop2_1"},
	{"textures/cop2_2", "textures/id1/cop2_2"},
	{"textures/cop2_3", "textures/id1/cop2_3"},
	{"textures/cop2_4", "textures/id1/cop2_4"},
	{"textures/cop2_5", "textures/id1/cop2_5"},
	{"textures/cop2_6", "textures/id1/cop2_6"},
	{"textures/cop3_1", "textures/id1/cop3_1"},
	{"textures/cop3_2", "textures/id1/cop3_2"},
	{"textures/cop3_4", "textures/id1/cop3_4"},
	{"textures/cop4_3", "textures/id1/cop4_3"},
	{"textures/cop4_5", "textures/id1/cop4_5"},
	{"textures/crate0_side", "textures/id1/crate0_side"},
	{"textures/crate0_top", "textures/id1/crate0_top"},
	{"textures/crate1_side", "textures/id1/crate1_side"},
	{"textures/crate1_top", "textures/id1/crate1_top"},
	{"textures/dem4_1", "textures/id1/dem4_1"},
	{"textures/dem4_4", "textures/id1/dem4_4"},
	{"textures/dem5_3", "textures/id1/dem5_3"},
	{"textures/demc4_4", "textures/id1/demc4_4"},
	{"textures/door01_2", "textures/id1/door01_2"},
	{"textures/door02_1", "textures/id1/door02_1"},
	{"textures/door02_2", "textures/id1/door02_2"},
	{"textures/door02_3", "textures/id1/door02_3"},
	{"textures/door02_7", "textures/id1/door02_7"},
	{"textures/door03_2", "textures/id1/door03_2"},
	{"textures/door03_3", "textures/id1/door03_3"},
	{"textures/door03_4", "textures/id1/door03_4"},
	{"textures/door03_5", "textures/id1/door03_5"},
	{"textures/door04_1", "textures/id1/door04_1"},
	{"textures/door04_2", "textures/id1/door04_2"},
	{"textures/door05_2", "textures/id1/door05_2"},
	{"textures/door05_3", "textures/id1/door05_3"},
	{"textures/dopeback", "textures/id1/dopeback"},
	{"textures/dopefish", "textures/id1/dopefish"},
	{"textures/dr01_1", "textures/id1/dr01_1"},
	{"textures/dr01_2", "textures/id1/dr01_2"},
	{"textures/dr02_1", "textures/id1/dr02_1"},
	{"textures/dr02_2", "textures/id1/dr02_2"},
	{"textures/dr03_1", "textures/id1/dr03_1"},
	{"textures/dr05_2", "textures/id1/dr05_2"},
	{"textures/dr07_1", "textures/id1/dr07_1"},
	{"textures/dung01_1", "textures/id1/dung01_1"},
	{"textures/dung01_2", "textures/id1/dung01_2"},
	{"textures/dung01_3", "textures/id1/dung01_3"},
	{"textures/dung01_4", "textures/id1/dung01_4"},
	{"textures/dung01_5", "textures/id1/dung01_5"},
	{"textures/dung02_1", "textures/id1/dung02_1"},
	{"textures/dung02_5", "textures/id1/dung02_5"},
	{"textures/ecop1_1", "textures/id1/ecop1_1"},
	{"textures/ecop1_4", "textures/id1/ecop1_4"},
	{"textures/ecop1_6", "textures/id1/ecop1_6"},
	{"textures/ecop1_7", "textures/id1/ecop1_7"},
	{"textures/ecop1_8", "textures/id1/ecop1_8"},
	{"textures/edoor01_1", "textures/id1/edoor01_1"},
	{"textures/elwall1_1", "textures/id1/elwall1_1"},
	{"textures/elwall2_4", "textures/id1/elwall2_4"},
	{"textures/emetal1_3", "textures/id1/emetal1_3"},
	{"textures/enter01", "textures/id1/enter01"},
	{"textures/exit01", "textures/id1/exit01"},
	{"textures/exit02_2", "textures/id1/exit02_2"},
	{"textures/exit02_3", "textures/id1/exit02_3"},
	{"textures/floor01_5", "textures/id1/floor01_5"},
	{"textures/grave01_1", "textures/id1/grave01_1"},
	{"textures/grave01_3", "textures/id1/grave01_3"},
	{"textures/grave02_1", "textures/id1/grave02_1"},
	{"textures/grave02_2", "textures/id1/grave02_2"},
	{"textures/grave02_3", "textures/id1/grave02_3"},
	{"textures/grave02_4", "textures/id1/grave02_4"},
	{"textures/grave02_5", "textures/id1/grave02_5"},
	{"textures/grave02_6", "textures/id1/grave02_6"},
	{"textures/grave02_7", "textures/id1/grave02_7"},
	{"textures/grave03_1", "textures/id1/grave03_1"},
	{"textures/grave03_2", "textures/id1/grave03_2"},
	{"textures/grave03_3", "textures/id1/grave03_3"},
	{"textures/grave03_4", "textures/id1/grave03_4"},
	{"textures/grave03_5", "textures/id1/grave03_5"},
	{"textures/grave03_6", "textures/id1/grave03_6"},
	{"textures/grave03_7", "textures/id1/grave03_7"},
	{"textures/ground1_1", "textures/id1/ground1_1"},
	{"textures/ground1_2", "textures/id1/ground1_2"},
	{"textures/ground1_5", "textures/id1/ground1_5"},
	{"textures/ground1_6", "textures/id1/ground1_6"},
	{"textures/ground1_7", "textures/id1/ground1_7"},
	{"textures/ground1_8", "textures/id1/ground1_8"},
	{"textures/key01_1", "textures/id1/key01_1"},
	{"textures/key01_2", "textures/id1/key01_2"},
	{"textures/key01_3", "textures/id1/key01_3"},
	{"textures/key02_1", "textures/id1/key02_1"},
	{"textures/key02_2", "textures/id1/key02_2"},
	{"textures/key03_1", "textures/id1/key03_1"},
	{"textures/key03_2", "textures/id1/key03_2"},
	{"textures/key03_3", "textures/id1/key03_3"},
	{"textures/lgmetal", "textures/id1/lgmetal"},
	{"textures/lgmetal2", "textures/id1/lgmetal2"},
	{"textures/lgmetal3", "textures/id1/lgmetal3"},
	{"textures/lgmetal4", "textures/id1/lgmetal4"},
	{"textures/light1_1", "textures/id1/light1_1"},
	{"textures/light1_2", "textures/id1/light1_2"},
	{"textures/light1_3", "textures/id1/light1_3"},
	{"textures/light1_4", "textures/id1/light1_4"},
	{"textures/light1_5", "textures/id1/light1_5"},
	{"textures/light1_7", "textures/id1/light1_7"},
	{"textures/light1_8", "textures/id1/light1_8"},
	{"textures/light3_3", "textures/id1/light3_3"},
	{"textures/light3_5", "textures/id1/light3_5"},
	{"textures/light3_6", "textures/id1/light3_6"},
	{"textures/light3_7", "textures/id1/light3_7"},
	{"textures/light3_8", "textures/id1/light3_8"},
	{"textures/m5_3", "textures/id1/m5_3"},
	{"textures/m5_5", "textures/id1/m5_5"},
	{"textures/m5_8", "textures/id1/m5_8"},
	{"textures/med100", "textures/id1/med100"},
	{"textures/med3_0", "textures/id1/med3_0"},
	{"textures/med3_1", "textures/id1/med3_1"},
	{"textures/met5_1", "textures/id1/met5_1"},
	{"textures/met5_2", "textures/id1/met5_2"},
	{"textures/met5_3", "textures/id1/met5_3"},
	{"textures/metal1_1", "textures/id1/metal1_1"},
	{"textures/metal1_2", "textures/id1/metal1_2"},
	{"textures/metal1_3", "textures/id1/metal1_3"},
	{"textures/metal1_4", "textures/id1/metal1_4"},
	{"textures/metal1_5", "textures/id1/metal1_5"},
	{"textures/metal1_6", "textures/id1/metal1_6"},
	{"textures/metal1_7", "textures/id1/metal1_7"},
	{"textures/metal2_1", "textures/id1/metal2_1"},
	{"textures/metal2_2", "textures/id1/metal2_2"},
	{"textures/metal2_3", "textures/id1/metal2_3"},
	{"textures/metal2_4", "textures/id1/metal2_4"},
	{"textures/metal2_5", "textures/id1/metal2_5"},
	{"textures/metal2_6", "textures/id1/metal2_6"},
	{"textures/metal2_7", "textures/id1/metal2_7"},
	{"textures/metal2_8", "textures/id1/metal2_8"},
	{"textures/metal3_2", "textures/id1/metal3_2"},
	{"textures/metal4_2", "textures/id1/metal4_2"},
	{"textures/metal4_3", "textures/id1/metal4_3"},
	{"textures/metal4_4", "textures/id1/metal4_4"},
	{"textures/metal4_5", "textures/id1/metal4_5"},
	{"textures/metal4_6", "textures/id1/metal4_6"},
	{"textures/metal4_7", "textures/id1/metal4_7"},
	{"textures/metal4_8", "textures/id1/metal4_8"},
	{"textures/metal5_1", "textures/id1/metal5_1"},
	{"textures/metal5_2", "textures/id1/metal5_2"},
	{"textures/metal5_3", "textures/id1/metal5_3"},
	{"textures/metal5_4", "textures/id1/metal5_4"},
	{"textures/metal5_5", "textures/id1/metal5_5"},
	{"textures/metal5_6", "textures/id1/metal5_6"},
	{"textures/metal5_8", "textures/id1/metal5_8"},
	{"textures/metal6_1", "textures/id1/metal6_1"},
	{"textures/metal6_2", "textures/id1/metal6_2"},
	{"textures/metal6_3", "textures/id1/metal6_3"},
	{"textures/metal6_4", "textures/id1/metal6_4"},
	{"textures/metalt1_1", "textures/id1/metalt1_1"},
	{"textures/metalt1_2", "textures/id1/metalt1_2"},
	{"textures/metalt1_7", "textures/id1/metalt1_7"},
	{"textures/metalt2_1", "textures/id1/metalt2_1"},
	{"textures/metalt2_2", "textures/id1/metalt2_2"},
	{"textures/metalt2_3", "textures/id1/metalt2_3"},
	{"textures/metalt2_4", "textures/id1/metalt2_4"},
	{"textures/metalt2_5", "textures/id1/metalt2_5"},
	{"textures/metalt2_6", "textures/id1/metalt2_6"},
	{"textures/metalt2_7", "textures/id1/metalt2_7"},
	{"textures/metalt2_8", "textures/id1/metalt2_8"},
	{"textures/metflor2_1", "textures/id1/metflor2_1"},
	{"textures/mmetal1_1", "textures/id1/mmetal1_1"},
	{"textures/mmetal1_2", "textures/id1/mmetal1_2"},
	{"textures/mmetal1_3", "textures/id1/mmetal1_3"},
	{"textures/mmetal1_5", "textures/id1/mmetal1_5"},
	{"textures/mmetal1_6", "textures/id1/mmetal1_6"},
	{"textures/mmetal1_7", "textures/id1/mmetal1_7"},
	{"textures/mmetal1_8", "textures/id1/mmetal1_8"},
	{"textures/mswtch_2", "textures/id1/mswtch_2"},
	{"textures/mswtch_3", "textures/id1/mswtch_3"},
	{"textures/mswtch_4", "textures/id1/mswtch_4"},
	{"textures/muh_bad", "textures/id1/muh_bad"},
	{"textures/nail0sid", "textures/id1/nail0sid"},
	{"textures/nail0top", "textures/id1/nail0top"},
	{"textures/nail1sid", "textures/id1/nail1sid"},
	{"textures/nail1top", "textures/id1/nail1top"},
	{"textures/nmetal2_1", "textures/id1/nmetal2_1"},
	{"textures/nmetal2_6", "textures/id1/nmetal2_6"},
	{"textures/plat_side1", "textures/id1/plat_side1"},
	{"textures/plat_stem", "textures/id1/plat_stem"},
	{"textures/plat_top1", "textures/id1/plat_top1"},
	{"textures/plat_top2", "textures/id1/plat_top2"},
	{"textures/quake", "textures/id1/quake"},
	{"textures/raven", "textures/id1/raven"},
	{"textures/rock0sid", "textures/id1/rock0sid"},
	{"textures/rock1_2", "textures/id1/rock1_2"},
	{"textures/rock1sid", "textures/id1/rock1sid"},
	{"textures/rock3_2", "textures/id1/rock3_2"},
	{"textures/rock3_7", "textures/id1/rock3_7"},
	{"textures/rock3_8", "textures/id1/rock3_8"},
	{"textures/rock4_1", "textures/id1/rock4_1"},
	{"textures/rock4_2", "textures/id1/rock4_2"},
	{"textures/rock5_2", "textures/id1/rock5_2"},
	{"textures/rockettop", "textures/id1/rockettop"},
	{"textures/rune1_1", "textures/id1/rune1_1"},
	{"textures/rune1_4", "textures/id1/rune1_4"},
	{"textures/rune1_5", "textures/id1/rune1_5"},
	{"textures/rune1_6", "textures/id1/rune1_6"},
	{"textures/rune1_7", "textures/id1/rune1_7"},
	{"textures/rune2_1", "textures/id1/rune2_1"},
	{"textures/rune2_2", "textures/id1/rune2_2"},
	{"textures/rune2_3", "textures/id1/rune2_3"},
	{"textures/rune2_4", "textures/id1/rune2_4"},
	{"textures/rune2_5", "textures/id1/rune2_5"},
	{"textures/rune_a", "textures/id1/rune_a"},
	{"textures/sfloor1_2", "textures/id1/sfloor1_2"},
	{"textures/sfloor3_2", "textures/id1/sfloor3_2"},
	{"textures/sfloor4_1", "textures/id1/sfloor4_1"},
	{"textures/sfloor4_2", "textures/id1/sfloor4_2"},
	{"textures/sfloor4_4", "textures/id1/sfloor4_4"},
	{"textures/sfloor4_5", "textures/id1/sfloor4_5"},
	{"textures/sfloor4_6", "textures/id1/sfloor4_6"},
	{"textures/sfloor4_7", "textures/id1/sfloor4_7"},
	{"textures/sfloor4_8", "textures/id1/sfloor4_8"},
	{"textures/shot0sid", "textures/id1/shot0sid"},
	{"textures/shot0top", "textures/id1/shot0top"},
	{"textures/shot1sid", "textures/id1/shot1sid"},
	{"textures/shot1top", "textures/id1/shot1top"},
	{"textures/skill0", "textures/id1/skill0"},
	{"textures/skill1", "textures/id1/skill1"},
	{"textures/skill2", "textures/id1/skill2"},
	{"textures/skill3", "textures/id1/skill3"},
	{"textures/sky1", "textures/id1/sky1"},
	{"textures/sky4", "textures/id1/sky4"},
	{"textures/slip1", "textures/id1/slip1"},
	{"textures/slip2", "textures/id1/slip2"},
	{"textures/slipbotsd", "textures/id1/slipbotsd"},
	{"textures/sliplite", "textures/id1/sliplite"},
	{"textures/slipside", "textures/id1/slipside"},
	{"textures/sliptopsd", "textures/id1/sliptopsd"},
	{"textures/stone1_3", "textures/id1/stone1_3"},
	{"textures/stone1_5", "textures/id1/stone1_5"},
	{"textures/stone1_7", "textures/id1/stone1_7"},
	{"textures/switch_1", "textures/id1/switch_1"},
	{"textures/swtch1_1", "textures/id1/swtch1_1"},
	{"textures/tech01_1", "textures/id1/tech01_1"},
	{"textures/tech01_2", "textures/id1/tech01_2"},
	{"textures/tech01_3", "textures/id1/tech01_3"},
	{"textures/tech01_5", "textures/id1/tech01_5"},
	{"textures/tech01_6", "textures/id1/tech01_6"},
	{"textures/tech01_7", "textures/id1/tech01_7"},
	{"textures/tech01_9", "textures/id1/tech01_9"},
	{"textures/tech02_1", "textures/id1/tech02_1"},
	{"textures/tech02_2", "textures/id1/tech02_2"},
	{"textures/tech02_3", "textures/id1/tech02_3"},
	{"textures/tech02_5", "textures/id1/tech02_5"},
	{"textures/tech02_6", "textures/id1/tech02_6"},
	{"textures/tech02_7", "textures/id1/tech02_7"},
	{"textures/tech03_1", "textures/id1/tech03_1"},
	{"textures/tech03_2", "textures/id1/tech03_2"},
	{"textures/tech04_1", "textures/id1/tech04_1"},
	{"textures/tech04_2", "textures/id1/tech04_2"},
	{"textures/tech04_3", "textures/id1/tech04_3"},
	{"textures/tech04_4", "textures/id1/tech04_4"},
	{"textures/tech04_5", "textures/id1/tech04_5"},
	{"textures/tech04_6", "textures/id1/tech04_6"},
	{"textures/tech04_7", "textures/id1/tech04_7"},
	{"textures/tech04_8", "textures/id1/tech04_8"},
	{"textures/tech05_1", "textures/id1/tech05_1"},
	{"textures/tech05_2", "textures/id1/tech05_2"},
	{"textures/tech06_1", "textures/id1/tech06_1"},
	{"textures/tech06_2", "textures/id1/tech06_2"},
	{"textures/tech07_1", "textures/id1/tech07_1"},
	{"textures/tech07_2", "textures/id1/tech07_2"},
	{"textures/tech08_1", "textures/id1/tech08_1"},
	{"textures/tech08_2", "textures/id1/tech08_2"},
	{"textures/tech09_3", "textures/id1/tech09_3"},
	{"textures/tech09_4", "textures/id1/tech09_4"},
	{"textures/tech10_1", "textures/id1/tech10_1"},
	{"textures/tech10_3", "textures/id1/tech10_3"},
	{"textures/tech11_1", "textures/id1/tech11_1"},
	{"textures/tech11_2", "textures/id1/tech11_2"},
	{"textures/tech12_1", "textures/id1/tech12_1"},
	{"textures/tech13_2", "textures/id1/tech13_2"},
	{"textures/tech14_1", "textures/id1/tech14_1"},
	{"textures/tech14_2", "textures/id1/tech14_2"},
	{"textures/tele_top", "textures/id1/tele_top"},
	{"textures/tlight01", "textures/id1/tlight01"},
	{"textures/tlight01_2", "textures/id1/tlight01_2"},
	{"textures/tlight02", "textures/id1/tlight02"},
	{"textures/tlight03", "textures/id1/tlight03"},
	{"textures/tlight05", "textures/id1/tlight05"},
	{"textures/tlight07", "textures/id1/tlight07"},
	{"textures/tlight08", "textures/id1/tlight08"},
	{"textures/tlight09", "textures/id1/tlight09"},
	{"textures/tlight10", "textures/id1/tlight10"},
	{"textures/tlight11", "textures/id1/tlight11"},
	{"textures/trigger", "textures/common/trigger"},
	{"textures/twall1_1", "textures/id1/twall1_1"},
	{"textures/twall1_2", "textures/id1/twall1_2"},
	{"textures/twall1_4", "textures/id1/twall1_4"},
	{"textures/twall2_1", "textures/id1/twall2_1"},
	{"textures/twall2_2", "textures/id1/twall2_2"},
	{"textures/twall2_3", "textures/id1/twall2_3"},
	{"textures/twall2_5", "textures/id1/twall2_5"},
	{"textures/twall2_6", "textures/id1/twall2_6"},
	{"textures/twall3_1", "textures/id1/twall3_1"},
	{"textures/twall5_1", "textures/id1/twall5_1"},
	{"textures/twall5_2", "textures/id1/twall5_2"},
	{"textures/twall5_3", "textures/id1/twall5_3"},
	{"textures/unwall1_8", "textures/id1/unwall1_8"},
	{"textures/uwall1_2", "textures/id1/uwall1_2"},
	{"textures/uwall1_3", "textures/id1/uwall1_3"},
	{"textures/uwall1_4", "textures/id1/uwall1_4"},
	{"textures/vine1_2", "textures/id1/vine1_2"},
	{"textures/wall11_2", "textures/id1/wall11_2"},
	{"textures/wall11_6", "textures/id1/wall11_6"},
	{"textures/wall14_5", "textures/id1/wall14_5"},
	{"textures/wall14_6", "textures/id1/wall14_6"},
	{"textures/wall16_7", "textures/id1/wall16_7"},
	{"textures/wall3_4", "textures/id1/wall3_4"},
	{"textures/wall5_4", "textures/id1/wall5_4"},
	{"textures/wall9_3", "textures/id1/wall9_3"},
	{"textures/wall9_8", "textures/id1/wall9_8"},
	{"textures/warch05", "textures/id1/warch05"},
	{"textures/wbrick1_4", "textures/id1/wbrick1_4"},
	{"textures/wbrick1_5", "textures/id1/wbrick1_5"},
	{"textures/wceiling4", "textures/id1/wceiling4"},
	{"textures/wceiling5", "textures/id1/wceiling5"},
	{"textures/wenter01", "textures/id1/wenter01"},
	{"textures/wexit01", "textures/id1/wexit01"},
	{"textures/wgrass1_1", "textures/id1/wgrass1_1"},
	{"textures/wgrnd1_5", "textures/id1/wgrnd1_5"},
	{"textures/wgrnd1_6", "textures/id1/wgrnd1_6"},
	{"textures/wgrnd1_8", "textures/id1/wgrnd1_8"},
	{"textures/window01_1", "textures/id1/window01_1"},
	{"textures/window01_2", "textures/id1/window01_2"},
	{"textures/window01_3", "textures/id1/window01_3"},
	{"textures/window01_4", "textures/id1/window01_4"},
	{"textures/window02_1", "textures/id1/window02_1"},
	{"textures/window03", "textures/id1/window03"},
	{"textures/window1_2", "textures/id1/window1_2"},
	{"textures/window1_3", "textures/id1/window1_3"},
	{"textures/window1_4", "textures/id1/window1_4"},
	{"textures/wiz1_1", "textures/id1/wiz1_1"},
	{"textures/wiz1_4", "textures/id1/wiz1_4"},
	{"textures/wizmet1_1", "textures/id1/wizmet1_1"},
	{"textures/wizmet1_2", "textures/id1/wizmet1_2"},
	{"textures/wizmet1_3", "textures/id1/wizmet1_3"},
	{"textures/wizmet1_4", "textures/id1/wizmet1_4"},
	{"textures/wizmet1_5", "textures/id1/wizmet1_5"},
	{"textures/wizmet1_6", "textures/id1/wizmet1_6"},
	{"textures/wizmet1_7", "textures/id1/wizmet1_7"},
	{"textures/wizmet1_8", "textures/id1/wizmet1_8"},
	{"textures/wizwin1_2", "textures/id1/wizwin1_2"},
	{"textures/wizwin1_8", "textures/id1/wizwin1_8"},
	{"textures/wizwood1_2", "textures/id1/wizwood1_2"},
	{"textures/wizwood1_3", "textures/id1/wizwood1_3"},
	{"textures/wizwood1_4", "textures/id1/wizwood1_4"},
	{"textures/wizwood1_5", "textures/id1/wizwood1_5"},
	{"textures/wizwood1_6", "textures/id1/wizwood1_6"},
	{"textures/wizwood1_7", "textures/id1/wizwood1_7"},
	{"textures/wizwood1_8", "textures/id1/wizwood1_8"},
	{"textures/wkey02_1", "textures/id1/wkey02_1"},
	{"textures/wkey02_2", "textures/id1/wkey02_2"},
	{"textures/wkey02_3", "textures/id1/wkey02_3"},
	{"textures/wmet1_1", "textures/id1/wmet1_1"},
	{"textures/wmet2_1", "textures/id1/wmet2_1"},
	{"textures/wmet2_2", "textures/id1/wmet2_2"},
	{"textures/wmet2_3", "textures/id1/wmet2_3"},
	{"textures/wmet2_4", "textures/id1/wmet2_4"},
	{"textures/wmet2_6", "textures/id1/wmet2_6"},
	{"textures/wmet3_1", "textures/id1/wmet3_1"},
	{"textures/wmet3_3", "textures/id1/wmet3_3"},
	{"textures/wmet3_4", "textures/id1/wmet3_4"},
	{"textures/wmet4_2", "textures/id1/wmet4_2"},
	{"textures/wmet4_3", "textures/id1/wmet4_3"},
	{"textures/wmet4_4", "textures/id1/wmet4_4"},
	{"textures/wmet4_5", "textures/id1/wmet4_5"},
	{"textures/wmet4_6", "textures/id1/wmet4_6"},
	{"textures/wmet4_7", "textures/id1/wmet4_7"},
	{"textures/wmet4_8", "textures/id1/wmet4_8"},
	{"textures/wood1_1", "textures/id1/wood1_1"},
	{"textures/wood1_5", "textures/id1/wood1_5"},
	{"textures/wood1_7", "textures/id1/wood1_7"},
	{"textures/wood1_8", "textures/id1/wood1_8"},
	{"textures/woodflr1_2", "textures/id1/woodflr1_2"},
	{"textures/woodflr1_4", "textures/id1/woodflr1_4"},
	{"textures/woodflr1_5", "textures/id1/woodflr1_5"},
	{"textures/wswamp1_2", "textures/id1/wswamp1_2"},
	{"textures/wswamp1_4", "textures/id1/wswamp1_4"},
	{"textures/wswamp2_1", "textures/id1/wswamp2_1"},
	{"textures/wswamp2_2", "textures/id1/wswamp2_2"},
	{"textures/wswitch1", "textures/id1/wswitch1"},
	{"textures/wwall1_1", "textures/id1/wwall1_1"},
	{"textures/wwood1_5", "textures/id1/wwood1_5"},
	{"textures/wwood1_7", "textures/id1/wwood1_7"},
	{"textures/z_exit", "textures/id1/z_exit"},

	// prototype.wad
	{"textures/#64_water1_1", "textures/prototype/64_water1_1"},
	{"textures/#64_water1_2", "textures/prototype/64_water1_2"},
	{"textures/#64_water1_3", "textures/prototype/64_water1_3"},
	{"textures/#64_water2_1", "textures/prototype/64_water2_1"},
	{"textures/#64_water2_2", "textures/prototype/64_water2_2"},
	{"textures/#64_water2_3", "textures/prototype/64_water2_3"},
	{"textures/#lava2_1", "textures/prototype/lava2_1"},
	{"textures/#lava2_2", "textures/prototype/lava2_2"},
	{"textures/#lava2_3", "textures/prototype/lava2_3"},
	{"textures/#lava_64_1", "textures/prototype/lava_64_1"},
	{"textures/#lava_64_2", "textures/prototype/lava_64_2"},
	{"textures/#lava_64_3", "textures/prototype/lava_64_3"},
	{"textures/#lavaskip", "textures/prototype/lavaskip"},
	{"textures/#slime2_1", "textures/prototype/slime2_1"},
	{"textures/#slime2_2", "textures/prototype/slime2_2"},
	{"textures/#slime2_3", "textures/prototype/slime2_3"},
	{"textures/#slimeskip", "textures/prototype/slimeskip"},
	{"textures/#tele01", "textures/prototype/tele01"},
	{"textures/#water3_1", "textures/prototype/water3_1"},
	{"textures/#water3_2", "textures/prototype/water3_2"},
	{"textures/#water3_3", "textures/prototype/water3_3"},
	{"textures/#waterskip", "textures/prototype/waterskip"},
	{"textures/+0button_0", "textures/prototype/0button_0"},
	{"textures/+0button_1", "textures/prototype/0button_1"},
	{"textures/+0button_2", "textures/prototype/0button_2"},
	{"textures/+0button_3", "textures/prototype/0button_3"},
	{"textures/+0shbtn_1", "textures/prototype/0shbtn_1"},
	{"textures/+0shbtn_2", "textures/prototype/0shbtn_2"},
	{"textures/+0shbtn_3", "textures/prototype/0shbtn_3"},
	{"textures/+0shbtn_4", "textures/prototype/0shbtn_4"},
	{"textures/+0shbtn_5", "textures/prototype/0shbtn_5"},
	{"textures/+0shbtn_6", "textures/prototype/0shbtn_6"},
	{"textures/+0shbtn_7", "textures/prototype/0shbtn_7"},
	{"textures/+1button_0", "textures/prototype/1button_0"},
	{"textures/+1button_1", "textures/prototype/1button_1"},
	{"textures/+1button_2", "textures/prototype/1button_2"},
	{"textures/+1button_3", "textures/prototype/1button_3"},
	{"textures/+1shbtn_1", "textures/prototype/1shbtn_1"},
	{"textures/+1shbtn_2", "textures/prototype/1shbtn_2"},
	{"textures/+1shbtn_3", "textures/prototype/1shbtn_3"},
	{"textures/+1shbtn_4", "textures/prototype/1shbtn_4"},
	{"textures/+1shbtn_5", "textures/prototype/1shbtn_5"},
	{"textures/+1shbtn_6", "textures/prototype/1shbtn_6"},
	{"textures/+1shbtn_7", "textures/prototype/1shbtn_7"},
	{"textures/128_blood_1", "textures/prototype/128_blood_1"},
	{"textures/128_blood_2", "textures/prototype/128_blood_2"},
	{"textures/128_blood_3", "textures/prototype/128_blood_3"},
	{"textures/128_blue_1", "textures/prototype/128_blue_1"},
	{"textures/128_blue_2", "textures/prototype/128_blue_2"},
	{"textures/128_blue_3", "textures/prototype/128_blue_3"},
	{"textures/128_brown_1", "textures/prototype/128_brown_1"},
	{"textures/128_brown_2", "textures/prototype/128_brown_2"},
	{"textures/128_brown_3", "textures/prototype/128_brown_3"},
	{"textures/128_cyan_1", "textures/prototype/128_cyan_1"},
	{"textures/128_cyan_2", "textures/prototype/128_cyan_2"},
	{"textures/128_cyan_3", "textures/prototype/128_cyan_3"},
	{"textures/128_gold_1", "textures/prototype/128_gold_1"},
	{"textures/128_gold_2", "textures/prototype/128_gold_2"},
	{"textures/128_gold_3", "textures/prototype/128_gold_3"},
	{"textures/128_green_1", "textures/prototype/128_green_1"},
	{"textures/128_green_2", "textures/prototype/128_green_2"},
	{"textures/128_green_3", "textures/prototype/128_green_3"},
	{"textures/128_grey_1", "textures/prototype/128_grey_1"},
	{"textures/128_grey_2", "textures/prototype/128_grey_2"},
	{"textures/128_grey_3", "textures/prototype/128_grey_3"},
	{"textures/128_honey_1", "textures/prototype/128_honey_1"},
	{"textures/128_honey_2", "textures/prototype/128_honey_2"},
	{"textures/128_honey_3", "textures/prototype/128_honey_3"},
	{"textures/16_blood_1", "textures/prototype/16_blood_1"},
	{"textures/16_blood_2", "textures/prototype/16_blood_2"},
	{"textures/16_blood_3", "textures/prototype/16_blood_3"},
	{"textures/16_blue_1", "textures/prototype/16_blue_1"},
	{"textures/16_blue_2", "textures/prototype/16_blue_2"},
	{"textures/16_blue_3", "textures/prototype/16_blue_3"},
	{"textures/16_brown_1", "textures/prototype/16_brown_1"},
	{"textures/16_brown_2", "textures/prototype/16_brown_2"},
	{"textures/16_brown_3", "textures/prototype/16_brown_3"},
	{"textures/16_cyan_1", "textures/prototype/16_cyan_1"},
	{"textures/16_cyan_2", "textures/prototype/16_cyan_2"},
	{"textures/16_cyan_3", "textures/prototype/16_cyan_3"},
	{"textures/16_gold_1", "textures/prototype/16_gold_1"},
	{"textures/16_gold_2", "textures/prototype/16_gold_2"},
	{"textures/16_gold_3", "textures/prototype/16_gold_3"},
	{"textures/16_green_1", "textures/prototype/16_green_1"},
	{"textures/16_green_2", "textures/prototype/16_green_2"},
	{"textures/16_green_3", "textures/prototype/16_green_3"},
	{"textures/16_grey_1", "textures/prototype/16_grey_1"},
	{"textures/16_grey_2", "textures/prototype/16_grey_2"},
	{"textures/16_grey_3", "textures/prototype/16_grey_3"},
	{"textures/16_honey_1", "textures/prototype/16_honey_1"},
	{"textures/16_honey_2", "textures/prototype/16_honey_2"},
	{"textures/16_honey_3", "textures/prototype/16_honey_3"},
	{"textures/32_blood_1", "textures/prototype/32_blood_1"},
	{"textures/32_blood_2", "textures/prototype/32_blood_2"},
	{"textures/32_blood_3", "textures/prototype/32_blood_3"},
	{"textures/32_blue_1", "textures/prototype/32_blue_1"},
	{"textures/32_blue_2", "textures/prototype/32_blue_2"},
	{"textures/32_blue_3", "textures/prototype/32_blue_3"},
	{"textures/32_brown_1", "textures/prototype/32_brown_1"},
	{"textures/32_brown_2", "textures/prototype/32_brown_2"},
	{"textures/32_brown_3", "textures/prototype/32_brown_3"},
	{"textures/32_cyan_1", "textures/prototype/32_cyan_1"},
	{"textures/32_cyan_2", "textures/prototype/32_cyan_2"},
	{"textures/32_cyan_3", "textures/prototype/32_cyan_3"},
	{"textures/32_gold_1", "textures/prototype/32_gold_1"},
	{"textures/32_gold_2", "textures/prototype/32_gold_2"},
	{"textures/32_gold_3", "textures/prototype/32_gold_3"},
	{"textures/32_green_1", "textures/prototype/32_green_1"},
	{"textures/32_green_2", "textures/prototype/32_green_2"},
	{"textures/32_green_3", "textures/prototype/32_green_3"},
	{"textures/32_grey_1", "textures/prototype/32_grey_1"},
	{"textures/32_grey_2", "textures/prototype/32_grey_2"},
	{"textures/32_grey_3", "textures/prototype/32_grey_3"},
	{"textures/32_honey_1", "textures/prototype/32_honey_1"},
	{"textures/32_honey_2", "textures/prototype/32_honey_2"},
	{"textures/32_honey_3", "textures/prototype/32_honey_3"},
	{"textures/64_blood_1", "textures/prototype/64_blood_1"},
	{"textures/64_blood_2", "textures/prototype/64_blood_2"},
	{"textures/64_blood_3", "textures/prototype/64_blood_3"},
	{"textures/64_blue_1", "textures/prototype/64_blue_1"},
	{"textures/64_blue_2", "textures/prototype/64_blue_2"},
	{"textures/64_blue_3", "textures/prototype/64_blue_3"},
	{"textures/64_brown_1", "textures/prototype/64_brown_1"},
	{"textures/64_brown_2", "textures/prototype/64_brown_2"},
	{"textures/64_brown_3", "textures/prototype/64_brown_3"},
	{"textures/64_cyan_1", "textures/prototype/64_cyan_1"},
	{"textures/64_cyan_2", "textures/prototype/64_cyan_2"},
	{"textures/64_cyan_3", "textures/prototype/64_cyan_3"},
	{"textures/64_gold_1", "textures/prototype/64_gold_1"},
	{"textures/64_gold_2", "textures/prototype/64_gold_2"},
	{"textures/64_gold_3", "textures/prototype/64_gold_3"},
	{"textures/64_green_1", "textures/prototype/64_green_1"},
	{"textures/64_green_2", "textures/prototype/64_green_2"},
	{"textures/64_green_3", "textures/prototype/64_green_3"},
	{"textures/64_grey_1", "textures/prototype/64_grey_1"},
	{"textures/64_grey_3", "textures/prototype/64_grey_3"},
	{"textures/64_grey_4", "textures/prototype/64_grey_4"},
	{"textures/64_honey_1", "textures/prototype/64_honey_1"},
	{"textures/64_honey_2", "textures/prototype/64_honey_2"},
	{"textures/64_honey_3", "textures/prototype/64_honey_3"},
	{"textures/blood_1", "textures/prototype/blood_1"},
	{"textures/blood_2", "textures/prototype/blood_2"},
	{"textures/blood_3", "textures/prototype/blood_3"},
	{"textures/blue_1", "textures/prototype/blue_1"},
	{"textures/blue_2", "textures/prototype/blue_2"},
	{"textures/blue_3", "textures/prototype/blue_3"},
	{"textures/brown_1", "textures/prototype/brown_1"},
	{"textures/brown_2", "textures/prototype/brown_2"},
	{"textures/brown_3", "textures/prototype/brown_3"},
	{"textures/clip", "textures/prototype/clip"},
	{"textures/cyan_1", "textures/prototype/cyan_1"},
	{"textures/cyan_2", "textures/prototype/cyan_2"},
	{"textures/cyan_3", "textures/prototype/cyan_3"},
	{"textures/fullbright_1", "textures/prototype/fullbright_1"},
	{"textures/fullbright_2", "textures/prototype/fullbright_2"},
	{"textures/fullbright_3", "textures/prototype/fullbright_3"},
	{"textures/fullbright_4", "textures/prototype/fullbright_4"},
	{"textures/fullbright_5", "textures/prototype/fullbright_5"},
	{"textures/fullbright_6", "textures/prototype/fullbright_6"},
	{"textures/fullbright_7", "textures/prototype/fullbright_7"},
	{"textures/fullbright_8", "textures/prototype/fullbright_8"},
	{"textures/fullbright_9", "textures/prototype/fullbright_9"},
	{"textures/gold_1", "textures/prototype/gold_1"},
	{"textures/gold_2", "textures/prototype/gold_2"},
	{"textures/gold_3", "textures/prototype/gold_3"},
	{"textures/green_1", "textures/prototype/green_1"},
	{"textures/green_2", "textures/prototype/green_2"},
	{"textures/green_3", "textures/prototype/green_3"},
	{"textures/grey_1", "textures/prototype/grey_1"},
	{"textures/grey_2", "textures/prototype/grey_2"},
	{"textures/grey_3", "textures/prototype/grey_3"},
	{"textures/hint", "textures/prototype/hint"},
	{"textures/hintskip", "textures/prototype/hintskip"},
	{"textures/honey_1", "textures/prototype/honey_1"},
	{"textures/honey_2", "textures/prototype/honey_2"},
	{"textures/honey_3", "textures/prototype/honey_3"},
	{"textures/key_gold_1", "textures/prototype/key_gold_1"},
	{"textures/key_silver_1", "textures/prototype/key_silver_1"},
	{"textures/light1_big", "textures/prototype/light1_big"},
	{"textures/light1_grt", "textures/prototype/light1_grt"},
	{"textures/light1_sml", "textures/prototype/light1_sml"},
	{"textures/origin", "textures/common/origin"},
	{"textures/skip", "textures/prototype/skip"},
	{"textures/sky1", "textures/prototype/sky1"},
	{"textures/sky2", "textures/prototype/sky2"},
	{"textures/sky_blood_1", "textures/prototype/sky_blood_1"},
	{"textures/sky_blood_2", "textures/prototype/sky_blood_2"},
	{"textures/sky_blood_3", "textures/prototype/sky_blood_3"},
	{"textures/sky_blue_1", "textures/prototype/sky_blue_1"},
	{"textures/sky_blue_2", "textures/prototype/sky_blue_2"},
	{"textures/sky_blue_3", "textures/prototype/sky_blue_3"},
	{"textures/sky_brown_1", "textures/prototype/sky_brown_1"},
	{"textures/sky_brown_2", "textures/prototype/sky_brown_2"},
	{"textures/sky_brown_3", "textures/prototype/sky_brown_3"},
	{"textures/sky_cyan_1", "textures/prototype/sky_cyan_1"},
	{"textures/sky_cyan_2", "textures/prototype/sky_cyan_2"},
	{"textures/sky_cyan_3", "textures/prototype/sky_cyan_3"},
	{"textures/sky_gold_1", "textures/prototype/sky_gold_1"},
	{"textures/sky_gold_2", "textures/prototype/sky_gold_2"},
	{"textures/sky_gold_3", "textures/prototype/sky_gold_3"},
	{"textures/sky_green_1", "textures/prototype/sky_green_1"},
	{"textures/sky_green_2", "textures/prototype/sky_green_2"},
	{"textures/sky_green_3", "textures/prototype/sky_green_3"},
	{"textures/sky_grey_1", "textures/prototype/sky_grey_1"},
	{"textures/sky_grey_2", "textures/prototype/sky_grey_2"},
	{"textures/sky_grey_3", "textures/prototype/sky_grey_3"},
	{"textures/sky_honey_1", "textures/prototype/sky_honey_1"},
	{"textures/sky_honey_2", "textures/prototype/sky_honey_2"},
	{"textures/sky_honey_3", "textures/prototype/sky_honey_3"},
	{"textures/sky_void", "textures/prototype/sky_void"},
	{"textures/text_0", "textures/prototype/text_0"},
	{"textures/text_1", "textures/prototype/text_1"},
	{"textures/text_2", "textures/prototype/text_2"},
	{"textures/text_3", "textures/prototype/text_3"},
	{"textures/text_4", "textures/prototype/text_4"},
	{"textures/text_5", "textures/prototype/text_5"},
	{"textures/text_6", "textures/prototype/text_6"},
	{"textures/text_7", "textures/prototype/text_7"},
	{"textures/text_8", "textures/prototype/text_8"},
	{"textures/text_9", "textures/prototype/text_9"},
	{"textures/text_A", "textures/prototype/text_A"},
	{"textures/text_B", "textures/prototype/text_B"},
	{"textures/text_C", "textures/prototype/text_C"},
	{"textures/text_D", "textures/prototype/text_D"},
	{"textures/text_E", "textures/prototype/text_E"},
	{"textures/text_F", "textures/prototype/text_F"},
	{"textures/text_G", "textures/prototype/text_G"},
	{"textures/text_H", "textures/prototype/text_H"},
	{"textures/text_I", "textures/prototype/text_I"},
	{"textures/text_J", "textures/prototype/text_J"},
	{"textures/text_K", "textures/prototype/text_K"},
	{"textures/text_L", "textures/prototype/text_L"},
	{"textures/text_M", "textures/prototype/text_M"},
	{"textures/text_N", "textures/prototype/text_N"},
	{"textures/text_O", "textures/prototype/text_O"},
	{"textures/text_P", "textures/prototype/text_P"},
	{"textures/text_Q", "textures/prototype/text_Q"},
	{"textures/text_R", "textures/prototype/text_R"},
	{"textures/text_S", "textures/prototype/text_S"},
	{"textures/text_T", "textures/prototype/text_T"},
	{"textures/text_U", "textures/prototype/text_U"},
	{"textures/text_V", "textures/prototype/text_V"},
	{"textures/text_W", "textures/prototype/text_W"},
	{"textures/text_X", "textures/prototype/text_X"},
	{"textures/text_Y", "textures/prototype/text_Y"},
	{"textures/text_Z", "textures/prototype/text_Z"},
	{"textures/text_amper", "textures/prototype/text_amper"},
	{"textures/text_arrow", "textures/prototype/text_arrow"},
	{"textures/text_circle", "textures/prototype/text_circle"},
	{"textures/text_comma", "textures/prototype/text_comma"},
	{"textures/text_dash", "textures/prototype/text_dash"},
	{"textures/text_dot", "textures/prototype/text_dot"},
	{"textures/text_emark", "textures/prototype/text_emark"},
	{"textures/text_hash", "textures/prototype/text_hash"},
	{"textures/text_light", "textures/prototype/text_light"},
	{"textures/text_percent", "textures/prototype/text_percent"},
	{"textures/text_qmark", "textures/prototype/text_qmark"},
	{"textures/text_smiley", "textures/prototype/text_smiley"},
	{"textures/text_sound", "textures/prototype/text_sound"},
	{"textures/text_uscore", "textures/prototype/text_uscore"},
	{"textures/trigger", "textures/common/trigger"},
	{"textures/void", "textures/prototype/void"},

	// Makkon building 2022.8
	{"textures/roof_blck_1", "textures/makkon_building/roof_blck_1"},
	{"textures/roof_blue1_1", "textures/makkon_building/roof_blue1_1"},
	{"textures/roof_blue2_1", "textures/makkon_building/roof_blue2_1"},
	{"textures/roof_brwn1_1", "textures/makkon_building/roof_brwn1_1"},
	{"textures/roof_brwn2_1", "textures/makkon_building/roof_brwn2_1"},
	{"textures/roof_grey_1", "textures/makkon_building/roof_grey_1"},
	{"textures/roof_pnk_1", "textures/makkon_building/roof_pnk_1"},
	{"textures/roof_prpl_1", "textures/makkon_building/roof_prpl_1"},
	{"textures/roof_red1_1", "textures/makkon_building/roof_red1_1"},
	{"textures/roof_red2_1", "textures/makkon_building/roof_red2_1"},
	{"textures/roof_teal_1", "textures/makkon_building/roof_teal_1"},
	{"textures/roofc1_rbare01", "textures/makkon_building/roofc1_rbare01"},
	{"textures/roofc1_rbare02", "textures/makkon_building/roofc1_rbare02"},
	{"textures/roofc1_rbare03", "textures/makkon_building/roofc1_rbare03"},
	{"textures/roofc1_rbare04", "textures/makkon_building/roofc1_rbare04"},
	{"textures/roofc1_rbare05", "textures/makkon_building/roofc1_rbare05"},
	{"textures/roofc1_rblck01", "textures/makkon_building/roofc1_rblck01"},
	{"textures/roofc1_rblck02", "textures/makkon_building/roofc1_rblck02"},
	{"textures/roofc1_rblck03", "textures/makkon_building/roofc1_rblck03"},
	{"textures/roofc1_rblck04", "textures/makkon_building/roofc1_rblck04"},
	{"textures/roofc1_rblck05", "textures/makkon_building/roofc1_rblck05"},
	{"textures/roofc1_rblue01", "textures/makkon_building/roofc1_rblue01"},
	{"textures/roofc1_rblue02", "textures/makkon_building/roofc1_rblue02"},
	{"textures/roofc1_rblue03", "textures/makkon_building/roofc1_rblue03"},
	{"textures/roofc1_rblue04", "textures/makkon_building/roofc1_rblue04"},
	{"textures/roofc1_rblue05", "textures/makkon_building/roofc1_rblue05"},
	{"textures/roofc1_rdark01", "textures/makkon_building/roofc1_rdark01"},
	{"textures/roofc1_rdark02", "textures/makkon_building/roofc1_rdark02"},
	{"textures/roofc1_rdark03", "textures/makkon_building/roofc1_rdark03"},
	{"textures/roofc1_rdark04", "textures/makkon_building/roofc1_rdark04"},
	{"textures/roofc1_rdark05", "textures/makkon_building/roofc1_rdark05"},
	{"textures/roofc1_rprpl01", "textures/makkon_building/roofc1_rprpl01"},
	{"textures/roofc1_rprpl02", "textures/makkon_building/roofc1_rprpl02"},
	{"textures/roofc1_rprpl03", "textures/makkon_building/roofc1_rprpl03"},
	{"textures/roofc1_rprpl04", "textures/makkon_building/roofc1_rprpl04"},
	{"textures/roofc1_rprpl05", "textures/makkon_building/roofc1_rprpl05"},
	{"textures/roofc1_rred01", "textures/makkon_building/roofc1_rred01"},
	{"textures/roofc1_rred02", "textures/makkon_building/roofc1_rred02"},
	{"textures/roofc1_rred03", "textures/makkon_building/roofc1_rred03"},
	{"textures/roofc1_rred04", "textures/makkon_building/roofc1_rred04"},
	{"textures/roofc1_rred05", "textures/makkon_building/roofc1_rred05"},
	{"textures/roofc1_rteal01", "textures/makkon_building/roofc1_rteal01"},
	{"textures/roofc1_rteal02", "textures/makkon_building/roofc1_rteal02"},
	{"textures/roofc1_rteal03", "textures/makkon_building/roofc1_rteal03"},
	{"textures/roofc1_rteal04", "textures/makkon_building/roofc1_rteal04"},
	{"textures/roofc1_rteal05", "textures/makkon_building/roofc1_rteal05"},
	{"textures/roofc1_rwht01", "textures/makkon_building/roofc1_rwht01"},
	{"textures/roofc1_rwht02", "textures/makkon_building/roofc1_rwht02"},
	{"textures/roofc1_rwht03", "textures/makkon_building/roofc1_rwht03"},
	{"textures/roofc1_rwht04", "textures/makkon_building/roofc1_rwht04"},
	{"textures/roofc1_rwht05", "textures/makkon_building/roofc1_rwht05"},
	{"textures/roofc1_rylw01", "textures/makkon_building/roofc1_rylw01"},
	{"textures/roofc1_rylw02", "textures/makkon_building/roofc1_rylw02"},
	{"textures/roofc1_rylw03", "textures/makkon_building/roofc1_rylw03"},
	{"textures/roofc1_rylw04", "textures/makkon_building/roofc1_rylw04"},
	{"textures/roofc1_rylw05", "textures/makkon_building/roofc1_rylw05"},
	{"textures/roofc2_rbare01", "textures/makkon_building/roofc2_rbare01"},
	{"textures/roofc2_rbare02", "textures/makkon_building/roofc2_rbare02"},
	{"textures/roofc2_rbare03", "textures/makkon_building/roofc2_rbare03"},
	{"textures/roofc2_rbare04", "textures/makkon_building/roofc2_rbare04"},
	{"textures/roofc2_rbare05", "textures/makkon_building/roofc2_rbare05"},
	{"textures/roofc3_rbare01", "textures/makkon_building/roofc3_rbare01"},
	{"textures/roofc3_rbare02", "textures/makkon_building/roofc3_rbare02"},
	{"textures/roofc3_rbare03", "textures/makkon_building/roofc3_rbare03"},
	{"textures/roofc3_rbare04", "textures/makkon_building/roofc3_rbare04"},
	{"textures/roofc3_rbare05", "textures/makkon_building/roofc3_rbare05"},
	{"textures/win_base1_01", "textures/makkon_building/win_base1_01"},
	{"textures/win_base1_03", "textures/makkon_building/win_base1_03"},
	{"textures/win_base1_05", "textures/makkon_building/win_base1_05"},
	{"textures/win_base1_07", "textures/makkon_building/win_base1_07"},
	{"textures/win_base1_09", "textures/makkon_building/win_base1_09"},
	{"textures/win_blu1_01", "textures/makkon_building/win_blu1_01"},
	{"textures/win_blu1_02", "textures/makkon_building/win_blu1_02"},
	{"textures/win_blu1_03", "textures/makkon_building/win_blu1_03"},
	{"textures/win_blu1_04", "textures/makkon_building/win_blu1_04"},
	{"textures/win_blu1_05", "textures/makkon_building/win_blu1_05"},
	{"textures/win_blu1_06", "textures/makkon_building/win_blu1_06"},
	{"textures/win_blu1_07", "textures/makkon_building/win_blu1_07"},
	{"textures/win_blu1_08", "textures/makkon_building/win_blu1_08"},
	{"textures/win_blu1_09", "textures/makkon_building/win_blu1_09"},
	{"textures/win_blu1_10", "textures/makkon_building/win_blu1_10"},
	{"textures/win_blu2_01", "textures/makkon_building/win_blu2_01"},
	{"textures/win_blu2_02", "textures/makkon_building/win_blu2_02"},
	{"textures/win_blu2_03", "textures/makkon_building/win_blu2_03"},
	{"textures/win_blu2_04", "textures/makkon_building/win_blu2_04"},
	{"textures/win_blu2_05", "textures/makkon_building/win_blu2_05"},
	{"textures/win_blu2_06", "textures/makkon_building/win_blu2_06"},
	{"textures/win_blu2_07", "textures/makkon_building/win_blu2_07"},
	{"textures/win_blu2_08", "textures/makkon_building/win_blu2_08"},
	{"textures/win_blu2_09", "textures/makkon_building/win_blu2_09"},
	{"textures/win_blu2_10", "textures/makkon_building/win_blu2_10"},
	{"textures/win_fbblu_01", "textures/makkon_building/win_fbblu_01"},
	{"textures/win_fbblu_03", "textures/makkon_building/win_fbblu_03"},
	{"textures/win_fbblu_05", "textures/makkon_building/win_fbblu_05"},
	{"textures/win_fbblu_07", "textures/makkon_building/win_fbblu_07"},
	{"textures/win_fbblu_09", "textures/makkon_building/win_fbblu_09"},
	{"textures/win_fborng_01", "textures/makkon_building/win_fborng_01"},
	{"textures/win_fborng_03", "textures/makkon_building/win_fborng_03"},
	{"textures/win_fborng_05", "textures/makkon_building/win_fborng_05"},
	{"textures/win_fborng_07", "textures/makkon_building/win_fborng_07"},
	{"textures/win_fborng_09", "textures/makkon_building/win_fborng_09"},
	{"textures/win_fbred_01", "textures/makkon_building/win_fbred_01"},
	{"textures/win_fbred_03", "textures/makkon_building/win_fbred_03"},
	{"textures/win_fbred_05", "textures/makkon_building/win_fbred_05"},
	{"textures/win_fbred_07", "textures/makkon_building/win_fbred_07"},
	{"textures/win_fbred_09", "textures/makkon_building/win_fbred_09"},
	{"textures/win_fbylw_01", "textures/makkon_building/win_fbylw_01"},
	{"textures/win_fbylw_03", "textures/makkon_building/win_fbylw_03"},
	{"textures/win_fbylw_05", "textures/makkon_building/win_fbylw_05"},
	{"textures/win_fbylw_07", "textures/makkon_building/win_fbylw_07"},
	{"textures/win_fbylw_09", "textures/makkon_building/win_fbylw_09"},
	{"textures/win_grn1_01", "textures/makkon_building/win_grn1_01"},
	{"textures/win_grn1_02", "textures/makkon_building/win_grn1_02"},
	{"textures/win_grn1_03", "textures/makkon_building/win_grn1_03"},
	{"textures/win_grn1_04", "textures/makkon_building/win_grn1_04"},
	{"textures/win_grn1_05", "textures/makkon_building/win_grn1_05"},
	{"textures/win_grn1_06", "textures/makkon_building/win_grn1_06"},
	{"textures/win_grn1_07", "textures/makkon_building/win_grn1_07"},
	{"textures/win_grn1_08", "textures/makkon_building/win_grn1_08"},
	{"textures/win_grn1_09", "textures/makkon_building/win_grn1_09"},
	{"textures/win_grn1_10", "textures/makkon_building/win_grn1_10"},
	{"textures/win_orng1_01", "textures/makkon_building/win_orng1_01"},
	{"textures/win_orng1_02", "textures/makkon_building/win_orng1_02"},
	{"textures/win_orng1_03", "textures/makkon_building/win_orng1_03"},
	{"textures/win_orng1_04", "textures/makkon_building/win_orng1_04"},
	{"textures/win_orng1_05", "textures/makkon_building/win_orng1_05"},
	{"textures/win_orng1_06", "textures/makkon_building/win_orng1_06"},
	{"textures/win_orng1_07", "textures/makkon_building/win_orng1_07"},
	{"textures/win_orng1_08", "textures/makkon_building/win_orng1_08"},
	{"textures/win_orng1_09", "textures/makkon_building/win_orng1_09"},
	{"textures/win_orng1_10", "textures/makkon_building/win_orng1_10"},
	{"textures/win_pink1_01", "textures/makkon_building/win_pink1_01"},
	{"textures/win_pink1_03", "textures/makkon_building/win_pink1_03"},
	{"textures/win_pink1_05", "textures/makkon_building/win_pink1_05"},
	{"textures/win_pink1_07", "textures/makkon_building/win_pink1_07"},
	{"textures/win_pink1_09", "textures/makkon_building/win_pink1_09"},
	{"textures/win_prpl1_01", "textures/makkon_building/win_prpl1_01"},
	{"textures/win_prpl1_02", "textures/makkon_building/win_prpl1_02"},
	{"textures/win_prpl1_03", "textures/makkon_building/win_prpl1_03"},
	{"textures/win_prpl1_04", "textures/makkon_building/win_prpl1_04"},
	{"textures/win_prpl1_05", "textures/makkon_building/win_prpl1_05"},
	{"textures/win_prpl1_06", "textures/makkon_building/win_prpl1_06"},
	{"textures/win_prpl1_07", "textures/makkon_building/win_prpl1_07"},
	{"textures/win_prpl1_08", "textures/makkon_building/win_prpl1_08"},
	{"textures/win_prpl1_09", "textures/makkon_building/win_prpl1_09"},
	{"textures/win_prpl1_10", "textures/makkon_building/win_prpl1_10"},
	{"textures/win_red1_01", "textures/makkon_building/win_red1_01"},
	{"textures/win_red1_02", "textures/makkon_building/win_red1_02"},
	{"textures/win_red1_03", "textures/makkon_building/win_red1_03"},
	{"textures/win_red1_04", "textures/makkon_building/win_red1_04"},
	{"textures/win_red1_05", "textures/makkon_building/win_red1_05"},
	{"textures/win_red1_06", "textures/makkon_building/win_red1_06"},
	{"textures/win_red1_07", "textures/makkon_building/win_red1_07"},
	{"textures/win_red1_08", "textures/makkon_building/win_red1_08"},
	{"textures/win_red1_09", "textures/makkon_building/win_red1_09"},
	{"textures/win_red1_10", "textures/makkon_building/win_red1_10"},
	{"textures/win_rnbw1_01", "textures/makkon_building/win_rnbw1_01"},
	{"textures/win_rnbw1_03", "textures/makkon_building/win_rnbw1_03"},
	{"textures/win_rnbw1_05", "textures/makkon_building/win_rnbw1_05"},
	{"textures/win_rnbw1_07", "textures/makkon_building/win_rnbw1_07"},
	{"textures/win_rnbw1_09", "textures/makkon_building/win_rnbw1_09"},
	{"textures/win_teal1_01", "textures/makkon_building/win_teal1_01"},
	{"textures/win_teal1_03", "textures/makkon_building/win_teal1_03"},
	{"textures/win_teal1_04", "textures/makkon_building/win_teal1_04"},
	{"textures/win_teal1_05", "textures/makkon_building/win_teal1_05"},
	{"textures/win_teal1_06", "textures/makkon_building/win_teal1_06"},
	{"textures/win_teal1_07", "textures/makkon_building/win_teal1_07"},
	{"textures/win_teal1_08", "textures/makkon_building/win_teal1_08"},
	{"textures/win_teal1_09", "textures/makkon_building/win_teal1_09"},
	{"textures/win_teal1_1", "textures/makkon_building/win_teal1_1"},
	{"textures/win_teal1_10", "textures/makkon_building/win_teal1_10"},
	{"textures/win_wht1_01", "textures/makkon_building/win_wht1_01"},
	{"textures/win_wht1_02", "textures/makkon_building/win_wht1_02"},
	{"textures/win_wht1_03", "textures/makkon_building/win_wht1_03"},
	{"textures/win_wht1_04", "textures/makkon_building/win_wht1_04"},
	{"textures/win_wht1_05", "textures/makkon_building/win_wht1_05"},
	{"textures/win_wht1_06", "textures/makkon_building/win_wht1_06"},
	{"textures/win_wht1_07", "textures/makkon_building/win_wht1_07"},
	{"textures/win_wht1_08", "textures/makkon_building/win_wht1_08"},
	{"textures/win_wht1_09", "textures/makkon_building/win_wht1_09"},
	{"textures/win_wht1_10", "textures/makkon_building/win_wht1_10"},
	{"textures/window_blue1_1", "textures/makkon_building/window_blue1_1"},
	{"textures/window_blue1_2", "textures/makkon_building/window_blue1_2"},
	{"textures/window_blue2_1", "textures/makkon_building/window_blue2_1"},
	{"textures/window_blue2_2", "textures/makkon_building/window_blue2_2"},
	{"textures/window_orng_1", "textures/makkon_building/window_orng_1"},
	{"textures/window_orng_2", "textures/makkon_building/window_orng_2"},
	{"textures/window_prpl_1", "textures/makkon_building/window_prpl_1"},
	{"textures/window_prpl_2", "textures/makkon_building/window_prpl_2"},
	{"textures/window_red_1", "textures/makkon_building/window_red_1"},
	{"textures/window_red_2", "textures/makkon_building/window_red_2"},
	{"textures/window_teal_1", "textures/makkon_building/window_teal_1"},
	{"textures/window_teal_2", "textures/makkon_building/window_teal_2"},
	{"textures/window_yllw_1", "textures/makkon_building/window_yllw_1"},
	{"textures/window_yllw_2", "textures/makkon_building/window_yllw_2"},
	{"textures/woodc1_cbare01", "textures/makkon_building/woodc1_cbare01"},
	{"textures/woodc1_cbare02", "textures/makkon_building/woodc1_cbare02"},
	{"textures/woodc1_cbare03", "textures/makkon_building/woodc1_cbare03"},
	{"textures/woodc1_cblck01", "textures/makkon_building/woodc1_cblck01"},
	{"textures/woodc1_cblck02", "textures/makkon_building/woodc1_cblck02"},
	{"textures/woodc1_cblck03", "textures/makkon_building/woodc1_cblck03"},
	{"textures/woodc1_cblue01", "textures/makkon_building/woodc1_cblue01"},
	{"textures/woodc1_cblue02", "textures/makkon_building/woodc1_cblue02"},
	{"textures/woodc1_cblue03", "textures/makkon_building/woodc1_cblue03"},
	{"textures/woodc1_cdark01", "textures/makkon_building/woodc1_cdark01"},
	{"textures/woodc1_cdark02", "textures/makkon_building/woodc1_cdark02"},
	{"textures/woodc1_cdark03", "textures/makkon_building/woodc1_cdark03"},
	{"textures/woodc1_cprpl01", "textures/makkon_building/woodc1_cprpl01"},
	{"textures/woodc1_cprpl02", "textures/makkon_building/woodc1_cprpl02"},
	{"textures/woodc1_cprpl03", "textures/makkon_building/woodc1_cprpl03"},
	{"textures/woodc1_cred01", "textures/makkon_building/woodc1_cred01"},
	{"textures/woodc1_cred02", "textures/makkon_building/woodc1_cred02"},
	{"textures/woodc1_cred03", "textures/makkon_building/woodc1_cred03"},
	{"textures/woodc1_cteal01", "textures/makkon_building/woodc1_cteal01"},
	{"textures/woodc1_cteal02", "textures/makkon_building/woodc1_cteal02"},
	{"textures/woodc1_cteal03", "textures/makkon_building/woodc1_cteal03"},
	{"textures/woodc1_cwht01", "textures/makkon_building/woodc1_cwht01"},
	{"textures/woodc1_cwht02", "textures/makkon_building/woodc1_cwht02"},
	{"textures/woodc1_cwht03", "textures/makkon_building/woodc1_cwht03"},
	{"textures/woodc1_cylw01", "textures/makkon_building/woodc1_cylw01"},
	{"textures/woodc1_cylw02", "textures/makkon_building/woodc1_cylw02"},
	{"textures/woodc1_cylw03", "textures/makkon_building/woodc1_cylw03"},
	{"textures/woodc1_fbare01", "textures/makkon_building/woodc1_fbare01"},
	{"textures/woodc1_fbare02", "textures/makkon_building/woodc1_fbare02"},
	{"textures/woodc1_fbare03", "textures/makkon_building/woodc1_fbare03"},
	{"textures/woodc1_fbare04", "textures/makkon_building/woodc1_fbare04"},
	{"textures/woodc1_fbare05", "textures/makkon_building/woodc1_fbare05"},
	{"textures/woodc1_fblck01", "textures/makkon_building/woodc1_fblck01"},
	{"textures/woodc1_fblck02", "textures/makkon_building/woodc1_fblck02"},
	{"textures/woodc1_fblck03", "textures/makkon_building/woodc1_fblck03"},
	{"textures/woodc1_fblck04", "textures/makkon_building/woodc1_fblck04"},
	{"textures/woodc1_fblck05", "textures/makkon_building/woodc1_fblck05"},
	{"textures/woodc1_fblue01", "textures/makkon_building/woodc1_fblue01"},
	{"textures/woodc1_fblue02", "textures/makkon_building/woodc1_fblue02"},
	{"textures/woodc1_fblue03", "textures/makkon_building/woodc1_fblue03"},
	{"textures/woodc1_fblue04", "textures/makkon_building/woodc1_fblue04"},
	{"textures/woodc1_fblue05", "textures/makkon_building/woodc1_fblue05"},
	{"textures/woodc1_fdark01", "textures/makkon_building/woodc1_fdark01"},
	{"textures/woodc1_fdark02", "textures/makkon_building/woodc1_fdark02"},
	{"textures/woodc1_fdark03", "textures/makkon_building/woodc1_fdark03"},
	{"textures/woodc1_fdark04", "textures/makkon_building/woodc1_fdark04"},
	{"textures/woodc1_fdark05", "textures/makkon_building/woodc1_fdark05"},
	{"textures/woodc1_fprpl01", "textures/makkon_building/woodc1_fprpl01"},
	{"textures/woodc1_fprpl02", "textures/makkon_building/woodc1_fprpl02"},
	{"textures/woodc1_fprpl03", "textures/makkon_building/woodc1_fprpl03"},
	{"textures/woodc1_fprpl04", "textures/makkon_building/woodc1_fprpl04"},
	{"textures/woodc1_fprpl05", "textures/makkon_building/woodc1_fprpl05"},
	{"textures/woodc1_fred01", "textures/makkon_building/woodc1_fred01"},
	{"textures/woodc1_fred02", "textures/makkon_building/woodc1_fred02"},
	{"textures/woodc1_fred03", "textures/makkon_building/woodc1_fred03"},
	{"textures/woodc1_fred04", "textures/makkon_building/woodc1_fred04"},
	{"textures/woodc1_fred05", "textures/makkon_building/woodc1_fred05"},
	{"textures/woodc1_fteal01", "textures/makkon_building/woodc1_fteal01"},
	{"textures/woodc1_fteal02", "textures/makkon_building/woodc1_fteal02"},
	{"textures/woodc1_fteal03", "textures/makkon_building/woodc1_fteal03"},
	{"textures/woodc1_fteal04", "textures/makkon_building/woodc1_fteal04"},
	{"textures/woodc1_fteal05", "textures/makkon_building/woodc1_fteal05"},
	{"textures/woodc1_fwht01", "textures/makkon_building/woodc1_fwht01"},
	{"textures/woodc1_fwht02", "textures/makkon_building/woodc1_fwht02"},
	{"textures/woodc1_fwht03", "textures/makkon_building/woodc1_fwht03"},
	{"textures/woodc1_fwht04", "textures/makkon_building/woodc1_fwht04"},
	{"textures/woodc1_fwht05", "textures/makkon_building/woodc1_fwht05"},
	{"textures/woodc1_fylw01", "textures/makkon_building/woodc1_fylw01"},
	{"textures/woodc1_fylw02", "textures/makkon_building/woodc1_fylw02"},
	{"textures/woodc1_fylw03", "textures/makkon_building/woodc1_fylw03"},
	{"textures/woodc1_fylw04", "textures/makkon_building/woodc1_fylw04"},
	{"textures/woodc1_fylw05", "textures/makkon_building/woodc1_fylw05"},
	{"textures/woodc1_pbare01", "textures/makkon_building/woodc1_pbare01"},
	{"textures/woodc1_pbare02", "textures/makkon_building/woodc1_pbare02"},
	{"textures/woodc1_pblck01", "textures/makkon_building/woodc1_pblck01"},
	{"textures/woodc1_pblck02", "textures/makkon_building/woodc1_pblck02"},
	{"textures/woodc1_pblue01", "textures/makkon_building/woodc1_pblue01"},
	{"textures/woodc1_pblue02", "textures/makkon_building/woodc1_pblue02"},
	{"textures/woodc1_pdark01", "textures/makkon_building/woodc1_pdark01"},
	{"textures/woodc1_pdark02", "textures/makkon_building/woodc1_pdark02"},
	{"textures/woodc1_pprpl01", "textures/makkon_building/woodc1_pprpl01"},
	{"textures/woodc1_pprpl02", "textures/makkon_building/woodc1_pprpl02"},
	{"textures/woodc1_pred01", "textures/makkon_building/woodc1_pred01"},
	{"textures/woodc1_pred02", "textures/makkon_building/woodc1_pred02"},
	{"textures/woodc1_pteal01", "textures/makkon_building/woodc1_pteal01"},
	{"textures/woodc1_pteal02", "textures/makkon_building/woodc1_pteal02"},
	{"textures/woodc1_pwht01", "textures/makkon_building/woodc1_pwht01"},
	{"textures/woodc1_pwht02", "textures/makkon_building/woodc1_pwht02"},
	{"textures/woodc1_pylw01", "textures/makkon_building/woodc1_pylw01"},
	{"textures/woodc1_pylw02", "textures/makkon_building/woodc1_pylw02"},
	{"textures/woodc1_tbare01", "textures/makkon_building/woodc1_tbare01"},
	{"textures/woodc1_tbare02", "textures/makkon_building/woodc1_tbare02"},
	{"textures/woodc1_tbare03", "textures/makkon_building/woodc1_tbare03"},
	{"textures/woodc1_tblck01", "textures/makkon_building/woodc1_tblck01"},
	{"textures/woodc1_tblck02", "textures/makkon_building/woodc1_tblck02"},
	{"textures/woodc1_tblck03", "textures/makkon_building/woodc1_tblck03"},
	{"textures/woodc1_tblue01", "textures/makkon_building/woodc1_tblue01"},
	{"textures/woodc1_tblue02", "textures/makkon_building/woodc1_tblue02"},
	{"textures/woodc1_tblue03", "textures/makkon_building/woodc1_tblue03"},
	{"textures/woodc1_tdark01", "textures/makkon_building/woodc1_tdark01"},
	{"textures/woodc1_tdark02", "textures/makkon_building/woodc1_tdark02"},
	{"textures/woodc1_tdark03", "textures/makkon_building/woodc1_tdark03"},
	{"textures/woodc1_tprpl01", "textures/makkon_building/woodc1_tprpl01"},
	{"textures/woodc1_tprpl02", "textures/makkon_building/woodc1_tprpl02"},
	{"textures/woodc1_tprpl03", "textures/makkon_building/woodc1_tprpl03"},
	{"textures/woodc1_tred01", "textures/makkon_building/woodc1_tred01"},
	{"textures/woodc1_tred02", "textures/makkon_building/woodc1_tred02"},
	{"textures/woodc1_tred03", "textures/makkon_building/woodc1_tred03"},
	{"textures/woodc1_tteal01", "textures/makkon_building/woodc1_tteal01"},
	{"textures/woodc1_tteal02", "textures/makkon_building/woodc1_tteal02"},
	{"textures/woodc1_tteal03", "textures/makkon_building/woodc1_tteal03"},
	{"textures/woodc1_twht01", "textures/makkon_building/woodc1_twht01"},
	{"textures/woodc1_twht02", "textures/makkon_building/woodc1_twht02"},
	{"textures/woodc1_twht03", "textures/makkon_building/woodc1_twht03"},
	{"textures/woodc1_tylw01", "textures/makkon_building/woodc1_tylw01"},
	{"textures/woodc1_tylw02", "textures/makkon_building/woodc1_tylw02"},
	{"textures/woodc1_tylw03", "textures/makkon_building/woodc1_tylw03"},
	{"textures/woodc1_wbare01", "textures/makkon_building/woodc1_wbare01"},
	{"textures/woodc1_wbare02", "textures/makkon_building/woodc1_wbare02"},
	{"textures/woodc1_wbare05", "textures/makkon_building/woodc1_wbare05"},
	{"textures/woodc1_wbare06", "textures/makkon_building/woodc1_wbare06"},
	{"textures/woodc1_wbare07", "textures/makkon_building/woodc1_wbare07"},
	{"textures/woodc1_wbare08", "textures/makkon_building/woodc1_wbare08"},
	{"textures/woodc1_wblck01", "textures/makkon_building/woodc1_wblck01"},
	{"textures/woodc1_wblck02", "textures/makkon_building/woodc1_wblck02"},
	{"textures/woodc1_wblck03", "textures/makkon_building/woodc1_wblck03"},
	{"textures/woodc1_wblck04", "textures/makkon_building/woodc1_wblck04"},
	{"textures/woodc1_wblck05", "textures/makkon_building/woodc1_wblck05"},
	{"textures/woodc1_wblck06", "textures/makkon_building/woodc1_wblck06"},
	{"textures/woodc1_wblck07", "textures/makkon_building/woodc1_wblck07"},
	{"textures/woodc1_wblck08", "textures/makkon_building/woodc1_wblck08"},
	{"textures/woodc1_wblue01", "textures/makkon_building/woodc1_wblue01"},
	{"textures/woodc1_wblue02", "textures/makkon_building/woodc1_wblue02"},
	{"textures/woodc1_wblue03", "textures/makkon_building/woodc1_wblue03"},
	{"textures/woodc1_wblue04", "textures/makkon_building/woodc1_wblue04"},
	{"textures/woodc1_wblue05", "textures/makkon_building/woodc1_wblue05"},
	{"textures/woodc1_wblue06", "textures/makkon_building/woodc1_wblue06"},
	{"textures/woodc1_wblue07", "textures/makkon_building/woodc1_wblue07"},
	{"textures/woodc1_wblue08", "textures/makkon_building/woodc1_wblue08"},
	{"textures/woodc1_wdark01", "textures/makkon_building/woodc1_wdark01"},
	{"textures/woodc1_wdark02", "textures/makkon_building/woodc1_wdark02"},
	{"textures/woodc1_wdark05", "textures/makkon_building/woodc1_wdark05"},
	{"textures/woodc1_wdark06", "textures/makkon_building/woodc1_wdark06"},
	{"textures/woodc1_wdark07", "textures/makkon_building/woodc1_wdark07"},
	{"textures/woodc1_wdark08", "textures/makkon_building/woodc1_wdark08"},
	{"textures/woodc1_wprpl01", "textures/makkon_building/woodc1_wprpl01"},
	{"textures/woodc1_wprpl02", "textures/makkon_building/woodc1_wprpl02"},
	{"textures/woodc1_wprpl03", "textures/makkon_building/woodc1_wprpl03"},
	{"textures/woodc1_wprpl04", "textures/makkon_building/woodc1_wprpl04"},
	{"textures/woodc1_wprpl05", "textures/makkon_building/woodc1_wprpl05"},
	{"textures/woodc1_wprpl06", "textures/makkon_building/woodc1_wprpl06"},
	{"textures/woodc1_wprpl07", "textures/makkon_building/woodc1_wprpl07"},
	{"textures/woodc1_wprpl08", "textures/makkon_building/woodc1_wprpl08"},
	{"textures/woodc1_wred01", "textures/makkon_building/woodc1_wred01"},
	{"textures/woodc1_wred02", "textures/makkon_building/woodc1_wred02"},
	{"textures/woodc1_wred03", "textures/makkon_building/woodc1_wred03"},
	{"textures/woodc1_wred04", "textures/makkon_building/woodc1_wred04"},
	{"textures/woodc1_wred05", "textures/makkon_building/woodc1_wred05"},
	{"textures/woodc1_wred06", "textures/makkon_building/woodc1_wred06"},
	{"textures/woodc1_wred07", "textures/makkon_building/woodc1_wred07"},
	{"textures/woodc1_wred08", "textures/makkon_building/woodc1_wred08"},
	{"textures/woodc1_wtan03", "textures/makkon_building/woodc1_wtan03"},
	{"textures/woodc1_wtan04", "textures/makkon_building/woodc1_wtan04"},
	{"textures/woodc1_wteal01", "textures/makkon_building/woodc1_wteal01"},
	{"textures/woodc1_wteal02", "textures/makkon_building/woodc1_wteal02"},
	{"textures/woodc1_wteal03", "textures/makkon_building/woodc1_wteal03"},
	{"textures/woodc1_wteal04", "textures/makkon_building/woodc1_wteal04"},
	{"textures/woodc1_wteal05", "textures/makkon_building/woodc1_wteal05"},
	{"textures/woodc1_wteal06", "textures/makkon_building/woodc1_wteal06"},
	{"textures/woodc1_wteal07", "textures/makkon_building/woodc1_wteal07"},
	{"textures/woodc1_wteal08", "textures/makkon_building/woodc1_wteal08"},
	{"textures/woodc1_wwht01", "textures/makkon_building/woodc1_wwht01"},
	{"textures/woodc1_wwht02", "textures/makkon_building/woodc1_wwht02"},
	{"textures/woodc1_wwht03", "textures/makkon_building/woodc1_wwht03"},
	{"textures/woodc1_wwht04", "textures/makkon_building/woodc1_wwht04"},
	{"textures/woodc1_wwht05", "textures/makkon_building/woodc1_wwht05"},
	{"textures/woodc1_wwht06", "textures/makkon_building/woodc1_wwht06"},
	{"textures/woodc1_wwht07", "textures/makkon_building/woodc1_wwht07"},
	{"textures/woodc1_wwht08", "textures/makkon_building/woodc1_wwht08"},
	{"textures/woodc1_wylw01", "textures/makkon_building/woodc1_wylw01"},
	{"textures/woodc1_wylw02", "textures/makkon_building/woodc1_wylw02"},
	{"textures/woodc1_wylw03", "textures/makkon_building/woodc1_wylw03"},
	{"textures/woodc1_wylw04", "textures/makkon_building/woodc1_wylw04"},
	{"textures/woodc1_wylw05", "textures/makkon_building/woodc1_wylw05"},
	{"textures/woodc1_wylw06", "textures/makkon_building/woodc1_wylw06"},
	{"textures/woodc1_wylw07", "textures/makkon_building/woodc1_wylw07"},
	{"textures/woodc1_wylw08", "textures/makkon_building/woodc1_wylw08"},
	{"textures/woodc2_cbare01", "textures/makkon_building/woodc2_cbare01"},
	{"textures/woodc2_cbare02", "textures/makkon_building/woodc2_cbare02"},
	{"textures/woodc2_cbare03", "textures/makkon_building/woodc2_cbare03"},
	{"textures/woodc2_fbare01", "textures/makkon_building/woodc2_fbare01"},
	{"textures/woodc2_fbare02", "textures/makkon_building/woodc2_fbare02"},
	{"textures/woodc2_fbare03", "textures/makkon_building/woodc2_fbare03"},
	{"textures/woodc2_fbare04", "textures/makkon_building/woodc2_fbare04"},
	{"textures/woodc2_fbare05", "textures/makkon_building/woodc2_fbare05"},
	{"textures/woodc2_pbare01", "textures/makkon_building/woodc2_pbare01"},
	{"textures/woodc2_pbare02", "textures/makkon_building/woodc2_pbare02"},
	{"textures/woodc2_pblck01", "textures/makkon_building/woodc2_pblck01"},
	{"textures/woodc2_pblue01", "textures/makkon_building/woodc2_pblue01"},
	{"textures/woodc2_pdark01", "textures/makkon_building/woodc2_pdark01"},
	{"textures/woodc2_pprpl01", "textures/makkon_building/woodc2_pprpl01"},
	{"textures/woodc2_pred01", "textures/makkon_building/woodc2_pred01"},
	{"textures/woodc2_pteal01", "textures/makkon_building/woodc2_pteal01"},
	{"textures/woodc2_pwht01", "textures/makkon_building/woodc2_pwht01"},
	{"textures/woodc2_pylw01", "textures/makkon_building/woodc2_pylw01"},
	{"textures/woodc2_tbare01", "textures/makkon_building/woodc2_tbare01"},
	{"textures/woodc2_tbare02", "textures/makkon_building/woodc2_tbare02"},
	{"textures/woodc2_tbare03", "textures/makkon_building/woodc2_tbare03"},
	{"textures/woodc2_wbare01", "textures/makkon_building/woodc2_wbare01"},
	{"textures/woodc2_wbare05", "textures/makkon_building/woodc2_wbare05"},
	{"textures/woodc2_wbare06", "textures/makkon_building/woodc2_wbare06"},
	{"textures/woodc2_wbare07", "textures/makkon_building/woodc2_wbare07"},
	{"textures/woodc2_wbare08", "textures/makkon_building/woodc2_wbare08"},
	{"textures/woodc2_wblck01", "textures/makkon_building/woodc2_wblck01"},
	{"textures/woodc2_wblck03", "textures/makkon_building/woodc2_wblck03"},
	{"textures/woodc2_wblck04", "textures/makkon_building/woodc2_wblck04"},
	{"textures/woodc2_wblue004", "textures/makkon_building/woodc2_wblue004"},
	{"textures/woodc2_wblue01", "textures/makkon_building/woodc2_wblue01"},
	{"textures/woodc2_wblue03", "textures/makkon_building/woodc2_wblue03"},
	{"textures/woodc2_wdark01", "textures/makkon_building/woodc2_wdark01"},
	{"textures/woodc2_wprpl01", "textures/makkon_building/woodc2_wprpl01"},
	{"textures/woodc2_wprpl03", "textures/makkon_building/woodc2_wprpl03"},
	{"textures/woodc2_wprpl04", "textures/makkon_building/woodc2_wprpl04"},
	{"textures/woodc2_wred01", "textures/makkon_building/woodc2_wred01"},
	{"textures/woodc2_wred03", "textures/makkon_building/woodc2_wred03"},
	{"textures/woodc2_wred04", "textures/makkon_building/woodc2_wred04"},
	{"textures/woodc2_wtan03", "textures/makkon_building/woodc2_wtan03"},
	{"textures/woodc2_wtan04", "textures/makkon_building/woodc2_wtan04"},
	{"textures/woodc2_wteal01", "textures/makkon_building/woodc2_wteal01"},
	{"textures/woodc2_wteal03", "textures/makkon_building/woodc2_wteal03"},
	{"textures/woodc2_wteal04", "textures/makkon_building/woodc2_wteal04"},
	{"textures/woodc2_wwht01", "textures/makkon_building/woodc2_wwht01"},
	{"textures/woodc2_wwht03", "textures/makkon_building/woodc2_wwht03"},
	{"textures/woodc2_wwht04", "textures/makkon_building/woodc2_wwht04"},
	{"textures/woodc2_wylw01", "textures/makkon_building/woodc2_wylw01"},
	{"textures/woodc2_wylw03", "textures/makkon_building/woodc2_wylw03"},
	{"textures/woodc2_wylw04", "textures/makkon_building/woodc2_wylw04"},
	{"textures/woodc3_cbare01", "textures/makkon_building/woodc3_cbare01"},
	{"textures/woodc3_cbare02", "textures/makkon_building/woodc3_cbare02"},
	{"textures/woodc3_cbare03", "textures/makkon_building/woodc3_cbare03"},
	{"textures/woodc3_fbare01", "textures/makkon_building/woodc3_fbare01"},
	{"textures/woodc3_fbare02", "textures/makkon_building/woodc3_fbare02"},
	{"textures/woodc3_fbare03", "textures/makkon_building/woodc3_fbare03"},
	{"textures/woodc3_fbare04", "textures/makkon_building/woodc3_fbare04"},
	{"textures/woodc3_fbare05", "textures/makkon_building/woodc3_fbare05"},
	{"textures/woodc3_pbare01", "textures/makkon_building/woodc3_pbare01"},
	{"textures/woodc3_pbare02", "textures/makkon_building/woodc3_pbare02"},
	{"textures/woodc3_pblck01", "textures/makkon_building/woodc3_pblck01"},
	{"textures/woodc3_pblue01", "textures/makkon_building/woodc3_pblue01"},
	{"textures/woodc3_pdark01", "textures/makkon_building/woodc3_pdark01"},
	{"textures/woodc3_pprpl01", "textures/makkon_building/woodc3_pprpl01"},
	{"textures/woodc3_pred01", "textures/makkon_building/woodc3_pred01"},
	{"textures/woodc3_pteal01", "textures/makkon_building/woodc3_pteal01"},
	{"textures/woodc3_pwht01", "textures/makkon_building/woodc3_pwht01"},
	{"textures/woodc3_pylw01", "textures/makkon_building/woodc3_pylw01"},
	{"textures/woodc3_tbare01", "textures/makkon_building/woodc3_tbare01"},
	{"textures/woodc3_tbare02", "textures/makkon_building/woodc3_tbare02"},
	{"textures/woodc3_tbare03", "textures/makkon_building/woodc3_tbare03"},
	{"textures/woodc3_wbare01", "textures/makkon_building/woodc3_wbare01"},
	{"textures/woodc3_wbare05", "textures/makkon_building/woodc3_wbare05"},
	{"textures/woodc3_wbare06", "textures/makkon_building/woodc3_wbare06"},
	{"textures/woodc3_wbare07", "textures/makkon_building/woodc3_wbare07"},
	{"textures/woodc3_wbare08", "textures/makkon_building/woodc3_wbare08"},
	{"textures/woodc3_wblck01", "textures/makkon_building/woodc3_wblck01"},
	{"textures/woodc3_wblck03", "textures/makkon_building/woodc3_wblck03"},
	{"textures/woodc3_wblck04", "textures/makkon_building/woodc3_wblck04"},
	{"textures/woodc3_wblue01", "textures/makkon_building/woodc3_wblue01"},
	{"textures/woodc3_wblue03", "textures/makkon_building/woodc3_wblue03"},
	{"textures/woodc3_wblue04", "textures/makkon_building/woodc3_wblue04"},
	{"textures/woodc3_wdark01", "textures/makkon_building/woodc3_wdark01"},
	{"textures/woodc3_wprpl01", "textures/makkon_building/woodc3_wprpl01"},
	{"textures/woodc3_wprpl03", "textures/makkon_building/woodc3_wprpl03"},
	{"textures/woodc3_wprpl04", "textures/makkon_building/woodc3_wprpl04"},
	{"textures/woodc3_wreal03", "textures/makkon_building/woodc3_wreal03"},
	{"textures/woodc3_wreal04", "textures/makkon_building/woodc3_wreal04"},
	{"textures/woodc3_wred01", "textures/makkon_building/woodc3_wred01"},
	{"textures/woodc3_wred03", "textures/makkon_building/woodc3_wred03"},
	{"textures/woodc3_wred04", "textures/makkon_building/woodc3_wred04"},
	{"textures/woodc3_wtan03", "textures/makkon_building/woodc3_wtan03"},
	{"textures/woodc3_wtan04", "textures/makkon_building/woodc3_wtan04"},
	{"textures/woodc3_wteal01", "textures/makkon_building/woodc3_wteal01"},
	{"textures/woodc3_wwht01", "textures/makkon_building/woodc3_wwht01"},
	{"textures/woodc3_wwht03", "textures/makkon_building/woodc3_wwht03"},
	{"textures/woodc3_wwht04", "textures/makkon_building/woodc3_wwht04"},
	{"textures/woodc3_wylw01", "textures/makkon_building/woodc3_wylw01"},
	{"textures/_win_base1_01", "textures/makkon_building/_win_base1_01"},
	{"textures/_win_base1_03", "textures/makkon_building/_win_base1_03"},
	{"textures/_win_base1_05", "textures/makkon_building/_win_base1_05"},
	{"textures/_win_base1_07", "textures/makkon_building/_win_base1_07"},
	{"textures/_win_base1_09", "textures/makkon_building/_win_base1_09"},
	{"textures/_woodc1_wbare07", "textures/makkon_building/_woodc1_wbare07"},
	{"textures/_woodc1_wbare08", "textures/makkon_building/_woodc1_wbare08"},
	{"textures/_woodc1_wblck07", "textures/makkon_building/_woodc1_wblck07"},
	{"textures/_woodc1_wblck08", "textures/makkon_building/_woodc1_wblck08"},
	{"textures/_woodc1_wblue07", "textures/makkon_building/_woodc1_wblue07"},
	{"textures/_woodc1_wblue08", "textures/makkon_building/_woodc1_wblue08"},
	{"textures/_woodc1_wdark07", "textures/makkon_building/_woodc1_wdark07"},
	{"textures/_woodc1_wdark08", "textures/makkon_building/_woodc1_wdark08"},
	{"textures/_woodc1_wprpl07", "textures/makkon_building/_woodc1_wprpl07"},
	{"textures/_woodc1_wprpl08", "textures/makkon_building/_woodc1_wprpl08"},
	{"textures/_woodc1_wred07", "textures/makkon_building/_woodc1_wred07"},
	{"textures/_woodc1_wred08", "textures/makkon_building/_woodc1_wred08"},
	{"textures/_woodc1_wteal07", "textures/makkon_building/_woodc1_wteal07"},
	{"textures/_woodc1_wteal08", "textures/makkon_building/_woodc1_wteal08"},
	{"textures/_woodc1_wwht07", "textures/makkon_building/_woodc1_wwht07"},
	{"textures/_woodc1_wwht08", "textures/makkon_building/_woodc1_wwht08"},
	{"textures/_woodc1_wylw07", "textures/makkon_building/_woodc1_wylw07"},
	{"textures/_woodc1_wylw08", "textures/makkon_building/_woodc1_wylw08"},
	{"textures/_woodc2_wbare07", "textures/makkon_building/_woodc2_wbare07"},
	{"textures/_woodc2_wbare08", "textures/makkon_building/_woodc2_wbare08"},
	{"textures/_woodc3_wbare07", "textures/makkon_building/_woodc3_wbare07"},
	{"textures/_woodc3_wbare08", "textures/makkon_building/_woodc3_wbare08"},

	// Makkon concrete 2022.8
	{"textures/conc_b01_blk", "textures/makkon_concrete/conc_b01_blk"},
	{"textures/conc_b01_blu", "textures/makkon_concrete/conc_b01_blu"},
	{"textures/conc_b01_brwn", "textures/makkon_concrete/conc_b01_brwn"},
	{"textures/conc_b01_grn", "textures/makkon_concrete/conc_b01_grn"},
	{"textures/conc_b01_gry1", "textures/makkon_concrete/conc_b01_gry1"},
	{"textures/conc_b01_gry2", "textures/makkon_concrete/conc_b01_gry2"},
	{"textures/conc_b01_gry3", "textures/makkon_concrete/conc_b01_gry3"},
	{"textures/conc_b01_light", "textures/makkon_concrete/conc_b01_light"},
	{"textures/conc_b01_pink", "textures/makkon_concrete/conc_b01_pink"},
	{"textures/conc_b01_red", "textures/makkon_concrete/conc_b01_red"},
	{"textures/conc_b01_red2", "textures/makkon_concrete/conc_b01_red2"},
	{"textures/conc_b01_teal", "textures/makkon_concrete/conc_b01_teal"},
	{"textures/conc_b01_wht", "textures/makkon_concrete/conc_b01_wht"},
	{"textures/conc_b01_ylw", "textures/makkon_concrete/conc_b01_ylw"},
	{"textures/conc_b02_blk", "textures/makkon_concrete/conc_b02_blk"},
	{"textures/conc_b02_blu", "textures/makkon_concrete/conc_b02_blu"},
	{"textures/conc_b02_brwn", "textures/makkon_concrete/conc_b02_brwn"},
	{"textures/conc_b02_grn", "textures/makkon_concrete/conc_b02_grn"},
	{"textures/conc_b02_gry1", "textures/makkon_concrete/conc_b02_gry1"},
	{"textures/conc_b02_gry2", "textures/makkon_concrete/conc_b02_gry2"},
	{"textures/conc_b02_gry3", "textures/makkon_concrete/conc_b02_gry3"},
	{"textures/conc_b02_light", "textures/makkon_concrete/conc_b02_light"},
	{"textures/conc_b02_pink", "textures/makkon_concrete/conc_b02_pink"},
	{"textures/conc_b02_red", "textures/makkon_concrete/conc_b02_red"},
	{"textures/conc_b02_red2", "textures/makkon_concrete/conc_b02_red2"},
	{"textures/conc_b02_teal", "textures/makkon_concrete/conc_b02_teal"},
	{"textures/conc_b02_wht", "textures/makkon_concrete/conc_b02_wht"},
	{"textures/conc_b02_ylw", "textures/makkon_concrete/conc_b02_ylw"},
	{"textures/conc_b03_blk", "textures/makkon_concrete/conc_b03_blk"},
	{"textures/conc_b03_blu", "textures/makkon_concrete/conc_b03_blu"},
	{"textures/conc_b03_brwn", "textures/makkon_concrete/conc_b03_brwn"},
	{"textures/conc_b03_grn", "textures/makkon_concrete/conc_b03_grn"},
	{"textures/conc_b03_gry1", "textures/makkon_concrete/conc_b03_gry1"},
	{"textures/conc_b03_gry2", "textures/makkon_concrete/conc_b03_gry2"},
	{"textures/conc_b03_gry3", "textures/makkon_concrete/conc_b03_gry3"},
	{"textures/conc_b03_light", "textures/makkon_concrete/conc_b03_light"},
	{"textures/conc_b03_pink", "textures/makkon_concrete/conc_b03_pink"},
	{"textures/conc_b03_red", "textures/makkon_concrete/conc_b03_red"},
	{"textures/conc_b03_red2", "textures/makkon_concrete/conc_b03_red2"},
	{"textures/conc_b03_teal", "textures/makkon_concrete/conc_b03_teal"},
	{"textures/conc_b03_wht", "textures/makkon_concrete/conc_b03_wht"},
	{"textures/conc_b03_ylw", "textures/makkon_concrete/conc_b03_ylw"},
	{"textures/conc_b04_blk", "textures/makkon_concrete/conc_b04_blk"},
	{"textures/conc_b04_blu", "textures/makkon_concrete/conc_b04_blu"},
	{"textures/conc_b04_brwn", "textures/makkon_concrete/conc_b04_brwn"},
	{"textures/conc_b04_grn", "textures/makkon_concrete/conc_b04_grn"},
	{"textures/conc_b04_gry1", "textures/makkon_concrete/conc_b04_gry1"},
	{"textures/conc_b04_gry2", "textures/makkon_concrete/conc_b04_gry2"},
	{"textures/conc_b04_gry3", "textures/makkon_concrete/conc_b04_gry3"},
	{"textures/conc_b04_light", "textures/makkon_concrete/conc_b04_light"},
	{"textures/conc_b04_pink", "textures/makkon_concrete/conc_b04_pink"},
	{"textures/conc_b04_red", "textures/makkon_concrete/conc_b04_red"},
	{"textures/conc_b04_red2", "textures/makkon_concrete/conc_b04_red2"},
	{"textures/conc_b04_teal", "textures/makkon_concrete/conc_b04_teal"},
	{"textures/conc_b04_wht", "textures/makkon_concrete/conc_b04_wht"},
	{"textures/conc_b04_ylw", "textures/makkon_concrete/conc_b04_ylw"},
	{"textures/conc_b05_blk", "textures/makkon_concrete/conc_b05_blk"},
	{"textures/conc_b05_blu", "textures/makkon_concrete/conc_b05_blu"},
	{"textures/conc_b05_brwn", "textures/makkon_concrete/conc_b05_brwn"},
	{"textures/conc_b05_grn", "textures/makkon_concrete/conc_b05_grn"},
	{"textures/conc_b05_gry1", "textures/makkon_concrete/conc_b05_gry1"},
	{"textures/conc_b05_gry2", "textures/makkon_concrete/conc_b05_gry2"},
	{"textures/conc_b05_gry3", "textures/makkon_concrete/conc_b05_gry3"},
	{"textures/conc_b05_light", "textures/makkon_concrete/conc_b05_light"},
	{"textures/conc_b05_pink", "textures/makkon_concrete/conc_b05_pink"},
	{"textures/conc_b05_red", "textures/makkon_concrete/conc_b05_red"},
	{"textures/conc_b05_red2", "textures/makkon_concrete/conc_b05_red2"},
	{"textures/conc_b05_teal", "textures/makkon_concrete/conc_b05_teal"},
	{"textures/conc_b05_wht", "textures/makkon_concrete/conc_b05_wht"},
	{"textures/conc_b05_ylw", "textures/makkon_concrete/conc_b05_ylw"},
	{"textures/conc_b06_blk", "textures/makkon_concrete/conc_b06_blk"},
	{"textures/conc_b06_blu", "textures/makkon_concrete/conc_b06_blu"},
	{"textures/conc_b06_brwn", "textures/makkon_concrete/conc_b06_brwn"},
	{"textures/conc_b06_grn", "textures/makkon_concrete/conc_b06_grn"},
	{"textures/conc_b06_gry1", "textures/makkon_concrete/conc_b06_gry1"},
	{"textures/conc_b06_gry2", "textures/makkon_concrete/conc_b06_gry2"},
	{"textures/conc_b06_gry3", "textures/makkon_concrete/conc_b06_gry3"},
	{"textures/conc_b06_light", "textures/makkon_concrete/conc_b06_light"},
	{"textures/conc_b06_pink", "textures/makkon_concrete/conc_b06_pink"},
	{"textures/conc_b06_red", "textures/makkon_concrete/conc_b06_red"},
	{"textures/conc_b06_red2", "textures/makkon_concrete/conc_b06_red2"},
	{"textures/conc_b06_teal", "textures/makkon_concrete/conc_b06_teal"},
	{"textures/conc_b06_wht", "textures/makkon_concrete/conc_b06_wht"},
	{"textures/conc_b06_ylw", "textures/makkon_concrete/conc_b06_ylw"},
	{"textures/conc_c01_blk", "textures/makkon_concrete/conc_c01_blk"},
	{"textures/conc_c01_blu", "textures/makkon_concrete/conc_c01_blu"},
	{"textures/conc_c01_brwn", "textures/makkon_concrete/conc_c01_brwn"},
	{"textures/conc_c01_grn", "textures/makkon_concrete/conc_c01_grn"},
	{"textures/conc_c01_gry1", "textures/makkon_concrete/conc_c01_gry1"},
	{"textures/conc_c01_gry2", "textures/makkon_concrete/conc_c01_gry2"},
	{"textures/conc_c01_gry3", "textures/makkon_concrete/conc_c01_gry3"},
	{"textures/conc_c01_pink", "textures/makkon_concrete/conc_c01_pink"},
	{"textures/conc_c01_red", "textures/makkon_concrete/conc_c01_red"},
	{"textures/conc_c01_red2", "textures/makkon_concrete/conc_c01_red2"},
	{"textures/conc_c01_teal", "textures/makkon_concrete/conc_c01_teal"},
	{"textures/conc_c01_wht", "textures/makkon_concrete/conc_c01_wht"},
	{"textures/conc_c01_ylw", "textures/makkon_concrete/conc_c01_ylw"},
	{"textures/conc_c02_blk", "textures/makkon_concrete/conc_c02_blk"},
	{"textures/conc_c02_blu", "textures/makkon_concrete/conc_c02_blu"},
	{"textures/conc_c02_brwn", "textures/makkon_concrete/conc_c02_brwn"},
	{"textures/conc_c02_grn", "textures/makkon_concrete/conc_c02_grn"},
	{"textures/conc_c02_gry1", "textures/makkon_concrete/conc_c02_gry1"},
	{"textures/conc_c02_gry2", "textures/makkon_concrete/conc_c02_gry2"},
	{"textures/conc_c02_gry3", "textures/makkon_concrete/conc_c02_gry3"},
	{"textures/conc_c02_pink", "textures/makkon_concrete/conc_c02_pink"},
	{"textures/conc_c02_red", "textures/makkon_concrete/conc_c02_red"},
	{"textures/conc_c02_red2", "textures/makkon_concrete/conc_c02_red2"},
	{"textures/conc_c02_teal", "textures/makkon_concrete/conc_c02_teal"},
	{"textures/conc_c02_wht", "textures/makkon_concrete/conc_c02_wht"},
	{"textures/conc_c02_ylw", "textures/makkon_concrete/conc_c02_ylw"},
	{"textures/conc_c03_blk", "textures/makkon_concrete/conc_c03_blk"},
	{"textures/conc_c03_blu", "textures/makkon_concrete/conc_c03_blu"},
	{"textures/conc_c03_brwn", "textures/makkon_concrete/conc_c03_brwn"},
	{"textures/conc_c03_grn", "textures/makkon_concrete/conc_c03_grn"},
	{"textures/conc_c03_gry1", "textures/makkon_concrete/conc_c03_gry1"},
	{"textures/conc_c03_gry2", "textures/makkon_concrete/conc_c03_gry2"},
	{"textures/conc_c03_gry3", "textures/makkon_concrete/conc_c03_gry3"},
	{"textures/conc_c03_pink", "textures/makkon_concrete/conc_c03_pink"},
	{"textures/conc_c03_red", "textures/makkon_concrete/conc_c03_red"},
	{"textures/conc_c03_red2", "textures/makkon_concrete/conc_c03_red2"},
	{"textures/conc_c03_teal", "textures/makkon_concrete/conc_c03_teal"},
	{"textures/conc_c03_wht", "textures/makkon_concrete/conc_c03_wht"},
	{"textures/conc_c03_ylw", "textures/makkon_concrete/conc_c03_ylw"},
	{"textures/conc_c_04_brwn", "textures/makkon_concrete/conc_c_04_brwn"},
	{"textures/conc_c_04_gre1", "textures/makkon_concrete/conc_c_04_gre1"},
	{"textures/conc_c_04_gry1", "textures/makkon_concrete/conc_c_04_gry1"},
	{"textures/conc_c_04_gry2", "textures/makkon_concrete/conc_c_04_gry2"},
	{"textures/conc_c_04_gry3", "textures/makkon_concrete/conc_c_04_gry3"},
	{"textures/conc_f01_blk", "textures/makkon_concrete/conc_f01_blk"},
	{"textures/conc_f01_blu", "textures/makkon_concrete/conc_f01_blu"},
	{"textures/conc_f01_brwn", "textures/makkon_concrete/conc_f01_brwn"},
	{"textures/conc_f01_grn", "textures/makkon_concrete/conc_f01_grn"},
	{"textures/conc_f01_gry1", "textures/makkon_concrete/conc_f01_gry1"},
	{"textures/conc_f01_gry2", "textures/makkon_concrete/conc_f01_gry2"},
	{"textures/conc_f01_gry3", "textures/makkon_concrete/conc_f01_gry3"},
	{"textures/conc_f01_pink", "textures/makkon_concrete/conc_f01_pink"},
	{"textures/conc_f01_red", "textures/makkon_concrete/conc_f01_red"},
	{"textures/conc_f01_red2", "textures/makkon_concrete/conc_f01_red2"},
	{"textures/conc_f01_teal", "textures/makkon_concrete/conc_f01_teal"},
	{"textures/conc_f01_wht", "textures/makkon_concrete/conc_f01_wht"},
	{"textures/conc_f01_ylw", "textures/makkon_concrete/conc_f01_ylw"},
	{"textures/conc_f02_blk", "textures/makkon_concrete/conc_f02_blk"},
	{"textures/conc_f02_blu", "textures/makkon_concrete/conc_f02_blu"},
	{"textures/conc_f02_brwn", "textures/makkon_concrete/conc_f02_brwn"},
	{"textures/conc_f02_grn", "textures/makkon_concrete/conc_f02_grn"},
	{"textures/conc_f02_gry1", "textures/makkon_concrete/conc_f02_gry1"},
	{"textures/conc_f02_gry2", "textures/makkon_concrete/conc_f02_gry2"},
	{"textures/conc_f02_gry3", "textures/makkon_concrete/conc_f02_gry3"},
	{"textures/conc_f02_pink", "textures/makkon_concrete/conc_f02_pink"},
	{"textures/conc_f02_red", "textures/makkon_concrete/conc_f02_red"},
	{"textures/conc_f02_red2", "textures/makkon_concrete/conc_f02_red2"},
	{"textures/conc_f02_teal", "textures/makkon_concrete/conc_f02_teal"},
	{"textures/conc_f02_wht", "textures/makkon_concrete/conc_f02_wht"},
	{"textures/conc_f02_ylw", "textures/makkon_concrete/conc_f02_ylw"},
	{"textures/conc_f03_blk", "textures/makkon_concrete/conc_f03_blk"},
	{"textures/conc_f03_blu", "textures/makkon_concrete/conc_f03_blu"},
	{"textures/conc_f03_brwn", "textures/makkon_concrete/conc_f03_brwn"},
	{"textures/conc_f03_grn", "textures/makkon_concrete/conc_f03_grn"},
	{"textures/conc_f03_gry1", "textures/makkon_concrete/conc_f03_gry1"},
	{"textures/conc_f03_gry2", "textures/makkon_concrete/conc_f03_gry2"},
	{"textures/conc_f03_gry3", "textures/makkon_concrete/conc_f03_gry3"},
	{"textures/conc_f03_pink", "textures/makkon_concrete/conc_f03_pink"},
	{"textures/conc_f03_red", "textures/makkon_concrete/conc_f03_red"},
	{"textures/conc_f03_red2", "textures/makkon_concrete/conc_f03_red2"},
	{"textures/conc_f03_teal", "textures/makkon_concrete/conc_f03_teal"},
	{"textures/conc_f03_wht", "textures/makkon_concrete/conc_f03_wht"},
	{"textures/conc_f03_ylw", "textures/makkon_concrete/conc_f03_ylw"},
	{"textures/conc_f04_blk", "textures/makkon_concrete/conc_f04_blk"},
	{"textures/conc_f04_blu", "textures/makkon_concrete/conc_f04_blu"},
	{"textures/conc_f04_brwn", "textures/makkon_concrete/conc_f04_brwn"},
	{"textures/conc_f04_grn", "textures/makkon_concrete/conc_f04_grn"},
	{"textures/conc_f04_gry1", "textures/makkon_concrete/conc_f04_gry1"},
	{"textures/conc_f04_gry2", "textures/makkon_concrete/conc_f04_gry2"},
	{"textures/conc_f04_gry3", "textures/makkon_concrete/conc_f04_gry3"},
	{"textures/conc_f04_pink", "textures/makkon_concrete/conc_f04_pink"},
	{"textures/conc_f04_red", "textures/makkon_concrete/conc_f04_red"},
	{"textures/conc_f04_red2", "textures/makkon_concrete/conc_f04_red2"},
	{"textures/conc_f04_teal", "textures/makkon_concrete/conc_f04_teal"},
	{"textures/conc_f04_wht", "textures/makkon_concrete/conc_f04_wht"},
	{"textures/conc_f04_ylw", "textures/makkon_concrete/conc_f04_ylw"},
	{"textures/conc_f05_blk", "textures/makkon_concrete/conc_f05_blk"},
	{"textures/conc_f05_blu", "textures/makkon_concrete/conc_f05_blu"},
	{"textures/conc_f05_brwn", "textures/makkon_concrete/conc_f05_brwn"},
	{"textures/conc_f05_grn", "textures/makkon_concrete/conc_f05_grn"},
	{"textures/conc_f05_gry1", "textures/makkon_concrete/conc_f05_gry1"},
	{"textures/conc_f05_gry2", "textures/makkon_concrete/conc_f05_gry2"},
	{"textures/conc_f05_gry3", "textures/makkon_concrete/conc_f05_gry3"},
	{"textures/conc_f05_pink", "textures/makkon_concrete/conc_f05_pink"},
	{"textures/conc_f05_red", "textures/makkon_concrete/conc_f05_red"},
	{"textures/conc_f05_red2", "textures/makkon_concrete/conc_f05_red2"},
	{"textures/conc_f05_teal", "textures/makkon_concrete/conc_f05_teal"},
	{"textures/conc_f05_wht", "textures/makkon_concrete/conc_f05_wht"},
	{"textures/conc_f05_ylw", "textures/makkon_concrete/conc_f05_ylw"},
	{"textures/conc_t01_blk", "textures/makkon_concrete/conc_t01_blk"},
	{"textures/conc_t01_blu", "textures/makkon_concrete/conc_t01_blu"},
	{"textures/conc_t01_brwn", "textures/makkon_concrete/conc_t01_brwn"},
	{"textures/conc_t01_grn", "textures/makkon_concrete/conc_t01_grn"},
	{"textures/conc_t01_gry1", "textures/makkon_concrete/conc_t01_gry1"},
	{"textures/conc_t01_gry2", "textures/makkon_concrete/conc_t01_gry2"},
	{"textures/conc_t01_gry3", "textures/makkon_concrete/conc_t01_gry3"},
	{"textures/conc_t01_pink", "textures/makkon_concrete/conc_t01_pink"},
	{"textures/conc_t01_red", "textures/makkon_concrete/conc_t01_red"},
	{"textures/conc_t01_red2", "textures/makkon_concrete/conc_t01_red2"},
	{"textures/conc_t01_teal", "textures/makkon_concrete/conc_t01_teal"},
	{"textures/conc_t01_wht", "textures/makkon_concrete/conc_t01_wht"},
	{"textures/conc_t01_ylw", "textures/makkon_concrete/conc_t01_ylw"},
	{"textures/conc_t02_blk", "textures/makkon_concrete/conc_t02_blk"},
	{"textures/conc_t02_blu", "textures/makkon_concrete/conc_t02_blu"},
	{"textures/conc_t02_brwn", "textures/makkon_concrete/conc_t02_brwn"},
	{"textures/conc_t02_grn", "textures/makkon_concrete/conc_t02_grn"},
	{"textures/conc_t02_gry1", "textures/makkon_concrete/conc_t02_gry1"},
	{"textures/conc_t02_gry2", "textures/makkon_concrete/conc_t02_gry2"},
	{"textures/conc_t02_gry3", "textures/makkon_concrete/conc_t02_gry3"},
	{"textures/conc_t02_pink", "textures/makkon_concrete/conc_t02_pink"},
	{"textures/conc_t02_red", "textures/makkon_concrete/conc_t02_red"},
	{"textures/conc_t02_red2", "textures/makkon_concrete/conc_t02_red2"},
	{"textures/conc_t02_teal", "textures/makkon_concrete/conc_t02_teal"},
	{"textures/conc_t02_wht", "textures/makkon_concrete/conc_t02_wht"},
	{"textures/conc_t02_ylw", "textures/makkon_concrete/conc_t02_ylw"},
	{"textures/conc_t03_blk", "textures/makkon_concrete/conc_t03_blk"},
	{"textures/conc_t03_blu", "textures/makkon_concrete/conc_t03_blu"},
	{"textures/conc_t03_brwn", "textures/makkon_concrete/conc_t03_brwn"},
	{"textures/conc_t03_grn", "textures/makkon_concrete/conc_t03_grn"},
	{"textures/conc_t03_gry1", "textures/makkon_concrete/conc_t03_gry1"},
	{"textures/conc_t03_gry2", "textures/makkon_concrete/conc_t03_gry2"},
	{"textures/conc_t03_gry3", "textures/makkon_concrete/conc_t03_gry3"},
	{"textures/conc_t03_pink", "textures/makkon_concrete/conc_t03_pink"},
	{"textures/conc_t03_red", "textures/makkon_concrete/conc_t03_red"},
	{"textures/conc_t03_red2", "textures/makkon_concrete/conc_t03_red2"},
	{"textures/conc_t03_teal", "textures/makkon_concrete/conc_t03_teal"},
	{"textures/conc_t03_wht", "textures/makkon_concrete/conc_t03_wht"},
	{"textures/conc_t03_ylw", "textures/makkon_concrete/conc_t03_ylw"},
	{"textures/conc_t04_blk", "textures/makkon_concrete/conc_t04_blk"},
	{"textures/conc_t04_blu", "textures/makkon_concrete/conc_t04_blu"},
	{"textures/conc_t04_brwn", "textures/makkon_concrete/conc_t04_brwn"},
	{"textures/conc_t04_grn", "textures/makkon_concrete/conc_t04_grn"},
	{"textures/conc_t04_gry1", "textures/makkon_concrete/conc_t04_gry1"},
	{"textures/conc_t04_gry2", "textures/makkon_concrete/conc_t04_gry2"},
	{"textures/conc_t04_gry3", "textures/makkon_concrete/conc_t04_gry3"},
	{"textures/conc_t04_pink", "textures/makkon_concrete/conc_t04_pink"},
	{"textures/conc_t04_red", "textures/makkon_concrete/conc_t04_red"},
	{"textures/conc_t04_red2", "textures/makkon_concrete/conc_t04_red2"},
	{"textures/conc_t04_teal", "textures/makkon_concrete/conc_t04_teal"},
	{"textures/conc_t04_wht", "textures/makkon_concrete/conc_t04_wht"},
	{"textures/conc_t04_ylw", "textures/makkon_concrete/conc_t04_ylw"},
	{"textures/conc_t06_blk", "textures/makkon_concrete/conc_t06_blk"},
	{"textures/conc_t06_blu", "textures/makkon_concrete/conc_t06_blu"},
	{"textures/conc_t06_brwn", "textures/makkon_concrete/conc_t06_brwn"},
	{"textures/conc_t06_grn", "textures/makkon_concrete/conc_t06_grn"},
	{"textures/conc_t06_gry1", "textures/makkon_concrete/conc_t06_gry1"},
	{"textures/conc_t06_gry2", "textures/makkon_concrete/conc_t06_gry2"},
	{"textures/conc_t06_gry3", "textures/makkon_concrete/conc_t06_gry3"},
	{"textures/conc_t06_pink", "textures/makkon_concrete/conc_t06_pink"},
	{"textures/conc_t06_red", "textures/makkon_concrete/conc_t06_red"},
	{"textures/conc_t06_red2", "textures/makkon_concrete/conc_t06_red2"},
	{"textures/conc_t06_teal", "textures/makkon_concrete/conc_t06_teal"},
	{"textures/conc_t06_wht", "textures/makkon_concrete/conc_t06_wht"},
	{"textures/conc_t06_ylw", "textures/makkon_concrete/conc_t06_ylw"},
	{"textures/conc_t07_blk", "textures/makkon_concrete/conc_t07_blk"},
	{"textures/conc_t07_blu", "textures/makkon_concrete/conc_t07_blu"},
	{"textures/conc_t07_brwn", "textures/makkon_concrete/conc_t07_brwn"},
	{"textures/conc_t07_grn", "textures/makkon_concrete/conc_t07_grn"},
	{"textures/conc_t07_gry1", "textures/makkon_concrete/conc_t07_gry1"},
	{"textures/conc_t07_gry2", "textures/makkon_concrete/conc_t07_gry2"},
	{"textures/conc_t07_gry3", "textures/makkon_concrete/conc_t07_gry3"},
	{"textures/conc_t07_pink", "textures/makkon_concrete/conc_t07_pink"},
	{"textures/conc_t07_red", "textures/makkon_concrete/conc_t07_red"},
	{"textures/conc_t07_red2", "textures/makkon_concrete/conc_t07_red2"},
	{"textures/conc_t07_teal", "textures/makkon_concrete/conc_t07_teal"},
	{"textures/conc_t07_wht", "textures/makkon_concrete/conc_t07_wht"},
	{"textures/conc_t07_ylw", "textures/makkon_concrete/conc_t07_ylw"},
	{"textures/conc_t08_blk", "textures/makkon_concrete/conc_t08_blk"},
	{"textures/conc_t08_blu", "textures/makkon_concrete/conc_t08_blu"},
	{"textures/conc_t08_brwn", "textures/makkon_concrete/conc_t08_brwn"},
	{"textures/conc_t08_grn", "textures/makkon_concrete/conc_t08_grn"},
	{"textures/conc_t08_gry1", "textures/makkon_concrete/conc_t08_gry1"},
	{"textures/conc_t08_gry2", "textures/makkon_concrete/conc_t08_gry2"},
	{"textures/conc_t08_gry3", "textures/makkon_concrete/conc_t08_gry3"},
	{"textures/conc_t08_pink", "textures/makkon_concrete/conc_t08_pink"},
	{"textures/conc_t08_red", "textures/makkon_concrete/conc_t08_red"},
	{"textures/conc_t08_red2", "textures/makkon_concrete/conc_t08_red2"},
	{"textures/conc_t08_teal", "textures/makkon_concrete/conc_t08_teal"},
	{"textures/conc_t08_wht", "textures/makkon_concrete/conc_t08_wht"},
	{"textures/conc_t08_ylw", "textures/makkon_concrete/conc_t08_ylw"},
	{"textures/conc_t09_blk", "textures/makkon_concrete/conc_t09_blk"},
	{"textures/conc_t09_blu", "textures/makkon_concrete/conc_t09_blu"},
	{"textures/conc_t09_brwn", "textures/makkon_concrete/conc_t09_brwn"},
	{"textures/conc_t09_grn", "textures/makkon_concrete/conc_t09_grn"},
	{"textures/conc_t09_gry1", "textures/makkon_concrete/conc_t09_gry1"},
	{"textures/conc_t09_gry2", "textures/makkon_concrete/conc_t09_gry2"},
	{"textures/conc_t09_gry3", "textures/makkon_concrete/conc_t09_gry3"},
	{"textures/conc_t09_pink", "textures/makkon_concrete/conc_t09_pink"},
	{"textures/conc_t09_red", "textures/makkon_concrete/conc_t09_red"},
	{"textures/conc_t09_red2", "textures/makkon_concrete/conc_t09_red2"},
	{"textures/conc_t09_teal", "textures/makkon_concrete/conc_t09_teal"},
	{"textures/conc_t09_wht", "textures/makkon_concrete/conc_t09_wht"},
	{"textures/conc_t09_ylw", "textures/makkon_concrete/conc_t09_ylw"},
	{"textures/conc_w01_blk", "textures/makkon_concrete/conc_w01_blk"},
	{"textures/conc_w01_blu", "textures/makkon_concrete/conc_w01_blu"},
	{"textures/conc_w01_brwn", "textures/makkon_concrete/conc_w01_brwn"},
	{"textures/conc_w01_grn", "textures/makkon_concrete/conc_w01_grn"},
	{"textures/conc_w01_gry1", "textures/makkon_concrete/conc_w01_gry1"},
	{"textures/conc_w01_gry2", "textures/makkon_concrete/conc_w01_gry2"},
	{"textures/conc_w01_gry3", "textures/makkon_concrete/conc_w01_gry3"},
	{"textures/conc_w01_pink", "textures/makkon_concrete/conc_w01_pink"},
	{"textures/conc_w01_red", "textures/makkon_concrete/conc_w01_red"},
	{"textures/conc_w01_red2", "textures/makkon_concrete/conc_w01_red2"},
	{"textures/conc_w01_teal", "textures/makkon_concrete/conc_w01_teal"},
	{"textures/conc_w01_wht", "textures/makkon_concrete/conc_w01_wht"},
	{"textures/conc_w01_ylw", "textures/makkon_concrete/conc_w01_ylw"},
	{"textures/conc_w02_blk", "textures/makkon_concrete/conc_w02_blk"},
	{"textures/conc_w02_blu", "textures/makkon_concrete/conc_w02_blu"},
	{"textures/conc_w02_brwn", "textures/makkon_concrete/conc_w02_brwn"},
	{"textures/conc_w02_grn", "textures/makkon_concrete/conc_w02_grn"},
	{"textures/conc_w02_gry1", "textures/makkon_concrete/conc_w02_gry1"},
	{"textures/conc_w02_gry2", "textures/makkon_concrete/conc_w02_gry2"},
	{"textures/conc_w02_gry3", "textures/makkon_concrete/conc_w02_gry3"},
	{"textures/conc_w02_pink", "textures/makkon_concrete/conc_w02_pink"},
	{"textures/conc_w02_red", "textures/makkon_concrete/conc_w02_red"},
	{"textures/conc_w02_red2", "textures/makkon_concrete/conc_w02_red2"},
	{"textures/conc_w02_teal", "textures/makkon_concrete/conc_w02_teal"},
	{"textures/conc_w02_wht", "textures/makkon_concrete/conc_w02_wht"},
	{"textures/conc_w02_ylw", "textures/makkon_concrete/conc_w02_ylw"},
	{"textures/conc_w03_blk", "textures/makkon_concrete/conc_w03_blk"},
	{"textures/conc_w03_blu", "textures/makkon_concrete/conc_w03_blu"},
	{"textures/conc_w03_brwn", "textures/makkon_concrete/conc_w03_brwn"},
	{"textures/conc_w03_grn", "textures/makkon_concrete/conc_w03_grn"},
	{"textures/conc_w03_gry1", "textures/makkon_concrete/conc_w03_gry1"},
	{"textures/conc_w03_gry2", "textures/makkon_concrete/conc_w03_gry2"},
	{"textures/conc_w03_gry3", "textures/makkon_concrete/conc_w03_gry3"},
	{"textures/conc_w03_pink", "textures/makkon_concrete/conc_w03_pink"},
	{"textures/conc_w03_red", "textures/makkon_concrete/conc_w03_red"},
	{"textures/conc_w03_red2", "textures/makkon_concrete/conc_w03_red2"},
	{"textures/conc_w03_teal", "textures/makkon_concrete/conc_w03_teal"},
	{"textures/conc_w03_wht", "textures/makkon_concrete/conc_w03_wht"},
	{"textures/conc_w03_ylw", "textures/makkon_concrete/conc_w03_ylw"},
	{"textures/conc_w04_blk", "textures/makkon_concrete/conc_w04_blk"},
	{"textures/conc_w04_blu", "textures/makkon_concrete/conc_w04_blu"},
	{"textures/conc_w04_brwn", "textures/makkon_concrete/conc_w04_brwn"},
	{"textures/conc_w04_grn", "textures/makkon_concrete/conc_w04_grn"},
	{"textures/conc_w04_gry1", "textures/makkon_concrete/conc_w04_gry1"},
	{"textures/conc_w04_gry2", "textures/makkon_concrete/conc_w04_gry2"},
	{"textures/conc_w04_gry3", "textures/makkon_concrete/conc_w04_gry3"},
	{"textures/conc_w04_pink", "textures/makkon_concrete/conc_w04_pink"},
	{"textures/conc_w04_red", "textures/makkon_concrete/conc_w04_red"},
	{"textures/conc_w04_red2", "textures/makkon_concrete/conc_w04_red2"},
	{"textures/conc_w04_teal", "textures/makkon_concrete/conc_w04_teal"},
	{"textures/conc_w04_wht", "textures/makkon_concrete/conc_w04_wht"},
	{"textures/conc_w04_ylw", "textures/makkon_concrete/conc_w04_ylw"},
	{"textures/conc_w05_blk", "textures/makkon_concrete/conc_w05_blk"},
	{"textures/conc_w05_blu", "textures/makkon_concrete/conc_w05_blu"},
	{"textures/conc_w05_brwn", "textures/makkon_concrete/conc_w05_brwn"},
	{"textures/conc_w05_grn", "textures/makkon_concrete/conc_w05_grn"},
	{"textures/conc_w05_gry1", "textures/makkon_concrete/conc_w05_gry1"},
	{"textures/conc_w05_gry2", "textures/makkon_concrete/conc_w05_gry2"},
	{"textures/conc_w05_gry3", "textures/makkon_concrete/conc_w05_gry3"},
	{"textures/conc_w05_pink", "textures/makkon_concrete/conc_w05_pink"},
	{"textures/conc_w05_red", "textures/makkon_concrete/conc_w05_red"},
	{"textures/conc_w05_red2", "textures/makkon_concrete/conc_w05_red2"},
	{"textures/conc_w05_teal", "textures/makkon_concrete/conc_w05_teal"},
	{"textures/conc_w05_wht", "textures/makkon_concrete/conc_w05_wht"},
	{"textures/conc_w05_ylw", "textures/makkon_concrete/conc_w05_ylw"},
	{"textures/conc_w06_blk", "textures/makkon_concrete/conc_w06_blk"},
	{"textures/conc_w06_blu", "textures/makkon_concrete/conc_w06_blu"},
	{"textures/conc_w06_brwn", "textures/makkon_concrete/conc_w06_brwn"},
	{"textures/conc_w06_grn", "textures/makkon_concrete/conc_w06_grn"},
	{"textures/conc_w06_gry1", "textures/makkon_concrete/conc_w06_gry1"},
	{"textures/conc_w06_gry2", "textures/makkon_concrete/conc_w06_gry2"},
	{"textures/conc_w06_gry3", "textures/makkon_concrete/conc_w06_gry3"},
	{"textures/conc_w06_pink", "textures/makkon_concrete/conc_w06_pink"},
	{"textures/conc_w06_red", "textures/makkon_concrete/conc_w06_red"},
	{"textures/conc_w06_red2", "textures/makkon_concrete/conc_w06_red2"},
	{"textures/conc_w06_teal", "textures/makkon_concrete/conc_w06_teal"},
	{"textures/conc_w06_wht", "textures/makkon_concrete/conc_w06_wht"},
	{"textures/conc_w06_ylw", "textures/makkon_concrete/conc_w06_ylw"},
	{"textures/conc_w07_blk", "textures/makkon_concrete/conc_w07_blk"},
	{"textures/conc_w07_blu", "textures/makkon_concrete/conc_w07_blu"},
	{"textures/conc_w07_brwn", "textures/makkon_concrete/conc_w07_brwn"},
	{"textures/conc_w07_grn", "textures/makkon_concrete/conc_w07_grn"},
	{"textures/conc_w07_gry1", "textures/makkon_concrete/conc_w07_gry1"},
	{"textures/conc_w07_gry2", "textures/makkon_concrete/conc_w07_gry2"},
	{"textures/conc_w07_gry3", "textures/makkon_concrete/conc_w07_gry3"},
	{"textures/conc_w07_pink", "textures/makkon_concrete/conc_w07_pink"},
	{"textures/conc_w07_red", "textures/makkon_concrete/conc_w07_red"},
	{"textures/conc_w07_red2", "textures/makkon_concrete/conc_w07_red2"},
	{"textures/conc_w07_teal", "textures/makkon_concrete/conc_w07_teal"},
	{"textures/conc_w07_wht", "textures/makkon_concrete/conc_w07_wht"},
	{"textures/conc_w07_ylw", "textures/makkon_concrete/conc_w07_ylw"},
	{"textures/conc_w08_blk", "textures/makkon_concrete/conc_w08_blk"},
	{"textures/conc_w08_blu", "textures/makkon_concrete/conc_w08_blu"},
	{"textures/conc_w08_brwn", "textures/makkon_concrete/conc_w08_brwn"},
	{"textures/conc_w08_grn", "textures/makkon_concrete/conc_w08_grn"},
	{"textures/conc_w08_gry1", "textures/makkon_concrete/conc_w08_gry1"},
	{"textures/conc_w08_gry2", "textures/makkon_concrete/conc_w08_gry2"},
	{"textures/conc_w08_gry3", "textures/makkon_concrete/conc_w08_gry3"},
	{"textures/conc_w08_pink", "textures/makkon_concrete/conc_w08_pink"},
	{"textures/conc_w08_red", "textures/makkon_concrete/conc_w08_red"},
	{"textures/conc_w08_red2", "textures/makkon_concrete/conc_w08_red2"},
	{"textures/conc_w08_teal", "textures/makkon_concrete/conc_w08_teal"},
	{"textures/conc_w08_wht", "textures/makkon_concrete/conc_w08_wht"},
	{"textures/conc_w08_ylw", "textures/makkon_concrete/conc_w08_ylw"},
	{"textures/conc_w09_blk", "textures/makkon_concrete/conc_w09_blk"},
	{"textures/conc_w09_blu", "textures/makkon_concrete/conc_w09_blu"},
	{"textures/conc_w09_brwn", "textures/makkon_concrete/conc_w09_brwn"},
	{"textures/conc_w09_grn", "textures/makkon_concrete/conc_w09_grn"},
	{"textures/conc_w09_gry1", "textures/makkon_concrete/conc_w09_gry1"},
	{"textures/conc_w09_gry2", "textures/makkon_concrete/conc_w09_gry2"},
	{"textures/conc_w09_gry3", "textures/makkon_concrete/conc_w09_gry3"},
	{"textures/conc_w09_pink", "textures/makkon_concrete/conc_w09_pink"},
	{"textures/conc_w09_red", "textures/makkon_concrete/conc_w09_red"},
	{"textures/conc_w09_red2", "textures/makkon_concrete/conc_w09_red2"},
	{"textures/conc_w09_teal", "textures/makkon_concrete/conc_w09_teal"},
	{"textures/conc_w09_wht", "textures/makkon_concrete/conc_w09_wht"},
	{"textures/conc_w09_ylw", "textures/makkon_concrete/conc_w09_ylw"},
	{"textures/conc_w10_blk", "textures/makkon_concrete/conc_w10_blk"},
	{"textures/conc_w10_blu", "textures/makkon_concrete/conc_w10_blu"},
	{"textures/conc_w10_brwn", "textures/makkon_concrete/conc_w10_brwn"},
	{"textures/conc_w10_grn", "textures/makkon_concrete/conc_w10_grn"},
	{"textures/conc_w10_gry1", "textures/makkon_concrete/conc_w10_gry1"},
	{"textures/conc_w10_gry2", "textures/makkon_concrete/conc_w10_gry2"},
	{"textures/conc_w10_gry3", "textures/makkon_concrete/conc_w10_gry3"},
	{"textures/conc_w10_pink", "textures/makkon_concrete/conc_w10_pink"},
	{"textures/conc_w10_red", "textures/makkon_concrete/conc_w10_red"},
	{"textures/conc_w10_red2", "textures/makkon_concrete/conc_w10_red2"},
	{"textures/conc_w10_teal", "textures/makkon_concrete/conc_w10_teal"},
	{"textures/conc_w10_wht", "textures/makkon_concrete/conc_w10_wht"},
	{"textures/conc_w10_ylw", "textures/makkon_concrete/conc_w10_ylw"},
	{"textures/conc_w11_blk", "textures/makkon_concrete/conc_w11_blk"},
	{"textures/conc_w11_blu", "textures/makkon_concrete/conc_w11_blu"},
	{"textures/conc_w11_brwn", "textures/makkon_concrete/conc_w11_brwn"},
	{"textures/conc_w11_grn", "textures/makkon_concrete/conc_w11_grn"},
	{"textures/conc_w11_gry1", "textures/makkon_concrete/conc_w11_gry1"},
	{"textures/conc_w11_gry2", "textures/makkon_concrete/conc_w11_gry2"},
	{"textures/conc_w11_gry3", "textures/makkon_concrete/conc_w11_gry3"},
	{"textures/conc_w11_pink", "textures/makkon_concrete/conc_w11_pink"},
	{"textures/conc_w11_red", "textures/makkon_concrete/conc_w11_red"},
	{"textures/conc_w11_red2", "textures/makkon_concrete/conc_w11_red2"},
	{"textures/conc_w11_teal", "textures/makkon_concrete/conc_w11_teal"},
	{"textures/conc_w11_wht", "textures/makkon_concrete/conc_w11_wht"},
	{"textures/conc_w11_ylw", "textures/makkon_concrete/conc_w11_ylw"},
	{"textures/conc_w12_blk", "textures/makkon_concrete/conc_w12_blk"},
	{"textures/conc_w12_blu", "textures/makkon_concrete/conc_w12_blu"},
	{"textures/conc_w12_brwn", "textures/makkon_concrete/conc_w12_brwn"},
	{"textures/conc_w12_grn", "textures/makkon_concrete/conc_w12_grn"},
	{"textures/conc_w12_gry1", "textures/makkon_concrete/conc_w12_gry1"},
	{"textures/conc_w12_gry2", "textures/makkon_concrete/conc_w12_gry2"},
	{"textures/conc_w12_gry3", "textures/makkon_concrete/conc_w12_gry3"},
	{"textures/conc_w12_pink", "textures/makkon_concrete/conc_w12_pink"},
	{"textures/conc_w12_red", "textures/makkon_concrete/conc_w12_red"},
	{"textures/conc_w12_red2", "textures/makkon_concrete/conc_w12_red2"},
	{"textures/conc_w12_teal", "textures/makkon_concrete/conc_w12_teal"},
	{"textures/conc_w12_wht", "textures/makkon_concrete/conc_w12_wht"},
	{"textures/conc_w12_ylw", "textures/makkon_concrete/conc_w12_ylw"},
	{"textures/conc_w13_blk", "textures/makkon_concrete/conc_w13_blk"},
	{"textures/conc_w13_blu", "textures/makkon_concrete/conc_w13_blu"},
	{"textures/conc_w13_brwn", "textures/makkon_concrete/conc_w13_brwn"},
	{"textures/conc_w13_grn", "textures/makkon_concrete/conc_w13_grn"},
	{"textures/conc_w13_gry1", "textures/makkon_concrete/conc_w13_gry1"},
	{"textures/conc_w13_gry2", "textures/makkon_concrete/conc_w13_gry2"},
	{"textures/conc_w13_gry3", "textures/makkon_concrete/conc_w13_gry3"},
	{"textures/conc_w13_pink", "textures/makkon_concrete/conc_w13_pink"},
	{"textures/conc_w13_red", "textures/makkon_concrete/conc_w13_red"},
	{"textures/conc_w13_red2", "textures/makkon_concrete/conc_w13_red2"},
	{"textures/conc_w13_teal", "textures/makkon_concrete/conc_w13_teal"},
	{"textures/conc_w13_wht", "textures/makkon_concrete/conc_w13_wht"},
	{"textures/conc_w13_ylw", "textures/makkon_concrete/conc_w13_ylw"},
	{"textures/conc_w14_blk", "textures/makkon_concrete/conc_w14_blk"},
	{"textures/conc_w14_blu", "textures/makkon_concrete/conc_w14_blu"},
	{"textures/conc_w14_brwn", "textures/makkon_concrete/conc_w14_brwn"},
	{"textures/conc_w14_grn", "textures/makkon_concrete/conc_w14_grn"},
	{"textures/conc_w14_gry1", "textures/makkon_concrete/conc_w14_gry1"},
	{"textures/conc_w14_gry2", "textures/makkon_concrete/conc_w14_gry2"},
	{"textures/conc_w14_gry3", "textures/makkon_concrete/conc_w14_gry3"},
	{"textures/conc_w14_pink", "textures/makkon_concrete/conc_w14_pink"},
	{"textures/conc_w14_red", "textures/makkon_concrete/conc_w14_red"},
	{"textures/conc_w14_red2", "textures/makkon_concrete/conc_w14_red2"},
	{"textures/conc_w14_teal", "textures/makkon_concrete/conc_w14_teal"},
	{"textures/conc_w14_wht", "textures/makkon_concrete/conc_w14_wht"},
	{"textures/conc_w14_ylw", "textures/makkon_concrete/conc_w14_ylw"},
	{"textures/conc_w_15_blk", "textures/makkon_concrete/conc_w_15_blk"},
	{"textures/conc_w_15_blu", "textures/makkon_concrete/conc_w_15_blu"},
	{"textures/conc_w_15_brwn", "textures/makkon_concrete/conc_w_15_brwn"},
	{"textures/conc_w_15_grn", "textures/makkon_concrete/conc_w_15_grn"},
	{"textures/conc_w_15_gry1", "textures/makkon_concrete/conc_w_15_gry1"},
	{"textures/conc_w_15_gry2", "textures/makkon_concrete/conc_w_15_gry2"},
	{"textures/conc_w_15_gry3", "textures/makkon_concrete/conc_w_15_gry3"},
	{"textures/conc_w_15_pink", "textures/makkon_concrete/conc_w_15_pink"},
	{"textures/conc_w_15_red", "textures/makkon_concrete/conc_w_15_red"},
	{"textures/conc_w_15_red2", "textures/makkon_concrete/conc_w_15_red2"},
	{"textures/conc_w_15_teal", "textures/makkon_concrete/conc_w_15_teal"},
	{"textures/conc_w_15_wht", "textures/makkon_concrete/conc_w_15_wht"},
	{"textures/conc_w_15_ylw", "textures/makkon_concrete/conc_w_15_ylw"},
	{"textures/conc_w_16_blk", "textures/makkon_concrete/conc_w_16_blk"},
	{"textures/conc_w_16_blu", "textures/makkon_concrete/conc_w_16_blu"},
	{"textures/conc_w_16_brwn", "textures/makkon_concrete/conc_w_16_brwn"},
	{"textures/conc_w_16_grn", "textures/makkon_concrete/conc_w_16_grn"},
	{"textures/conc_w_16_gry1", "textures/makkon_concrete/conc_w_16_gry1"},
	{"textures/conc_w_16_gry2", "textures/makkon_concrete/conc_w_16_gry2"},
	{"textures/conc_w_16_gry3", "textures/makkon_concrete/conc_w_16_gry3"},
	{"textures/conc_w_16_pink", "textures/makkon_concrete/conc_w_16_pink"},
	{"textures/conc_w_16_red", "textures/makkon_concrete/conc_w_16_red"},
	{"textures/conc_w_16_red2", "textures/makkon_concrete/conc_w_16_red2"},
	{"textures/conc_w_16_teal", "textures/makkon_concrete/conc_w_16_teal"},
	{"textures/conc_w_16_wht", "textures/makkon_concrete/conc_w_16_wht"},
	{"textures/conc_w_16_ylw", "textures/makkon_concrete/conc_w_16_ylw"},
	{"textures/conc_w_17_blk", "textures/makkon_concrete/conc_w_17_blk"},
	{"textures/conc_w_17_blu", "textures/makkon_concrete/conc_w_17_blu"},
	{"textures/conc_w_17_brwn", "textures/makkon_concrete/conc_w_17_brwn"},
	{"textures/conc_w_17_grn", "textures/makkon_concrete/conc_w_17_grn"},
	{"textures/conc_w_17_gry1", "textures/makkon_concrete/conc_w_17_gry1"},
	{"textures/conc_w_17_gry2", "textures/makkon_concrete/conc_w_17_gry2"},
	{"textures/conc_w_17_gry3", "textures/makkon_concrete/conc_w_17_gry3"},
	{"textures/conc_w_17_pink", "textures/makkon_concrete/conc_w_17_pink"},
	{"textures/conc_w_17_red", "textures/makkon_concrete/conc_w_17_red"},
	{"textures/conc_w_17_red2", "textures/makkon_concrete/conc_w_17_red2"},
	{"textures/conc_w_17_teal", "textures/makkon_concrete/conc_w_17_teal"},
	{"textures/conc_w_17_wht", "textures/makkon_concrete/conc_w_17_wht"},
	{"textures/conc_w_17_ylw", "textures/makkon_concrete/conc_w_17_ylw"},
	{"textures/_conc_d01_brwn", "textures/makkon_concrete/_conc_d01_brwn"},
	{"textures/_conc_d01_gry1", "textures/makkon_concrete/_conc_d01_gry1"},
	{"textures/_conc_d01_gry2", "textures/makkon_concrete/_conc_d01_gry2"},
	{"textures/_conc_d01_gry3", "textures/makkon_concrete/_conc_d01_gry3"},
	{"textures/_conc_d01_rebar", "textures/makkon_concrete/_conc_d01_rebar"},
	{"textures/_conc_d02_brwn", "textures/makkon_concrete/_conc_d02_brwn"},
	{"textures/_conc_d02_gry1", "textures/makkon_concrete/_conc_d02_gry1"},
	{"textures/_conc_d02_gry2", "textures/makkon_concrete/_conc_d02_gry2"},
	{"textures/_conc_d02_gry3", "textures/makkon_concrete/_conc_d02_gry3"},
	{"textures/_conc_d02_rebar", "textures/makkon_concrete/_conc_d02_rebar"},
	{"textures/_conc_d03_brwn", "textures/makkon_concrete/_conc_d03_brwn"},
	{"textures/_conc_d03_gry1", "textures/makkon_concrete/_conc_d03_gry1"},
	{"textures/_conc_d03_gry2", "textures/makkon_concrete/_conc_d03_gry2"},
	{"textures/_conc_d03_gry3", "textures/makkon_concrete/_conc_d03_gry3"},
	{"textures/_conc_d04_brwn", "textures/makkon_concrete/_conc_d04_brwn"},
	{"textures/_conc_d04_gry1", "textures/makkon_concrete/_conc_d04_gry1"},
	{"textures/_conc_d04_gry2", "textures/makkon_concrete/_conc_d04_gry2"},
	{"textures/_conc_d04_gry3", "textures/makkon_concrete/_conc_d04_gry3"},
	{"textures/_conc_d05_brwn", "textures/makkon_concrete/_conc_d05_brwn"},
	{"textures/_conc_d05_gry1", "textures/makkon_concrete/_conc_d05_gry1"},
	{"textures/_conc_d05_gry2", "textures/makkon_concrete/_conc_d05_gry2"},
	{"textures/_conc_d05_gry3", "textures/makkon_concrete/_conc_d05_gry3"},
	{"textures/_conc_d06_brwn", "textures/makkon_concrete/_conc_d06_brwn"},
	{"textures/_conc_d06_gry1", "textures/makkon_concrete/_conc_d06_gry1"},
	{"textures/_conc_d06_gry2", "textures/makkon_concrete/_conc_d06_gry2"},
	{"textures/_conc_d06_gry3", "textures/makkon_concrete/_conc_d06_gry3"},
	{"textures/_conc_d07_brwn", "textures/makkon_concrete/_conc_d07_brwn"},
	{"textures/_conc_d07_gry1", "textures/makkon_concrete/_conc_d07_gry1"},
	{"textures/_conc_d07_gry2", "textures/makkon_concrete/_conc_d07_gry2"},
	{"textures/_conc_d07_gry3", "textures/makkon_concrete/_conc_d07_gry3"},
	{"textures/_conc_d08_brwn", "textures/makkon_concrete/_conc_d08_brwn"},
	{"textures/_conc_d08_gry1", "textures/makkon_concrete/_conc_d08_gry1"},
	{"textures/_conc_d08_gry2", "textures/makkon_concrete/_conc_d08_gry2"},
	{"textures/_conc_d08_gry3", "textures/makkon_concrete/_conc_d08_gry3"},

	// Makkon horror 2022.8
	{"textures/bek_blck1_w01_1", "textures/makkon_horror/bek_blck1_w01_1"},
	{"textures/bek_blck1_w01_2", "textures/makkon_horror/bek_blck1_w01_2"},
	{"textures/bek_blck1_w01_3", "textures/makkon_horror/bek_blck1_w01_3"},
	{"textures/bek_blck1_w01_4", "textures/makkon_horror/bek_blck1_w01_4"},
	{"textures/bek_blck1_w02_1", "textures/makkon_horror/bek_blck1_w02_1"},
	{"textures/bek_blck1_w03_1", "textures/makkon_horror/bek_blck1_w03_1"},
	{"textures/bek_blck1_w03_2", "textures/makkon_horror/bek_blck1_w03_2"},
	{"textures/bek_blck1_w03_3", "textures/makkon_horror/bek_blck1_w03_3"},
	{"textures/bek_blck1_w03_4", "textures/makkon_horror/bek_blck1_w03_4"},
	{"textures/bek_blck1_w03_6", "textures/makkon_horror/bek_blck1_w03_6"},
	{"textures/bek_blck1_w03_7", "textures/makkon_horror/bek_blck1_w03_7"},
	{"textures/bek_blck1_w03_8", "textures/makkon_horror/bek_blck1_w03_8"},
	{"textures/bek_blck1_w04_1", "textures/makkon_horror/bek_blck1_w04_1"},
	{"textures/bek_blck1_w05_1", "textures/makkon_horror/bek_blck1_w05_1"},
	{"textures/bek_blck1_w05_2", "textures/makkon_horror/bek_blck1_w05_2"},
	{"textures/bek_blck1_w05_3", "textures/makkon_horror/bek_blck1_w05_3"},
	{"textures/bek_blck1_w06_1", "textures/makkon_horror/bek_blck1_w06_1"},
	{"textures/bek_blck1_w06_2", "textures/makkon_horror/bek_blck1_w06_2"},
	{"textures/bek_blck1_w06_3", "textures/makkon_horror/bek_blck1_w06_3"},
	{"textures/bek_blck1_w07_1", "textures/makkon_horror/bek_blck1_w07_1"},
	{"textures/bek_blck1_w07_2", "textures/makkon_horror/bek_blck1_w07_2"},
	{"textures/bek_blck1_w07_3", "textures/makkon_horror/bek_blck1_w07_3"},
	{"textures/bek_blck1_w08_1", "textures/makkon_horror/bek_blck1_w08_1"},
	{"textures/bek_blck1_w09_1", "textures/makkon_horror/bek_blck1_w09_1"},
	{"textures/bek_blck1_w10_1", "textures/makkon_horror/bek_blck1_w10_1"},
	{"textures/bek_blck1_w11_1", "textures/makkon_horror/bek_blck1_w11_1"},
	{"textures/bek_blck1_w12_1", "textures/makkon_horror/bek_blck1_w12_1"},
	{"textures/bek_blck1_w12_2", "textures/makkon_horror/bek_blck1_w12_2"},
	{"textures/bek_blck1_w12_3", "textures/makkon_horror/bek_blck1_w12_3"},
	{"textures/bek_blck_t01_1", "textures/makkon_horror/bek_blck_t01_1"},
	{"textures/bek_blue1_t01_1", "textures/makkon_horror/bek_blue1_t01_1"},
	{"textures/bek_blue1_t02_1", "textures/makkon_horror/bek_blue1_t02_1"},
	{"textures/bek_blue1_w01_1", "textures/makkon_horror/bek_blue1_w01_1"},
	{"textures/bek_blue1_w01_2", "textures/makkon_horror/bek_blue1_w01_2"},
	{"textures/bek_blue1_w01_3", "textures/makkon_horror/bek_blue1_w01_3"},
	{"textures/bek_blue1_w01_4", "textures/makkon_horror/bek_blue1_w01_4"},
	{"textures/bek_blue1_w02_1", "textures/makkon_horror/bek_blue1_w02_1"},
	{"textures/bek_blue1_w03_1", "textures/makkon_horror/bek_blue1_w03_1"},
	{"textures/bek_blue1_w03_2", "textures/makkon_horror/bek_blue1_w03_2"},
	{"textures/bek_blue1_w03_3", "textures/makkon_horror/bek_blue1_w03_3"},
	{"textures/bek_blue1_w03_4", "textures/makkon_horror/bek_blue1_w03_4"},
	{"textures/bek_blue1_w03_5", "textures/makkon_horror/bek_blue1_w03_5"},
	{"textures/bek_blue1_w03_6", "textures/makkon_horror/bek_blue1_w03_6"},
	{"textures/bek_blue1_w03_7", "textures/makkon_horror/bek_blue1_w03_7"},
	{"textures/bek_blue1_w03_8", "textures/makkon_horror/bek_blue1_w03_8"},
	{"textures/bek_blue1_w04_1", "textures/makkon_horror/bek_blue1_w04_1"},
	{"textures/bek_blue1_w05_1", "textures/makkon_horror/bek_blue1_w05_1"},
	{"textures/bek_blue1_w05_2", "textures/makkon_horror/bek_blue1_w05_2"},
	{"textures/bek_blue1_w05_3", "textures/makkon_horror/bek_blue1_w05_3"},
	{"textures/bek_blue1_w06_1", "textures/makkon_horror/bek_blue1_w06_1"},
	{"textures/bek_blue1_w06_2", "textures/makkon_horror/bek_blue1_w06_2"},
	{"textures/bek_blue1_w07_1", "textures/makkon_horror/bek_blue1_w07_1"},
	{"textures/bek_blue1_w07_2", "textures/makkon_horror/bek_blue1_w07_2"},
	{"textures/bek_blue1_w08_1", "textures/makkon_horror/bek_blue1_w08_1"},
	{"textures/bek_blue1_w09_1", "textures/makkon_horror/bek_blue1_w09_1"},
	{"textures/bek_blue1_w10_1", "textures/makkon_horror/bek_blue1_w10_1"},
	{"textures/bek_blue1_w11_1", "textures/makkon_horror/bek_blue1_w11_1"},
	{"textures/bek_blue1_w12_1", "textures/makkon_horror/bek_blue1_w12_1"},
	{"textures/bek_blue1_w12_2", "textures/makkon_horror/bek_blue1_w12_2"},
	{"textures/bek_blue1_w12_3", "textures/makkon_horror/bek_blue1_w12_3"},
	{"textures/bek_brwn1_t01_1", "textures/makkon_horror/bek_brwn1_t01_1"},
	{"textures/bek_brwn1_t02_1", "textures/makkon_horror/bek_brwn1_t02_1"},
	{"textures/bek_brwn1_w01_1", "textures/makkon_horror/bek_brwn1_w01_1"},
	{"textures/bek_brwn1_w01_2", "textures/makkon_horror/bek_brwn1_w01_2"},
	{"textures/bek_brwn1_w01_3", "textures/makkon_horror/bek_brwn1_w01_3"},
	{"textures/bek_brwn1_w01_4", "textures/makkon_horror/bek_brwn1_w01_4"},
	{"textures/bek_brwn1_w02_1", "textures/makkon_horror/bek_brwn1_w02_1"},
	{"textures/bek_brwn1_w03_1", "textures/makkon_horror/bek_brwn1_w03_1"},
	{"textures/bek_brwn1_w03_2", "textures/makkon_horror/bek_brwn1_w03_2"},
	{"textures/bek_brwn1_w03_3", "textures/makkon_horror/bek_brwn1_w03_3"},
	{"textures/bek_brwn1_w03_4", "textures/makkon_horror/bek_brwn1_w03_4"},
	{"textures/bek_brwn1_w03_5", "textures/makkon_horror/bek_brwn1_w03_5"},
	{"textures/bek_brwn1_w03_6", "textures/makkon_horror/bek_brwn1_w03_6"},
	{"textures/bek_brwn1_w03_7", "textures/makkon_horror/bek_brwn1_w03_7"},
	{"textures/bek_brwn1_w03_8", "textures/makkon_horror/bek_brwn1_w03_8"},
	{"textures/bek_brwn1_w04_1", "textures/makkon_horror/bek_brwn1_w04_1"},
	{"textures/bek_brwn1_w04_2", "textures/makkon_horror/bek_brwn1_w04_2"},
	{"textures/bek_brwn1_w04_3", "textures/makkon_horror/bek_brwn1_w04_3"},
	{"textures/bek_brwn1_w04_4", "textures/makkon_horror/bek_brwn1_w04_4"},
	{"textures/bek_brwn1_w06_1", "textures/makkon_horror/bek_brwn1_w06_1"},
	{"textures/bek_brwn1_w06_2", "textures/makkon_horror/bek_brwn1_w06_2"},
	{"textures/bek_brwn1_w07_1", "textures/makkon_horror/bek_brwn1_w07_1"},
	{"textures/bek_brwn1_w07_2", "textures/makkon_horror/bek_brwn1_w07_2"},
	{"textures/bek_brwn1_w08_1", "textures/makkon_horror/bek_brwn1_w08_1"},
	{"textures/bek_brwn1_w09_1", "textures/makkon_horror/bek_brwn1_w09_1"},
	{"textures/bek_brwn1_w10_1", "textures/makkon_horror/bek_brwn1_w10_1"},
	{"textures/bek_brwn1_w11_1", "textures/makkon_horror/bek_brwn1_w11_1"},
	{"textures/bek_brwn1_w12_1", "textures/makkon_horror/bek_brwn1_w12_1"},
	{"textures/bek_brwn1_w12_2", "textures/makkon_horror/bek_brwn1_w12_2"},
	{"textures/bek_brwn1_w12_3", "textures/makkon_horror/bek_brwn1_w12_3"},
	{"textures/bek_gen1_w01_1", "textures/makkon_horror/bek_gen1_w01_1"},
	{"textures/bek_grey1_t01_1", "textures/makkon_horror/bek_grey1_t01_1"},
	{"textures/bek_grey1_t02_1", "textures/makkon_horror/bek_grey1_t02_1"},
	{"textures/bek_grey1_w01_1", "textures/makkon_horror/bek_grey1_w01_1"},
	{"textures/bek_grey1_w01_2", "textures/makkon_horror/bek_grey1_w01_2"},
	{"textures/bek_grey1_w01_3", "textures/makkon_horror/bek_grey1_w01_3"},
	{"textures/bek_grey1_w01_4", "textures/makkon_horror/bek_grey1_w01_4"},
	{"textures/bek_grey1_w02_1", "textures/makkon_horror/bek_grey1_w02_1"},
	{"textures/bek_grey1_w03_1", "textures/makkon_horror/bek_grey1_w03_1"},
	{"textures/bek_grey1_w03_2", "textures/makkon_horror/bek_grey1_w03_2"},
	{"textures/bek_grey1_w03_3", "textures/makkon_horror/bek_grey1_w03_3"},
	{"textures/bek_grey1_w03_4", "textures/makkon_horror/bek_grey1_w03_4"},
	{"textures/bek_grey1_w03_5", "textures/makkon_horror/bek_grey1_w03_5"},
	{"textures/bek_grey1_w03_6", "textures/makkon_horror/bek_grey1_w03_6"},
	{"textures/bek_grey1_w03_7", "textures/makkon_horror/bek_grey1_w03_7"},
	{"textures/bek_grey1_w03_8", "textures/makkon_horror/bek_grey1_w03_8"},
	{"textures/bek_grey1_w04_1", "textures/makkon_horror/bek_grey1_w04_1"},
	{"textures/bek_grey1_w05_1", "textures/makkon_horror/bek_grey1_w05_1"},
	{"textures/bek_grey1_w05_2", "textures/makkon_horror/bek_grey1_w05_2"},
	{"textures/bek_grey1_w05_3", "textures/makkon_horror/bek_grey1_w05_3"},
	{"textures/bek_grey1_w06_1", "textures/makkon_horror/bek_grey1_w06_1"},
	{"textures/bek_grey1_w06_2", "textures/makkon_horror/bek_grey1_w06_2"},
	{"textures/bek_grey1_w07_1", "textures/makkon_horror/bek_grey1_w07_1"},
	{"textures/bek_grey1_w07_2", "textures/makkon_horror/bek_grey1_w07_2"},
	{"textures/bek_grey1_w08_1", "textures/makkon_horror/bek_grey1_w08_1"},
	{"textures/bek_grey1_w09_1", "textures/makkon_horror/bek_grey1_w09_1"},
	{"textures/bek_grey1_w10_1", "textures/makkon_horror/bek_grey1_w10_1"},
	{"textures/bek_grey1_w11_1", "textures/makkon_horror/bek_grey1_w11_1"},
	{"textures/bek_grey1_w12_1", "textures/makkon_horror/bek_grey1_w12_1"},
	{"textures/bek_grey1_w12_2", "textures/makkon_horror/bek_grey1_w12_2"},
	{"textures/bek_grey1_w12_3", "textures/makkon_horror/bek_grey1_w12_3"},
	{"textures/bek_meat1_t01_1", "textures/makkon_horror/bek_meat1_t01_1"},
	{"textures/bek_meat1_t02_1", "textures/makkon_horror/bek_meat1_t02_1"},
	{"textures/bek_meat1_w01_1", "textures/makkon_horror/bek_meat1_w01_1"},
	{"textures/bek_meat1_w01_2", "textures/makkon_horror/bek_meat1_w01_2"},
	{"textures/bek_meat1_w01_3", "textures/makkon_horror/bek_meat1_w01_3"},
	{"textures/bek_meat1_w01_4", "textures/makkon_horror/bek_meat1_w01_4"},
	{"textures/bek_meat1_w02_1", "textures/makkon_horror/bek_meat1_w02_1"},
	{"textures/bek_meat1_w03_1", "textures/makkon_horror/bek_meat1_w03_1"},
	{"textures/bek_meat1_w03_2", "textures/makkon_horror/bek_meat1_w03_2"},
	{"textures/bek_meat1_w03_3", "textures/makkon_horror/bek_meat1_w03_3"},
	{"textures/bek_meat1_w03_4", "textures/makkon_horror/bek_meat1_w03_4"},
	{"textures/bek_meat1_w03_5", "textures/makkon_horror/bek_meat1_w03_5"},
	{"textures/bek_meat1_w03_6", "textures/makkon_horror/bek_meat1_w03_6"},
	{"textures/bek_meat1_w03_7", "textures/makkon_horror/bek_meat1_w03_7"},
	{"textures/bek_meat1_w04_1", "textures/makkon_horror/bek_meat1_w04_1"},
	{"textures/bek_meat1_w05_1", "textures/makkon_horror/bek_meat1_w05_1"},
	{"textures/bek_meat1_w05_2", "textures/makkon_horror/bek_meat1_w05_2"},
	{"textures/bek_meat1_w05_3", "textures/makkon_horror/bek_meat1_w05_3"},
	{"textures/bek_meat1_w06_1", "textures/makkon_horror/bek_meat1_w06_1"},
	{"textures/bek_meat1_w07_1", "textures/makkon_horror/bek_meat1_w07_1"},
	{"textures/bek_meat1_w08_1", "textures/makkon_horror/bek_meat1_w08_1"},
	{"textures/bek_meat1_w09_1", "textures/makkon_horror/bek_meat1_w09_1"},
	{"textures/bek_meat1_w10_1", "textures/makkon_horror/bek_meat1_w10_1"},
	{"textures/bek_meat1_w11_1", "textures/makkon_horror/bek_meat1_w11_1"},
	{"textures/bek_meat1_w12_1", "textures/makkon_horror/bek_meat1_w12_1"},
	{"textures/bek_meat1_w12_2", "textures/makkon_horror/bek_meat1_w12_2"},
	{"textures/bek_meat1_w12_3", "textures/makkon_horror/bek_meat1_w12_3"},
	{"textures/bek_pink1_t01_1", "textures/makkon_horror/bek_pink1_t01_1"},
	{"textures/bek_pink1_t02_1", "textures/makkon_horror/bek_pink1_t02_1"},
	{"textures/bek_pink1_w01_1", "textures/makkon_horror/bek_pink1_w01_1"},
	{"textures/bek_pink1_w01_2", "textures/makkon_horror/bek_pink1_w01_2"},
	{"textures/bek_pink1_w01_3", "textures/makkon_horror/bek_pink1_w01_3"},
	{"textures/bek_pink1_w01_4", "textures/makkon_horror/bek_pink1_w01_4"},
	{"textures/bek_pink1_w02_1", "textures/makkon_horror/bek_pink1_w02_1"},
	{"textures/bek_pink1_w03_1", "textures/makkon_horror/bek_pink1_w03_1"},
	{"textures/bek_pink1_w03_2", "textures/makkon_horror/bek_pink1_w03_2"},
	{"textures/bek_pink1_w03_3", "textures/makkon_horror/bek_pink1_w03_3"},
	{"textures/bek_pink1_w03_4", "textures/makkon_horror/bek_pink1_w03_4"},
	{"textures/bek_pink1_w03_5", "textures/makkon_horror/bek_pink1_w03_5"},
	{"textures/bek_pink1_w03_6", "textures/makkon_horror/bek_pink1_w03_6"},
	{"textures/bek_pink1_w03_7", "textures/makkon_horror/bek_pink1_w03_7"},
	{"textures/bek_pink1_w03_8", "textures/makkon_horror/bek_pink1_w03_8"},
	{"textures/bek_pink1_w03_9", "textures/makkon_horror/bek_pink1_w03_9"},
	{"textures/bek_pink1_w04_1", "textures/makkon_horror/bek_pink1_w04_1"},
	{"textures/bek_pink1_w05_1", "textures/makkon_horror/bek_pink1_w05_1"},
	{"textures/bek_pink1_w05_2", "textures/makkon_horror/bek_pink1_w05_2"},
	{"textures/bek_pink1_w05_3", "textures/makkon_horror/bek_pink1_w05_3"},
	{"textures/bek_pink1_w06_1", "textures/makkon_horror/bek_pink1_w06_1"},
	{"textures/bek_pink1_w06_2", "textures/makkon_horror/bek_pink1_w06_2"},
	{"textures/bek_pink1_w06_3", "textures/makkon_horror/bek_pink1_w06_3"},
	{"textures/bek_pink1_w07_1", "textures/makkon_horror/bek_pink1_w07_1"},
	{"textures/bek_pink1_w07_2", "textures/makkon_horror/bek_pink1_w07_2"},
	{"textures/bek_pink1_w07_3", "textures/makkon_horror/bek_pink1_w07_3"},
	{"textures/bek_pink1_w08_1", "textures/makkon_horror/bek_pink1_w08_1"},
	{"textures/bek_pink1_w09_1", "textures/makkon_horror/bek_pink1_w09_1"},
	{"textures/bek_pink1_w10_1", "textures/makkon_horror/bek_pink1_w10_1"},
	{"textures/bek_pink1_w11_1", "textures/makkon_horror/bek_pink1_w11_1"},
	{"textures/bek_pink1_w12_1", "textures/makkon_horror/bek_pink1_w12_1"},
	{"textures/bek_pink1_w12_2", "textures/makkon_horror/bek_pink1_w12_2"},
	{"textures/bek_pink1_w12_3", "textures/makkon_horror/bek_pink1_w12_3"},
	{"textures/bek_prpl1_t01_1", "textures/makkon_horror/bek_prpl1_t01_1"},
	{"textures/bek_prpl1_t02_1", "textures/makkon_horror/bek_prpl1_t02_1"},
	{"textures/bek_prpl1_w01_1", "textures/makkon_horror/bek_prpl1_w01_1"},
	{"textures/bek_prpl1_w01_2", "textures/makkon_horror/bek_prpl1_w01_2"},
	{"textures/bek_prpl1_w01_3", "textures/makkon_horror/bek_prpl1_w01_3"},
	{"textures/bek_prpl1_w01_4", "textures/makkon_horror/bek_prpl1_w01_4"},
	{"textures/bek_prpl1_w02_1", "textures/makkon_horror/bek_prpl1_w02_1"},
	{"textures/bek_prpl1_w03_1", "textures/makkon_horror/bek_prpl1_w03_1"},
	{"textures/bek_prpl1_w03_2", "textures/makkon_horror/bek_prpl1_w03_2"},
	{"textures/bek_prpl1_w03_3", "textures/makkon_horror/bek_prpl1_w03_3"},
	{"textures/bek_prpl1_w03_4", "textures/makkon_horror/bek_prpl1_w03_4"},
	{"textures/bek_prpl1_w03_5", "textures/makkon_horror/bek_prpl1_w03_5"},
	{"textures/bek_prpl1_w03_6", "textures/makkon_horror/bek_prpl1_w03_6"},
	{"textures/bek_prpl1_w03_7", "textures/makkon_horror/bek_prpl1_w03_7"},
	{"textures/bek_prpl1_w03_8", "textures/makkon_horror/bek_prpl1_w03_8"},
	{"textures/bek_prpl1_w04_1", "textures/makkon_horror/bek_prpl1_w04_1"},
	{"textures/bek_prpl1_w05_1", "textures/makkon_horror/bek_prpl1_w05_1"},
	{"textures/bek_prpl1_w05_2", "textures/makkon_horror/bek_prpl1_w05_2"},
	{"textures/bek_prpl1_w05_3", "textures/makkon_horror/bek_prpl1_w05_3"},
	{"textures/bek_prpl1_w06_1", "textures/makkon_horror/bek_prpl1_w06_1"},
	{"textures/bek_prpl1_w06_2", "textures/makkon_horror/bek_prpl1_w06_2"},
	{"textures/bek_prpl1_w06_3", "textures/makkon_horror/bek_prpl1_w06_3"},
	{"textures/bek_prpl1_w07_1", "textures/makkon_horror/bek_prpl1_w07_1"},
	{"textures/bek_prpl1_w07_2", "textures/makkon_horror/bek_prpl1_w07_2"},
	{"textures/bek_prpl1_w07_3", "textures/makkon_horror/bek_prpl1_w07_3"},
	{"textures/bek_prpl1_w08_1", "textures/makkon_horror/bek_prpl1_w08_1"},
	{"textures/bek_prpl1_w09_1", "textures/makkon_horror/bek_prpl1_w09_1"},
	{"textures/bek_prpl1_w10_1", "textures/makkon_horror/bek_prpl1_w10_1"},
	{"textures/bek_prpl1_w11_1", "textures/makkon_horror/bek_prpl1_w11_1"},
	{"textures/bek_prpl1_w12_1", "textures/makkon_horror/bek_prpl1_w12_1"},
	{"textures/bek_prpl1_w12_2", "textures/makkon_horror/bek_prpl1_w12_2"},
	{"textures/bek_prpl1_w12_3", "textures/makkon_horror/bek_prpl1_w12_3"},
	{"textures/bek_red1_t01_1", "textures/makkon_horror/bek_red1_t01_1"},
	{"textures/bek_red1_t02_1", "textures/makkon_horror/bek_red1_t02_1"},
	{"textures/bek_red1_w01_1", "textures/makkon_horror/bek_red1_w01_1"},
	{"textures/bek_red1_w01_2", "textures/makkon_horror/bek_red1_w01_2"},
	{"textures/bek_red1_w01_3", "textures/makkon_horror/bek_red1_w01_3"},
	{"textures/bek_red1_w01_4", "textures/makkon_horror/bek_red1_w01_4"},
	{"textures/bek_red1_w02_1", "textures/makkon_horror/bek_red1_w02_1"},
	{"textures/bek_red1_w03_1", "textures/makkon_horror/bek_red1_w03_1"},
	{"textures/bek_red1_w03_2", "textures/makkon_horror/bek_red1_w03_2"},
	{"textures/bek_red1_w03_3", "textures/makkon_horror/bek_red1_w03_3"},
	{"textures/bek_red1_w03_4", "textures/makkon_horror/bek_red1_w03_4"},
	{"textures/bek_red1_w03_5", "textures/makkon_horror/bek_red1_w03_5"},
	{"textures/bek_red1_w03_6", "textures/makkon_horror/bek_red1_w03_6"},
	{"textures/bek_red1_w03_7", "textures/makkon_horror/bek_red1_w03_7"},
	{"textures/bek_red1_w03_8", "textures/makkon_horror/bek_red1_w03_8"},
	{"textures/bek_red1_w04_1", "textures/makkon_horror/bek_red1_w04_1"},
	{"textures/bek_red1_w05_1", "textures/makkon_horror/bek_red1_w05_1"},
	{"textures/bek_red1_w05_2", "textures/makkon_horror/bek_red1_w05_2"},
	{"textures/bek_red1_w05_3", "textures/makkon_horror/bek_red1_w05_3"},
	{"textures/bek_red1_w06_1", "textures/makkon_horror/bek_red1_w06_1"},
	{"textures/bek_red1_w06_2", "textures/makkon_horror/bek_red1_w06_2"},
	{"textures/bek_red1_w07_1", "textures/makkon_horror/bek_red1_w07_1"},
	{"textures/bek_red1_w07_2", "textures/makkon_horror/bek_red1_w07_2"},
	{"textures/bek_red1_w08_1", "textures/makkon_horror/bek_red1_w08_1"},
	{"textures/bek_red1_w09_1", "textures/makkon_horror/bek_red1_w09_1"},
	{"textures/bek_red1_w10_1", "textures/makkon_horror/bek_red1_w10_1"},
	{"textures/bek_red1_w11_1", "textures/makkon_horror/bek_red1_w11_1"},
	{"textures/bek_red1_w12_1", "textures/makkon_horror/bek_red1_w12_1"},
	{"textures/bek_red1_w12_2", "textures/makkon_horror/bek_red1_w12_2"},
	{"textures/bek_red1_w12_3", "textures/makkon_horror/bek_red1_w12_3"},
	{"textures/bek_red2_t01_1", "textures/makkon_horror/bek_red2_t01_1"},
	{"textures/bek_red2_t02_1", "textures/makkon_horror/bek_red2_t02_1"},
	{"textures/bek_red2_w01_1", "textures/makkon_horror/bek_red2_w01_1"},
	{"textures/bek_red2_w01_2", "textures/makkon_horror/bek_red2_w01_2"},
	{"textures/bek_red2_w01_3", "textures/makkon_horror/bek_red2_w01_3"},
	{"textures/bek_red2_w01_4", "textures/makkon_horror/bek_red2_w01_4"},
	{"textures/bek_red2_w02_1", "textures/makkon_horror/bek_red2_w02_1"},
	{"textures/bek_red2_w03_1", "textures/makkon_horror/bek_red2_w03_1"},
	{"textures/bek_red2_w03_2", "textures/makkon_horror/bek_red2_w03_2"},
	{"textures/bek_red2_w03_3", "textures/makkon_horror/bek_red2_w03_3"},
	{"textures/bek_red2_w03_4", "textures/makkon_horror/bek_red2_w03_4"},
	{"textures/bek_red2_w03_5", "textures/makkon_horror/bek_red2_w03_5"},
	{"textures/bek_red2_w03_6", "textures/makkon_horror/bek_red2_w03_6"},
	{"textures/bek_red2_w03_7", "textures/makkon_horror/bek_red2_w03_7"},
	{"textures/bek_red2_w03_8", "textures/makkon_horror/bek_red2_w03_8"},
	{"textures/bek_red2_w04_1", "textures/makkon_horror/bek_red2_w04_1"},
	{"textures/bek_red2_w05_1", "textures/makkon_horror/bek_red2_w05_1"},
	{"textures/bek_red2_w05_2", "textures/makkon_horror/bek_red2_w05_2"},
	{"textures/bek_red2_w05_3", "textures/makkon_horror/bek_red2_w05_3"},
	{"textures/bek_red2_w06_1", "textures/makkon_horror/bek_red2_w06_1"},
	{"textures/bek_red2_w06_2", "textures/makkon_horror/bek_red2_w06_2"},
	{"textures/bek_red2_w07_1", "textures/makkon_horror/bek_red2_w07_1"},
	{"textures/bek_red2_w07_2", "textures/makkon_horror/bek_red2_w07_2"},
	{"textures/bek_red2_w08_1", "textures/makkon_horror/bek_red2_w08_1"},
	{"textures/bek_red2_w09_1", "textures/makkon_horror/bek_red2_w09_1"},
	{"textures/bek_red2_w10_1", "textures/makkon_horror/bek_red2_w10_1"},
	{"textures/bek_red2_w11_1", "textures/makkon_horror/bek_red2_w11_1"},
	{"textures/bek_red2_w12_1", "textures/makkon_horror/bek_red2_w12_1"},
	{"textures/bek_red2_w12_2", "textures/makkon_horror/bek_red2_w12_2"},
	{"textures/bek_red2_w12_3", "textures/makkon_horror/bek_red2_w12_3"},
	{"textures/bek_tan1_t01_1", "textures/makkon_horror/bek_tan1_t01_1"},
	{"textures/bek_tan1_t02_1", "textures/makkon_horror/bek_tan1_t02_1"},
	{"textures/bek_tan1_w01_1", "textures/makkon_horror/bek_tan1_w01_1"},
	{"textures/bek_tan1_w01_2", "textures/makkon_horror/bek_tan1_w01_2"},
	{"textures/bek_tan1_w01_3", "textures/makkon_horror/bek_tan1_w01_3"},
	{"textures/bek_tan1_w01_4", "textures/makkon_horror/bek_tan1_w01_4"},
	{"textures/bek_tan1_w02_1", "textures/makkon_horror/bek_tan1_w02_1"},
	{"textures/bek_tan1_w03_1", "textures/makkon_horror/bek_tan1_w03_1"},
	{"textures/bek_tan1_w03_2", "textures/makkon_horror/bek_tan1_w03_2"},
	{"textures/bek_tan1_w03_3", "textures/makkon_horror/bek_tan1_w03_3"},
	{"textures/bek_tan1_w03_4", "textures/makkon_horror/bek_tan1_w03_4"},
	{"textures/bek_tan1_w03_5", "textures/makkon_horror/bek_tan1_w03_5"},
	{"textures/bek_tan1_w03_6", "textures/makkon_horror/bek_tan1_w03_6"},
	{"textures/bek_tan1_w03_7", "textures/makkon_horror/bek_tan1_w03_7"},
	{"textures/bek_tan1_w03_8", "textures/makkon_horror/bek_tan1_w03_8"},
	{"textures/bek_tan1_w04_1", "textures/makkon_horror/bek_tan1_w04_1"},
	{"textures/bek_tan1_w05_1", "textures/makkon_horror/bek_tan1_w05_1"},
	{"textures/bek_tan1_w05_2", "textures/makkon_horror/bek_tan1_w05_2"},
	{"textures/bek_tan1_w05_3", "textures/makkon_horror/bek_tan1_w05_3"},
	{"textures/bek_tan1_w06_1", "textures/makkon_horror/bek_tan1_w06_1"},
	{"textures/bek_tan1_w06_2", "textures/makkon_horror/bek_tan1_w06_2"},
	{"textures/bek_tan1_w07_1", "textures/makkon_horror/bek_tan1_w07_1"},
	{"textures/bek_tan1_w07_2", "textures/makkon_horror/bek_tan1_w07_2"},
	{"textures/bek_tan1_w08_1", "textures/makkon_horror/bek_tan1_w08_1"},
	{"textures/bek_tan1_w09_1", "textures/makkon_horror/bek_tan1_w09_1"},
	{"textures/bek_tan1_w10_1", "textures/makkon_horror/bek_tan1_w10_1"},
	{"textures/bek_tan1_w11_1", "textures/makkon_horror/bek_tan1_w11_1"},
	{"textures/bek_tan1_w12_1", "textures/makkon_horror/bek_tan1_w12_1"},
	{"textures/bek_tan1_w12_2", "textures/makkon_horror/bek_tan1_w12_2"},
	{"textures/bek_tan1_w12_3", "textures/makkon_horror/bek_tan1_w12_3"},
	{"textures/bek_teal1_t01_1", "textures/makkon_horror/bek_teal1_t01_1"},
	{"textures/bek_teal1_t02_1", "textures/makkon_horror/bek_teal1_t02_1"},
	{"textures/bek_teal1_w01_1", "textures/makkon_horror/bek_teal1_w01_1"},
	{"textures/bek_teal1_w01_2", "textures/makkon_horror/bek_teal1_w01_2"},
	{"textures/bek_teal1_w01_3", "textures/makkon_horror/bek_teal1_w01_3"},
	{"textures/bek_teal1_w01_4", "textures/makkon_horror/bek_teal1_w01_4"},
	{"textures/bek_teal1_w02_1", "textures/makkon_horror/bek_teal1_w02_1"},
	{"textures/bek_teal1_w03_1", "textures/makkon_horror/bek_teal1_w03_1"},
	{"textures/bek_teal1_w03_2", "textures/makkon_horror/bek_teal1_w03_2"},
	{"textures/bek_teal1_w03_3", "textures/makkon_horror/bek_teal1_w03_3"},
	{"textures/bek_teal1_w03_4", "textures/makkon_horror/bek_teal1_w03_4"},
	{"textures/bek_teal1_w03_5", "textures/makkon_horror/bek_teal1_w03_5"},
	{"textures/bek_teal1_w03_6", "textures/makkon_horror/bek_teal1_w03_6"},
	{"textures/bek_teal1_w03_7", "textures/makkon_horror/bek_teal1_w03_7"},
	{"textures/bek_teal1_w03_8", "textures/makkon_horror/bek_teal1_w03_8"},
	{"textures/bek_teal1_w04_1", "textures/makkon_horror/bek_teal1_w04_1"},
	{"textures/bek_teal1_w05_1", "textures/makkon_horror/bek_teal1_w05_1"},
	{"textures/bek_teal1_w05_2", "textures/makkon_horror/bek_teal1_w05_2"},
	{"textures/bek_teal1_w05_3", "textures/makkon_horror/bek_teal1_w05_3"},
	{"textures/bek_teal1_w06_1", "textures/makkon_horror/bek_teal1_w06_1"},
	{"textures/bek_teal1_w06_2", "textures/makkon_horror/bek_teal1_w06_2"},
	{"textures/bek_teal1_w07_1", "textures/makkon_horror/bek_teal1_w07_1"},
	{"textures/bek_teal1_w07_2", "textures/makkon_horror/bek_teal1_w07_2"},
	{"textures/bek_teal1_w08_1", "textures/makkon_horror/bek_teal1_w08_1"},
	{"textures/bek_teal1_w09_1", "textures/makkon_horror/bek_teal1_w09_1"},
	{"textures/bek_teal1_w10_1", "textures/makkon_horror/bek_teal1_w10_1"},
	{"textures/bek_teal1_w11_1", "textures/makkon_horror/bek_teal1_w11_1"},
	{"textures/bek_teal1_w12_1", "textures/makkon_horror/bek_teal1_w12_1"},
	{"textures/bek_teal1_w12_2", "textures/makkon_horror/bek_teal1_w12_2"},
	{"textures/bek_teal1_w12_3", "textures/makkon_horror/bek_teal1_w12_3"},
	{"textures/bek_wht1_t01_1", "textures/makkon_horror/bek_wht1_t01_1"},
	{"textures/bek_wht1_t02_1", "textures/makkon_horror/bek_wht1_t02_1"},
	{"textures/bek_wht1_w01_1", "textures/makkon_horror/bek_wht1_w01_1"},
	{"textures/bek_wht1_w01_2", "textures/makkon_horror/bek_wht1_w01_2"},
	{"textures/bek_wht1_w01_3", "textures/makkon_horror/bek_wht1_w01_3"},
	{"textures/bek_wht1_w01_5", "textures/makkon_horror/bek_wht1_w01_5"},
	{"textures/bek_wht1_w02_1", "textures/makkon_horror/bek_wht1_w02_1"},
	{"textures/bek_wht1_w03_1", "textures/makkon_horror/bek_wht1_w03_1"},
	{"textures/bek_wht1_w03_2", "textures/makkon_horror/bek_wht1_w03_2"},
	{"textures/bek_wht1_w03_3", "textures/makkon_horror/bek_wht1_w03_3"},
	{"textures/bek_wht1_w03_4", "textures/makkon_horror/bek_wht1_w03_4"},
	{"textures/bek_wht1_w03_5", "textures/makkon_horror/bek_wht1_w03_5"},
	{"textures/bek_wht1_w03_6", "textures/makkon_horror/bek_wht1_w03_6"},
	{"textures/bek_wht1_w03_7", "textures/makkon_horror/bek_wht1_w03_7"},
	{"textures/bek_wht1_w03_8", "textures/makkon_horror/bek_wht1_w03_8"},
	{"textures/bek_wht1_w04_1", "textures/makkon_horror/bek_wht1_w04_1"},
	{"textures/bek_wht1_w05_1", "textures/makkon_horror/bek_wht1_w05_1"},
	{"textures/bek_wht1_w05_2", "textures/makkon_horror/bek_wht1_w05_2"},
	{"textures/bek_wht1_w05_3", "textures/makkon_horror/bek_wht1_w05_3"},
	{"textures/bek_wht1_w06_1", "textures/makkon_horror/bek_wht1_w06_1"},
	{"textures/bek_wht1_w06_2", "textures/makkon_horror/bek_wht1_w06_2"},
	{"textures/bek_wht1_w07_1", "textures/makkon_horror/bek_wht1_w07_1"},
	{"textures/bek_wht1_w07_2", "textures/makkon_horror/bek_wht1_w07_2"},
	{"textures/bek_wht1_w08_1", "textures/makkon_horror/bek_wht1_w08_1"},
	{"textures/bek_wht1_w09_1", "textures/makkon_horror/bek_wht1_w09_1"},
	{"textures/bek_wht1_w10_1", "textures/makkon_horror/bek_wht1_w10_1"},
	{"textures/bek_wht1_w11_1", "textures/makkon_horror/bek_wht1_w11_1"},
	{"textures/bek_wht1_w12_1", "textures/makkon_horror/bek_wht1_w12_1"},
	{"textures/bek_wht1_w12_2", "textures/makkon_horror/bek_wht1_w12_2"},
	{"textures/bek_wht1_w12_3", "textures/makkon_horror/bek_wht1_w12_3"},
	{"textures/_bekblck1_w01_1", "textures/makkon_horror/_bekblck1_w01_1"},
	{"textures/_bekblck1_w02_1", "textures/makkon_horror/_bekblck1_w02_1"},
	{"textures/_bekblck1_w02_2", "textures/makkon_horror/_bekblck1_w02_2"},
	{"textures/_bekblck1_w04_1", "textures/makkon_horror/_bekblck1_w04_1"},
	{"textures/_bekblck1_w04_2", "textures/makkon_horror/_bekblck1_w04_2"},
	{"textures/_bekblck1_w04_3", "textures/makkon_horror/_bekblck1_w04_3"},
	{"textures/_bekblck1_w05_1", "textures/makkon_horror/_bekblck1_w05_1"},
	{"textures/_bekblck1_w05_2", "textures/makkon_horror/_bekblck1_w05_2"},
	{"textures/_bekblck1_w05_3", "textures/makkon_horror/_bekblck1_w05_3"},
	{"textures/_bekblck_t02_1", "textures/makkon_horror/_bekblck_t02_1"},
	{"textures/_bekblue1_t02_1", "textures/makkon_horror/_bekblue1_t02_1"},
	{"textures/_bekblue1_w01_1", "textures/makkon_horror/_bekblue1_w01_1"},
	{"textures/_bekblue1_w02_1", "textures/makkon_horror/_bekblue1_w02_1"},
	{"textures/_bekblue1_w02_2", "textures/makkon_horror/_bekblue1_w02_2"},
	{"textures/_bekblue1_w04_1", "textures/makkon_horror/_bekblue1_w04_1"},
	{"textures/_bekblue1_w04_2", "textures/makkon_horror/_bekblue1_w04_2"},
	{"textures/_bekblue1_w04_3", "textures/makkon_horror/_bekblue1_w04_3"},
	{"textures/_bekblue1_w05_1", "textures/makkon_horror/_bekblue1_w05_1"},
	{"textures/_bekblue1_w05_2", "textures/makkon_horror/_bekblue1_w05_2"},
	{"textures/_bekblue1_w05_3", "textures/makkon_horror/_bekblue1_w05_3"},
	{"textures/_bekbrwn1_t02_1", "textures/makkon_horror/_bekbrwn1_t02_1"},
	{"textures/_bekbrwn1_w01_1", "textures/makkon_horror/_bekbrwn1_w01_1"},
	{"textures/_bekbrwn1_w02_1", "textures/makkon_horror/_bekbrwn1_w02_1"},
	{"textures/_bekbrwn1_w02_2", "textures/makkon_horror/_bekbrwn1_w02_2"},
	{"textures/_bekbrwn1_w04_1", "textures/makkon_horror/_bekbrwn1_w04_1"},
	{"textures/_bekbrwn1_w04_2", "textures/makkon_horror/_bekbrwn1_w04_2"},
	{"textures/_bekbrwn1_w04_3", "textures/makkon_horror/_bekbrwn1_w04_3"},
	{"textures/_bekbrwn1_w04_4", "textures/makkon_horror/_bekbrwn1_w04_4"},
	{"textures/_bekbrwn1_w04_5", "textures/makkon_horror/_bekbrwn1_w04_5"},
	{"textures/_bekbrwn1_w04_6", "textures/makkon_horror/_bekbrwn1_w04_6"},
	{"textures/_bekgrey1_t02_1", "textures/makkon_horror/_bekgrey1_t02_1"},
	{"textures/_bekgrey1_w01_1", "textures/makkon_horror/_bekgrey1_w01_1"},
	{"textures/_bekgrey1_w02_1", "textures/makkon_horror/_bekgrey1_w02_1"},
	{"textures/_bekgrey1_w02_2", "textures/makkon_horror/_bekgrey1_w02_2"},
	{"textures/_bekgrey1_w04_1", "textures/makkon_horror/_bekgrey1_w04_1"},
	{"textures/_bekgrey1_w04_2", "textures/makkon_horror/_bekgrey1_w04_2"},
	{"textures/_bekgrey1_w04_3", "textures/makkon_horror/_bekgrey1_w04_3"},
	{"textures/_bekgrey1_w05_1", "textures/makkon_horror/_bekgrey1_w05_1"},
	{"textures/_bekgrey1_w05_2", "textures/makkon_horror/_bekgrey1_w05_2"},
	{"textures/_bekgrey1_w05_3", "textures/makkon_horror/_bekgrey1_w05_3"},
	{"textures/_bekmeat1_t02_1", "textures/makkon_horror/_bekmeat1_t02_1"},
	{"textures/_bekmeat1_w01_1", "textures/makkon_horror/_bekmeat1_w01_1"},
	{"textures/_bekmeat1_w02_1", "textures/makkon_horror/_bekmeat1_w02_1"},
	{"textures/_bekmeat1_w04_1", "textures/makkon_horror/_bekmeat1_w04_1"},
	{"textures/_bekmeat1_w04_2", "textures/makkon_horror/_bekmeat1_w04_2"},
	{"textures/_bekmeat1_w04_3", "textures/makkon_horror/_bekmeat1_w04_3"},
	{"textures/_bekmeat1_w05_1", "textures/makkon_horror/_bekmeat1_w05_1"},
	{"textures/_bekmeat1_w05_2", "textures/makkon_horror/_bekmeat1_w05_2"},
	{"textures/_bekmeat1_w05_3", "textures/makkon_horror/_bekmeat1_w05_3"},
	{"textures/_bekpink1_t02_1", "textures/makkon_horror/_bekpink1_t02_1"},
	{"textures/_bekpink1_w01_1", "textures/makkon_horror/_bekpink1_w01_1"},
	{"textures/_bekpink1_w02_1", "textures/makkon_horror/_bekpink1_w02_1"},
	{"textures/_bekpink1_w02_2", "textures/makkon_horror/_bekpink1_w02_2"},
	{"textures/_bekpink1_w04_1", "textures/makkon_horror/_bekpink1_w04_1"},
	{"textures/_bekpink1_w04_2", "textures/makkon_horror/_bekpink1_w04_2"},
	{"textures/_bekpink1_w04_3", "textures/makkon_horror/_bekpink1_w04_3"},
	{"textures/_bekpink1_w05_1", "textures/makkon_horror/_bekpink1_w05_1"},
	{"textures/_bekpink1_w05_2", "textures/makkon_horror/_bekpink1_w05_2"},
	{"textures/_bekpink1_w05_3", "textures/makkon_horror/_bekpink1_w05_3"},
	{"textures/_bekprpl1_t02_1", "textures/makkon_horror/_bekprpl1_t02_1"},
	{"textures/_bekprpl1_w01_1", "textures/makkon_horror/_bekprpl1_w01_1"},
	{"textures/_bekprpl1_w02_1", "textures/makkon_horror/_bekprpl1_w02_1"},
	{"textures/_bekprpl1_w02_2", "textures/makkon_horror/_bekprpl1_w02_2"},
	{"textures/_bekprpl1_w04_1", "textures/makkon_horror/_bekprpl1_w04_1"},
	{"textures/_bekprpl1_w04_2", "textures/makkon_horror/_bekprpl1_w04_2"},
	{"textures/_bekprpl1_w04_3", "textures/makkon_horror/_bekprpl1_w04_3"},
	{"textures/_bekprpl1_w05_1", "textures/makkon_horror/_bekprpl1_w05_1"},
	{"textures/_bekprpl1_w05_2", "textures/makkon_horror/_bekprpl1_w05_2"},
	{"textures/_bekprpl1_w05_3", "textures/makkon_horror/_bekprpl1_w05_3"},
	{"textures/_bekred1_t02_1", "textures/makkon_horror/_bekred1_t02_1"},
	{"textures/_bekred1_w01_1", "textures/makkon_horror/_bekred1_w01_1"},
	{"textures/_bekred1_w02_1", "textures/makkon_horror/_bekred1_w02_1"},
	{"textures/_bekred1_w02_2", "textures/makkon_horror/_bekred1_w02_2"},
	{"textures/_bekred1_w04_1", "textures/makkon_horror/_bekred1_w04_1"},
	{"textures/_bekred1_w04_2", "textures/makkon_horror/_bekred1_w04_2"},
	{"textures/_bekred1_w04_3", "textures/makkon_horror/_bekred1_w04_3"},
	{"textures/_bekred1_w05_1", "textures/makkon_horror/_bekred1_w05_1"},
	{"textures/_bekred1_w05_2", "textures/makkon_horror/_bekred1_w05_2"},
	{"textures/_bekred1_w05_3", "textures/makkon_horror/_bekred1_w05_3"},
	{"textures/_bekred2_t02_1", "textures/makkon_horror/_bekred2_t02_1"},
	{"textures/_bekred2_w01_1", "textures/makkon_horror/_bekred2_w01_1"},
	{"textures/_bekred2_w02_1", "textures/makkon_horror/_bekred2_w02_1"},
	{"textures/_bekred2_w02_2", "textures/makkon_horror/_bekred2_w02_2"},
	{"textures/_bekred2_w04_1", "textures/makkon_horror/_bekred2_w04_1"},
	{"textures/_bekred2_w04_2", "textures/makkon_horror/_bekred2_w04_2"},
	{"textures/_bekred2_w04_3", "textures/makkon_horror/_bekred2_w04_3"},
	{"textures/_bekred2_w05_1", "textures/makkon_horror/_bekred2_w05_1"},
	{"textures/_bekred2_w05_2", "textures/makkon_horror/_bekred2_w05_2"},
	{"textures/_bekred2_w05_3", "textures/makkon_horror/_bekred2_w05_3"},
	{"textures/_bektan1_t02_1", "textures/makkon_horror/_bektan1_t02_1"},
	{"textures/_bektan1_w01_1", "textures/makkon_horror/_bektan1_w01_1"},
	{"textures/_bektan1_w02_1", "textures/makkon_horror/_bektan1_w02_1"},
	{"textures/_bektan1_w02_2", "textures/makkon_horror/_bektan1_w02_2"},
	{"textures/_bektan1_w04_1", "textures/makkon_horror/_bektan1_w04_1"},
	{"textures/_bektan1_w04_2", "textures/makkon_horror/_bektan1_w04_2"},
	{"textures/_bektan1_w04_3", "textures/makkon_horror/_bektan1_w04_3"},
	{"textures/_bektan1_w05_1", "textures/makkon_horror/_bektan1_w05_1"},
	{"textures/_bektan1_w05_2", "textures/makkon_horror/_bektan1_w05_2"},
	{"textures/_bektan1_w05_3", "textures/makkon_horror/_bektan1_w05_3"},
	{"textures/_bekteal1_t02_1", "textures/makkon_horror/_bekteal1_t02_1"},
	{"textures/_bekteal1_w01_1", "textures/makkon_horror/_bekteal1_w01_1"},
	{"textures/_bekteal1_w02_1", "textures/makkon_horror/_bekteal1_w02_1"},
	{"textures/_bekteal1_w02_2", "textures/makkon_horror/_bekteal1_w02_2"},
	{"textures/_bekteal1_w04_1", "textures/makkon_horror/_bekteal1_w04_1"},
	{"textures/_bekteal1_w04_2", "textures/makkon_horror/_bekteal1_w04_2"},
	{"textures/_bekteal1_w04_3", "textures/makkon_horror/_bekteal1_w04_3"},
	{"textures/_bekteal1_w05_1", "textures/makkon_horror/_bekteal1_w05_1"},
	{"textures/_bekteal1_w05_2", "textures/makkon_horror/_bekteal1_w05_2"},
	{"textures/_bekteal1_w05_3", "textures/makkon_horror/_bekteal1_w05_3"},
	{"textures/_bekwht1_t02_1", "textures/makkon_horror/_bekwht1_t02_1"},
	{"textures/_bekwht1_w01_1", "textures/makkon_horror/_bekwht1_w01_1"},
	{"textures/_bekwht1_w02_1", "textures/makkon_horror/_bekwht1_w02_1"},
	{"textures/_bekwht1_w02_2", "textures/makkon_horror/_bekwht1_w02_2"},
	{"textures/_bekwht1_w04_1", "textures/makkon_horror/_bekwht1_w04_1"},
	{"textures/_bekwht1_w04_2", "textures/makkon_horror/_bekwht1_w04_2"},
	{"textures/_bekwht1_w04_3", "textures/makkon_horror/_bekwht1_w04_3"},
	{"textures/_bekwht1_w05_1", "textures/makkon_horror/_bekwht1_w05_1"},
	{"textures/_bekwht1_w05_2", "textures/makkon_horror/_bekwht1_w05_2"},
	{"textures/_bekwht1_w05_3", "textures/makkon_horror/_bekwht1_w05_3"},

	// Makkon marble 2022.8
	{"textures/__0mrbl_b1_1", "textures/makkon_marble/__0mrbl_b1_1"},
	{"textures/__0mrbl_b1_2", "textures/makkon_marble/__0mrbl_b1_2"},
	{"textures/__0mrbl_b1_3", "textures/makkon_marble/__0mrbl_b1_3"},
	{"textures/__0mrbl_b1_4", "textures/makkon_marble/__0mrbl_b1_4"},
	{"textures/__0mrbl_b1_5", "textures/makkon_marble/__0mrbl_b1_5"},
	{"textures/__0mrbl_b1_6", "textures/makkon_marble/__0mrbl_b1_6"},
	{"textures/__0mrbl_b1_7", "textures/makkon_marble/__0mrbl_b1_7"},
	{"textures/__0mrbl_b1_8", "textures/makkon_marble/__0mrbl_b1_8"},
	{"textures/__amrbl_b1_1", "textures/makkon_marble/__amrbl_b1_1"},
	{"textures/__amrbl_b1_2", "textures/makkon_marble/__amrbl_b1_2"},
	{"textures/__amrbl_b1_3", "textures/makkon_marble/__amrbl_b1_3"},
	{"textures/__amrbl_b1_4", "textures/makkon_marble/__amrbl_b1_4"},
	{"textures/__amrbl_b1_5", "textures/makkon_marble/__amrbl_b1_5"},
	{"textures/__amrbl_b1_6", "textures/makkon_marble/__amrbl_b1_6"},
	{"textures/__amrbl_b1_7", "textures/makkon_marble/__amrbl_b1_7"},
	{"textures/__amrbl_b1_8", "textures/makkon_marble/__amrbl_b1_8"},
	{"textures/mrbl_c1_1", "textures/makkon_marble/mrbl_c1_1"},
	{"textures/mrbl_c1_2", "textures/makkon_marble/mrbl_c1_2"},
	{"textures/mrbl_c1_3", "textures/makkon_marble/mrbl_c1_3"},
	{"textures/mrbl_c1_4", "textures/makkon_marble/mrbl_c1_4"},
	{"textures/mrbl_c1_5", "textures/makkon_marble/mrbl_c1_5"},
	{"textures/mrbl_c1_6", "textures/makkon_marble/mrbl_c1_6"},
	{"textures/mrbl_c1_7", "textures/makkon_marble/mrbl_c1_7"},
	{"textures/mrbl_c1_8", "textures/makkon_marble/mrbl_c1_8"},
	{"textures/mrbl_c2_1", "textures/makkon_marble/mrbl_c2_1"},
	{"textures/mrbl_c2_2", "textures/makkon_marble/mrbl_c2_2"},
	{"textures/mrbl_c2_3", "textures/makkon_marble/mrbl_c2_3"},
	{"textures/mrbl_c2_4", "textures/makkon_marble/mrbl_c2_4"},
	{"textures/mrbl_c2_5", "textures/makkon_marble/mrbl_c2_5"},
	{"textures/mrbl_c2_6", "textures/makkon_marble/mrbl_c2_6"},
	{"textures/mrbl_c2_7", "textures/makkon_marble/mrbl_c2_7"},
	{"textures/mrbl_c2_8", "textures/makkon_marble/mrbl_c2_8"},
	{"textures/mrbl_c3_1", "textures/makkon_marble/mrbl_c3_1"},
	{"textures/mrbl_c3_2", "textures/makkon_marble/mrbl_c3_2"},
	{"textures/mrbl_c3_3", "textures/makkon_marble/mrbl_c3_3"},
	{"textures/mrbl_c3_4", "textures/makkon_marble/mrbl_c3_4"},
	{"textures/mrbl_c3_5", "textures/makkon_marble/mrbl_c3_5"},
	{"textures/mrbl_c3_6", "textures/makkon_marble/mrbl_c3_6"},
	{"textures/mrbl_c3_7", "textures/makkon_marble/mrbl_c3_7"},
	{"textures/mrbl_c3_8", "textures/makkon_marble/mrbl_c3_8"},
	{"textures/mrbl_fp1_1", "textures/makkon_marble/mrbl_fp1_1"},
	{"textures/mrbl_fp1_2", "textures/makkon_marble/mrbl_fp1_2"},
	{"textures/mrbl_fp1_3", "textures/makkon_marble/mrbl_fp1_3"},
	{"textures/mrbl_fp1_4", "textures/makkon_marble/mrbl_fp1_4"},
	{"textures/mrbl_fp1_5", "textures/makkon_marble/mrbl_fp1_5"},
	{"textures/mrbl_fp1_6", "textures/makkon_marble/mrbl_fp1_6"},
	{"textures/mrbl_fp1_7", "textures/makkon_marble/mrbl_fp1_7"},
	{"textures/mrbl_fp1_8", "textures/makkon_marble/mrbl_fp1_8"},
	{"textures/mrbl_fp2_1", "textures/makkon_marble/mrbl_fp2_1"},
	{"textures/mrbl_fp2_10", "textures/makkon_marble/mrbl_fp2_10"},
	{"textures/mrbl_fp2_11", "textures/makkon_marble/mrbl_fp2_11"},
	{"textures/mrbl_fp2_12", "textures/makkon_marble/mrbl_fp2_12"},
	{"textures/mrbl_fp2_13", "textures/makkon_marble/mrbl_fp2_13"},
	{"textures/mrbl_fp2_14", "textures/makkon_marble/mrbl_fp2_14"},
	{"textures/mrbl_fp2_15", "textures/makkon_marble/mrbl_fp2_15"},
	{"textures/mrbl_fp2_16", "textures/makkon_marble/mrbl_fp2_16"},
	{"textures/mrbl_fp2_17", "textures/makkon_marble/mrbl_fp2_17"},
	{"textures/mrbl_fp2_18", "textures/makkon_marble/mrbl_fp2_18"},
	{"textures/mrbl_fp2_19", "textures/makkon_marble/mrbl_fp2_19"},
	{"textures/mrbl_fp2_2", "textures/makkon_marble/mrbl_fp2_2"},
	{"textures/mrbl_fp2_20", "textures/makkon_marble/mrbl_fp2_20"},
	{"textures/mrbl_fp2_21", "textures/makkon_marble/mrbl_fp2_21"},
	{"textures/mrbl_fp2_3", "textures/makkon_marble/mrbl_fp2_3"},
	{"textures/mrbl_fp2_4", "textures/makkon_marble/mrbl_fp2_4"},
	{"textures/mrbl_fp2_5", "textures/makkon_marble/mrbl_fp2_5"},
	{"textures/mrbl_fp2_6", "textures/makkon_marble/mrbl_fp2_6"},
	{"textures/mrbl_fp2_7", "textures/makkon_marble/mrbl_fp2_7"},
	{"textures/mrbl_fp2_8", "textures/makkon_marble/mrbl_fp2_8"},
	{"textures/mrbl_fp2_9", "textures/makkon_marble/mrbl_fp2_9"},
	{"textures/mrbl_fp3_1", "textures/makkon_marble/mrbl_fp3_1"},
	{"textures/mrbl_fp3_2", "textures/makkon_marble/mrbl_fp3_2"},
	{"textures/mrbl_fp3_3", "textures/makkon_marble/mrbl_fp3_3"},
	{"textures/mrbl_fp3_4", "textures/makkon_marble/mrbl_fp3_4"},
	{"textures/mrbl_fp3_5", "textures/makkon_marble/mrbl_fp3_5"},
	{"textures/mrbl_fp3_6", "textures/makkon_marble/mrbl_fp3_6"},
	{"textures/mrbl_fp3_7", "textures/makkon_marble/mrbl_fp3_7"},
	{"textures/mrbl_fp3_8", "textures/makkon_marble/mrbl_fp3_8"},
	{"textures/mrbl_fp_1", "textures/makkon_marble/mrbl_fp_1"},
	{"textures/mrbl_fp_10", "textures/makkon_marble/mrbl_fp_10"},
	{"textures/mrbl_fp_11", "textures/makkon_marble/mrbl_fp_11"},
	{"textures/mrbl_fp_12", "textures/makkon_marble/mrbl_fp_12"},
	{"textures/mrbl_fp_13", "textures/makkon_marble/mrbl_fp_13"},
	{"textures/mrbl_fp_14", "textures/makkon_marble/mrbl_fp_14"},
	{"textures/mrbl_fp_15", "textures/makkon_marble/mrbl_fp_15"},
	{"textures/mrbl_fp_16", "textures/makkon_marble/mrbl_fp_16"},
	{"textures/mrbl_fp_17", "textures/makkon_marble/mrbl_fp_17"},
	{"textures/mrbl_fp_18", "textures/makkon_marble/mrbl_fp_18"},
	{"textures/mrbl_fp_19", "textures/makkon_marble/mrbl_fp_19"},
	{"textures/mrbl_fp_2", "textures/makkon_marble/mrbl_fp_2"},
	{"textures/mrbl_fp_20", "textures/makkon_marble/mrbl_fp_20"},
	{"textures/mrbl_fp_21", "textures/makkon_marble/mrbl_fp_21"},
	{"textures/mrbl_fp_22", "textures/makkon_marble/mrbl_fp_22"},
	{"textures/mrbl_fp_23", "textures/makkon_marble/mrbl_fp_23"},
	{"textures/mrbl_fp_24", "textures/makkon_marble/mrbl_fp_24"},
	{"textures/mrbl_fp_25", "textures/makkon_marble/mrbl_fp_25"},
	{"textures/mrbl_fp_26", "textures/makkon_marble/mrbl_fp_26"},
	{"textures/mrbl_fp_27", "textures/makkon_marble/mrbl_fp_27"},
	{"textures/mrbl_fp_28", "textures/makkon_marble/mrbl_fp_28"},
	{"textures/mrbl_fp_29", "textures/makkon_marble/mrbl_fp_29"},
	{"textures/mrbl_fp_3", "textures/makkon_marble/mrbl_fp_3"},
	{"textures/mrbl_fp_4", "textures/makkon_marble/mrbl_fp_4"},
	{"textures/mrbl_fp_5", "textures/makkon_marble/mrbl_fp_5"},
	{"textures/mrbl_fp_6", "textures/makkon_marble/mrbl_fp_6"},
	{"textures/mrbl_fp_7", "textures/makkon_marble/mrbl_fp_7"},
	{"textures/mrbl_fp_8", "textures/makkon_marble/mrbl_fp_8"},
	{"textures/mrbl_fp_9", "textures/makkon_marble/mrbl_fp_9"},
	{"textures/mrbl_orn1_1", "textures/makkon_marble/mrbl_orn1_1"},
	{"textures/mrbl_orn1_10", "textures/makkon_marble/mrbl_orn1_10"},
	{"textures/mrbl_orn1_11", "textures/makkon_marble/mrbl_orn1_11"},
	{"textures/mrbl_orn1_12", "textures/makkon_marble/mrbl_orn1_12"},
	{"textures/mrbl_orn1_13", "textures/makkon_marble/mrbl_orn1_13"},
	{"textures/mrbl_orn1_14", "textures/makkon_marble/mrbl_orn1_14"},
	{"textures/mrbl_orn1_15", "textures/makkon_marble/mrbl_orn1_15"},
	{"textures/mrbl_orn1_16", "textures/makkon_marble/mrbl_orn1_16"},
	{"textures/mrbl_orn1_17", "textures/makkon_marble/mrbl_orn1_17"},
	{"textures/mrbl_orn1_18", "textures/makkon_marble/mrbl_orn1_18"},
	{"textures/mrbl_orn1_19", "textures/makkon_marble/mrbl_orn1_19"},
	{"textures/mrbl_orn1_2", "textures/makkon_marble/mrbl_orn1_2"},
	{"textures/mrbl_orn1_20", "textures/makkon_marble/mrbl_orn1_20"},
	{"textures/mrbl_orn1_21", "textures/makkon_marble/mrbl_orn1_21"},
	{"textures/mrbl_orn1_22", "textures/makkon_marble/mrbl_orn1_22"},
	{"textures/mrbl_orn1_23", "textures/makkon_marble/mrbl_orn1_23"},
	{"textures/mrbl_orn1_24", "textures/makkon_marble/mrbl_orn1_24"},
	{"textures/mrbl_orn1_25", "textures/makkon_marble/mrbl_orn1_25"},
	{"textures/mrbl_orn1_26", "textures/makkon_marble/mrbl_orn1_26"},
	{"textures/mrbl_orn1_27", "textures/makkon_marble/mrbl_orn1_27"},
	{"textures/mrbl_orn1_28", "textures/makkon_marble/mrbl_orn1_28"},
	{"textures/mrbl_orn1_29", "textures/makkon_marble/mrbl_orn1_29"},
	{"textures/mrbl_orn1_3", "textures/makkon_marble/mrbl_orn1_3"},
	{"textures/mrbl_orn1_30", "textures/makkon_marble/mrbl_orn1_30"},
	{"textures/mrbl_orn1_31", "textures/makkon_marble/mrbl_orn1_31"},
	{"textures/mrbl_orn1_4", "textures/makkon_marble/mrbl_orn1_4"},
	{"textures/mrbl_orn1_5", "textures/makkon_marble/mrbl_orn1_5"},
	{"textures/mrbl_orn1_6", "textures/makkon_marble/mrbl_orn1_6"},
	{"textures/mrbl_orn1_7", "textures/makkon_marble/mrbl_orn1_7"},
	{"textures/mrbl_orn1_8", "textures/makkon_marble/mrbl_orn1_8"},
	{"textures/mrbl_orn1_9", "textures/makkon_marble/mrbl_orn1_9"},
	{"textures/mrbl_orn2_1", "textures/makkon_marble/mrbl_orn2_1"},
	{"textures/mrbl_orn2_10", "textures/makkon_marble/mrbl_orn2_10"},
	{"textures/mrbl_orn2_11", "textures/makkon_marble/mrbl_orn2_11"},
	{"textures/mrbl_orn2_12", "textures/makkon_marble/mrbl_orn2_12"},
	{"textures/mrbl_orn2_13", "textures/makkon_marble/mrbl_orn2_13"},
	{"textures/mrbl_orn2_14", "textures/makkon_marble/mrbl_orn2_14"},
	{"textures/mrbl_orn2_15", "textures/makkon_marble/mrbl_orn2_15"},
	{"textures/mrbl_orn2_16", "textures/makkon_marble/mrbl_orn2_16"},
	{"textures/mrbl_orn2_17", "textures/makkon_marble/mrbl_orn2_17"},
	{"textures/mrbl_orn2_18", "textures/makkon_marble/mrbl_orn2_18"},
	{"textures/mrbl_orn2_19", "textures/makkon_marble/mrbl_orn2_19"},
	{"textures/mrbl_orn2_2", "textures/makkon_marble/mrbl_orn2_2"},
	{"textures/mrbl_orn2_20", "textures/makkon_marble/mrbl_orn2_20"},
	{"textures/mrbl_orn2_21", "textures/makkon_marble/mrbl_orn2_21"},
	{"textures/mrbl_orn2_3", "textures/makkon_marble/mrbl_orn2_3"},
	{"textures/mrbl_orn2_4", "textures/makkon_marble/mrbl_orn2_4"},
	{"textures/mrbl_orn2_5", "textures/makkon_marble/mrbl_orn2_5"},
	{"textures/mrbl_orn2_6", "textures/makkon_marble/mrbl_orn2_6"},
	{"textures/mrbl_orn2_7", "textures/makkon_marble/mrbl_orn2_7"},
	{"textures/mrbl_orn2_8", "textures/makkon_marble/mrbl_orn2_8"},
	{"textures/mrbl_orn2_9", "textures/makkon_marble/mrbl_orn2_9"},
	{"textures/mrbl_t10_1", "textures/makkon_marble/mrbl_t10_1"},
	{"textures/mrbl_t10_2", "textures/makkon_marble/mrbl_t10_2"},
	{"textures/mrbl_t10_3", "textures/makkon_marble/mrbl_t10_3"},
	{"textures/mrbl_t10_4", "textures/makkon_marble/mrbl_t10_4"},
	{"textures/mrbl_t10_5", "textures/makkon_marble/mrbl_t10_5"},
	{"textures/mrbl_t10_6", "textures/makkon_marble/mrbl_t10_6"},
	{"textures/mrbl_t10_7", "textures/makkon_marble/mrbl_t10_7"},
	{"textures/mrbl_t10_8", "textures/makkon_marble/mrbl_t10_8"},
	{"textures/mrbl_t11_1", "textures/makkon_marble/mrbl_t11_1"},
	{"textures/mrbl_t11_2", "textures/makkon_marble/mrbl_t11_2"},
	{"textures/mrbl_t11_3", "textures/makkon_marble/mrbl_t11_3"},
	{"textures/mrbl_t11_4", "textures/makkon_marble/mrbl_t11_4"},
	{"textures/mrbl_t11_5", "textures/makkon_marble/mrbl_t11_5"},
	{"textures/mrbl_t11_6", "textures/makkon_marble/mrbl_t11_6"},
	{"textures/mrbl_t11_7", "textures/makkon_marble/mrbl_t11_7"},
	{"textures/mrbl_t11_8", "textures/makkon_marble/mrbl_t11_8"},
	{"textures/mrbl_t12_1", "textures/makkon_marble/mrbl_t12_1"},
	{"textures/mrbl_t12_2", "textures/makkon_marble/mrbl_t12_2"},
	{"textures/mrbl_t12_3", "textures/makkon_marble/mrbl_t12_3"},
	{"textures/mrbl_t12_4", "textures/makkon_marble/mrbl_t12_4"},
	{"textures/mrbl_t12_5", "textures/makkon_marble/mrbl_t12_5"},
	{"textures/mrbl_t12_6", "textures/makkon_marble/mrbl_t12_6"},
	{"textures/mrbl_t12_7", "textures/makkon_marble/mrbl_t12_7"},
	{"textures/mrbl_t12_8", "textures/makkon_marble/mrbl_t12_8"},
	{"textures/mrbl_t13_1", "textures/makkon_marble/mrbl_t13_1"},
	{"textures/mrbl_t13_2", "textures/makkon_marble/mrbl_t13_2"},
	{"textures/mrbl_t13_3", "textures/makkon_marble/mrbl_t13_3"},
	{"textures/mrbl_t13_4", "textures/makkon_marble/mrbl_t13_4"},
	{"textures/mrbl_t13_5", "textures/makkon_marble/mrbl_t13_5"},
	{"textures/mrbl_t13_6", "textures/makkon_marble/mrbl_t13_6"},
	{"textures/mrbl_t13_7", "textures/makkon_marble/mrbl_t13_7"},
	{"textures/mrbl_t13_8", "textures/makkon_marble/mrbl_t13_8"},
	{"textures/mrbl_t14_1", "textures/makkon_marble/mrbl_t14_1"},
	{"textures/mrbl_t14_2", "textures/makkon_marble/mrbl_t14_2"},
	{"textures/mrbl_t14_3", "textures/makkon_marble/mrbl_t14_3"},
	{"textures/mrbl_t14_4", "textures/makkon_marble/mrbl_t14_4"},
	{"textures/mrbl_t14_5", "textures/makkon_marble/mrbl_t14_5"},
	{"textures/mrbl_t14_6", "textures/makkon_marble/mrbl_t14_6"},
	{"textures/mrbl_t14_7", "textures/makkon_marble/mrbl_t14_7"},
	{"textures/mrbl_t14_8", "textures/makkon_marble/mrbl_t14_8"},
	{"textures/mrbl_t1_1", "textures/makkon_marble/mrbl_t1_1"},
	{"textures/mrbl_t1_2", "textures/makkon_marble/mrbl_t1_2"},
	{"textures/mrbl_t1_3", "textures/makkon_marble/mrbl_t1_3"},
	{"textures/mrbl_t1_4", "textures/makkon_marble/mrbl_t1_4"},
	{"textures/mrbl_t1_5", "textures/makkon_marble/mrbl_t1_5"},
	{"textures/mrbl_t1_6", "textures/makkon_marble/mrbl_t1_6"},
	{"textures/mrbl_t1_7", "textures/makkon_marble/mrbl_t1_7"},
	{"textures/mrbl_t1_8", "textures/makkon_marble/mrbl_t1_8"},
	{"textures/mrbl_t2_1", "textures/makkon_marble/mrbl_t2_1"},
	{"textures/mrbl_t2_2", "textures/makkon_marble/mrbl_t2_2"},
	{"textures/mrbl_t2_3", "textures/makkon_marble/mrbl_t2_3"},
	{"textures/mrbl_t2_4", "textures/makkon_marble/mrbl_t2_4"},
	{"textures/mrbl_t2_5", "textures/makkon_marble/mrbl_t2_5"},
	{"textures/mrbl_t2_6", "textures/makkon_marble/mrbl_t2_6"},
	{"textures/mrbl_t2_7", "textures/makkon_marble/mrbl_t2_7"},
	{"textures/mrbl_t2_8", "textures/makkon_marble/mrbl_t2_8"},
	{"textures/mrbl_t3_1", "textures/makkon_marble/mrbl_t3_1"},
	{"textures/mrbl_t3_2", "textures/makkon_marble/mrbl_t3_2"},
	{"textures/mrbl_t3_3", "textures/makkon_marble/mrbl_t3_3"},
	{"textures/mrbl_t3_4", "textures/makkon_marble/mrbl_t3_4"},
	{"textures/mrbl_t3_5", "textures/makkon_marble/mrbl_t3_5"},
	{"textures/mrbl_t3_6", "textures/makkon_marble/mrbl_t3_6"},
	{"textures/mrbl_t3_7", "textures/makkon_marble/mrbl_t3_7"},
	{"textures/mrbl_t3_8", "textures/makkon_marble/mrbl_t3_8"},
	{"textures/mrbl_t4_1", "textures/makkon_marble/mrbl_t4_1"},
	{"textures/mrbl_t4_2", "textures/makkon_marble/mrbl_t4_2"},
	{"textures/mrbl_t4_3", "textures/makkon_marble/mrbl_t4_3"},
	{"textures/mrbl_t4_4", "textures/makkon_marble/mrbl_t4_4"},
	{"textures/mrbl_t4_5", "textures/makkon_marble/mrbl_t4_5"},
	{"textures/mrbl_t4_6", "textures/makkon_marble/mrbl_t4_6"},
	{"textures/mrbl_t4_7", "textures/makkon_marble/mrbl_t4_7"},
	{"textures/mrbl_t4_8", "textures/makkon_marble/mrbl_t4_8"},
	{"textures/mrbl_t5_1", "textures/makkon_marble/mrbl_t5_1"},
	{"textures/mrbl_t5_2", "textures/makkon_marble/mrbl_t5_2"},
	{"textures/mrbl_t5_3", "textures/makkon_marble/mrbl_t5_3"},
	{"textures/mrbl_t5_4", "textures/makkon_marble/mrbl_t5_4"},
	{"textures/mrbl_t5_5", "textures/makkon_marble/mrbl_t5_5"},
	{"textures/mrbl_t5_6", "textures/makkon_marble/mrbl_t5_6"},
	{"textures/mrbl_t5_7", "textures/makkon_marble/mrbl_t5_7"},
	{"textures/mrbl_t5_8", "textures/makkon_marble/mrbl_t5_8"},
	{"textures/mrbl_t6_1", "textures/makkon_marble/mrbl_t6_1"},
	{"textures/mrbl_t6_2", "textures/makkon_marble/mrbl_t6_2"},
	{"textures/mrbl_t6_3", "textures/makkon_marble/mrbl_t6_3"},
	{"textures/mrbl_t6_4", "textures/makkon_marble/mrbl_t6_4"},
	{"textures/mrbl_t6_5", "textures/makkon_marble/mrbl_t6_5"},
	{"textures/mrbl_t6_6", "textures/makkon_marble/mrbl_t6_6"},
	{"textures/mrbl_t6_7", "textures/makkon_marble/mrbl_t6_7"},
	{"textures/mrbl_t6_8", "textures/makkon_marble/mrbl_t6_8"},
	{"textures/mrbl_t7_1", "textures/makkon_marble/mrbl_t7_1"},
	{"textures/mrbl_t7_2", "textures/makkon_marble/mrbl_t7_2"},
	{"textures/mrbl_t7_3", "textures/makkon_marble/mrbl_t7_3"},
	{"textures/mrbl_t7_4", "textures/makkon_marble/mrbl_t7_4"},
	{"textures/mrbl_t7_5", "textures/makkon_marble/mrbl_t7_5"},
	{"textures/mrbl_t7_6", "textures/makkon_marble/mrbl_t7_6"},
	{"textures/mrbl_t7_7", "textures/makkon_marble/mrbl_t7_7"},
	{"textures/mrbl_t7_8", "textures/makkon_marble/mrbl_t7_8"},
	{"textures/mrbl_t8_1", "textures/makkon_marble/mrbl_t8_1"},
	{"textures/mrbl_t8_2", "textures/makkon_marble/mrbl_t8_2"},
	{"textures/mrbl_t8_3", "textures/makkon_marble/mrbl_t8_3"},
	{"textures/mrbl_t8_4", "textures/makkon_marble/mrbl_t8_4"},
	{"textures/mrbl_t8_5", "textures/makkon_marble/mrbl_t8_5"},
	{"textures/mrbl_t8_6", "textures/makkon_marble/mrbl_t8_6"},
	{"textures/mrbl_t8_7", "textures/makkon_marble/mrbl_t8_7"},
	{"textures/mrbl_t8_8", "textures/makkon_marble/mrbl_t8_8"},
	{"textures/mrbl_t9_1", "textures/makkon_marble/mrbl_t9_1"},
	{"textures/mrbl_t9_2", "textures/makkon_marble/mrbl_t9_2"},
	{"textures/mrbl_t9_3", "textures/makkon_marble/mrbl_t9_3"},
	{"textures/mrbl_t9_4", "textures/makkon_marble/mrbl_t9_4"},
	{"textures/mrbl_t9_5", "textures/makkon_marble/mrbl_t9_5"},
	{"textures/mrbl_t9_6", "textures/makkon_marble/mrbl_t9_6"},
	{"textures/mrbl_t9_7", "textures/makkon_marble/mrbl_t9_7"},
	{"textures/mrbl_t9_8", "textures/makkon_marble/mrbl_t9_8"},
	{"textures/_mrbl_f1_1", "textures/makkon_marble/_mrbl_f1_1"},
	{"textures/_mrbl_f1_2", "textures/makkon_marble/_mrbl_f1_2"},
	{"textures/_mrbl_f1_3", "textures/makkon_marble/_mrbl_f1_3"},
	{"textures/_mrbl_f1_4", "textures/makkon_marble/_mrbl_f1_4"},
	{"textures/_mrbl_f1_5", "textures/makkon_marble/_mrbl_f1_5"},
	{"textures/_mrbl_f1_6", "textures/makkon_marble/_mrbl_f1_6"},
	{"textures/_mrbl_f1_7", "textures/makkon_marble/_mrbl_f1_7"},
	{"textures/_mrbl_f1_8", "textures/makkon_marble/_mrbl_f1_8"},
	{"textures/_mrbl_f2_1", "textures/makkon_marble/_mrbl_f2_1"},
	{"textures/_mrbl_f2_2", "textures/makkon_marble/_mrbl_f2_2"},
	{"textures/_mrbl_f2_3", "textures/makkon_marble/_mrbl_f2_3"},
	{"textures/_mrbl_f2_4", "textures/makkon_marble/_mrbl_f2_4"},
	{"textures/_mrbl_f2_5", "textures/makkon_marble/_mrbl_f2_5"},
	{"textures/_mrbl_f2_6", "textures/makkon_marble/_mrbl_f2_6"},
	{"textures/_mrbl_f2_7", "textures/makkon_marble/_mrbl_f2_7"},
	{"textures/_mrbl_f2_8", "textures/makkon_marble/_mrbl_f2_8"},

	// Makkon metal 2022.8
	{"textures/__0metal_brass1", "textures/makkon_marble/__0metal_brass1"},
	{"textures/__0metal_brass2", "textures/makkon_marble/__0metal_brass2"},
	{"textures/__0metal_brnz1", "textures/makkon_marble/__0metal_brnz1"},
	{"textures/__0metal_copp1", "textures/makkon_marble/__0metal_copp1"},
	{"textures/__0metal_gold1", "textures/makkon_marble/__0metal_gold1"},
	{"textures/__0metal_gold2", "textures/makkon_marble/__0metal_gold2"},
	{"textures/__0metal_iron1", "textures/makkon_marble/__0metal_iron1"},
	{"textures/__0metal_iron2", "textures/makkon_marble/__0metal_iron2"},
	{"textures/__0metal_iron3", "textures/makkon_marble/__0metal_iron3"},
	{"textures/__0metal_rust1", "textures/makkon_marble/__0metal_rust1"},
	{"textures/__0metal_slvr1", "textures/makkon_marble/__0metal_slvr1"},
	{"textures/__ametal_brass1", "textures/makkon_marble/__ametal_brass1"},
	{"textures/__ametal_brass2", "textures/makkon_marble/__ametal_brass2"},
	{"textures/__ametal_brnz1", "textures/makkon_marble/__ametal_brnz1"},
	{"textures/__ametal_copp1", "textures/makkon_marble/__ametal_copp1"},
	{"textures/__ametal_gold1", "textures/makkon_marble/__ametal_gold1"},
	{"textures/__ametal_gold2", "textures/makkon_marble/__ametal_gold2"},
	{"textures/__ametal_iron1", "textures/makkon_marble/__ametal_iron1"},
	{"textures/__ametal_iron2", "textures/makkon_marble/__ametal_iron2"},
	{"textures/__ametal_iron3", "textures/makkon_marble/__ametal_iron3"},
	{"textures/__ametal_rust1", "textures/makkon_marble/__ametal_rust1"},
	{"textures/__ametal_slvr1", "textures/makkon_marble/__ametal_slvr1"},
	{"textures/metal_bras1_t7", "textures/makkon_marble/metal_bras1_t7"},
	{"textures/metal_bras2_02", "textures/makkon_marble/metal_bras2_02"},
	{"textures/metal_bras2_03", "textures/makkon_marble/metal_bras2_03"},
	{"textures/metal_bras2_04", "textures/makkon_marble/metal_bras2_04"},
	{"textures/metal_bras2_05", "textures/makkon_marble/metal_bras2_05"},
	{"textures/metal_bras2_06", "textures/makkon_marble/metal_bras2_06"},
	{"textures/metal_bras2_t7", "textures/makkon_marble/metal_bras2_t7"},
	{"textures/metal_bras_02", "textures/makkon_marble/metal_bras_02"},
	{"textures/metal_bras_03", "textures/makkon_marble/metal_bras_03"},
	{"textures/metal_bras_04", "textures/makkon_marble/metal_bras_04"},
	{"textures/metal_bras_05", "textures/makkon_marble/metal_bras_05"},
	{"textures/metal_bras_06", "textures/makkon_marble/metal_bras_06"},
	{"textures/metal_brass2_01", "textures/makkon_marble/metal_brass2_01"},
	{"textures/metal_brass2_07", "textures/makkon_marble/metal_brass2_07"},
	{"textures/metal_brass2_08", "textures/makkon_marble/metal_brass2_08"},
	{"textures/metal_brass2_09", "textures/makkon_marble/metal_brass2_09"},
	{"textures/metal_brass2_c1", "textures/makkon_marble/metal_brass2_c1"},
	{"textures/metal_brass2_t1", "textures/makkon_marble/metal_brass2_t1"},
	{"textures/metal_brass2_t2", "textures/makkon_marble/metal_brass2_t2"},
	{"textures/metal_brass2_t3", "textures/makkon_marble/metal_brass2_t3"},
	{"textures/metal_brass2_t4", "textures/makkon_marble/metal_brass2_t4"},
	{"textures/metal_brass2_t5", "textures/makkon_marble/metal_brass2_t5"},
	{"textures/metal_brass2_t6", "textures/makkon_marble/metal_brass2_t6"},
	{"textures/metal_brass2_t7", "textures/makkon_marble/metal_brass2_t7"},
	{"textures/metal_brass2_t8", "textures/makkon_marble/metal_brass2_t8"},
	{"textures/metal_brass_01", "textures/makkon_marble/metal_brass_01"},
	{"textures/metal_brass_07", "textures/makkon_marble/metal_brass_07"},
	{"textures/metal_brass_08", "textures/makkon_marble/metal_brass_08"},
	{"textures/metal_brass_09", "textures/makkon_marble/metal_brass_09"},
	{"textures/metal_brass_c1", "textures/makkon_marble/metal_brass_c1"},
	{"textures/metal_brass_t1", "textures/makkon_marble/metal_brass_t1"},
	{"textures/metal_brass_t2", "textures/makkon_marble/metal_brass_t2"},
	{"textures/metal_brass_t3", "textures/makkon_marble/metal_brass_t3"},
	{"textures/metal_brass_t4", "textures/makkon_marble/metal_brass_t4"},
	{"textures/metal_brass_t5", "textures/makkon_marble/metal_brass_t5"},
	{"textures/metal_brass_t6", "textures/makkon_marble/metal_brass_t6"},
	{"textures/metal_brass_t7", "textures/makkon_marble/metal_brass_t7"},
	{"textures/metal_brass_t8", "textures/makkon_marble/metal_brass_t8"},
	{"textures/metal_brnz_01", "textures/makkon_marble/metal_brnz_01"},
	{"textures/metal_brnz_02", "textures/makkon_marble/metal_brnz_02"},
	{"textures/metal_brnz_03", "textures/makkon_marble/metal_brnz_03"},
	{"textures/metal_brnz_04", "textures/makkon_marble/metal_brnz_04"},
	{"textures/metal_brnz_05", "textures/makkon_marble/metal_brnz_05"},
	{"textures/metal_brnz_06", "textures/makkon_marble/metal_brnz_06"},
	{"textures/metal_brnz_07", "textures/makkon_marble/metal_brnz_07"},
	{"textures/metal_brnz_08", "textures/makkon_marble/metal_brnz_08"},
	{"textures/metal_brnz_09", "textures/makkon_marble/metal_brnz_09"},
	{"textures/metal_brnz_c1", "textures/makkon_marble/metal_brnz_c1"},
	{"textures/metal_brnz_t1", "textures/makkon_marble/metal_brnz_t1"},
	{"textures/metal_brnz_t2", "textures/makkon_marble/metal_brnz_t2"},
	{"textures/metal_brnz_t3", "textures/makkon_marble/metal_brnz_t3"},
	{"textures/metal_brnz_t4", "textures/makkon_marble/metal_brnz_t4"},
	{"textures/metal_brnz_t5", "textures/makkon_marble/metal_brnz_t5"},
	{"textures/metal_brnz_t6", "textures/makkon_marble/metal_brnz_t6"},
	{"textures/metal_brnz_t7", "textures/makkon_marble/metal_brnz_t7"},
	{"textures/metal_brnz_t8", "textures/makkon_marble/metal_brnz_t8"},
	{"textures/metal_copp_01", "textures/makkon_marble/metal_copp_01"},
	{"textures/metal_copp_02", "textures/makkon_marble/metal_copp_02"},
	{"textures/metal_copp_03", "textures/makkon_marble/metal_copp_03"},
	{"textures/metal_copp_04", "textures/makkon_marble/metal_copp_04"},
	{"textures/metal_copp_05", "textures/makkon_marble/metal_copp_05"},
	{"textures/metal_copp_06", "textures/makkon_marble/metal_copp_06"},
	{"textures/metal_copp_07", "textures/makkon_marble/metal_copp_07"},
	{"textures/metal_copp_08", "textures/makkon_marble/metal_copp_08"},
	{"textures/metal_copp_09", "textures/makkon_marble/metal_copp_09"},
	{"textures/metal_copp_c1", "textures/makkon_marble/metal_copp_c1"},
	{"textures/metal_copp_t1", "textures/makkon_marble/metal_copp_t1"},
	{"textures/metal_copp_t2", "textures/makkon_marble/metal_copp_t2"},
	{"textures/metal_copp_t3", "textures/makkon_marble/metal_copp_t3"},
	{"textures/metal_copp_t4", "textures/makkon_marble/metal_copp_t4"},
	{"textures/metal_copp_t5", "textures/makkon_marble/metal_copp_t5"},
	{"textures/metal_copp_t6", "textures/makkon_marble/metal_copp_t6"},
	{"textures/metal_copp_t7", "textures/makkon_marble/metal_copp_t7"},
	{"textures/metal_copp_t8", "textures/makkon_marble/metal_copp_t8"},
	{"textures/metal_gold1_t7", "textures/makkon_marble/metal_gold1_t7"},
	{"textures/metal_gold2_01", "textures/makkon_marble/metal_gold2_01"},
	{"textures/metal_gold2_02", "textures/makkon_marble/metal_gold2_02"},
	{"textures/metal_gold2_03", "textures/makkon_marble/metal_gold2_03"},
	{"textures/metal_gold2_04", "textures/makkon_marble/metal_gold2_04"},
	{"textures/metal_gold2_05", "textures/makkon_marble/metal_gold2_05"},
	{"textures/metal_gold2_06", "textures/makkon_marble/metal_gold2_06"},
	{"textures/metal_gold2_07", "textures/makkon_marble/metal_gold2_07"},
	{"textures/metal_gold2_08", "textures/makkon_marble/metal_gold2_08"},
	{"textures/metal_gold2_09", "textures/makkon_marble/metal_gold2_09"},
	{"textures/metal_gold2_c1", "textures/makkon_marble/metal_gold2_c1"},
	{"textures/metal_gold2_t1", "textures/makkon_marble/metal_gold2_t1"},
	{"textures/metal_gold2_t2", "textures/makkon_marble/metal_gold2_t2"},
	{"textures/metal_gold2_t3", "textures/makkon_marble/metal_gold2_t3"},
	{"textures/metal_gold2_t4", "textures/makkon_marble/metal_gold2_t4"},
	{"textures/metal_gold2_t5", "textures/makkon_marble/metal_gold2_t5"},
	{"textures/metal_gold2_t6", "textures/makkon_marble/metal_gold2_t6"},
	{"textures/metal_gold2_t7", "textures/makkon_marble/metal_gold2_t7"},
	{"textures/metal_gold2_t8", "textures/makkon_marble/metal_gold2_t8"},
	{"textures/metal_gold_01", "textures/makkon_marble/metal_gold_01"},
	{"textures/metal_gold_02", "textures/makkon_marble/metal_gold_02"},
	{"textures/metal_gold_03", "textures/makkon_marble/metal_gold_03"},
	{"textures/metal_gold_04", "textures/makkon_marble/metal_gold_04"},
	{"textures/metal_gold_05", "textures/makkon_marble/metal_gold_05"},
	{"textures/metal_gold_06", "textures/makkon_marble/metal_gold_06"},
	{"textures/metal_gold_07", "textures/makkon_marble/metal_gold_07"},
	{"textures/metal_gold_08", "textures/makkon_marble/metal_gold_08"},
	{"textures/metal_gold_09", "textures/makkon_marble/metal_gold_09"},
	{"textures/metal_gold_c1", "textures/makkon_marble/metal_gold_c1"},
	{"textures/metal_gold_t1", "textures/makkon_marble/metal_gold_t1"},
	{"textures/metal_gold_t2", "textures/makkon_marble/metal_gold_t2"},
	{"textures/metal_gold_t3", "textures/makkon_marble/metal_gold_t3"},
	{"textures/metal_gold_t4", "textures/makkon_marble/metal_gold_t4"},
	{"textures/metal_gold_t5", "textures/makkon_marble/metal_gold_t5"},
	{"textures/metal_gold_t6", "textures/makkon_marble/metal_gold_t6"},
	{"textures/metal_gold_t7", "textures/makkon_marble/metal_gold_t7"},
	{"textures/metal_gold_t8", "textures/makkon_marble/metal_gold_t8"},
	{"textures/metal_iron1_01", "textures/makkon_marble/metal_iron1_01"},
	{"textures/metal_iron1_02", "textures/makkon_marble/metal_iron1_02"},
	{"textures/metal_iron1_03", "textures/makkon_marble/metal_iron1_03"},
	{"textures/metal_iron1_04", "textures/makkon_marble/metal_iron1_04"},
	{"textures/metal_iron1_05", "textures/makkon_marble/metal_iron1_05"},
	{"textures/metal_iron1_06", "textures/makkon_marble/metal_iron1_06"},
	{"textures/metal_iron1_07", "textures/makkon_marble/metal_iron1_07"},
	{"textures/metal_iron1_08", "textures/makkon_marble/metal_iron1_08"},
	{"textures/metal_iron1_09", "textures/makkon_marble/metal_iron1_09"},
	{"textures/metal_iron1_c1", "textures/makkon_marble/metal_iron1_c1"},
	{"textures/metal_iron1_t1", "textures/makkon_marble/metal_iron1_t1"},
	{"textures/metal_iron1_t2", "textures/makkon_marble/metal_iron1_t2"},
	{"textures/metal_iron1_t3", "textures/makkon_marble/metal_iron1_t3"},
	{"textures/metal_iron1_t4", "textures/makkon_marble/metal_iron1_t4"},
	{"textures/metal_iron1_t5", "textures/makkon_marble/metal_iron1_t5"},
	{"textures/metal_iron1_t6", "textures/makkon_marble/metal_iron1_t6"},
	{"textures/metal_iron1_t7", "textures/makkon_marble/metal_iron1_t7"},
	{"textures/metal_iron1_t8", "textures/makkon_marble/metal_iron1_t8"},
	{"textures/metal_iron2_01", "textures/makkon_marble/metal_iron2_01"},
	{"textures/metal_iron2_02", "textures/makkon_marble/metal_iron2_02"},
	{"textures/metal_iron2_03", "textures/makkon_marble/metal_iron2_03"},
	{"textures/metal_iron2_04", "textures/makkon_marble/metal_iron2_04"},
	{"textures/metal_iron2_05", "textures/makkon_marble/metal_iron2_05"},
	{"textures/metal_iron2_06", "textures/makkon_marble/metal_iron2_06"},
	{"textures/metal_iron2_07", "textures/makkon_marble/metal_iron2_07"},
	{"textures/metal_iron2_08", "textures/makkon_marble/metal_iron2_08"},
	{"textures/metal_iron2_09", "textures/makkon_marble/metal_iron2_09"},
	{"textures/metal_iron2_c1", "textures/makkon_marble/metal_iron2_c1"},
	{"textures/metal_iron2_t1", "textures/makkon_marble/metal_iron2_t1"},
	{"textures/metal_iron2_t2", "textures/makkon_marble/metal_iron2_t2"},
	{"textures/metal_iron2_t3", "textures/makkon_marble/metal_iron2_t3"},
	{"textures/metal_iron2_t4", "textures/makkon_marble/metal_iron2_t4"},
	{"textures/metal_iron2_t5", "textures/makkon_marble/metal_iron2_t5"},
	{"textures/metal_iron2_t6", "textures/makkon_marble/metal_iron2_t6"},
	{"textures/metal_iron2_t7", "textures/makkon_marble/metal_iron2_t7"},
	{"textures/metal_iron2_t8", "textures/makkon_marble/metal_iron2_t8"},
	{"textures/metal_iron3_01", "textures/makkon_marble/metal_iron3_01"},
	{"textures/metal_iron3_02", "textures/makkon_marble/metal_iron3_02"},
	{"textures/metal_iron3_03", "textures/makkon_marble/metal_iron3_03"},
	{"textures/metal_iron3_04", "textures/makkon_marble/metal_iron3_04"},
	{"textures/metal_iron3_05", "textures/makkon_marble/metal_iron3_05"},
	{"textures/metal_iron3_06", "textures/makkon_marble/metal_iron3_06"},
	{"textures/metal_iron3_07", "textures/makkon_marble/metal_iron3_07"},
	{"textures/metal_iron3_08", "textures/makkon_marble/metal_iron3_08"},
	{"textures/metal_iron3_09", "textures/makkon_marble/metal_iron3_09"},
	{"textures/metal_iron3_c1", "textures/makkon_marble/metal_iron3_c1"},
	{"textures/metal_iron3_t1", "textures/makkon_marble/metal_iron3_t1"},
	{"textures/metal_iron3_t2", "textures/makkon_marble/metal_iron3_t2"},
	{"textures/metal_iron3_t3", "textures/makkon_marble/metal_iron3_t3"},
	{"textures/metal_iron3_t4", "textures/makkon_marble/metal_iron3_t4"},
	{"textures/metal_iron3_t5", "textures/makkon_marble/metal_iron3_t5"},
	{"textures/metal_iron3_t6", "textures/makkon_marble/metal_iron3_t6"},
	{"textures/metal_iron3_t7", "textures/makkon_marble/metal_iron3_t7"},
	{"textures/metal_iron3_t8", "textures/makkon_marble/metal_iron3_t8"},
	{"textures/metal_rust_01", "textures/makkon_marble/metal_rust_01"},
	{"textures/metal_rust_02", "textures/makkon_marble/metal_rust_02"},
	{"textures/metal_rust_03", "textures/makkon_marble/metal_rust_03"},
	{"textures/metal_rust_04", "textures/makkon_marble/metal_rust_04"},
	{"textures/metal_rust_05", "textures/makkon_marble/metal_rust_05"},
	{"textures/metal_rust_06", "textures/makkon_marble/metal_rust_06"},
	{"textures/metal_rust_07", "textures/makkon_marble/metal_rust_07"},
	{"textures/metal_rust_08", "textures/makkon_marble/metal_rust_08"},
	{"textures/metal_rust_09", "textures/makkon_marble/metal_rust_09"},
	{"textures/metal_rust_c1", "textures/makkon_marble/metal_rust_c1"},
	{"textures/metal_rust_t1", "textures/makkon_marble/metal_rust_t1"},
	{"textures/metal_rust_t2", "textures/makkon_marble/metal_rust_t2"},
	{"textures/metal_rust_t3", "textures/makkon_marble/metal_rust_t3"},
	{"textures/metal_rust_t4", "textures/makkon_marble/metal_rust_t4"},
	{"textures/metal_rust_t5", "textures/makkon_marble/metal_rust_t5"},
	{"textures/metal_rust_t6", "textures/makkon_marble/metal_rust_t6"},
	{"textures/metal_rust_t7", "textures/makkon_marble/metal_rust_t7"},
	{"textures/metal_rust_t8", "textures/makkon_marble/metal_rust_t8"},
	{"textures/metal_slvr_01", "textures/makkon_marble/metal_slvr_01"},
	{"textures/metal_slvr_02", "textures/makkon_marble/metal_slvr_02"},
	{"textures/metal_slvr_03", "textures/makkon_marble/metal_slvr_03"},
	{"textures/metal_slvr_04", "textures/makkon_marble/metal_slvr_04"},
	{"textures/metal_slvr_05", "textures/makkon_marble/metal_slvr_05"},
	{"textures/metal_slvr_06", "textures/makkon_marble/metal_slvr_06"},
	{"textures/metal_slvr_07", "textures/makkon_marble/metal_slvr_07"},
	{"textures/metal_slvr_09", "textures/makkon_marble/metal_slvr_09"},
	{"textures/metal_slvr__08", "textures/makkon_marble/metal_slvr__08"},
	{"textures/metal_slvr_c1", "textures/makkon_marble/metal_slvr_c1"},
	{"textures/metal_slvr_t1", "textures/makkon_marble/metal_slvr_t1"},
	{"textures/metal_slvr_t2", "textures/makkon_marble/metal_slvr_t2"},
	{"textures/metal_slvr_t3", "textures/makkon_marble/metal_slvr_t3"},
	{"textures/metal_slvr_t4", "textures/makkon_marble/metal_slvr_t4"},
	{"textures/metal_slvr_t5", "textures/makkon_marble/metal_slvr_t5"},
	{"textures/metal_slvr_t6", "textures/makkon_marble/metal_slvr_t6"},
	{"textures/metal_slvr_t7", "textures/makkon_marble/metal_slvr_t7"},
	{"textures/metal_slvr_t8", "textures/makkon_marble/metal_slvr_t8"},
	{"textures/_metal_bras1_c2", "textures/makkon_marble/_metal_bras1_c2"},
	{"textures/_metal_bras1_t7", "textures/makkon_marble/_metal_bras1_t7"},
	{"textures/_metal_bras2_c2", "textures/makkon_marble/_metal_bras2_c2"},
	{"textures/_metal_bras2_t7", "textures/makkon_marble/_metal_bras2_t7"},
	{"textures/_metal_bras2f01", "textures/makkon_marble/_metal_bras2f01"},
	{"textures/_metal_bras2f02", "textures/makkon_marble/_metal_bras2f02"},
	{"textures/_metal_bras2f03", "textures/makkon_marble/_metal_bras2f03"},
	{"textures/_metal_bras2f04", "textures/makkon_marble/_metal_bras2f04"},
	{"textures/_metal_bras2g01", "textures/makkon_marble/_metal_bras2g01"},
	{"textures/_metal_bras2g02", "textures/makkon_marble/_metal_bras2g02"},
	{"textures/_metal_bras2g03", "textures/makkon_marble/_metal_bras2g03"},
	{"textures/_metal_bras2g04", "textures/makkon_marble/_metal_bras2g04"},
	{"textures/_metal_brasf01", "textures/makkon_marble/_metal_brasf01"},
	{"textures/_metal_brasf02", "textures/makkon_marble/_metal_brasf02"},
	{"textures/_metal_brasf03", "textures/makkon_marble/_metal_brasf03"},
	{"textures/_metal_brasf04", "textures/makkon_marble/_metal_brasf04"},
	{"textures/_metal_brasg01", "textures/makkon_marble/_metal_brasg01"},
	{"textures/_metal_brasg02", "textures/makkon_marble/_metal_brasg02"},
	{"textures/_metal_brasg03", "textures/makkon_marble/_metal_brasg03"},
	{"textures/_metal_brasg04", "textures/makkon_marble/_metal_brasg04"},
	{"textures/_metal_brass2_t5", "textures/makkon_marble/_metal_brass2_t5"},
	{"textures/_metal_brass_t5", "textures/makkon_marble/_metal_brass_t5"},
	{"textures/_metal_brnz_c2", "textures/makkon_marble/_metal_brnz_c2"},
	{"textures/_metal_brnz_t5", "textures/makkon_marble/_metal_brnz_t5"},
	{"textures/_metal_brnz_t7", "textures/makkon_marble/_metal_brnz_t7"},
	{"textures/_metal_brnzf01", "textures/makkon_marble/_metal_brnzf01"},
	{"textures/_metal_brnzf02", "textures/makkon_marble/_metal_brnzf02"},
	{"textures/_metal_brnzf03", "textures/makkon_marble/_metal_brnzf03"},
	{"textures/_metal_brnzf04", "textures/makkon_marble/_metal_brnzf04"},
	{"textures/_metal_brnzg01", "textures/makkon_marble/_metal_brnzg01"},
	{"textures/_metal_brnzg02", "textures/makkon_marble/_metal_brnzg02"},
	{"textures/_metal_brnzg03", "textures/makkon_marble/_metal_brnzg03"},
	{"textures/_metal_brnzg04", "textures/makkon_marble/_metal_brnzg04"},
	{"textures/_metal_copp_c2", "textures/makkon_marble/_metal_copp_c2"},
	{"textures/_metal_copp_t5", "textures/makkon_marble/_metal_copp_t5"},
	{"textures/_metal_copp_t7", "textures/makkon_marble/_metal_copp_t7"},
	{"textures/_metal_coppf01", "textures/makkon_marble/_metal_coppf01"},
	{"textures/_metal_coppf02", "textures/makkon_marble/_metal_coppf02"},
	{"textures/_metal_coppf03", "textures/makkon_marble/_metal_coppf03"},
	{"textures/_metal_coppf04", "textures/makkon_marble/_metal_coppf04"},
	{"textures/_metal_coppg01", "textures/makkon_marble/_metal_coppg01"},
	{"textures/_metal_coppg02", "textures/makkon_marble/_metal_coppg02"},
	{"textures/_metal_coppg03", "textures/makkon_marble/_metal_coppg03"},
	{"textures/_metal_coppg04", "textures/makkon_marble/_metal_coppg04"},
	{"textures/_metal_gold1_c2", "textures/makkon_marble/_metal_gold1_c2"},
	{"textures/_metal_gold1_t7", "textures/makkon_marble/_metal_gold1_t7"},
	{"textures/_metal_gold2_c2", "textures/makkon_marble/_metal_gold2_c2"},
	{"textures/_metal_gold2_t5", "textures/makkon_marble/_metal_gold2_t5"},
	{"textures/_metal_gold2_t7", "textures/makkon_marble/_metal_gold2_t7"},
	{"textures/_metal_gold2f01", "textures/makkon_marble/_metal_gold2f01"},
	{"textures/_metal_gold2f02", "textures/makkon_marble/_metal_gold2f02"},
	{"textures/_metal_gold2f03", "textures/makkon_marble/_metal_gold2f03"},
	{"textures/_metal_gold2f04", "textures/makkon_marble/_metal_gold2f04"},
	{"textures/_metal_gold2f91", "textures/makkon_marble/_metal_gold2f91"},
	{"textures/_metal_gold2g01", "textures/makkon_marble/_metal_gold2g01"},
	{"textures/_metal_gold2g02", "textures/makkon_marble/_metal_gold2g02"},
	{"textures/_metal_gold2g03", "textures/makkon_marble/_metal_gold2g03"},
	{"textures/_metal_gold2g04", "textures/makkon_marble/_metal_gold2g04"},
	{"textures/_metal_gold_t5", "textures/makkon_marble/_metal_gold_t5"},
	{"textures/_metal_goldf01", "textures/makkon_marble/_metal_goldf01"},
	{"textures/_metal_goldf02", "textures/makkon_marble/_metal_goldf02"},
	{"textures/_metal_goldf03", "textures/makkon_marble/_metal_goldf03"},
	{"textures/_metal_goldf04", "textures/makkon_marble/_metal_goldf04"},
	{"textures/_metal_goldg01", "textures/makkon_marble/_metal_goldg01"},
	{"textures/_metal_goldg02", "textures/makkon_marble/_metal_goldg02"},
	{"textures/_metal_goldg03", "textures/makkon_marble/_metal_goldg03"},
	{"textures/_metal_goldg04", "textures/makkon_marble/_metal_goldg04"},
	{"textures/_metal_iron1_c2", "textures/makkon_marble/_metal_iron1_c2"},
	{"textures/_metal_iron1_t5", "textures/makkon_marble/_metal_iron1_t5"},
	{"textures/_metal_iron1_t7", "textures/makkon_marble/_metal_iron1_t7"},
	{"textures/_metal_iron1f01", "textures/makkon_marble/_metal_iron1f01"},
	{"textures/_metal_iron1f02", "textures/makkon_marble/_metal_iron1f02"},
	{"textures/_metal_iron1f03", "textures/makkon_marble/_metal_iron1f03"},
	{"textures/_metal_iron1f04", "textures/makkon_marble/_metal_iron1f04"},
	{"textures/_metal_iron1g01", "textures/makkon_marble/_metal_iron1g01"},
	{"textures/_metal_iron1g02", "textures/makkon_marble/_metal_iron1g02"},
	{"textures/_metal_iron1g03", "textures/makkon_marble/_metal_iron1g03"},
	{"textures/_metal_iron1g04", "textures/makkon_marble/_metal_iron1g04"},
	{"textures/_metal_iron2_c2", "textures/makkon_marble/_metal_iron2_c2"},
	{"textures/_metal_iron2_t5", "textures/makkon_marble/_metal_iron2_t5"},
	{"textures/_metal_iron2_t7", "textures/makkon_marble/_metal_iron2_t7"},
	{"textures/_metal_iron2f01", "textures/makkon_marble/_metal_iron2f01"},
	{"textures/_metal_iron2f02", "textures/makkon_marble/_metal_iron2f02"},
	{"textures/_metal_iron2f03", "textures/makkon_marble/_metal_iron2f03"},
	{"textures/_metal_iron2f04", "textures/makkon_marble/_metal_iron2f04"},
	{"textures/_metal_iron2g01", "textures/makkon_marble/_metal_iron2g01"},
	{"textures/_metal_iron2g02", "textures/makkon_marble/_metal_iron2g02"},
	{"textures/_metal_iron2g03", "textures/makkon_marble/_metal_iron2g03"},
	{"textures/_metal_iron2g04", "textures/makkon_marble/_metal_iron2g04"},
	{"textures/_metal_iron3_c2", "textures/makkon_marble/_metal_iron3_c2"},
	{"textures/_metal_iron3_t5", "textures/makkon_marble/_metal_iron3_t5"},
	{"textures/_metal_iron3_t7", "textures/makkon_marble/_metal_iron3_t7"},
	{"textures/_metal_iron3f01", "textures/makkon_marble/_metal_iron3f01"},
	{"textures/_metal_iron3f02", "textures/makkon_marble/_metal_iron3f02"},
	{"textures/_metal_iron3f03", "textures/makkon_marble/_metal_iron3f03"},
	{"textures/_metal_iron3f04", "textures/makkon_marble/_metal_iron3f04"},
	{"textures/_metal_iron3g01", "textures/makkon_marble/_metal_iron3g01"},
	{"textures/_metal_iron3g02", "textures/makkon_marble/_metal_iron3g02"},
	{"textures/_metal_iron3g03", "textures/makkon_marble/_metal_iron3g03"},
	{"textures/_metal_iron3g04", "textures/makkon_marble/_metal_iron3g04"},
	{"textures/_metal_rust_c2", "textures/makkon_marble/_metal_rust_c2"},
	{"textures/_metal_rust_t5", "textures/makkon_marble/_metal_rust_t5"},
	{"textures/_metal_rust_t7", "textures/makkon_marble/_metal_rust_t7"},
	{"textures/_metal_rustf01", "textures/makkon_marble/_metal_rustf01"},
	{"textures/_metal_rustf02", "textures/makkon_marble/_metal_rustf02"},
	{"textures/_metal_rustf03", "textures/makkon_marble/_metal_rustf03"},
	{"textures/_metal_rustf04", "textures/makkon_marble/_metal_rustf04"},
	{"textures/_metal_rustg01", "textures/makkon_marble/_metal_rustg01"},
	{"textures/_metal_rustg02", "textures/makkon_marble/_metal_rustg02"},
	{"textures/_metal_rustg03", "textures/makkon_marble/_metal_rustg03"},
	{"textures/_metal_rustg04", "textures/makkon_marble/_metal_rustg04"},
	{"textures/_metal_slvr_c2", "textures/makkon_marble/_metal_slvr_c2"},
	{"textures/_metal_slvr_t5", "textures/makkon_marble/_metal_slvr_t5"},
	{"textures/_metal_slvr_t7", "textures/makkon_marble/_metal_slvr_t7"},
	{"textures/_metal_slvrf01", "textures/makkon_marble/_metal_slvrf01"},
	{"textures/_metal_slvrf02", "textures/makkon_marble/_metal_slvrf02"},
	{"textures/_metal_slvrf03", "textures/makkon_marble/_metal_slvrf03"},
	{"textures/_metal_slvrf04", "textures/makkon_marble/_metal_slvrf04"},
	{"textures/_metal_slvrg01", "textures/makkon_marble/_metal_slvrg01"},
	{"textures/_metal_slvrg02", "textures/makkon_marble/_metal_slvrg02"},
	{"textures/_metal_slvrg03", "textures/makkon_marble/_metal_slvrg03"},
	{"textures/_metal_slvrg04", "textures/makkon_marble/_metal_slvrg04"},

	// Makkon nature 2022.8
	{"textures/Cliff_B_blk", "textures/makkon_nature/Cliff_B_blk"},
	{"textures/Cliff_B_blue", "textures/makkon_nature/Cliff_B_blue"},
	{"textures/Cliff_B_brwn", "textures/makkon_nature/Cliff_B_brwn"},
	{"textures/Cliff_B_grey", "textures/makkon_nature/Cliff_B_grey"},
	{"textures/Cliff_B_pink", "textures/makkon_nature/Cliff_B_pink"},
	{"textures/Cliff_B_prpl", "textures/makkon_nature/Cliff_B_prpl"},
	{"textures/Cliff_B_red", "textures/makkon_nature/Cliff_B_red"},
	{"textures/Cliff_B_tan", "textures/makkon_nature/Cliff_B_tan"},
	{"textures/Cliff_B_teal", "textures/makkon_nature/Cliff_B_teal"},
	{"textures/Cliff_B_wht", "textures/makkon_nature/Cliff_B_wht"},
	{"textures/cliff_B2_blk_7", "textures/makkon_nature/cliff_B2_blk_7"},
	{"textures/cliff_B2_blue_7", "textures/makkon_nature/cliff_B2_blue_7"},
	{"textures/cliff_B2_brwn_7", "textures/makkon_nature/cliff_B2_brwn_7"},
	{"textures/cliff_B2_grey_7", "textures/makkon_nature/cliff_B2_grey_7"},
	{"textures/cliff_B2_pink_7", "textures/makkon_nature/cliff_B2_pink_7"},
	{"textures/cliff_B2_prpl_7", "textures/makkon_nature/cliff_B2_prpl_7"},
	{"textures/cliff_B2_red_7", "textures/makkon_nature/cliff_B2_red_7"},
	{"textures/cliff_B2_tan_7", "textures/makkon_nature/cliff_B2_tan_7"},
	{"textures/cliff_B2_teal_7", "textures/makkon_nature/cliff_B2_teal_7"},
	{"textures/cliff_B2_wht_7", "textures/makkon_nature/cliff_B2_wht_7"},
	{"textures/cliff_blck_14", "textures/makkon_nature/cliff_blck_14"},
	{"textures/cliff_blck_15", "textures/makkon_nature/cliff_blck_15"},
	{"textures/cliff_blck_16", "textures/makkon_nature/cliff_blck_16"},
	{"textures/cliff_blck_17", "textures/makkon_nature/cliff_blck_17"},
	{"textures/cliff_blck_18", "textures/makkon_nature/cliff_blck_18"},
	{"textures/cliff_blck_19", "textures/makkon_nature/cliff_blck_19"},
	{"textures/cliff_blck_20", "textures/makkon_nature/cliff_blck_20"},
	{"textures/cliff_blck_21", "textures/makkon_nature/cliff_blck_21"},
	{"textures/cliff_blck_22", "textures/makkon_nature/cliff_blck_22"},
	{"textures/cliff_blck_23", "textures/makkon_nature/cliff_blck_23"},
	{"textures/cliff_blck_24", "textures/makkon_nature/cliff_blck_24"},
	{"textures/cliff_blck_25", "textures/makkon_nature/cliff_blck_25"},
	{"textures/cliff_blck_26", "textures/makkon_nature/cliff_blck_26"},
	{"textures/cliff_blck_27", "textures/makkon_nature/cliff_blck_27"},
	{"textures/cliff_blck_5", "textures/makkon_nature/cliff_blck_5"},
	{"textures/cliff_blck_6", "textures/makkon_nature/cliff_blck_6"},
	{"textures/cliff_blk_1", "textures/makkon_nature/cliff_blk_1"},
	{"textures/cliff_blk_10", "textures/makkon_nature/cliff_blk_10"},
	{"textures/cliff_blk_11", "textures/makkon_nature/cliff_blk_11"},
	{"textures/cliff_blk_12", "textures/makkon_nature/cliff_blk_12"},
	{"textures/cliff_blk_13", "textures/makkon_nature/cliff_blk_13"},
	{"textures/cliff_blk_2", "textures/makkon_nature/cliff_blk_2"},
	{"textures/cliff_blk_3", "textures/makkon_nature/cliff_blk_3"},
	{"textures/cliff_blk_4", "textures/makkon_nature/cliff_blk_4"},
	{"textures/cliff_blk_7", "textures/makkon_nature/cliff_blk_7"},
	{"textures/cliff_blk_8", "textures/makkon_nature/cliff_blk_8"},
	{"textures/cliff_blk_9", "textures/makkon_nature/cliff_blk_9"},
	{"textures/cliff_blue_1", "textures/makkon_nature/cliff_blue_1"},
	{"textures/cliff_blue_10", "textures/makkon_nature/cliff_blue_10"},
	{"textures/cliff_blue_11", "textures/makkon_nature/cliff_blue_11"},
	{"textures/cliff_blue_12", "textures/makkon_nature/cliff_blue_12"},
	{"textures/cliff_blue_13", "textures/makkon_nature/cliff_blue_13"},
	{"textures/cliff_blue_14", "textures/makkon_nature/cliff_blue_14"},
	{"textures/cliff_blue_15", "textures/makkon_nature/cliff_blue_15"},
	{"textures/cliff_blue_16", "textures/makkon_nature/cliff_blue_16"},
	{"textures/cliff_blue_17", "textures/makkon_nature/cliff_blue_17"},
	{"textures/cliff_blue_18", "textures/makkon_nature/cliff_blue_18"},
	{"textures/cliff_blue_19", "textures/makkon_nature/cliff_blue_19"},
	{"textures/cliff_blue_2", "textures/makkon_nature/cliff_blue_2"},
	{"textures/cliff_blue_20", "textures/makkon_nature/cliff_blue_20"},
	{"textures/cliff_blue_21", "textures/makkon_nature/cliff_blue_21"},
	{"textures/cliff_blue_22", "textures/makkon_nature/cliff_blue_22"},
	{"textures/cliff_blue_23", "textures/makkon_nature/cliff_blue_23"},
	{"textures/cliff_blue_24", "textures/makkon_nature/cliff_blue_24"},
	{"textures/cliff_blue_25", "textures/makkon_nature/cliff_blue_25"},
	{"textures/cliff_blue_26", "textures/makkon_nature/cliff_blue_26"},
	{"textures/cliff_blue_27", "textures/makkon_nature/cliff_blue_27"},
	{"textures/cliff_blue_3", "textures/makkon_nature/cliff_blue_3"},
	{"textures/cliff_blue_4", "textures/makkon_nature/cliff_blue_4"},
	{"textures/cliff_blue_5", "textures/makkon_nature/cliff_blue_5"},
	{"textures/cliff_blue_6", "textures/makkon_nature/cliff_blue_6"},
	{"textures/cliff_blue_7", "textures/makkon_nature/cliff_blue_7"},
	{"textures/cliff_blue_8", "textures/makkon_nature/cliff_blue_8"},
	{"textures/cliff_blue_9", "textures/makkon_nature/cliff_blue_9"},
	{"textures/cliff_brwn_1", "textures/makkon_nature/cliff_brwn_1"},
	{"textures/cliff_brwn_10", "textures/makkon_nature/cliff_brwn_10"},
	{"textures/cliff_brwn_11", "textures/makkon_nature/cliff_brwn_11"},
	{"textures/cliff_brwn_12", "textures/makkon_nature/cliff_brwn_12"},
	{"textures/cliff_brwn_13", "textures/makkon_nature/cliff_brwn_13"},
	{"textures/cliff_brwn_14", "textures/makkon_nature/cliff_brwn_14"},
	{"textures/cliff_brwn_15", "textures/makkon_nature/cliff_brwn_15"},
	{"textures/cliff_brwn_16", "textures/makkon_nature/cliff_brwn_16"},
	{"textures/cliff_brwn_17", "textures/makkon_nature/cliff_brwn_17"},
	{"textures/cliff_brwn_18", "textures/makkon_nature/cliff_brwn_18"},
	{"textures/cliff_brwn_19", "textures/makkon_nature/cliff_brwn_19"},
	{"textures/cliff_brwn_2", "textures/makkon_nature/cliff_brwn_2"},
	{"textures/cliff_brwn_20", "textures/makkon_nature/cliff_brwn_20"},
	{"textures/cliff_brwn_21", "textures/makkon_nature/cliff_brwn_21"},
	{"textures/cliff_brwn_22", "textures/makkon_nature/cliff_brwn_22"},
	{"textures/cliff_brwn_23", "textures/makkon_nature/cliff_brwn_23"},
	{"textures/cliff_brwn_24", "textures/makkon_nature/cliff_brwn_24"},
	{"textures/cliff_brwn_25", "textures/makkon_nature/cliff_brwn_25"},
	{"textures/cliff_brwn_26", "textures/makkon_nature/cliff_brwn_26"},
	{"textures/cliff_brwn_27", "textures/makkon_nature/cliff_brwn_27"},
	{"textures/cliff_brwn_3", "textures/makkon_nature/cliff_brwn_3"},
	{"textures/cliff_brwn_4", "textures/makkon_nature/cliff_brwn_4"},
	{"textures/cliff_brwn_5", "textures/makkon_nature/cliff_brwn_5"},
	{"textures/cliff_brwn_6", "textures/makkon_nature/cliff_brwn_6"},
	{"textures/cliff_brwn_7", "textures/makkon_nature/cliff_brwn_7"},
	{"textures/cliff_brwn_8", "textures/makkon_nature/cliff_brwn_8"},
	{"textures/cliff_brwn_9", "textures/makkon_nature/cliff_brwn_9"},
	{"textures/cliff_grey_1", "textures/makkon_nature/cliff_grey_1"},
	{"textures/cliff_grey_10", "textures/makkon_nature/cliff_grey_10"},
	{"textures/cliff_grey_11", "textures/makkon_nature/cliff_grey_11"},
	{"textures/cliff_grey_12", "textures/makkon_nature/cliff_grey_12"},
	{"textures/cliff_grey_13", "textures/makkon_nature/cliff_grey_13"},
	{"textures/cliff_grey_14", "textures/makkon_nature/cliff_grey_14"},
	{"textures/cliff_grey_15", "textures/makkon_nature/cliff_grey_15"},
	{"textures/cliff_grey_16", "textures/makkon_nature/cliff_grey_16"},
	{"textures/cliff_grey_17", "textures/makkon_nature/cliff_grey_17"},
	{"textures/cliff_grey_18", "textures/makkon_nature/cliff_grey_18"},
	{"textures/cliff_grey_19", "textures/makkon_nature/cliff_grey_19"},
	{"textures/cliff_grey_2", "textures/makkon_nature/cliff_grey_2"},
	{"textures/cliff_grey_20", "textures/makkon_nature/cliff_grey_20"},
	{"textures/cliff_grey_21", "textures/makkon_nature/cliff_grey_21"},
	{"textures/cliff_grey_22", "textures/makkon_nature/cliff_grey_22"},
	{"textures/cliff_grey_23", "textures/makkon_nature/cliff_grey_23"},
	{"textures/cliff_grey_24", "textures/makkon_nature/cliff_grey_24"},
	{"textures/cliff_grey_25", "textures/makkon_nature/cliff_grey_25"},
	{"textures/cliff_grey_26", "textures/makkon_nature/cliff_grey_26"},
	{"textures/cliff_grey_27", "textures/makkon_nature/cliff_grey_27"},
	{"textures/cliff_grey_3", "textures/makkon_nature/cliff_grey_3"},
	{"textures/cliff_grey_4", "textures/makkon_nature/cliff_grey_4"},
	{"textures/cliff_grey_5", "textures/makkon_nature/cliff_grey_5"},
	{"textures/cliff_grey_6", "textures/makkon_nature/cliff_grey_6"},
	{"textures/cliff_grey_7", "textures/makkon_nature/cliff_grey_7"},
	{"textures/cliff_grey_8", "textures/makkon_nature/cliff_grey_8"},
	{"textures/cliff_grey_9", "textures/makkon_nature/cliff_grey_9"},
	{"textures/cliff_pink_10", "textures/makkon_nature/cliff_pink_10"},
	{"textures/cliff_pink_11", "textures/makkon_nature/cliff_pink_11"},
	{"textures/cliff_pink_14", "textures/makkon_nature/cliff_pink_14"},
	{"textures/cliff_pink_15", "textures/makkon_nature/cliff_pink_15"},
	{"textures/cliff_pink_16", "textures/makkon_nature/cliff_pink_16"},
	{"textures/cliff_pink_17", "textures/makkon_nature/cliff_pink_17"},
	{"textures/cliff_pink_18", "textures/makkon_nature/cliff_pink_18"},
	{"textures/cliff_pink_19", "textures/makkon_nature/cliff_pink_19"},
	{"textures/cliff_pink_20", "textures/makkon_nature/cliff_pink_20"},
	{"textures/cliff_pink_21", "textures/makkon_nature/cliff_pink_21"},
	{"textures/cliff_pink_22", "textures/makkon_nature/cliff_pink_22"},
	{"textures/cliff_pink_23", "textures/makkon_nature/cliff_pink_23"},
	{"textures/cliff_pink_24", "textures/makkon_nature/cliff_pink_24"},
	{"textures/cliff_pink_25", "textures/makkon_nature/cliff_pink_25"},
	{"textures/cliff_pink_26", "textures/makkon_nature/cliff_pink_26"},
	{"textures/cliff_pink_27", "textures/makkon_nature/cliff_pink_27"},
	{"textures/cliff_pink_9", "textures/makkon_nature/cliff_pink_9"},
	{"textures/cliff_prpl_10", "textures/makkon_nature/cliff_prpl_10"},
	{"textures/cliff_prpl_11", "textures/makkon_nature/cliff_prpl_11"},
	{"textures/cliff_prpl_14", "textures/makkon_nature/cliff_prpl_14"},
	{"textures/cliff_prpl_15", "textures/makkon_nature/cliff_prpl_15"},
	{"textures/cliff_prpl_16", "textures/makkon_nature/cliff_prpl_16"},
	{"textures/cliff_prpl_17", "textures/makkon_nature/cliff_prpl_17"},
	{"textures/cliff_prpl_18", "textures/makkon_nature/cliff_prpl_18"},
	{"textures/cliff_prpl_19", "textures/makkon_nature/cliff_prpl_19"},
	{"textures/cliff_prpl_20", "textures/makkon_nature/cliff_prpl_20"},
	{"textures/cliff_prpl_21", "textures/makkon_nature/cliff_prpl_21"},
	{"textures/cliff_prpl_22", "textures/makkon_nature/cliff_prpl_22"},
	{"textures/cliff_prpl_23", "textures/makkon_nature/cliff_prpl_23"},
	{"textures/cliff_prpl_24", "textures/makkon_nature/cliff_prpl_24"},
	{"textures/cliff_prpl_25", "textures/makkon_nature/cliff_prpl_25"},
	{"textures/cliff_prpl_26", "textures/makkon_nature/cliff_prpl_26"},
	{"textures/cliff_prpl_27", "textures/makkon_nature/cliff_prpl_27"},
	{"textures/cliff_prpl_9", "textures/makkon_nature/cliff_prpl_9"},
	{"textures/cliff_red1_14", "textures/makkon_nature/cliff_red1_14"},
	{"textures/cliff_red2_14", "textures/makkon_nature/cliff_red2_14"},
	{"textures/cliff_red_1", "textures/makkon_nature/cliff_red_1"},
	{"textures/cliff_red_10", "textures/makkon_nature/cliff_red_10"},
	{"textures/cliff_red_11", "textures/makkon_nature/cliff_red_11"},
	{"textures/cliff_red_12", "textures/makkon_nature/cliff_red_12"},
	{"textures/cliff_red_13", "textures/makkon_nature/cliff_red_13"},
	{"textures/cliff_red_15", "textures/makkon_nature/cliff_red_15"},
	{"textures/cliff_red_16", "textures/makkon_nature/cliff_red_16"},
	{"textures/cliff_red_17", "textures/makkon_nature/cliff_red_17"},
	{"textures/cliff_red_18", "textures/makkon_nature/cliff_red_18"},
	{"textures/cliff_red_19", "textures/makkon_nature/cliff_red_19"},
	{"textures/cliff_red_2", "textures/makkon_nature/cliff_red_2"},
	{"textures/cliff_red_20", "textures/makkon_nature/cliff_red_20"},
	{"textures/cliff_red_21", "textures/makkon_nature/cliff_red_21"},
	{"textures/cliff_red_22", "textures/makkon_nature/cliff_red_22"},
	{"textures/cliff_red_23", "textures/makkon_nature/cliff_red_23"},
	{"textures/cliff_red_24", "textures/makkon_nature/cliff_red_24"},
	{"textures/cliff_red_25", "textures/makkon_nature/cliff_red_25"},
	{"textures/cliff_red_26", "textures/makkon_nature/cliff_red_26"},
	{"textures/cliff_red_27", "textures/makkon_nature/cliff_red_27"},
	{"textures/cliff_red_3", "textures/makkon_nature/cliff_red_3"},
	{"textures/cliff_red_4", "textures/makkon_nature/cliff_red_4"},
	{"textures/cliff_red_5", "textures/makkon_nature/cliff_red_5"},
	{"textures/cliff_red_6", "textures/makkon_nature/cliff_red_6"},
	{"textures/cliff_red_7", "textures/makkon_nature/cliff_red_7"},
	{"textures/cliff_red_8", "textures/makkon_nature/cliff_red_8"},
	{"textures/cliff_red_9", "textures/makkon_nature/cliff_red_9"},
	{"textures/cliff_tan_1", "textures/makkon_nature/cliff_tan_1"},
	{"textures/cliff_tan_10", "textures/makkon_nature/cliff_tan_10"},
	{"textures/cliff_tan_11", "textures/makkon_nature/cliff_tan_11"},
	{"textures/cliff_tan_12", "textures/makkon_nature/cliff_tan_12"},
	{"textures/cliff_tan_13", "textures/makkon_nature/cliff_tan_13"},
	{"textures/cliff_tan_14", "textures/makkon_nature/cliff_tan_14"},
	{"textures/cliff_tan_15", "textures/makkon_nature/cliff_tan_15"},
	{"textures/cliff_tan_16", "textures/makkon_nature/cliff_tan_16"},
	{"textures/cliff_tan_17", "textures/makkon_nature/cliff_tan_17"},
	{"textures/cliff_tan_18", "textures/makkon_nature/cliff_tan_18"},
	{"textures/cliff_tan_19", "textures/makkon_nature/cliff_tan_19"},
	{"textures/cliff_tan_2", "textures/makkon_nature/cliff_tan_2"},
	{"textures/cliff_tan_20", "textures/makkon_nature/cliff_tan_20"},
	{"textures/cliff_tan_21", "textures/makkon_nature/cliff_tan_21"},
	{"textures/cliff_tan_22", "textures/makkon_nature/cliff_tan_22"},
	{"textures/cliff_tan_23", "textures/makkon_nature/cliff_tan_23"},
	{"textures/cliff_tan_24", "textures/makkon_nature/cliff_tan_24"},
	{"textures/cliff_tan_25", "textures/makkon_nature/cliff_tan_25"},
	{"textures/cliff_tan_26", "textures/makkon_nature/cliff_tan_26"},
	{"textures/cliff_tan_27", "textures/makkon_nature/cliff_tan_27"},
	{"textures/cliff_tan_3", "textures/makkon_nature/cliff_tan_3"},
	{"textures/cliff_tan_4", "textures/makkon_nature/cliff_tan_4"},
	{"textures/cliff_tan_5", "textures/makkon_nature/cliff_tan_5"},
	{"textures/cliff_tan_6", "textures/makkon_nature/cliff_tan_6"},
	{"textures/cliff_tan_7", "textures/makkon_nature/cliff_tan_7"},
	{"textures/cliff_tan_8", "textures/makkon_nature/cliff_tan_8"},
	{"textures/cliff_tan_9", "textures/makkon_nature/cliff_tan_9"},
	{"textures/cliff_teal_1", "textures/makkon_nature/cliff_teal_1"},
	{"textures/cliff_teal_10", "textures/makkon_nature/cliff_teal_10"},
	{"textures/cliff_teal_11", "textures/makkon_nature/cliff_teal_11"},
	{"textures/cliff_teal_12", "textures/makkon_nature/cliff_teal_12"},
	{"textures/cliff_teal_13", "textures/makkon_nature/cliff_teal_13"},
	{"textures/cliff_teal_14", "textures/makkon_nature/cliff_teal_14"},
	{"textures/cliff_teal_15", "textures/makkon_nature/cliff_teal_15"},
	{"textures/cliff_teal_16", "textures/makkon_nature/cliff_teal_16"},
	{"textures/cliff_teal_17", "textures/makkon_nature/cliff_teal_17"},
	{"textures/cliff_teal_18", "textures/makkon_nature/cliff_teal_18"},
	{"textures/cliff_teal_19", "textures/makkon_nature/cliff_teal_19"},
	{"textures/cliff_teal_2", "textures/makkon_nature/cliff_teal_2"},
	{"textures/cliff_teal_20", "textures/makkon_nature/cliff_teal_20"},
	{"textures/cliff_teal_21", "textures/makkon_nature/cliff_teal_21"},
	{"textures/cliff_teal_22", "textures/makkon_nature/cliff_teal_22"},
	{"textures/cliff_teal_23", "textures/makkon_nature/cliff_teal_23"},
	{"textures/cliff_teal_24", "textures/makkon_nature/cliff_teal_24"},
	{"textures/cliff_teal_25", "textures/makkon_nature/cliff_teal_25"},
	{"textures/cliff_teal_26", "textures/makkon_nature/cliff_teal_26"},
	{"textures/cliff_teal_27", "textures/makkon_nature/cliff_teal_27"},
	{"textures/cliff_teal_3", "textures/makkon_nature/cliff_teal_3"},
	{"textures/cliff_teal_4", "textures/makkon_nature/cliff_teal_4"},
	{"textures/cliff_teal_5", "textures/makkon_nature/cliff_teal_5"},
	{"textures/cliff_teal_6", "textures/makkon_nature/cliff_teal_6"},
	{"textures/cliff_teal_7", "textures/makkon_nature/cliff_teal_7"},
	{"textures/cliff_teal_8", "textures/makkon_nature/cliff_teal_8"},
	{"textures/cliff_teal_9", "textures/makkon_nature/cliff_teal_9"},
	{"textures/cliff_wht_1", "textures/makkon_nature/cliff_wht_1"},
	{"textures/cliff_wht_10", "textures/makkon_nature/cliff_wht_10"},
	{"textures/cliff_wht_11", "textures/makkon_nature/cliff_wht_11"},
	{"textures/cliff_wht_12", "textures/makkon_nature/cliff_wht_12"},
	{"textures/cliff_wht_13", "textures/makkon_nature/cliff_wht_13"},
	{"textures/cliff_wht_14", "textures/makkon_nature/cliff_wht_14"},
	{"textures/cliff_wht_15", "textures/makkon_nature/cliff_wht_15"},
	{"textures/cliff_wht_16", "textures/makkon_nature/cliff_wht_16"},
	{"textures/cliff_wht_17", "textures/makkon_nature/cliff_wht_17"},
	{"textures/cliff_wht_18", "textures/makkon_nature/cliff_wht_18"},
	{"textures/cliff_wht_19", "textures/makkon_nature/cliff_wht_19"},
	{"textures/cliff_wht_2", "textures/makkon_nature/cliff_wht_2"},
	{"textures/cliff_wht_20", "textures/makkon_nature/cliff_wht_20"},
	{"textures/cliff_wht_21", "textures/makkon_nature/cliff_wht_21"},
	{"textures/cliff_wht_22", "textures/makkon_nature/cliff_wht_22"},
	{"textures/cliff_wht_23", "textures/makkon_nature/cliff_wht_23"},
	{"textures/cliff_wht_24", "textures/makkon_nature/cliff_wht_24"},
	{"textures/cliff_wht_25", "textures/makkon_nature/cliff_wht_25"},
	{"textures/cliff_wht_26", "textures/makkon_nature/cliff_wht_26"},
	{"textures/cliff_wht_27", "textures/makkon_nature/cliff_wht_27"},
	{"textures/cliff_wht_3", "textures/makkon_nature/cliff_wht_3"},
	{"textures/cliff_wht_4", "textures/makkon_nature/cliff_wht_4"},
	{"textures/cliff_wht_5", "textures/makkon_nature/cliff_wht_5"},
	{"textures/cliff_wht_6", "textures/makkon_nature/cliff_wht_6"},
	{"textures/cliff_wht_7", "textures/makkon_nature/cliff_wht_7"},
	{"textures/cliff_wht_8", "textures/makkon_nature/cliff_wht_8"},
	{"textures/cliff_wht_9", "textures/makkon_nature/cliff_wht_9"},
	{"textures/dirt_blck01", "textures/makkon_nature/dirt_blck01"},
	{"textures/dirt_blcksm01", "textures/makkon_nature/dirt_blcksm01"},
	{"textures/dirt_blue01", "textures/makkon_nature/dirt_blue01"},
	{"textures/dirt_bluesm01", "textures/makkon_nature/dirt_bluesm01"},
	{"textures/dirt_brwn01", "textures/makkon_nature/dirt_brwn01"},
	{"textures/dirt_brwnsm01", "textures/makkon_nature/dirt_brwnsm01"},
	{"textures/dirt_grey01", "textures/makkon_nature/dirt_grey01"},
	{"textures/dirt_greysm01", "textures/makkon_nature/dirt_greysm01"},
	{"textures/dirt_pink01", "textures/makkon_nature/dirt_pink01"},
	{"textures/dirt_pinksm01", "textures/makkon_nature/dirt_pinksm01"},
	{"textures/dirt_prpl01", "textures/makkon_nature/dirt_prpl01"},
	{"textures/dirt_prplsm01", "textures/makkon_nature/dirt_prplsm01"},
	{"textures/dirt_red01", "textures/makkon_nature/dirt_red01"},
	{"textures/dirt_redsm01", "textures/makkon_nature/dirt_redsm01"},
	{"textures/dirt_tan01", "textures/makkon_nature/dirt_tan01"},
	{"textures/dirt_tan01_1", "textures/makkon_nature/dirt_tan01_1"},
	{"textures/dirt_tansm01", "textures/makkon_nature/dirt_tansm01"},
	{"textures/dirt_tansm01_1", "textures/makkon_nature/dirt_tansm01_1"},
	{"textures/dirt_wht01", "textures/makkon_nature/dirt_wht01"},
	{"textures/dirt_whtsm01", "textures/makkon_nature/dirt_whtsm01"},
	{"textures/ground_blck01", "textures/makkon_nature/ground_blck01"},
	{"textures/ground_blck02", "textures/makkon_nature/ground_blck02"},
	{"textures/ground_blck03", "textures/makkon_nature/ground_blck03"},
	{"textures/ground_blck04", "textures/makkon_nature/ground_blck04"},
	{"textures/ground_blck05", "textures/makkon_nature/ground_blck05"},
	{"textures/ground_blck06", "textures/makkon_nature/ground_blck06"},
	{"textures/ground_blck07", "textures/makkon_nature/ground_blck07"},
	{"textures/ground_blk_1", "textures/makkon_nature/ground_blk_1"},
	{"textures/ground_blue01", "textures/makkon_nature/ground_blue01"},
	{"textures/ground_blue02", "textures/makkon_nature/ground_blue02"},
	{"textures/ground_blue03", "textures/makkon_nature/ground_blue03"},
	{"textures/ground_blue04", "textures/makkon_nature/ground_blue04"},
	{"textures/ground_blue05", "textures/makkon_nature/ground_blue05"},
	{"textures/ground_blue06", "textures/makkon_nature/ground_blue06"},
	{"textures/ground_blue07", "textures/makkon_nature/ground_blue07"},
	{"textures/ground_blue08", "textures/makkon_nature/ground_blue08"},
	{"textures/ground_blue_1", "textures/makkon_nature/ground_blue_1"},
	{"textures/ground_brwn01", "textures/makkon_nature/ground_brwn01"},
	{"textures/ground_brwn02", "textures/makkon_nature/ground_brwn02"},
	{"textures/ground_brwn03", "textures/makkon_nature/ground_brwn03"},
	{"textures/ground_brwn04", "textures/makkon_nature/ground_brwn04"},
	{"textures/ground_brwn05", "textures/makkon_nature/ground_brwn05"},
	{"textures/ground_brwn06", "textures/makkon_nature/ground_brwn06"},
	{"textures/ground_brwn07", "textures/makkon_nature/ground_brwn07"},
	{"textures/ground_brwn_1", "textures/makkon_nature/ground_brwn_1"},
	{"textures/ground_grey01", "textures/makkon_nature/ground_grey01"},
	{"textures/ground_grey02", "textures/makkon_nature/ground_grey02"},
	{"textures/ground_grey03", "textures/makkon_nature/ground_grey03"},
	{"textures/ground_grey04", "textures/makkon_nature/ground_grey04"},
	{"textures/ground_grey_1", "textures/makkon_nature/ground_grey_1"},
	{"textures/ground_pink01", "textures/makkon_nature/ground_pink01"},
	{"textures/ground_pink02", "textures/makkon_nature/ground_pink02"},
	{"textures/ground_pink03", "textures/makkon_nature/ground_pink03"},
	{"textures/ground_pink04", "textures/makkon_nature/ground_pink04"},
	{"textures/ground_prpl01", "textures/makkon_nature/ground_prpl01"},
	{"textures/ground_prpl02", "textures/makkon_nature/ground_prpl02"},
	{"textures/ground_prpl03", "textures/makkon_nature/ground_prpl03"},
	{"textures/ground_prpl04", "textures/makkon_nature/ground_prpl04"},
	{"textures/ground_red01", "textures/makkon_nature/ground_red01"},
	{"textures/ground_red02", "textures/makkon_nature/ground_red02"},
	{"textures/ground_red03", "textures/makkon_nature/ground_red03"},
	{"textures/ground_red04", "textures/makkon_nature/ground_red04"},
	{"textures/ground_red_1", "textures/makkon_nature/ground_red_1"},
	{"textures/ground_tan01", "textures/makkon_nature/ground_tan01"},
	{"textures/ground_tan02", "textures/makkon_nature/ground_tan02"},
	{"textures/ground_tan03", "textures/makkon_nature/ground_tan03"},
	{"textures/ground_tan04", "textures/makkon_nature/ground_tan04"},
	{"textures/ground_tan_1", "textures/makkon_nature/ground_tan_1"},
	{"textures/ground_teal01", "textures/makkon_nature/ground_teal01"},
	{"textures/ground_teal02", "textures/makkon_nature/ground_teal02"},
	{"textures/ground_teal03", "textures/makkon_nature/ground_teal03"},
	{"textures/ground_teal04", "textures/makkon_nature/ground_teal04"},
	{"textures/ground_teal05", "textures/makkon_nature/ground_teal05"},
	{"textures/ground_teal06", "textures/makkon_nature/ground_teal06"},
	{"textures/ground_teal07", "textures/makkon_nature/ground_teal07"},
	{"textures/ground_teal_1", "textures/makkon_nature/ground_teal_1"},
	{"textures/ground_wht01", "textures/makkon_nature/ground_wht01"},
	{"textures/ground_wht02", "textures/makkon_nature/ground_wht02"},
	{"textures/ground_wht03", "textures/makkon_nature/ground_wht03"},
	{"textures/ground_wht04", "textures/makkon_nature/ground_wht04"},
	{"textures/ground_wht05", "textures/makkon_nature/ground_wht05"},
	{"textures/ground_wht06", "textures/makkon_nature/ground_wht06"},
	{"textures/ground_wht07", "textures/makkon_nature/ground_wht07"},
	{"textures/ground_wht_1", "textures/makkon_nature/ground_wht_1"},
	{"textures/groundsm_blck01", "textures/makkon_nature/groundsm_blck01"},
	{"textures/groundsm_blck02", "textures/makkon_nature/groundsm_blck02"},
	{"textures/groundsm_blck03", "textures/makkon_nature/groundsm_blck03"},
	{"textures/groundsm_blck04", "textures/makkon_nature/groundsm_blck04"},
	{"textures/groundsm_blck05", "textures/makkon_nature/groundsm_blck05"},
	{"textures/groundsm_blck06", "textures/makkon_nature/groundsm_blck06"},
	{"textures/groundsm_blck07", "textures/makkon_nature/groundsm_blck07"},
	{"textures/groundsm_blue01", "textures/makkon_nature/groundsm_blue01"},
	{"textures/groundsm_blue02", "textures/makkon_nature/groundsm_blue02"},
	{"textures/groundsm_blue03", "textures/makkon_nature/groundsm_blue03"},
	{"textures/groundsm_blue04", "textures/makkon_nature/groundsm_blue04"},
	{"textures/groundsm_blue05", "textures/makkon_nature/groundsm_blue05"},
	{"textures/groundsm_blue06", "textures/makkon_nature/groundsm_blue06"},
	{"textures/groundsm_blue07", "textures/makkon_nature/groundsm_blue07"},
	{"textures/groundsm_blue08", "textures/makkon_nature/groundsm_blue08"},
	{"textures/groundsm_brwn01", "textures/makkon_nature/groundsm_brwn01"},
	{"textures/groundsm_brwn02", "textures/makkon_nature/groundsm_brwn02"},
	{"textures/groundsm_brwn03", "textures/makkon_nature/groundsm_brwn03"},
	{"textures/groundsm_brwn04", "textures/makkon_nature/groundsm_brwn04"},
	{"textures/groundsm_brwn05", "textures/makkon_nature/groundsm_brwn05"},
	{"textures/groundsm_brwn06", "textures/makkon_nature/groundsm_brwn06"},
	{"textures/groundsm_brwn07", "textures/makkon_nature/groundsm_brwn07"},
	{"textures/groundsm_grey01", "textures/makkon_nature/groundsm_grey01"},
	{"textures/groundsm_grey02", "textures/makkon_nature/groundsm_grey02"},
	{"textures/groundsm_grey03", "textures/makkon_nature/groundsm_grey03"},
	{"textures/groundsm_grey04", "textures/makkon_nature/groundsm_grey04"},
	{"textures/groundsm_pink01", "textures/makkon_nature/groundsm_pink01"},
	{"textures/groundsm_pink02", "textures/makkon_nature/groundsm_pink02"},
	{"textures/groundsm_pink03", "textures/makkon_nature/groundsm_pink03"},
	{"textures/groundsm_pink04", "textures/makkon_nature/groundsm_pink04"},
	{"textures/groundsm_prpl01", "textures/makkon_nature/groundsm_prpl01"},
	{"textures/groundsm_prpl02", "textures/makkon_nature/groundsm_prpl02"},
	{"textures/groundsm_prpl03", "textures/makkon_nature/groundsm_prpl03"},
	{"textures/groundsm_prpl04", "textures/makkon_nature/groundsm_prpl04"},
	{"textures/groundsm_red01", "textures/makkon_nature/groundsm_red01"},
	{"textures/groundsm_red02", "textures/makkon_nature/groundsm_red02"},
	{"textures/groundsm_red03", "textures/makkon_nature/groundsm_red03"},
	{"textures/groundsm_red04", "textures/makkon_nature/groundsm_red04"},
	{"textures/groundsm_tan01", "textures/makkon_nature/groundsm_tan01"},
	{"textures/groundsm_tan02", "textures/makkon_nature/groundsm_tan02"},
	{"textures/groundsm_tan03", "textures/makkon_nature/groundsm_tan03"},
	{"textures/groundsm_tan04", "textures/makkon_nature/groundsm_tan04"},
	{"textures/groundsm_teal01", "textures/makkon_nature/groundsm_teal01"},
	{"textures/groundsm_teal02", "textures/makkon_nature/groundsm_teal02"},
	{"textures/groundsm_teal03", "textures/makkon_nature/groundsm_teal03"},
	{"textures/groundsm_teal04", "textures/makkon_nature/groundsm_teal04"},
	{"textures/groundsm_teal05", "textures/makkon_nature/groundsm_teal05"},
	{"textures/groundsm_teal06", "textures/makkon_nature/groundsm_teal06"},
	{"textures/groundsm_teal07", "textures/makkon_nature/groundsm_teal07"},
	{"textures/groundsm_wht01", "textures/makkon_nature/groundsm_wht01"},
	{"textures/groundsm_wht02", "textures/makkon_nature/groundsm_wht02"},
	{"textures/groundsm_wht03", "textures/makkon_nature/groundsm_wht03"},
	{"textures/groundsm_wht04", "textures/makkon_nature/groundsm_wht04"},
	{"textures/groundsm_wht05", "textures/makkon_nature/groundsm_wht05"},
	{"textures/groundsm_wht06", "textures/makkon_nature/groundsm_wht06"},
	{"textures/groundsm_wht07", "textures/makkon_nature/groundsm_wht07"},
	{"textures/mud_blck01", "textures/makkon_nature/mud_blck01"},
	{"textures/mud_blck02", "textures/makkon_nature/mud_blck02"},
	{"textures/mud_blck03", "textures/makkon_nature/mud_blck03"},
	{"textures/mud_blck04", "textures/makkon_nature/mud_blck04"},
	{"textures/mud_blck05", "textures/makkon_nature/mud_blck05"},
	{"textures/mud_blck06", "textures/makkon_nature/mud_blck06"},
	{"textures/mud_blck07", "textures/makkon_nature/mud_blck07"},
	{"textures/mud_blcksm02", "textures/makkon_nature/mud_blcksm02"},
	{"textures/mud_blcksm03", "textures/makkon_nature/mud_blcksm03"},
	{"textures/mud_blcksm04", "textures/makkon_nature/mud_blcksm04"},
	{"textures/mud_blcksm05", "textures/makkon_nature/mud_blcksm05"},
	{"textures/mud_blue01", "textures/makkon_nature/mud_blue01"},
	{"textures/mud_blue02", "textures/makkon_nature/mud_blue02"},
	{"textures/mud_blue03", "textures/makkon_nature/mud_blue03"},
	{"textures/mud_blue04", "textures/makkon_nature/mud_blue04"},
	{"textures/mud_blue05", "textures/makkon_nature/mud_blue05"},
	{"textures/mud_blue06", "textures/makkon_nature/mud_blue06"},
	{"textures/mud_blue07", "textures/makkon_nature/mud_blue07"},
	{"textures/mud_bluesm02", "textures/makkon_nature/mud_bluesm02"},
	{"textures/mud_bluesm03", "textures/makkon_nature/mud_bluesm03"},
	{"textures/mud_bluesm04", "textures/makkon_nature/mud_bluesm04"},
	{"textures/mud_bluesm05", "textures/makkon_nature/mud_bluesm05"},
	{"textures/mud_brwn01", "textures/makkon_nature/mud_brwn01"},
	{"textures/mud_brwn02", "textures/makkon_nature/mud_brwn02"},
	{"textures/mud_brwn03", "textures/makkon_nature/mud_brwn03"},
	{"textures/mud_brwn04", "textures/makkon_nature/mud_brwn04"},
	{"textures/mud_brwn05", "textures/makkon_nature/mud_brwn05"},
	{"textures/mud_brwn06", "textures/makkon_nature/mud_brwn06"},
	{"textures/mud_brwn07", "textures/makkon_nature/mud_brwn07"},
	{"textures/mud_brwnsm02", "textures/makkon_nature/mud_brwnsm02"},
	{"textures/mud_brwnsm03", "textures/makkon_nature/mud_brwnsm03"},
	{"textures/mud_brwnsm04", "textures/makkon_nature/mud_brwnsm04"},
	{"textures/mud_brwnsm05", "textures/makkon_nature/mud_brwnsm05"},
	{"textures/mud_grey01", "textures/makkon_nature/mud_grey01"},
	{"textures/mud_grey02", "textures/makkon_nature/mud_grey02"},
	{"textures/mud_grey03", "textures/makkon_nature/mud_grey03"},
	{"textures/mud_grey04", "textures/makkon_nature/mud_grey04"},
	{"textures/mud_grey05", "textures/makkon_nature/mud_grey05"},
	{"textures/mud_grey06", "textures/makkon_nature/mud_grey06"},
	{"textures/mud_grey07", "textures/makkon_nature/mud_grey07"},
	{"textures/mud_greysm02", "textures/makkon_nature/mud_greysm02"},
	{"textures/mud_greysm03", "textures/makkon_nature/mud_greysm03"},
	{"textures/mud_greysm04", "textures/makkon_nature/mud_greysm04"},
	{"textures/mud_greysm05", "textures/makkon_nature/mud_greysm05"},
	{"textures/mud_pink01", "textures/makkon_nature/mud_pink01"},
	{"textures/mud_pink02", "textures/makkon_nature/mud_pink02"},
	{"textures/mud_pink03", "textures/makkon_nature/mud_pink03"},
	{"textures/mud_pink04", "textures/makkon_nature/mud_pink04"},
	{"textures/mud_pink05", "textures/makkon_nature/mud_pink05"},
	{"textures/mud_pink06", "textures/makkon_nature/mud_pink06"},
	{"textures/mud_pink07", "textures/makkon_nature/mud_pink07"},
	{"textures/mud_pinksm02", "textures/makkon_nature/mud_pinksm02"},
	{"textures/mud_pinksm03", "textures/makkon_nature/mud_pinksm03"},
	{"textures/mud_pinksm04", "textures/makkon_nature/mud_pinksm04"},
	{"textures/mud_pinksm05", "textures/makkon_nature/mud_pinksm05"},
	{"textures/mud_prpl01", "textures/makkon_nature/mud_prpl01"},
	{"textures/mud_prpl02", "textures/makkon_nature/mud_prpl02"},
	{"textures/mud_prpl03", "textures/makkon_nature/mud_prpl03"},
	{"textures/mud_prpl04", "textures/makkon_nature/mud_prpl04"},
	{"textures/mud_prpl05", "textures/makkon_nature/mud_prpl05"},
	{"textures/mud_prpl06", "textures/makkon_nature/mud_prpl06"},
	{"textures/mud_prpl07", "textures/makkon_nature/mud_prpl07"},
	{"textures/mud_prplsm02", "textures/makkon_nature/mud_prplsm02"},
	{"textures/mud_prplsm03", "textures/makkon_nature/mud_prplsm03"},
	{"textures/mud_prplsm04", "textures/makkon_nature/mud_prplsm04"},
	{"textures/mud_prplsm05", "textures/makkon_nature/mud_prplsm05"},
	{"textures/mud_red01", "textures/makkon_nature/mud_red01"},
	{"textures/mud_red02", "textures/makkon_nature/mud_red02"},
	{"textures/mud_red03", "textures/makkon_nature/mud_red03"},
	{"textures/mud_red034", "textures/makkon_nature/mud_red034"},
	{"textures/mud_red035", "textures/makkon_nature/mud_red035"},
	{"textures/mud_red06", "textures/makkon_nature/mud_red06"},
	{"textures/mud_red07", "textures/makkon_nature/mud_red07"},
	{"textures/mud_redsm02", "textures/makkon_nature/mud_redsm02"},
	{"textures/mud_redsm03", "textures/makkon_nature/mud_redsm03"},
	{"textures/mud_redsm04", "textures/makkon_nature/mud_redsm04"},
	{"textures/mud_redsm05", "textures/makkon_nature/mud_redsm05"},
	{"textures/mud_tan01", "textures/makkon_nature/mud_tan01"},
	{"textures/mud_tan02", "textures/makkon_nature/mud_tan02"},
	{"textures/mud_tan03", "textures/makkon_nature/mud_tan03"},
	{"textures/mud_tan04", "textures/makkon_nature/mud_tan04"},
	{"textures/mud_tan05", "textures/makkon_nature/mud_tan05"},
	{"textures/mud_tan06", "textures/makkon_nature/mud_tan06"},
	{"textures/mud_tan07", "textures/makkon_nature/mud_tan07"},
	{"textures/mud_tansm02", "textures/makkon_nature/mud_tansm02"},
	{"textures/mud_tansm03", "textures/makkon_nature/mud_tansm03"},
	{"textures/mud_tansm04", "textures/makkon_nature/mud_tansm04"},
	{"textures/mud_tansm05", "textures/makkon_nature/mud_tansm05"},
	{"textures/mud_teal01", "textures/makkon_nature/mud_teal01"},
	{"textures/mud_teal02", "textures/makkon_nature/mud_teal02"},
	{"textures/mud_teal03", "textures/makkon_nature/mud_teal03"},
	{"textures/mud_teal04", "textures/makkon_nature/mud_teal04"},
	{"textures/mud_teal05", "textures/makkon_nature/mud_teal05"},
	{"textures/mud_teal06", "textures/makkon_nature/mud_teal06"},
	{"textures/mud_teal07", "textures/makkon_nature/mud_teal07"},
	{"textures/mud_tealsm02", "textures/makkon_nature/mud_tealsm02"},
	{"textures/mud_tealsm03", "textures/makkon_nature/mud_tealsm03"},
	{"textures/mud_tealsm04", "textures/makkon_nature/mud_tealsm04"},
	{"textures/mud_tealsm05", "textures/makkon_nature/mud_tealsm05"},
	{"textures/mud_wht01", "textures/makkon_nature/mud_wht01"},
	{"textures/mud_wht02", "textures/makkon_nature/mud_wht02"},
	{"textures/mud_wht03", "textures/makkon_nature/mud_wht03"},
	{"textures/mud_wht04", "textures/makkon_nature/mud_wht04"},
	{"textures/mud_wht05", "textures/makkon_nature/mud_wht05"},
	{"textures/mud_wht06", "textures/makkon_nature/mud_wht06"},
	{"textures/mud_wht07", "textures/makkon_nature/mud_wht07"},
	{"textures/mud_whtsm02", "textures/makkon_nature/mud_whtsm02"},
	{"textures/mud_whtsm03", "textures/makkon_nature/mud_whtsm03"},
	{"textures/mud_whtsm04", "textures/makkon_nature/mud_whtsm04"},
	{"textures/mud_whtsm05", "textures/makkon_nature/mud_whtsm05"},
	{"textures/sand_blck01", "textures/makkon_nature/sand_blck01"},
	{"textures/sand_blck02", "textures/makkon_nature/sand_blck02"},
	{"textures/sand_blcksm01", "textures/makkon_nature/sand_blcksm01"},
	{"textures/sand_blcksm02", "textures/makkon_nature/sand_blcksm02"},
	{"textures/sand_blue01", "textures/makkon_nature/sand_blue01"},
	{"textures/sand_blue02", "textures/makkon_nature/sand_blue02"},
	{"textures/sand_bluesm01", "textures/makkon_nature/sand_bluesm01"},
	{"textures/sand_bluesm02", "textures/makkon_nature/sand_bluesm02"},
	{"textures/sand_brwn01", "textures/makkon_nature/sand_brwn01"},
	{"textures/sand_brwn02", "textures/makkon_nature/sand_brwn02"},
	{"textures/sand_brwnsm01", "textures/makkon_nature/sand_brwnsm01"},
	{"textures/sand_brwnsm02", "textures/makkon_nature/sand_brwnsm02"},
	{"textures/sand_grey01", "textures/makkon_nature/sand_grey01"},
	{"textures/sand_grey02", "textures/makkon_nature/sand_grey02"},
	{"textures/sand_greysm01", "textures/makkon_nature/sand_greysm01"},
	{"textures/sand_greysm02", "textures/makkon_nature/sand_greysm02"},
	{"textures/sand_pink01", "textures/makkon_nature/sand_pink01"},
	{"textures/sand_pink02", "textures/makkon_nature/sand_pink02"},
	{"textures/sand_pinksm01", "textures/makkon_nature/sand_pinksm01"},
	{"textures/sand_pinksm02", "textures/makkon_nature/sand_pinksm02"},
	{"textures/sand_prpl01", "textures/makkon_nature/sand_prpl01"},
	{"textures/sand_prpl02", "textures/makkon_nature/sand_prpl02"},
	{"textures/sand_prplsm01", "textures/makkon_nature/sand_prplsm01"},
	{"textures/sand_prplsm02", "textures/makkon_nature/sand_prplsm02"},
	{"textures/sand_red101", "textures/makkon_nature/sand_red101"},
	{"textures/sand_red102", "textures/makkon_nature/sand_red102"},
	{"textures/sand_red1sm01", "textures/makkon_nature/sand_red1sm01"},
	{"textures/sand_red1sm02", "textures/makkon_nature/sand_red1sm02"},
	{"textures/sand_red201", "textures/makkon_nature/sand_red201"},
	{"textures/sand_red202", "textures/makkon_nature/sand_red202"},
	{"textures/sand_red2sm01", "textures/makkon_nature/sand_red2sm01"},
	{"textures/sand_red2sm02", "textures/makkon_nature/sand_red2sm02"},
	{"textures/sand_tan01", "textures/makkon_nature/sand_tan01"},
	{"textures/sand_tan02", "textures/makkon_nature/sand_tan02"},
	{"textures/sand_tansm01", "textures/makkon_nature/sand_tansm01"},
	{"textures/sand_tansm02", "textures/makkon_nature/sand_tansm02"},
	{"textures/sand_teal01", "textures/makkon_nature/sand_teal01"},
	{"textures/sand_teal02", "textures/makkon_nature/sand_teal02"},
	{"textures/sand_tealsm01", "textures/makkon_nature/sand_tealsm01"},
	{"textures/sand_tealsm02", "textures/makkon_nature/sand_tealsm02"},
	{"textures/sand_wht101", "textures/makkon_nature/sand_wht101"},
	{"textures/sand_wht102", "textures/makkon_nature/sand_wht102"},
	{"textures/sand_wht1sm01", "textures/makkon_nature/sand_wht1sm01"},
	{"textures/sand_wht1sm02", "textures/makkon_nature/sand_wht1sm02"},
	{"textures/sand_wht201", "textures/makkon_nature/sand_wht201"},
	{"textures/sand_wht202", "textures/makkon_nature/sand_wht202"},
	{"textures/sand_wht2sm01", "textures/makkon_nature/sand_wht2sm01"},
	{"textures/sand_wht2sm02", "textures/makkon_nature/sand_wht2sm02"},
	{"textures/treebark_01", "textures/makkon_nature/treebark_01"},
	{"textures/treebark_02", "textures/makkon_nature/treebark_02"},
	{"textures/treebark_03", "textures/makkon_nature/treebark_03"},
	{"textures/treebark_04", "textures/makkon_nature/treebark_04"},
	{"textures/treebark_05", "textures/makkon_nature/treebark_05"},
	{"textures/treebark_06", "textures/makkon_nature/treebark_06"},
	{"textures/treebark_blck1", "textures/makkon_nature/treebark_blck1"},
	{"textures/treebark_blck2", "textures/makkon_nature/treebark_blck2"},
	{"textures/treebark_blck3", "textures/makkon_nature/treebark_blck3"},
	{"textures/treebark_blck4", "textures/makkon_nature/treebark_blck4"},
	{"textures/treebark_blue1", "textures/makkon_nature/treebark_blue1"},
	{"textures/treebark_blue2", "textures/makkon_nature/treebark_blue2"},
	{"textures/treebark_blue3", "textures/makkon_nature/treebark_blue3"},
	{"textures/treebark_blue4", "textures/makkon_nature/treebark_blue4"},
	{"textures/treebark_brwn1", "textures/makkon_nature/treebark_brwn1"},
	{"textures/treebark_brwn2", "textures/makkon_nature/treebark_brwn2"},
	{"textures/treebark_brwn3", "textures/makkon_nature/treebark_brwn3"},
	{"textures/treebark_brwn4", "textures/makkon_nature/treebark_brwn4"},
	{"textures/treebark_burn01", "textures/makkon_nature/treebark_burn01"},
	{"textures/treebark_burn02", "textures/makkon_nature/treebark_burn02"},
	{"textures/treebark_grey1", "textures/makkon_nature/treebark_grey1"},
	{"textures/treebark_grey2", "textures/makkon_nature/treebark_grey2"},
	{"textures/treebark_grey3", "textures/makkon_nature/treebark_grey3"},
	{"textures/treebark_grey4", "textures/makkon_nature/treebark_grey4"},
	{"textures/treebark_pink1", "textures/makkon_nature/treebark_pink1"},
	{"textures/treebark_pink2", "textures/makkon_nature/treebark_pink2"},
	{"textures/treebark_pink3", "textures/makkon_nature/treebark_pink3"},
	{"textures/treebark_pink4", "textures/makkon_nature/treebark_pink4"},
	{"textures/treebark_prpl1", "textures/makkon_nature/treebark_prpl1"},
	{"textures/treebark_prpl2", "textures/makkon_nature/treebark_prpl2"},
	{"textures/treebark_prpl3", "textures/makkon_nature/treebark_prpl3"},
	{"textures/treebark_prpl4", "textures/makkon_nature/treebark_prpl4"},
	{"textures/treebark_red1", "textures/makkon_nature/treebark_red1"},
	{"textures/treebark_red2", "textures/makkon_nature/treebark_red2"},
	{"textures/treebark_red3", "textures/makkon_nature/treebark_red3"},
	{"textures/treebark_red4", "textures/makkon_nature/treebark_red4"},
	{"textures/treebark_tan1", "textures/makkon_nature/treebark_tan1"},
	{"textures/treebark_tan2", "textures/makkon_nature/treebark_tan2"},
	{"textures/treebark_tan3", "textures/makkon_nature/treebark_tan3"},
	{"textures/treebark_tan4", "textures/makkon_nature/treebark_tan4"},
	{"textures/treebark_teal1", "textures/makkon_nature/treebark_teal1"},
	{"textures/treebark_teal2", "textures/makkon_nature/treebark_teal2"},
	{"textures/treebark_teal3", "textures/makkon_nature/treebark_teal3"},
	{"textures/treebark_teal4", "textures/makkon_nature/treebark_teal4"},
	{"textures/treebark_wht1", "textures/makkon_nature/treebark_wht1"},
	{"textures/treebark_wht2", "textures/makkon_nature/treebark_wht2"},
	{"textures/treebark_wht3", "textures/makkon_nature/treebark_wht3"},
	{"textures/treebark_wht4", "textures/makkon_nature/treebark_wht4"},

	// Makkon stone 2022.8
	{"textures/__0stone_blck01", "textures/makkon_stone/__0stone_blck01"},
	{"textures/__0stone_blue01", "textures/makkon_stone/__0stone_blue01"},
	{"textures/__0stone_brwn01", "textures/makkon_stone/__0stone_brwn01"},
	{"textures/__0stone_grey01", "textures/makkon_stone/__0stone_grey01"},
	{"textures/__0stone_pink01", "textures/makkon_stone/__0stone_pink01"},
	{"textures/__0stone_prpl01", "textures/makkon_stone/__0stone_prpl01"},
	{"textures/__0stone_red01", "textures/makkon_stone/__0stone_red01"},
	{"textures/__0stone_tan01", "textures/makkon_stone/__0stone_tan01"},
	{"textures/__0stone_teal01", "textures/makkon_stone/__0stone_teal01"},
	{"textures/__0stone_wht01", "textures/makkon_stone/__0stone_wht01"},
	{"textures/__astone_blck01", "textures/makkon_stone/__astone_blck01"},
	{"textures/__astone_blue01", "textures/makkon_stone/__astone_blue01"},
	{"textures/__astone_brwn01", "textures/makkon_stone/__astone_brwn01"},
	{"textures/__astone_grey01", "textures/makkon_stone/__astone_grey01"},
	{"textures/__astone_prpl01", "textures/makkon_stone/__astone_prpl01"},
	{"textures/__astone_red01", "textures/makkon_stone/__astone_red01"},
	{"textures/__astone_tan01", "textures/makkon_stone/__astone_tan01"},
	{"textures/__astone_teal01", "textures/makkon_stone/__astone_teal01"},
	{"textures/__astone_wht01", "textures/makkon_stone/__astone_wht01"},
	{"textures/__bstone_pink01", "textures/makkon_stone/__bstone_pink01"},
	{"textures/stone_blck2_1", "textures/makkon_stone/stone_blck2_1"},
	{"textures/stone_blck2_10", "textures/makkon_stone/stone_blck2_10"},
	{"textures/stone_blck2_11", "textures/makkon_stone/stone_blck2_11"},
	{"textures/stone_blck2_12", "textures/makkon_stone/stone_blck2_12"},
	{"textures/stone_blck2_13", "textures/makkon_stone/stone_blck2_13"},
	{"textures/stone_blck2_14", "textures/makkon_stone/stone_blck2_14"},
	{"textures/stone_blck2_15", "textures/makkon_stone/stone_blck2_15"},
	{"textures/stone_blck2_16", "textures/makkon_stone/stone_blck2_16"},
	{"textures/stone_blck2_17", "textures/makkon_stone/stone_blck2_17"},
	{"textures/stone_blck2_18", "textures/makkon_stone/stone_blck2_18"},
	{"textures/stone_blck2_19", "textures/makkon_stone/stone_blck2_19"},
	{"textures/stone_blck2_1_1", "textures/makkon_stone/stone_blck2_1_1"},
	{"textures/stone_blck2_1_2", "textures/makkon_stone/stone_blck2_1_2"},
	{"textures/stone_blck2_1_3", "textures/makkon_stone/stone_blck2_1_3"},
	{"textures/stone_blck2_1_4", "textures/makkon_stone/stone_blck2_1_4"},
	{"textures/stone_blck2_1_5", "textures/makkon_stone/stone_blck2_1_5"},
	{"textures/stone_blck2_2", "textures/makkon_stone/stone_blck2_2"},
	{"textures/stone_blck2_20", "textures/makkon_stone/stone_blck2_20"},
	{"textures/stone_blck2_21", "textures/makkon_stone/stone_blck2_21"},
	{"textures/stone_blck2_22", "textures/makkon_stone/stone_blck2_22"},
	{"textures/stone_blck2_23", "textures/makkon_stone/stone_blck2_23"},
	{"textures/stone_blck2_24", "textures/makkon_stone/stone_blck2_24"},
	{"textures/stone_blck2_25", "textures/makkon_stone/stone_blck2_25"},
	{"textures/stone_blck2_26", "textures/makkon_stone/stone_blck2_26"},
	{"textures/stone_blck2_3", "textures/makkon_stone/stone_blck2_3"},
	{"textures/stone_blck2_4", "textures/makkon_stone/stone_blck2_4"},
	{"textures/stone_blck2_5", "textures/makkon_stone/stone_blck2_5"},
	{"textures/stone_blck2_6", "textures/makkon_stone/stone_blck2_6"},
	{"textures/stone_blck2_7", "textures/makkon_stone/stone_blck2_7"},
	{"textures/stone_blck2_8", "textures/makkon_stone/stone_blck2_8"},
	{"textures/stone_blck2_9", "textures/makkon_stone/stone_blck2_9"},
	{"textures/stone_blck2_c1", "textures/makkon_stone/stone_blck2_c1"},
	{"textures/stone_blck2_c16", "textures/makkon_stone/stone_blck2_c16"},
	{"textures/stone_blck2_c17", "textures/makkon_stone/stone_blck2_c17"},
	{"textures/stone_blck2_c2", "textures/makkon_stone/stone_blck2_c2"},
	{"textures/stone_blck2_c3", "textures/makkon_stone/stone_blck2_c3"},
	{"textures/stone_blck2_c4", "textures/makkon_stone/stone_blck2_c4"},
	{"textures/stone_blck2_c5", "textures/makkon_stone/stone_blck2_c5"},
	{"textures/stone_blck2_t10", "textures/makkon_stone/stone_blck2_t10"},
	{"textures/stone_blck2_t11", "textures/makkon_stone/stone_blck2_t11"},
	{"textures/stone_blck2_t12", "textures/makkon_stone/stone_blck2_t12"},
	{"textures/stone_blck2_t13", "textures/makkon_stone/stone_blck2_t13"},
	{"textures/stone_blck2_t14", "textures/makkon_stone/stone_blck2_t14"},
	{"textures/stone_blck2_t15", "textures/makkon_stone/stone_blck2_t15"},
	{"textures/stone_blck2_t16", "textures/makkon_stone/stone_blck2_t16"},
	{"textures/stone_blck2_t17", "textures/makkon_stone/stone_blck2_t17"},
	{"textures/stone_blck2_t5", "textures/makkon_stone/stone_blck2_t5"},
	{"textures/stone_blck2_t6", "textures/makkon_stone/stone_blck2_t6"},
	{"textures/stone_blck2_t7", "textures/makkon_stone/stone_blck2_t7"},
	{"textures/stone_blck2_t8", "textures/makkon_stone/stone_blck2_t8"},
	{"textures/stone_blck_1", "textures/makkon_stone/stone_blck_1"},
	{"textures/stone_blck_10", "textures/makkon_stone/stone_blck_10"},
	{"textures/stone_blck_11", "textures/makkon_stone/stone_blck_11"},
	{"textures/stone_blck_12", "textures/makkon_stone/stone_blck_12"},
	{"textures/stone_blck_13", "textures/makkon_stone/stone_blck_13"},
	{"textures/stone_blck_14", "textures/makkon_stone/stone_blck_14"},
	{"textures/stone_blck_15", "textures/makkon_stone/stone_blck_15"},
	{"textures/stone_blck_16", "textures/makkon_stone/stone_blck_16"},
	{"textures/stone_blck_17", "textures/makkon_stone/stone_blck_17"},
	{"textures/stone_blck_18", "textures/makkon_stone/stone_blck_18"},
	{"textures/stone_blck_19", "textures/makkon_stone/stone_blck_19"},
	{"textures/stone_blck_1_1", "textures/makkon_stone/stone_blck_1_1"},
	{"textures/stone_blck_1_2", "textures/makkon_stone/stone_blck_1_2"},
	{"textures/stone_blck_1_3", "textures/makkon_stone/stone_blck_1_3"},
	{"textures/stone_blck_1_4", "textures/makkon_stone/stone_blck_1_4"},
	{"textures/stone_blck_2", "textures/makkon_stone/stone_blck_2"},
	{"textures/stone_blck_20", "textures/makkon_stone/stone_blck_20"},
	{"textures/stone_blck_21", "textures/makkon_stone/stone_blck_21"},
	{"textures/stone_blck_22", "textures/makkon_stone/stone_blck_22"},
	{"textures/stone_blck_23", "textures/makkon_stone/stone_blck_23"},
	{"textures/stone_blck_24", "textures/makkon_stone/stone_blck_24"},
	{"textures/stone_blck_25", "textures/makkon_stone/stone_blck_25"},
	{"textures/stone_blck_26", "textures/makkon_stone/stone_blck_26"},
	{"textures/stone_blck_3", "textures/makkon_stone/stone_blck_3"},
	{"textures/stone_blck_4", "textures/makkon_stone/stone_blck_4"},
	{"textures/stone_blck_5", "textures/makkon_stone/stone_blck_5"},
	{"textures/stone_blck_6", "textures/makkon_stone/stone_blck_6"},
	{"textures/stone_blck_7", "textures/makkon_stone/stone_blck_7"},
	{"textures/stone_blck_8", "textures/makkon_stone/stone_blck_8"},
	{"textures/stone_blck_9", "textures/makkon_stone/stone_blck_9"},
	{"textures/stone_blck_c1", "textures/makkon_stone/stone_blck_c1"},
	{"textures/stone_blck_c16", "textures/makkon_stone/stone_blck_c16"},
	{"textures/stone_blck_c17", "textures/makkon_stone/stone_blck_c17"},
	{"textures/stone_blck_c2", "textures/makkon_stone/stone_blck_c2"},
	{"textures/stone_blck_c3", "textures/makkon_stone/stone_blck_c3"},
	{"textures/stone_blck_c4", "textures/makkon_stone/stone_blck_c4"},
	{"textures/stone_blck_c5", "textures/makkon_stone/stone_blck_c5"},
	{"textures/stone_blck_t10", "textures/makkon_stone/stone_blck_t10"},
	{"textures/stone_blck_t11", "textures/makkon_stone/stone_blck_t11"},
	{"textures/stone_blck_t12", "textures/makkon_stone/stone_blck_t12"},
	{"textures/stone_blck_t13", "textures/makkon_stone/stone_blck_t13"},
	{"textures/stone_blck_t14", "textures/makkon_stone/stone_blck_t14"},
	{"textures/stone_blck_t15", "textures/makkon_stone/stone_blck_t15"},
	{"textures/stone_blck_t16", "textures/makkon_stone/stone_blck_t16"},
	{"textures/stone_blck_t17", "textures/makkon_stone/stone_blck_t17"},
	{"textures/stone_blck_t3", "textures/makkon_stone/stone_blck_t3"},
	{"textures/stone_blck_t4", "textures/makkon_stone/stone_blck_t4"},
	{"textures/stone_blck_t5", "textures/makkon_stone/stone_blck_t5"},
	{"textures/stone_blck_t6", "textures/makkon_stone/stone_blck_t6"},
	{"textures/stone_blck_t7", "textures/makkon_stone/stone_blck_t7"},
	{"textures/stone_blck_t8", "textures/makkon_stone/stone_blck_t8"},
	{"textures/stone_blk2_c1", "textures/makkon_stone/stone_blk2_c1"},
	{"textures/stone_blk2_c2", "textures/makkon_stone/stone_blk2_c2"},
	{"textures/stone_blk2_c3", "textures/makkon_stone/stone_blk2_c3"},
	{"textures/stone_blk2_t1", "textures/makkon_stone/stone_blk2_t1"},
	{"textures/stone_blk2_t2", "textures/makkon_stone/stone_blk2_t2"},
	{"textures/stone_blk2_t3", "textures/makkon_stone/stone_blk2_t3"},
	{"textures/stone_blk2_t4", "textures/makkon_stone/stone_blk2_t4"},
	{"textures/stone_blk_1", "textures/makkon_stone/stone_blk_1"},
	{"textures/stone_blk_2", "textures/makkon_stone/stone_blk_2"},
	{"textures/stone_blk_3", "textures/makkon_stone/stone_blk_3"},
	{"textures/stone_blk_4", "textures/makkon_stone/stone_blk_4"},
	{"textures/stone_blk_5", "textures/makkon_stone/stone_blk_5"},
	{"textures/stone_blk_c1", "textures/makkon_stone/stone_blk_c1"},
	{"textures/stone_blk_t1", "textures/makkon_stone/stone_blk_t1"},
	{"textures/stone_blk_t2", "textures/makkon_stone/stone_blk_t2"},
	{"textures/stone_blk_t3", "textures/makkon_stone/stone_blk_t3"},
	{"textures/stone_blk_t4", "textures/makkon_stone/stone_blk_t4"},
	{"textures/stone_blue2_1", "textures/makkon_stone/stone_blue2_1"},
	{"textures/stone_blue2_10", "textures/makkon_stone/stone_blue2_10"},
	{"textures/stone_blue2_11", "textures/makkon_stone/stone_blue2_11"},
	{"textures/stone_blue2_12", "textures/makkon_stone/stone_blue2_12"},
	{"textures/stone_blue2_13", "textures/makkon_stone/stone_blue2_13"},
	{"textures/stone_blue2_14", "textures/makkon_stone/stone_blue2_14"},
	{"textures/stone_blue2_15", "textures/makkon_stone/stone_blue2_15"},
	{"textures/stone_blue2_16", "textures/makkon_stone/stone_blue2_16"},
	{"textures/stone_blue2_17", "textures/makkon_stone/stone_blue2_17"},
	{"textures/stone_blue2_18", "textures/makkon_stone/stone_blue2_18"},
	{"textures/stone_blue2_19", "textures/makkon_stone/stone_blue2_19"},
	{"textures/stone_blue2_1_1", "textures/makkon_stone/stone_blue2_1_1"},
	{"textures/stone_blue2_1_2", "textures/makkon_stone/stone_blue2_1_2"},
	{"textures/stone_blue2_1_3", "textures/makkon_stone/stone_blue2_1_3"},
	{"textures/stone_blue2_1_4", "textures/makkon_stone/stone_blue2_1_4"},
	{"textures/stone_blue2_1_5", "textures/makkon_stone/stone_blue2_1_5"},
	{"textures/stone_blue2_2", "textures/makkon_stone/stone_blue2_2"},
	{"textures/stone_blue2_20", "textures/makkon_stone/stone_blue2_20"},
	{"textures/stone_blue2_21", "textures/makkon_stone/stone_blue2_21"},
	{"textures/stone_blue2_22", "textures/makkon_stone/stone_blue2_22"},
	{"textures/stone_blue2_23", "textures/makkon_stone/stone_blue2_23"},
	{"textures/stone_blue2_24", "textures/makkon_stone/stone_blue2_24"},
	{"textures/stone_blue2_25", "textures/makkon_stone/stone_blue2_25"},
	{"textures/stone_blue2_26", "textures/makkon_stone/stone_blue2_26"},
	{"textures/stone_blue2_3", "textures/makkon_stone/stone_blue2_3"},
	{"textures/stone_blue2_4", "textures/makkon_stone/stone_blue2_4"},
	{"textures/stone_blue2_5", "textures/makkon_stone/stone_blue2_5"},
	{"textures/stone_blue2_6", "textures/makkon_stone/stone_blue2_6"},
	{"textures/stone_blue2_7", "textures/makkon_stone/stone_blue2_7"},
	{"textures/stone_blue2_8", "textures/makkon_stone/stone_blue2_8"},
	{"textures/stone_blue2_9", "textures/makkon_stone/stone_blue2_9"},
	{"textures/stone_blue2_c1", "textures/makkon_stone/stone_blue2_c1"},
	{"textures/stone_blue2_c16", "textures/makkon_stone/stone_blue2_c16"},
	{"textures/stone_blue2_c17", "textures/makkon_stone/stone_blue2_c17"},
	{"textures/stone_blue2_c2", "textures/makkon_stone/stone_blue2_c2"},
	{"textures/stone_blue2_c3", "textures/makkon_stone/stone_blue2_c3"},
	{"textures/stone_blue2_c4", "textures/makkon_stone/stone_blue2_c4"},
	{"textures/stone_blue2_c5", "textures/makkon_stone/stone_blue2_c5"},
	{"textures/stone_blue2_t1", "textures/makkon_stone/stone_blue2_t1"},
	{"textures/stone_blue2_t10", "textures/makkon_stone/stone_blue2_t10"},
	{"textures/stone_blue2_t11", "textures/makkon_stone/stone_blue2_t11"},
	{"textures/stone_blue2_t12", "textures/makkon_stone/stone_blue2_t12"},
	{"textures/stone_blue2_t13", "textures/makkon_stone/stone_blue2_t13"},
	{"textures/stone_blue2_t14", "textures/makkon_stone/stone_blue2_t14"},
	{"textures/stone_blue2_t15", "textures/makkon_stone/stone_blue2_t15"},
	{"textures/stone_blue2_t16", "textures/makkon_stone/stone_blue2_t16"},
	{"textures/stone_blue2_t17", "textures/makkon_stone/stone_blue2_t17"},
	{"textures/stone_blue2_t2", "textures/makkon_stone/stone_blue2_t2"},
	{"textures/stone_blue2_t3", "textures/makkon_stone/stone_blue2_t3"},
	{"textures/stone_blue2_t4", "textures/makkon_stone/stone_blue2_t4"},
	{"textures/stone_blue2_t5", "textures/makkon_stone/stone_blue2_t5"},
	{"textures/stone_blue2_t6", "textures/makkon_stone/stone_blue2_t6"},
	{"textures/stone_blue2_t7", "textures/makkon_stone/stone_blue2_t7"},
	{"textures/stone_blue2_t8", "textures/makkon_stone/stone_blue2_t8"},
	{"textures/stone_blue_1", "textures/makkon_stone/stone_blue_1"},
	{"textures/stone_blue_10", "textures/makkon_stone/stone_blue_10"},
	{"textures/stone_blue_11", "textures/makkon_stone/stone_blue_11"},
	{"textures/stone_blue_12", "textures/makkon_stone/stone_blue_12"},
	{"textures/stone_blue_13", "textures/makkon_stone/stone_blue_13"},
	{"textures/stone_blue_14", "textures/makkon_stone/stone_blue_14"},
	{"textures/stone_blue_15", "textures/makkon_stone/stone_blue_15"},
	{"textures/stone_blue_16", "textures/makkon_stone/stone_blue_16"},
	{"textures/stone_blue_17", "textures/makkon_stone/stone_blue_17"},
	{"textures/stone_blue_18", "textures/makkon_stone/stone_blue_18"},
	{"textures/stone_blue_19", "textures/makkon_stone/stone_blue_19"},
	{"textures/stone_blue_1_1", "textures/makkon_stone/stone_blue_1_1"},
	{"textures/stone_blue_1_2", "textures/makkon_stone/stone_blue_1_2"},
	{"textures/stone_blue_1_3", "textures/makkon_stone/stone_blue_1_3"},
	{"textures/stone_blue_1_4", "textures/makkon_stone/stone_blue_1_4"},
	{"textures/stone_blue_2", "textures/makkon_stone/stone_blue_2"},
	{"textures/stone_blue_20", "textures/makkon_stone/stone_blue_20"},
	{"textures/stone_blue_21", "textures/makkon_stone/stone_blue_21"},
	{"textures/stone_blue_22", "textures/makkon_stone/stone_blue_22"},
	{"textures/stone_blue_23", "textures/makkon_stone/stone_blue_23"},
	{"textures/stone_blue_24", "textures/makkon_stone/stone_blue_24"},
	{"textures/stone_blue_25", "textures/makkon_stone/stone_blue_25"},
	{"textures/stone_blue_26", "textures/makkon_stone/stone_blue_26"},
	{"textures/stone_blue_3", "textures/makkon_stone/stone_blue_3"},
	{"textures/stone_blue_4", "textures/makkon_stone/stone_blue_4"},
	{"textures/stone_blue_5", "textures/makkon_stone/stone_blue_5"},
	{"textures/stone_blue_6", "textures/makkon_stone/stone_blue_6"},
	{"textures/stone_blue_7", "textures/makkon_stone/stone_blue_7"},
	{"textures/stone_blue_8", "textures/makkon_stone/stone_blue_8"},
	{"textures/stone_blue_9", "textures/makkon_stone/stone_blue_9"},
	{"textures/stone_blue_c1", "textures/makkon_stone/stone_blue_c1"},
	{"textures/stone_blue_c16", "textures/makkon_stone/stone_blue_c16"},
	{"textures/stone_blue_c17", "textures/makkon_stone/stone_blue_c17"},
	{"textures/stone_blue_c2", "textures/makkon_stone/stone_blue_c2"},
	{"textures/stone_blue_c3", "textures/makkon_stone/stone_blue_c3"},
	{"textures/stone_blue_c4", "textures/makkon_stone/stone_blue_c4"},
	{"textures/stone_blue_c5", "textures/makkon_stone/stone_blue_c5"},
	{"textures/stone_blue_t1", "textures/makkon_stone/stone_blue_t1"},
	{"textures/stone_blue_t10", "textures/makkon_stone/stone_blue_t10"},
	{"textures/stone_blue_t11", "textures/makkon_stone/stone_blue_t11"},
	{"textures/stone_blue_t12", "textures/makkon_stone/stone_blue_t12"},
	{"textures/stone_blue_t13", "textures/makkon_stone/stone_blue_t13"},
	{"textures/stone_blue_t14", "textures/makkon_stone/stone_blue_t14"},
	{"textures/stone_blue_t15", "textures/makkon_stone/stone_blue_t15"},
	{"textures/stone_blue_t16", "textures/makkon_stone/stone_blue_t16"},
	{"textures/stone_blue_t17", "textures/makkon_stone/stone_blue_t17"},
	{"textures/stone_blue_t2", "textures/makkon_stone/stone_blue_t2"},
	{"textures/stone_blue_t3", "textures/makkon_stone/stone_blue_t3"},
	{"textures/stone_blue_t4", "textures/makkon_stone/stone_blue_t4"},
	{"textures/stone_blue_t5", "textures/makkon_stone/stone_blue_t5"},
	{"textures/stone_blue_t6", "textures/makkon_stone/stone_blue_t6"},
	{"textures/stone_blue_t7", "textures/makkon_stone/stone_blue_t7"},
	{"textures/stone_blue_t8", "textures/makkon_stone/stone_blue_t8"},
	{"textures/stone_brwn2_1", "textures/makkon_stone/stone_brwn2_1"},
	{"textures/stone_brwn2_10", "textures/makkon_stone/stone_brwn2_10"},
	{"textures/stone_brwn2_11", "textures/makkon_stone/stone_brwn2_11"},
	{"textures/stone_brwn2_12", "textures/makkon_stone/stone_brwn2_12"},
	{"textures/stone_brwn2_13", "textures/makkon_stone/stone_brwn2_13"},
	{"textures/stone_brwn2_14", "textures/makkon_stone/stone_brwn2_14"},
	{"textures/stone_brwn2_15", "textures/makkon_stone/stone_brwn2_15"},
	{"textures/stone_brwn2_16", "textures/makkon_stone/stone_brwn2_16"},
	{"textures/stone_brwn2_17", "textures/makkon_stone/stone_brwn2_17"},
	{"textures/stone_brwn2_18", "textures/makkon_stone/stone_brwn2_18"},
	{"textures/stone_brwn2_19", "textures/makkon_stone/stone_brwn2_19"},
	{"textures/stone_brwn2_1_1", "textures/makkon_stone/stone_brwn2_1_1"},
	{"textures/stone_brwn2_1_2", "textures/makkon_stone/stone_brwn2_1_2"},
	{"textures/stone_brwn2_1_3", "textures/makkon_stone/stone_brwn2_1_3"},
	{"textures/stone_brwn2_1_4", "textures/makkon_stone/stone_brwn2_1_4"},
	{"textures/stone_brwn2_1_5", "textures/makkon_stone/stone_brwn2_1_5"},
	{"textures/stone_brwn2_2", "textures/makkon_stone/stone_brwn2_2"},
	{"textures/stone_brwn2_20", "textures/makkon_stone/stone_brwn2_20"},
	{"textures/stone_brwn2_21", "textures/makkon_stone/stone_brwn2_21"},
	{"textures/stone_brwn2_22", "textures/makkon_stone/stone_brwn2_22"},
	{"textures/stone_brwn2_23", "textures/makkon_stone/stone_brwn2_23"},
	{"textures/stone_brwn2_24", "textures/makkon_stone/stone_brwn2_24"},
	{"textures/stone_brwn2_25", "textures/makkon_stone/stone_brwn2_25"},
	{"textures/stone_brwn2_26", "textures/makkon_stone/stone_brwn2_26"},
	{"textures/stone_brwn2_3", "textures/makkon_stone/stone_brwn2_3"},
	{"textures/stone_brwn2_4", "textures/makkon_stone/stone_brwn2_4"},
	{"textures/stone_brwn2_5", "textures/makkon_stone/stone_brwn2_5"},
	{"textures/stone_brwn2_6", "textures/makkon_stone/stone_brwn2_6"},
	{"textures/stone_brwn2_7", "textures/makkon_stone/stone_brwn2_7"},
	{"textures/stone_brwn2_8", "textures/makkon_stone/stone_brwn2_8"},
	{"textures/stone_brwn2_9", "textures/makkon_stone/stone_brwn2_9"},
	{"textures/stone_brwn2_c1", "textures/makkon_stone/stone_brwn2_c1"},
	{"textures/stone_brwn2_c16", "textures/makkon_stone/stone_brwn2_c16"},
	{"textures/stone_brwn2_c17", "textures/makkon_stone/stone_brwn2_c17"},
	{"textures/stone_brwn2_c2", "textures/makkon_stone/stone_brwn2_c2"},
	{"textures/stone_brwn2_c3", "textures/makkon_stone/stone_brwn2_c3"},
	{"textures/stone_brwn2_c4", "textures/makkon_stone/stone_brwn2_c4"},
	{"textures/stone_brwn2_c5", "textures/makkon_stone/stone_brwn2_c5"},
	{"textures/stone_brwn2_t1", "textures/makkon_stone/stone_brwn2_t1"},
	{"textures/stone_brwn2_t10", "textures/makkon_stone/stone_brwn2_t10"},
	{"textures/stone_brwn2_t11", "textures/makkon_stone/stone_brwn2_t11"},
	{"textures/stone_brwn2_t12", "textures/makkon_stone/stone_brwn2_t12"},
	{"textures/stone_brwn2_t13", "textures/makkon_stone/stone_brwn2_t13"},
	{"textures/stone_brwn2_t14", "textures/makkon_stone/stone_brwn2_t14"},
	{"textures/stone_brwn2_t15", "textures/makkon_stone/stone_brwn2_t15"},
	{"textures/stone_brwn2_t16", "textures/makkon_stone/stone_brwn2_t16"},
	{"textures/stone_brwn2_t17", "textures/makkon_stone/stone_brwn2_t17"},
	{"textures/stone_brwn2_t2", "textures/makkon_stone/stone_brwn2_t2"},
	{"textures/stone_brwn2_t3", "textures/makkon_stone/stone_brwn2_t3"},
	{"textures/stone_brwn2_t4", "textures/makkon_stone/stone_brwn2_t4"},
	{"textures/stone_brwn2_t5", "textures/makkon_stone/stone_brwn2_t5"},
	{"textures/stone_brwn2_t6", "textures/makkon_stone/stone_brwn2_t6"},
	{"textures/stone_brwn2_t7", "textures/makkon_stone/stone_brwn2_t7"},
	{"textures/stone_brwn2_t8", "textures/makkon_stone/stone_brwn2_t8"},
	{"textures/stone_brwn_1", "textures/makkon_stone/stone_brwn_1"},
	{"textures/stone_brwn_10", "textures/makkon_stone/stone_brwn_10"},
	{"textures/stone_brwn_11", "textures/makkon_stone/stone_brwn_11"},
	{"textures/stone_brwn_12", "textures/makkon_stone/stone_brwn_12"},
	{"textures/stone_brwn_13", "textures/makkon_stone/stone_brwn_13"},
	{"textures/stone_brwn_14", "textures/makkon_stone/stone_brwn_14"},
	{"textures/stone_brwn_15", "textures/makkon_stone/stone_brwn_15"},
	{"textures/stone_brwn_16", "textures/makkon_stone/stone_brwn_16"},
	{"textures/stone_brwn_17", "textures/makkon_stone/stone_brwn_17"},
	{"textures/stone_brwn_18", "textures/makkon_stone/stone_brwn_18"},
	{"textures/stone_brwn_19", "textures/makkon_stone/stone_brwn_19"},
	{"textures/stone_brwn_1_1", "textures/makkon_stone/stone_brwn_1_1"},
	{"textures/stone_brwn_1_2", "textures/makkon_stone/stone_brwn_1_2"},
	{"textures/stone_brwn_1_3", "textures/makkon_stone/stone_brwn_1_3"},
	{"textures/stone_brwn_1_4", "textures/makkon_stone/stone_brwn_1_4"},
	{"textures/stone_brwn_2", "textures/makkon_stone/stone_brwn_2"},
	{"textures/stone_brwn_20", "textures/makkon_stone/stone_brwn_20"},
	{"textures/stone_brwn_21", "textures/makkon_stone/stone_brwn_21"},
	{"textures/stone_brwn_22", "textures/makkon_stone/stone_brwn_22"},
	{"textures/stone_brwn_23", "textures/makkon_stone/stone_brwn_23"},
	{"textures/stone_brwn_24", "textures/makkon_stone/stone_brwn_24"},
	{"textures/stone_brwn_25", "textures/makkon_stone/stone_brwn_25"},
	{"textures/stone_brwn_26", "textures/makkon_stone/stone_brwn_26"},
	{"textures/stone_brwn_3", "textures/makkon_stone/stone_brwn_3"},
	{"textures/stone_brwn_4", "textures/makkon_stone/stone_brwn_4"},
	{"textures/stone_brwn_5", "textures/makkon_stone/stone_brwn_5"},
	{"textures/stone_brwn_6", "textures/makkon_stone/stone_brwn_6"},
	{"textures/stone_brwn_7", "textures/makkon_stone/stone_brwn_7"},
	{"textures/stone_brwn_8", "textures/makkon_stone/stone_brwn_8"},
	{"textures/stone_brwn_9", "textures/makkon_stone/stone_brwn_9"},
	{"textures/stone_brwn_c1", "textures/makkon_stone/stone_brwn_c1"},
	{"textures/stone_brwn_c16", "textures/makkon_stone/stone_brwn_c16"},
	{"textures/stone_brwn_c17", "textures/makkon_stone/stone_brwn_c17"},
	{"textures/stone_brwn_c2", "textures/makkon_stone/stone_brwn_c2"},
	{"textures/stone_brwn_c3", "textures/makkon_stone/stone_brwn_c3"},
	{"textures/stone_brwn_c4", "textures/makkon_stone/stone_brwn_c4"},
	{"textures/stone_brwn_c5", "textures/makkon_stone/stone_brwn_c5"},
	{"textures/stone_brwn_t1", "textures/makkon_stone/stone_brwn_t1"},
	{"textures/stone_brwn_t10", "textures/makkon_stone/stone_brwn_t10"},
	{"textures/stone_brwn_t11", "textures/makkon_stone/stone_brwn_t11"},
	{"textures/stone_brwn_t12", "textures/makkon_stone/stone_brwn_t12"},
	{"textures/stone_brwn_t13", "textures/makkon_stone/stone_brwn_t13"},
	{"textures/stone_brwn_t14", "textures/makkon_stone/stone_brwn_t14"},
	{"textures/stone_brwn_t15", "textures/makkon_stone/stone_brwn_t15"},
	{"textures/stone_brwn_t16", "textures/makkon_stone/stone_brwn_t16"},
	{"textures/stone_brwn_t17", "textures/makkon_stone/stone_brwn_t17"},
	{"textures/stone_brwn_t2", "textures/makkon_stone/stone_brwn_t2"},
	{"textures/stone_brwn_t3", "textures/makkon_stone/stone_brwn_t3"},
	{"textures/stone_brwn_t4", "textures/makkon_stone/stone_brwn_t4"},
	{"textures/stone_brwn_t5", "textures/makkon_stone/stone_brwn_t5"},
	{"textures/stone_brwn_t6", "textures/makkon_stone/stone_brwn_t6"},
	{"textures/stone_brwn_t7", "textures/makkon_stone/stone_brwn_t7"},
	{"textures/stone_brwn_t8", "textures/makkon_stone/stone_brwn_t8"},
	{"textures/stone_grey2_1", "textures/makkon_stone/stone_grey2_1"},
	{"textures/stone_grey2_10", "textures/makkon_stone/stone_grey2_10"},
	{"textures/stone_grey2_11", "textures/makkon_stone/stone_grey2_11"},
	{"textures/stone_grey2_12", "textures/makkon_stone/stone_grey2_12"},
	{"textures/stone_grey2_13", "textures/makkon_stone/stone_grey2_13"},
	{"textures/stone_grey2_14", "textures/makkon_stone/stone_grey2_14"},
	{"textures/stone_grey2_15", "textures/makkon_stone/stone_grey2_15"},
	{"textures/stone_grey2_16", "textures/makkon_stone/stone_grey2_16"},
	{"textures/stone_grey2_17", "textures/makkon_stone/stone_grey2_17"},
	{"textures/stone_grey2_18", "textures/makkon_stone/stone_grey2_18"},
	{"textures/stone_grey2_19", "textures/makkon_stone/stone_grey2_19"},
	{"textures/stone_grey2_1_1", "textures/makkon_stone/stone_grey2_1_1"},
	{"textures/stone_grey2_1_2", "textures/makkon_stone/stone_grey2_1_2"},
	{"textures/stone_grey2_1_3", "textures/makkon_stone/stone_grey2_1_3"},
	{"textures/stone_grey2_1_4", "textures/makkon_stone/stone_grey2_1_4"},
	{"textures/stone_grey2_1_5", "textures/makkon_stone/stone_grey2_1_5"},
	{"textures/stone_grey2_2", "textures/makkon_stone/stone_grey2_2"},
	{"textures/stone_grey2_20", "textures/makkon_stone/stone_grey2_20"},
	{"textures/stone_grey2_21", "textures/makkon_stone/stone_grey2_21"},
	{"textures/stone_grey2_22", "textures/makkon_stone/stone_grey2_22"},
	{"textures/stone_grey2_23", "textures/makkon_stone/stone_grey2_23"},
	{"textures/stone_grey2_24", "textures/makkon_stone/stone_grey2_24"},
	{"textures/stone_grey2_25", "textures/makkon_stone/stone_grey2_25"},
	{"textures/stone_grey2_26", "textures/makkon_stone/stone_grey2_26"},
	{"textures/stone_grey2_3", "textures/makkon_stone/stone_grey2_3"},
	{"textures/stone_grey2_4", "textures/makkon_stone/stone_grey2_4"},
	{"textures/stone_grey2_5", "textures/makkon_stone/stone_grey2_5"},
	{"textures/stone_grey2_6", "textures/makkon_stone/stone_grey2_6"},
	{"textures/stone_grey2_7", "textures/makkon_stone/stone_grey2_7"},
	{"textures/stone_grey2_8", "textures/makkon_stone/stone_grey2_8"},
	{"textures/stone_grey2_9", "textures/makkon_stone/stone_grey2_9"},
	{"textures/stone_grey2_c1", "textures/makkon_stone/stone_grey2_c1"},
	{"textures/stone_grey2_c16", "textures/makkon_stone/stone_grey2_c16"},
	{"textures/stone_grey2_c17", "textures/makkon_stone/stone_grey2_c17"},
	{"textures/stone_grey2_c2", "textures/makkon_stone/stone_grey2_c2"},
	{"textures/stone_grey2_c3", "textures/makkon_stone/stone_grey2_c3"},
	{"textures/stone_grey2_c4", "textures/makkon_stone/stone_grey2_c4"},
	{"textures/stone_grey2_c5", "textures/makkon_stone/stone_grey2_c5"},
	{"textures/stone_grey2_t1", "textures/makkon_stone/stone_grey2_t1"},
	{"textures/stone_grey2_t10", "textures/makkon_stone/stone_grey2_t10"},
	{"textures/stone_grey2_t11", "textures/makkon_stone/stone_grey2_t11"},
	{"textures/stone_grey2_t12", "textures/makkon_stone/stone_grey2_t12"},
	{"textures/stone_grey2_t13", "textures/makkon_stone/stone_grey2_t13"},
	{"textures/stone_grey2_t14", "textures/makkon_stone/stone_grey2_t14"},
	{"textures/stone_grey2_t15", "textures/makkon_stone/stone_grey2_t15"},
	{"textures/stone_grey2_t16", "textures/makkon_stone/stone_grey2_t16"},
	{"textures/stone_grey2_t17", "textures/makkon_stone/stone_grey2_t17"},
	{"textures/stone_grey2_t2", "textures/makkon_stone/stone_grey2_t2"},
	{"textures/stone_grey2_t3", "textures/makkon_stone/stone_grey2_t3"},
	{"textures/stone_grey2_t4", "textures/makkon_stone/stone_grey2_t4"},
	{"textures/stone_grey2_t5", "textures/makkon_stone/stone_grey2_t5"},
	{"textures/stone_grey2_t6", "textures/makkon_stone/stone_grey2_t6"},
	{"textures/stone_grey2_t7", "textures/makkon_stone/stone_grey2_t7"},
	{"textures/stone_grey2_t8", "textures/makkon_stone/stone_grey2_t8"},
	{"textures/stone_grey_1", "textures/makkon_stone/stone_grey_1"},
	{"textures/stone_grey_10", "textures/makkon_stone/stone_grey_10"},
	{"textures/stone_grey_11", "textures/makkon_stone/stone_grey_11"},
	{"textures/stone_grey_12", "textures/makkon_stone/stone_grey_12"},
	{"textures/stone_grey_13", "textures/makkon_stone/stone_grey_13"},
	{"textures/stone_grey_14", "textures/makkon_stone/stone_grey_14"},
	{"textures/stone_grey_15", "textures/makkon_stone/stone_grey_15"},
	{"textures/stone_grey_16", "textures/makkon_stone/stone_grey_16"},
	{"textures/stone_grey_17", "textures/makkon_stone/stone_grey_17"},
	{"textures/stone_grey_18", "textures/makkon_stone/stone_grey_18"},
	{"textures/stone_grey_19", "textures/makkon_stone/stone_grey_19"},
	{"textures/stone_grey_1_1", "textures/makkon_stone/stone_grey_1_1"},
	{"textures/stone_grey_1_2", "textures/makkon_stone/stone_grey_1_2"},
	{"textures/stone_grey_1_3", "textures/makkon_stone/stone_grey_1_3"},
	{"textures/stone_grey_1_4", "textures/makkon_stone/stone_grey_1_4"},
	{"textures/stone_grey_2", "textures/makkon_stone/stone_grey_2"},
	{"textures/stone_grey_20", "textures/makkon_stone/stone_grey_20"},
	{"textures/stone_grey_21", "textures/makkon_stone/stone_grey_21"},
	{"textures/stone_grey_22", "textures/makkon_stone/stone_grey_22"},
	{"textures/stone_grey_23", "textures/makkon_stone/stone_grey_23"},
	{"textures/stone_grey_24", "textures/makkon_stone/stone_grey_24"},
	{"textures/stone_grey_25", "textures/makkon_stone/stone_grey_25"},
	{"textures/stone_grey_26", "textures/makkon_stone/stone_grey_26"},
	{"textures/stone_grey_3", "textures/makkon_stone/stone_grey_3"},
	{"textures/stone_grey_4", "textures/makkon_stone/stone_grey_4"},
	{"textures/stone_grey_5", "textures/makkon_stone/stone_grey_5"},
	{"textures/stone_grey_6", "textures/makkon_stone/stone_grey_6"},
	{"textures/stone_grey_7", "textures/makkon_stone/stone_grey_7"},
	{"textures/stone_grey_8", "textures/makkon_stone/stone_grey_8"},
	{"textures/stone_grey_9", "textures/makkon_stone/stone_grey_9"},
	{"textures/stone_grey_c1", "textures/makkon_stone/stone_grey_c1"},
	{"textures/stone_grey_c16", "textures/makkon_stone/stone_grey_c16"},
	{"textures/stone_grey_c17", "textures/makkon_stone/stone_grey_c17"},
	{"textures/stone_grey_c2", "textures/makkon_stone/stone_grey_c2"},
	{"textures/stone_grey_c3", "textures/makkon_stone/stone_grey_c3"},
	{"textures/stone_grey_c4", "textures/makkon_stone/stone_grey_c4"},
	{"textures/stone_grey_c5", "textures/makkon_stone/stone_grey_c5"},
	{"textures/stone_grey_t1", "textures/makkon_stone/stone_grey_t1"},
	{"textures/stone_grey_t10", "textures/makkon_stone/stone_grey_t10"},
	{"textures/stone_grey_t11", "textures/makkon_stone/stone_grey_t11"},
	{"textures/stone_grey_t12", "textures/makkon_stone/stone_grey_t12"},
	{"textures/stone_grey_t13", "textures/makkon_stone/stone_grey_t13"},
	{"textures/stone_grey_t14", "textures/makkon_stone/stone_grey_t14"},
	{"textures/stone_grey_t15", "textures/makkon_stone/stone_grey_t15"},
	{"textures/stone_grey_t16", "textures/makkon_stone/stone_grey_t16"},
	{"textures/stone_grey_t17", "textures/makkon_stone/stone_grey_t17"},
	{"textures/stone_grey_t2", "textures/makkon_stone/stone_grey_t2"},
	{"textures/stone_grey_t3", "textures/makkon_stone/stone_grey_t3"},
	{"textures/stone_grey_t4", "textures/makkon_stone/stone_grey_t4"},
	{"textures/stone_grey_t5", "textures/makkon_stone/stone_grey_t5"},
	{"textures/stone_grey_t6", "textures/makkon_stone/stone_grey_t6"},
	{"textures/stone_grey_t7", "textures/makkon_stone/stone_grey_t7"},
	{"textures/stone_grey_t8", "textures/makkon_stone/stone_grey_t8"},
	{"textures/stone_output_10", "textures/makkon_stone/stone_output_10"},
	{"textures/stone_pink2_1", "textures/makkon_stone/stone_pink2_1"},
	{"textures/stone_pink2_10", "textures/makkon_stone/stone_pink2_10"},
	{"textures/stone_pink2_11", "textures/makkon_stone/stone_pink2_11"},
	{"textures/stone_pink2_12", "textures/makkon_stone/stone_pink2_12"},
	{"textures/stone_pink2_13", "textures/makkon_stone/stone_pink2_13"},
	{"textures/stone_pink2_14", "textures/makkon_stone/stone_pink2_14"},
	{"textures/stone_pink2_15", "textures/makkon_stone/stone_pink2_15"},
	{"textures/stone_pink2_16", "textures/makkon_stone/stone_pink2_16"},
	{"textures/stone_pink2_17", "textures/makkon_stone/stone_pink2_17"},
	{"textures/stone_pink2_18", "textures/makkon_stone/stone_pink2_18"},
	{"textures/stone_pink2_19", "textures/makkon_stone/stone_pink2_19"},
	{"textures/stone_pink2_1_1", "textures/makkon_stone/stone_pink2_1_1"},
	{"textures/stone_pink2_1_2", "textures/makkon_stone/stone_pink2_1_2"},
	{"textures/stone_pink2_1_3", "textures/makkon_stone/stone_pink2_1_3"},
	{"textures/stone_pink2_1_4", "textures/makkon_stone/stone_pink2_1_4"},
	{"textures/stone_pink2_1_5", "textures/makkon_stone/stone_pink2_1_5"},
	{"textures/stone_pink2_2", "textures/makkon_stone/stone_pink2_2"},
	{"textures/stone_pink2_20", "textures/makkon_stone/stone_pink2_20"},
	{"textures/stone_pink2_21", "textures/makkon_stone/stone_pink2_21"},
	{"textures/stone_pink2_22", "textures/makkon_stone/stone_pink2_22"},
	{"textures/stone_pink2_23", "textures/makkon_stone/stone_pink2_23"},
	{"textures/stone_pink2_24", "textures/makkon_stone/stone_pink2_24"},
	{"textures/stone_pink2_25", "textures/makkon_stone/stone_pink2_25"},
	{"textures/stone_pink2_26", "textures/makkon_stone/stone_pink2_26"},
	{"textures/stone_pink2_3", "textures/makkon_stone/stone_pink2_3"},
	{"textures/stone_pink2_4", "textures/makkon_stone/stone_pink2_4"},
	{"textures/stone_pink2_5", "textures/makkon_stone/stone_pink2_5"},
	{"textures/stone_pink2_6", "textures/makkon_stone/stone_pink2_6"},
	{"textures/stone_pink2_7", "textures/makkon_stone/stone_pink2_7"},
	{"textures/stone_pink2_8", "textures/makkon_stone/stone_pink2_8"},
	{"textures/stone_pink2_9", "textures/makkon_stone/stone_pink2_9"},
	{"textures/stone_pink2_c1", "textures/makkon_stone/stone_pink2_c1"},
	{"textures/stone_pink2_c16", "textures/makkon_stone/stone_pink2_c16"},
	{"textures/stone_pink2_c17", "textures/makkon_stone/stone_pink2_c17"},
	{"textures/stone_pink2_c2", "textures/makkon_stone/stone_pink2_c2"},
	{"textures/stone_pink2_c3", "textures/makkon_stone/stone_pink2_c3"},
	{"textures/stone_pink2_c4", "textures/makkon_stone/stone_pink2_c4"},
	{"textures/stone_pink2_c5", "textures/makkon_stone/stone_pink2_c5"},
	{"textures/stone_pink2_t1", "textures/makkon_stone/stone_pink2_t1"},
	{"textures/stone_pink2_t10", "textures/makkon_stone/stone_pink2_t10"},
	{"textures/stone_pink2_t11", "textures/makkon_stone/stone_pink2_t11"},
	{"textures/stone_pink2_t12", "textures/makkon_stone/stone_pink2_t12"},
	{"textures/stone_pink2_t13", "textures/makkon_stone/stone_pink2_t13"},
	{"textures/stone_pink2_t14", "textures/makkon_stone/stone_pink2_t14"},
	{"textures/stone_pink2_t15", "textures/makkon_stone/stone_pink2_t15"},
	{"textures/stone_pink2_t16", "textures/makkon_stone/stone_pink2_t16"},
	{"textures/stone_pink2_t17", "textures/makkon_stone/stone_pink2_t17"},
	{"textures/stone_pink2_t3", "textures/makkon_stone/stone_pink2_t3"},
	{"textures/stone_pink2_t4", "textures/makkon_stone/stone_pink2_t4"},
	{"textures/stone_pink2_t5", "textures/makkon_stone/stone_pink2_t5"},
	{"textures/stone_pink2_t6", "textures/makkon_stone/stone_pink2_t6"},
	{"textures/stone_pink2_t7", "textures/makkon_stone/stone_pink2_t7"},
	{"textures/stone_pink2_t8", "textures/makkon_stone/stone_pink2_t8"},
	{"textures/stone_pink_1", "textures/makkon_stone/stone_pink_1"},
	{"textures/stone_pink_10", "textures/makkon_stone/stone_pink_10"},
	{"textures/stone_pink_11", "textures/makkon_stone/stone_pink_11"},
	{"textures/stone_pink_12", "textures/makkon_stone/stone_pink_12"},
	{"textures/stone_pink_13", "textures/makkon_stone/stone_pink_13"},
	{"textures/stone_pink_14", "textures/makkon_stone/stone_pink_14"},
	{"textures/stone_pink_15", "textures/makkon_stone/stone_pink_15"},
	{"textures/stone_pink_16", "textures/makkon_stone/stone_pink_16"},
	{"textures/stone_pink_17", "textures/makkon_stone/stone_pink_17"},
	{"textures/stone_pink_18", "textures/makkon_stone/stone_pink_18"},
	{"textures/stone_pink_19", "textures/makkon_stone/stone_pink_19"},
	{"textures/stone_pink_1_1", "textures/makkon_stone/stone_pink_1_1"},
	{"textures/stone_pink_1_2", "textures/makkon_stone/stone_pink_1_2"},
	{"textures/stone_pink_1_3", "textures/makkon_stone/stone_pink_1_3"},
	{"textures/stone_pink_1_4", "textures/makkon_stone/stone_pink_1_4"},
	{"textures/stone_pink_1_5", "textures/makkon_stone/stone_pink_1_5"},
	{"textures/stone_pink_2", "textures/makkon_stone/stone_pink_2"},
	{"textures/stone_pink_20", "textures/makkon_stone/stone_pink_20"},
	{"textures/stone_pink_21", "textures/makkon_stone/stone_pink_21"},
	{"textures/stone_pink_22", "textures/makkon_stone/stone_pink_22"},
	{"textures/stone_pink_23", "textures/makkon_stone/stone_pink_23"},
	{"textures/stone_pink_24", "textures/makkon_stone/stone_pink_24"},
	{"textures/stone_pink_25", "textures/makkon_stone/stone_pink_25"},
	{"textures/stone_pink_26", "textures/makkon_stone/stone_pink_26"},
	{"textures/stone_pink_3", "textures/makkon_stone/stone_pink_3"},
	{"textures/stone_pink_4", "textures/makkon_stone/stone_pink_4"},
	{"textures/stone_pink_5", "textures/makkon_stone/stone_pink_5"},
	{"textures/stone_pink_6", "textures/makkon_stone/stone_pink_6"},
	{"textures/stone_pink_7", "textures/makkon_stone/stone_pink_7"},
	{"textures/stone_pink_8", "textures/makkon_stone/stone_pink_8"},
	{"textures/stone_pink_9", "textures/makkon_stone/stone_pink_9"},
	{"textures/stone_pink_c1", "textures/makkon_stone/stone_pink_c1"},
	{"textures/stone_pink_c16", "textures/makkon_stone/stone_pink_c16"},
	{"textures/stone_pink_c17", "textures/makkon_stone/stone_pink_c17"},
	{"textures/stone_pink_c2", "textures/makkon_stone/stone_pink_c2"},
	{"textures/stone_pink_c3", "textures/makkon_stone/stone_pink_c3"},
	{"textures/stone_pink_c4", "textures/makkon_stone/stone_pink_c4"},
	{"textures/stone_pink_c5", "textures/makkon_stone/stone_pink_c5"},
	{"textures/stone_pink_t1", "textures/makkon_stone/stone_pink_t1"},
	{"textures/stone_pink_t10", "textures/makkon_stone/stone_pink_t10"},
	{"textures/stone_pink_t11", "textures/makkon_stone/stone_pink_t11"},
	{"textures/stone_pink_t12", "textures/makkon_stone/stone_pink_t12"},
	{"textures/stone_pink_t13", "textures/makkon_stone/stone_pink_t13"},
	{"textures/stone_pink_t15", "textures/makkon_stone/stone_pink_t15"},
	{"textures/stone_pink_t16", "textures/makkon_stone/stone_pink_t16"},
	{"textures/stone_pink_t17", "textures/makkon_stone/stone_pink_t17"},
	{"textures/stone_pink_t3", "textures/makkon_stone/stone_pink_t3"},
	{"textures/stone_pink_t4", "textures/makkon_stone/stone_pink_t4"},
	{"textures/stone_pink_t5", "textures/makkon_stone/stone_pink_t5"},
	{"textures/stone_pink_t6", "textures/makkon_stone/stone_pink_t6"},
	{"textures/stone_pink_t7", "textures/makkon_stone/stone_pink_t7"},
	{"textures/stone_pink_t8", "textures/makkon_stone/stone_pink_t8"},
	{"textures/stone_prpl2_1", "textures/makkon_stone/stone_prpl2_1"},
	{"textures/stone_prpl2_10", "textures/makkon_stone/stone_prpl2_10"},
	{"textures/stone_prpl2_11", "textures/makkon_stone/stone_prpl2_11"},
	{"textures/stone_prpl2_12", "textures/makkon_stone/stone_prpl2_12"},
	{"textures/stone_prpl2_13", "textures/makkon_stone/stone_prpl2_13"},
	{"textures/stone_prpl2_14", "textures/makkon_stone/stone_prpl2_14"},
	{"textures/stone_prpl2_15", "textures/makkon_stone/stone_prpl2_15"},
	{"textures/stone_prpl2_16", "textures/makkon_stone/stone_prpl2_16"},
	{"textures/stone_prpl2_17", "textures/makkon_stone/stone_prpl2_17"},
	{"textures/stone_prpl2_18", "textures/makkon_stone/stone_prpl2_18"},
	{"textures/stone_prpl2_19", "textures/makkon_stone/stone_prpl2_19"},
	{"textures/stone_prpl2_1_1", "textures/makkon_stone/stone_prpl2_1_1"},
	{"textures/stone_prpl2_1_2", "textures/makkon_stone/stone_prpl2_1_2"},
	{"textures/stone_prpl2_1_3", "textures/makkon_stone/stone_prpl2_1_3"},
	{"textures/stone_prpl2_1_4", "textures/makkon_stone/stone_prpl2_1_4"},
	{"textures/stone_prpl2_1_5", "textures/makkon_stone/stone_prpl2_1_5"},
	{"textures/stone_prpl2_2", "textures/makkon_stone/stone_prpl2_2"},
	{"textures/stone_prpl2_20", "textures/makkon_stone/stone_prpl2_20"},
	{"textures/stone_prpl2_21", "textures/makkon_stone/stone_prpl2_21"},
	{"textures/stone_prpl2_22", "textures/makkon_stone/stone_prpl2_22"},
	{"textures/stone_prpl2_23", "textures/makkon_stone/stone_prpl2_23"},
	{"textures/stone_prpl2_24", "textures/makkon_stone/stone_prpl2_24"},
	{"textures/stone_prpl2_25", "textures/makkon_stone/stone_prpl2_25"},
	{"textures/stone_prpl2_26", "textures/makkon_stone/stone_prpl2_26"},
	{"textures/stone_prpl2_3", "textures/makkon_stone/stone_prpl2_3"},
	{"textures/stone_prpl2_4", "textures/makkon_stone/stone_prpl2_4"},
	{"textures/stone_prpl2_5", "textures/makkon_stone/stone_prpl2_5"},
	{"textures/stone_prpl2_6", "textures/makkon_stone/stone_prpl2_6"},
	{"textures/stone_prpl2_7", "textures/makkon_stone/stone_prpl2_7"},
	{"textures/stone_prpl2_8", "textures/makkon_stone/stone_prpl2_8"},
	{"textures/stone_prpl2_9", "textures/makkon_stone/stone_prpl2_9"},
	{"textures/stone_prpl2_c1", "textures/makkon_stone/stone_prpl2_c1"},
	{"textures/stone_prpl2_c16", "textures/makkon_stone/stone_prpl2_c16"},
	{"textures/stone_prpl2_c17", "textures/makkon_stone/stone_prpl2_c17"},
	{"textures/stone_prpl2_c2", "textures/makkon_stone/stone_prpl2_c2"},
	{"textures/stone_prpl2_c3", "textures/makkon_stone/stone_prpl2_c3"},
	{"textures/stone_prpl2_c4", "textures/makkon_stone/stone_prpl2_c4"},
	{"textures/stone_prpl2_c5", "textures/makkon_stone/stone_prpl2_c5"},
	{"textures/stone_prpl2_t1", "textures/makkon_stone/stone_prpl2_t1"},
	{"textures/stone_prpl2_t10", "textures/makkon_stone/stone_prpl2_t10"},
	{"textures/stone_prpl2_t11", "textures/makkon_stone/stone_prpl2_t11"},
	{"textures/stone_prpl2_t12", "textures/makkon_stone/stone_prpl2_t12"},
	{"textures/stone_prpl2_t13", "textures/makkon_stone/stone_prpl2_t13"},
	{"textures/stone_prpl2_t14", "textures/makkon_stone/stone_prpl2_t14"},
	{"textures/stone_prpl2_t15", "textures/makkon_stone/stone_prpl2_t15"},
	{"textures/stone_prpl2_t16", "textures/makkon_stone/stone_prpl2_t16"},
	{"textures/stone_prpl2_t17", "textures/makkon_stone/stone_prpl2_t17"},
	{"textures/stone_prpl2_t3", "textures/makkon_stone/stone_prpl2_t3"},
	{"textures/stone_prpl2_t4", "textures/makkon_stone/stone_prpl2_t4"},
	{"textures/stone_prpl2_t5", "textures/makkon_stone/stone_prpl2_t5"},
	{"textures/stone_prpl2_t6", "textures/makkon_stone/stone_prpl2_t6"},
	{"textures/stone_prpl2_t7", "textures/makkon_stone/stone_prpl2_t7"},
	{"textures/stone_prpl2_t8", "textures/makkon_stone/stone_prpl2_t8"},
	{"textures/stone_prpl_1", "textures/makkon_stone/stone_prpl_1"},
	{"textures/stone_prpl_10", "textures/makkon_stone/stone_prpl_10"},
	{"textures/stone_prpl_11", "textures/makkon_stone/stone_prpl_11"},
	{"textures/stone_prpl_12", "textures/makkon_stone/stone_prpl_12"},
	{"textures/stone_prpl_13", "textures/makkon_stone/stone_prpl_13"},
	{"textures/stone_prpl_14", "textures/makkon_stone/stone_prpl_14"},
	{"textures/stone_prpl_15", "textures/makkon_stone/stone_prpl_15"},
	{"textures/stone_prpl_16", "textures/makkon_stone/stone_prpl_16"},
	{"textures/stone_prpl_17", "textures/makkon_stone/stone_prpl_17"},
	{"textures/stone_prpl_18", "textures/makkon_stone/stone_prpl_18"},
	{"textures/stone_prpl_19", "textures/makkon_stone/stone_prpl_19"},
	{"textures/stone_prpl_1_1", "textures/makkon_stone/stone_prpl_1_1"},
	{"textures/stone_prpl_1_2", "textures/makkon_stone/stone_prpl_1_2"},
	{"textures/stone_prpl_1_3", "textures/makkon_stone/stone_prpl_1_3"},
	{"textures/stone_prpl_1_4", "textures/makkon_stone/stone_prpl_1_4"},
	{"textures/stone_prpl_1_5", "textures/makkon_stone/stone_prpl_1_5"},
	{"textures/stone_prpl_2", "textures/makkon_stone/stone_prpl_2"},
	{"textures/stone_prpl_20", "textures/makkon_stone/stone_prpl_20"},
	{"textures/stone_prpl_21", "textures/makkon_stone/stone_prpl_21"},
	{"textures/stone_prpl_22", "textures/makkon_stone/stone_prpl_22"},
	{"textures/stone_prpl_23", "textures/makkon_stone/stone_prpl_23"},
	{"textures/stone_prpl_24", "textures/makkon_stone/stone_prpl_24"},
	{"textures/stone_prpl_25", "textures/makkon_stone/stone_prpl_25"},
	{"textures/stone_prpl_26", "textures/makkon_stone/stone_prpl_26"},
	{"textures/stone_prpl_3", "textures/makkon_stone/stone_prpl_3"},
	{"textures/stone_prpl_4", "textures/makkon_stone/stone_prpl_4"},
	{"textures/stone_prpl_5", "textures/makkon_stone/stone_prpl_5"},
	{"textures/stone_prpl_6", "textures/makkon_stone/stone_prpl_6"},
	{"textures/stone_prpl_7", "textures/makkon_stone/stone_prpl_7"},
	{"textures/stone_prpl_8", "textures/makkon_stone/stone_prpl_8"},
	{"textures/stone_prpl_9", "textures/makkon_stone/stone_prpl_9"},
	{"textures/stone_prpl_c1", "textures/makkon_stone/stone_prpl_c1"},
	{"textures/stone_prpl_c16", "textures/makkon_stone/stone_prpl_c16"},
	{"textures/stone_prpl_c17", "textures/makkon_stone/stone_prpl_c17"},
	{"textures/stone_prpl_c2", "textures/makkon_stone/stone_prpl_c2"},
	{"textures/stone_prpl_c3", "textures/makkon_stone/stone_prpl_c3"},
	{"textures/stone_prpl_c4", "textures/makkon_stone/stone_prpl_c4"},
	{"textures/stone_prpl_c5", "textures/makkon_stone/stone_prpl_c5"},
	{"textures/stone_prpl_t1", "textures/makkon_stone/stone_prpl_t1"},
	{"textures/stone_prpl_t10", "textures/makkon_stone/stone_prpl_t10"},
	{"textures/stone_prpl_t11", "textures/makkon_stone/stone_prpl_t11"},
	{"textures/stone_prpl_t12", "textures/makkon_stone/stone_prpl_t12"},
	{"textures/stone_prpl_t13", "textures/makkon_stone/stone_prpl_t13"},
	{"textures/stone_prpl_t14", "textures/makkon_stone/stone_prpl_t14"},
	{"textures/stone_prpl_t15", "textures/makkon_stone/stone_prpl_t15"},
	{"textures/stone_prpl_t16", "textures/makkon_stone/stone_prpl_t16"},
	{"textures/stone_prpl_t17", "textures/makkon_stone/stone_prpl_t17"},
	{"textures/stone_prpl_t3", "textures/makkon_stone/stone_prpl_t3"},
	{"textures/stone_prpl_t4", "textures/makkon_stone/stone_prpl_t4"},
	{"textures/stone_prpl_t5", "textures/makkon_stone/stone_prpl_t5"},
	{"textures/stone_prpl_t6", "textures/makkon_stone/stone_prpl_t6"},
	{"textures/stone_prpl_t7", "textures/makkon_stone/stone_prpl_t7"},
	{"textures/stone_prpl_t8", "textures/makkon_stone/stone_prpl_t8"},
	{"textures/stone_red2_1", "textures/makkon_stone/stone_red2_1"},
	{"textures/stone_red2_10", "textures/makkon_stone/stone_red2_10"},
	{"textures/stone_red2_11", "textures/makkon_stone/stone_red2_11"},
	{"textures/stone_red2_12", "textures/makkon_stone/stone_red2_12"},
	{"textures/stone_red2_13", "textures/makkon_stone/stone_red2_13"},
	{"textures/stone_red2_14", "textures/makkon_stone/stone_red2_14"},
	{"textures/stone_red2_15", "textures/makkon_stone/stone_red2_15"},
	{"textures/stone_red2_16", "textures/makkon_stone/stone_red2_16"},
	{"textures/stone_red2_17", "textures/makkon_stone/stone_red2_17"},
	{"textures/stone_red2_18", "textures/makkon_stone/stone_red2_18"},
	{"textures/stone_red2_19", "textures/makkon_stone/stone_red2_19"},
	{"textures/stone_red2_1_1", "textures/makkon_stone/stone_red2_1_1"},
	{"textures/stone_red2_1_2", "textures/makkon_stone/stone_red2_1_2"},
	{"textures/stone_red2_1_3", "textures/makkon_stone/stone_red2_1_3"},
	{"textures/stone_red2_1_4", "textures/makkon_stone/stone_red2_1_4"},
	{"textures/stone_red2_1_5", "textures/makkon_stone/stone_red2_1_5"},
	{"textures/stone_red2_2", "textures/makkon_stone/stone_red2_2"},
	{"textures/stone_red2_20", "textures/makkon_stone/stone_red2_20"},
	{"textures/stone_red2_21", "textures/makkon_stone/stone_red2_21"},
	{"textures/stone_red2_22", "textures/makkon_stone/stone_red2_22"},
	{"textures/stone_red2_23", "textures/makkon_stone/stone_red2_23"},
	{"textures/stone_red2_24", "textures/makkon_stone/stone_red2_24"},
	{"textures/stone_red2_25", "textures/makkon_stone/stone_red2_25"},
	{"textures/stone_red2_26", "textures/makkon_stone/stone_red2_26"},
	{"textures/stone_red2_3", "textures/makkon_stone/stone_red2_3"},
	{"textures/stone_red2_4", "textures/makkon_stone/stone_red2_4"},
	{"textures/stone_red2_5", "textures/makkon_stone/stone_red2_5"},
	{"textures/stone_red2_6", "textures/makkon_stone/stone_red2_6"},
	{"textures/stone_red2_7", "textures/makkon_stone/stone_red2_7"},
	{"textures/stone_red2_8", "textures/makkon_stone/stone_red2_8"},
	{"textures/stone_red2_9", "textures/makkon_stone/stone_red2_9"},
	{"textures/stone_red2_c1", "textures/makkon_stone/stone_red2_c1"},
	{"textures/stone_red2_c16", "textures/makkon_stone/stone_red2_c16"},
	{"textures/stone_red2_c17", "textures/makkon_stone/stone_red2_c17"},
	{"textures/stone_red2_c2", "textures/makkon_stone/stone_red2_c2"},
	{"textures/stone_red2_c3", "textures/makkon_stone/stone_red2_c3"},
	{"textures/stone_red2_c4", "textures/makkon_stone/stone_red2_c4"},
	{"textures/stone_red2_c5", "textures/makkon_stone/stone_red2_c5"},
	{"textures/stone_red2_t1", "textures/makkon_stone/stone_red2_t1"},
	{"textures/stone_red2_t10", "textures/makkon_stone/stone_red2_t10"},
	{"textures/stone_red2_t11", "textures/makkon_stone/stone_red2_t11"},
	{"textures/stone_red2_t12", "textures/makkon_stone/stone_red2_t12"},
	{"textures/stone_red2_t13", "textures/makkon_stone/stone_red2_t13"},
	{"textures/stone_red2_t14", "textures/makkon_stone/stone_red2_t14"},
	{"textures/stone_red2_t15", "textures/makkon_stone/stone_red2_t15"},
	{"textures/stone_red2_t16", "textures/makkon_stone/stone_red2_t16"},
	{"textures/stone_red2_t17", "textures/makkon_stone/stone_red2_t17"},
	{"textures/stone_red2_t2", "textures/makkon_stone/stone_red2_t2"},
	{"textures/stone_red2_t3", "textures/makkon_stone/stone_red2_t3"},
	{"textures/stone_red2_t4", "textures/makkon_stone/stone_red2_t4"},
	{"textures/stone_red2_t5", "textures/makkon_stone/stone_red2_t5"},
	{"textures/stone_red2_t6", "textures/makkon_stone/stone_red2_t6"},
	{"textures/stone_red2_t7", "textures/makkon_stone/stone_red2_t7"},
	{"textures/stone_red2_t8", "textures/makkon_stone/stone_red2_t8"},
	{"textures/stone_red_1", "textures/makkon_stone/stone_red_1"},
	{"textures/stone_red_10", "textures/makkon_stone/stone_red_10"},
	{"textures/stone_red_11", "textures/makkon_stone/stone_red_11"},
	{"textures/stone_red_12", "textures/makkon_stone/stone_red_12"},
	{"textures/stone_red_13", "textures/makkon_stone/stone_red_13"},
	{"textures/stone_red_14", "textures/makkon_stone/stone_red_14"},
	{"textures/stone_red_15", "textures/makkon_stone/stone_red_15"},
	{"textures/stone_red_16", "textures/makkon_stone/stone_red_16"},
	{"textures/stone_red_17", "textures/makkon_stone/stone_red_17"},
	{"textures/stone_red_18", "textures/makkon_stone/stone_red_18"},
	{"textures/stone_red_19", "textures/makkon_stone/stone_red_19"},
	{"textures/stone_red_1_1", "textures/makkon_stone/stone_red_1_1"},
	{"textures/stone_red_1_2", "textures/makkon_stone/stone_red_1_2"},
	{"textures/stone_red_1_3", "textures/makkon_stone/stone_red_1_3"},
	{"textures/stone_red_1_4", "textures/makkon_stone/stone_red_1_4"},
	{"textures/stone_red_2", "textures/makkon_stone/stone_red_2"},
	{"textures/stone_red_20", "textures/makkon_stone/stone_red_20"},
	{"textures/stone_red_21", "textures/makkon_stone/stone_red_21"},
	{"textures/stone_red_22", "textures/makkon_stone/stone_red_22"},
	{"textures/stone_red_23", "textures/makkon_stone/stone_red_23"},
	{"textures/stone_red_24", "textures/makkon_stone/stone_red_24"},
	{"textures/stone_red_25", "textures/makkon_stone/stone_red_25"},
	{"textures/stone_red_26", "textures/makkon_stone/stone_red_26"},
	{"textures/stone_red_3", "textures/makkon_stone/stone_red_3"},
	{"textures/stone_red_4", "textures/makkon_stone/stone_red_4"},
	{"textures/stone_red_5", "textures/makkon_stone/stone_red_5"},
	{"textures/stone_red_6", "textures/makkon_stone/stone_red_6"},
	{"textures/stone_red_7", "textures/makkon_stone/stone_red_7"},
	{"textures/stone_red_8", "textures/makkon_stone/stone_red_8"},
	{"textures/stone_red_9", "textures/makkon_stone/stone_red_9"},
	{"textures/stone_red_c1", "textures/makkon_stone/stone_red_c1"},
	{"textures/stone_red_c16", "textures/makkon_stone/stone_red_c16"},
	{"textures/stone_red_c17", "textures/makkon_stone/stone_red_c17"},
	{"textures/stone_red_c2", "textures/makkon_stone/stone_red_c2"},
	{"textures/stone_red_c3", "textures/makkon_stone/stone_red_c3"},
	{"textures/stone_red_c4", "textures/makkon_stone/stone_red_c4"},
	{"textures/stone_red_c5", "textures/makkon_stone/stone_red_c5"},
	{"textures/stone_red_t1", "textures/makkon_stone/stone_red_t1"},
	{"textures/stone_red_t10", "textures/makkon_stone/stone_red_t10"},
	{"textures/stone_red_t11", "textures/makkon_stone/stone_red_t11"},
	{"textures/stone_red_t12", "textures/makkon_stone/stone_red_t12"},
	{"textures/stone_red_t13", "textures/makkon_stone/stone_red_t13"},
	{"textures/stone_red_t14", "textures/makkon_stone/stone_red_t14"},
	{"textures/stone_red_t15", "textures/makkon_stone/stone_red_t15"},
	{"textures/stone_red_t16", "textures/makkon_stone/stone_red_t16"},
	{"textures/stone_red_t17", "textures/makkon_stone/stone_red_t17"},
	{"textures/stone_red_t2", "textures/makkon_stone/stone_red_t2"},
	{"textures/stone_red_t3", "textures/makkon_stone/stone_red_t3"},
	{"textures/stone_red_t4", "textures/makkon_stone/stone_red_t4"},
	{"textures/stone_red_t5", "textures/makkon_stone/stone_red_t5"},
	{"textures/stone_red_t6", "textures/makkon_stone/stone_red_t6"},
	{"textures/stone_red_t7", "textures/makkon_stone/stone_red_t7"},
	{"textures/stone_red_t8", "textures/makkon_stone/stone_red_t8"},
	{"textures/stone_tan2_1", "textures/makkon_stone/stone_tan2_1"},
	{"textures/stone_tan2_10", "textures/makkon_stone/stone_tan2_10"},
	{"textures/stone_tan2_11", "textures/makkon_stone/stone_tan2_11"},
	{"textures/stone_tan2_12", "textures/makkon_stone/stone_tan2_12"},
	{"textures/stone_tan2_13", "textures/makkon_stone/stone_tan2_13"},
	{"textures/stone_tan2_14", "textures/makkon_stone/stone_tan2_14"},
	{"textures/stone_tan2_15", "textures/makkon_stone/stone_tan2_15"},
	{"textures/stone_tan2_16", "textures/makkon_stone/stone_tan2_16"},
	{"textures/stone_tan2_17", "textures/makkon_stone/stone_tan2_17"},
	{"textures/stone_tan2_18", "textures/makkon_stone/stone_tan2_18"},
	{"textures/stone_tan2_19", "textures/makkon_stone/stone_tan2_19"},
	{"textures/stone_tan2_1_1", "textures/makkon_stone/stone_tan2_1_1"},
	{"textures/stone_tan2_1_2", "textures/makkon_stone/stone_tan2_1_2"},
	{"textures/stone_tan2_1_3", "textures/makkon_stone/stone_tan2_1_3"},
	{"textures/stone_tan2_1_4", "textures/makkon_stone/stone_tan2_1_4"},
	{"textures/stone_tan2_1_5", "textures/makkon_stone/stone_tan2_1_5"},
	{"textures/stone_tan2_2", "textures/makkon_stone/stone_tan2_2"},
	{"textures/stone_tan2_20", "textures/makkon_stone/stone_tan2_20"},
	{"textures/stone_tan2_21", "textures/makkon_stone/stone_tan2_21"},
	{"textures/stone_tan2_22", "textures/makkon_stone/stone_tan2_22"},
	{"textures/stone_tan2_23", "textures/makkon_stone/stone_tan2_23"},
	{"textures/stone_tan2_24", "textures/makkon_stone/stone_tan2_24"},
	{"textures/stone_tan2_25", "textures/makkon_stone/stone_tan2_25"},
	{"textures/stone_tan2_26", "textures/makkon_stone/stone_tan2_26"},
	{"textures/stone_tan2_3", "textures/makkon_stone/stone_tan2_3"},
	{"textures/stone_tan2_4", "textures/makkon_stone/stone_tan2_4"},
	{"textures/stone_tan2_5", "textures/makkon_stone/stone_tan2_5"},
	{"textures/stone_tan2_6", "textures/makkon_stone/stone_tan2_6"},
	{"textures/stone_tan2_7", "textures/makkon_stone/stone_tan2_7"},
	{"textures/stone_tan2_8", "textures/makkon_stone/stone_tan2_8"},
	{"textures/stone_tan2_9", "textures/makkon_stone/stone_tan2_9"},
	{"textures/stone_tan2_c1", "textures/makkon_stone/stone_tan2_c1"},
	{"textures/stone_tan2_c16", "textures/makkon_stone/stone_tan2_c16"},
	{"textures/stone_tan2_c17", "textures/makkon_stone/stone_tan2_c17"},
	{"textures/stone_tan2_c2", "textures/makkon_stone/stone_tan2_c2"},
	{"textures/stone_tan2_c3", "textures/makkon_stone/stone_tan2_c3"},
	{"textures/stone_tan2_c4", "textures/makkon_stone/stone_tan2_c4"},
	{"textures/stone_tan2_c5", "textures/makkon_stone/stone_tan2_c5"},
	{"textures/stone_tan2_t1", "textures/makkon_stone/stone_tan2_t1"},
	{"textures/stone_tan2_t10", "textures/makkon_stone/stone_tan2_t10"},
	{"textures/stone_tan2_t11", "textures/makkon_stone/stone_tan2_t11"},
	{"textures/stone_tan2_t12", "textures/makkon_stone/stone_tan2_t12"},
	{"textures/stone_tan2_t13", "textures/makkon_stone/stone_tan2_t13"},
	{"textures/stone_tan2_t14", "textures/makkon_stone/stone_tan2_t14"},
	{"textures/stone_tan2_t15", "textures/makkon_stone/stone_tan2_t15"},
	{"textures/stone_tan2_t16", "textures/makkon_stone/stone_tan2_t16"},
	{"textures/stone_tan2_t17", "textures/makkon_stone/stone_tan2_t17"},
	{"textures/stone_tan2_t2", "textures/makkon_stone/stone_tan2_t2"},
	{"textures/stone_tan2_t3", "textures/makkon_stone/stone_tan2_t3"},
	{"textures/stone_tan2_t4", "textures/makkon_stone/stone_tan2_t4"},
	{"textures/stone_tan2_t5", "textures/makkon_stone/stone_tan2_t5"},
	{"textures/stone_tan2_t6", "textures/makkon_stone/stone_tan2_t6"},
	{"textures/stone_tan2_t7", "textures/makkon_stone/stone_tan2_t7"},
	{"textures/stone_tan2_t8", "textures/makkon_stone/stone_tan2_t8"},
	{"textures/stone_tan_1", "textures/makkon_stone/stone_tan_1"},
	{"textures/stone_tan_10", "textures/makkon_stone/stone_tan_10"},
	{"textures/stone_tan_11", "textures/makkon_stone/stone_tan_11"},
	{"textures/stone_tan_12", "textures/makkon_stone/stone_tan_12"},
	{"textures/stone_tan_13", "textures/makkon_stone/stone_tan_13"},
	{"textures/stone_tan_14", "textures/makkon_stone/stone_tan_14"},
	{"textures/stone_tan_15", "textures/makkon_stone/stone_tan_15"},
	{"textures/stone_tan_16", "textures/makkon_stone/stone_tan_16"},
	{"textures/stone_tan_17", "textures/makkon_stone/stone_tan_17"},
	{"textures/stone_tan_18", "textures/makkon_stone/stone_tan_18"},
	{"textures/stone_tan_19", "textures/makkon_stone/stone_tan_19"},
	{"textures/stone_tan_1_1", "textures/makkon_stone/stone_tan_1_1"},
	{"textures/stone_tan_1_2", "textures/makkon_stone/stone_tan_1_2"},
	{"textures/stone_tan_1_3", "textures/makkon_stone/stone_tan_1_3"},
	{"textures/stone_tan_1_4", "textures/makkon_stone/stone_tan_1_4"},
	{"textures/stone_tan_2", "textures/makkon_stone/stone_tan_2"},
	{"textures/stone_tan_20", "textures/makkon_stone/stone_tan_20"},
	{"textures/stone_tan_21", "textures/makkon_stone/stone_tan_21"},
	{"textures/stone_tan_22", "textures/makkon_stone/stone_tan_22"},
	{"textures/stone_tan_23", "textures/makkon_stone/stone_tan_23"},
	{"textures/stone_tan_24", "textures/makkon_stone/stone_tan_24"},
	{"textures/stone_tan_25", "textures/makkon_stone/stone_tan_25"},
	{"textures/stone_tan_26", "textures/makkon_stone/stone_tan_26"},
	{"textures/stone_tan_3", "textures/makkon_stone/stone_tan_3"},
	{"textures/stone_tan_4", "textures/makkon_stone/stone_tan_4"},
	{"textures/stone_tan_5", "textures/makkon_stone/stone_tan_5"},
	{"textures/stone_tan_6", "textures/makkon_stone/stone_tan_6"},
	{"textures/stone_tan_7", "textures/makkon_stone/stone_tan_7"},
	{"textures/stone_tan_8", "textures/makkon_stone/stone_tan_8"},
	{"textures/stone_tan_9", "textures/makkon_stone/stone_tan_9"},
	{"textures/stone_tan_c1", "textures/makkon_stone/stone_tan_c1"},
	{"textures/stone_tan_c16", "textures/makkon_stone/stone_tan_c16"},
	{"textures/stone_tan_c17", "textures/makkon_stone/stone_tan_c17"},
	{"textures/stone_tan_c2", "textures/makkon_stone/stone_tan_c2"},
	{"textures/stone_tan_c3", "textures/makkon_stone/stone_tan_c3"},
	{"textures/stone_tan_c4", "textures/makkon_stone/stone_tan_c4"},
	{"textures/stone_tan_c5", "textures/makkon_stone/stone_tan_c5"},
	{"textures/stone_tan_t1", "textures/makkon_stone/stone_tan_t1"},
	{"textures/stone_tan_t10", "textures/makkon_stone/stone_tan_t10"},
	{"textures/stone_tan_t11", "textures/makkon_stone/stone_tan_t11"},
	{"textures/stone_tan_t12", "textures/makkon_stone/stone_tan_t12"},
	{"textures/stone_tan_t13", "textures/makkon_stone/stone_tan_t13"},
	{"textures/stone_tan_t14", "textures/makkon_stone/stone_tan_t14"},
	{"textures/stone_tan_t15", "textures/makkon_stone/stone_tan_t15"},
	{"textures/stone_tan_t16", "textures/makkon_stone/stone_tan_t16"},
	{"textures/stone_tan_t17", "textures/makkon_stone/stone_tan_t17"},
	{"textures/stone_tan_t2", "textures/makkon_stone/stone_tan_t2"},
	{"textures/stone_tan_t3", "textures/makkon_stone/stone_tan_t3"},
	{"textures/stone_tan_t4", "textures/makkon_stone/stone_tan_t4"},
	{"textures/stone_tan_t5", "textures/makkon_stone/stone_tan_t5"},
	{"textures/stone_tan_t6", "textures/makkon_stone/stone_tan_t6"},
	{"textures/stone_tan_t7", "textures/makkon_stone/stone_tan_t7"},
	{"textures/stone_tan_t8", "textures/makkon_stone/stone_tan_t8"},
	{"textures/stone_teal2_1", "textures/makkon_stone/stone_teal2_1"},
	{"textures/stone_teal2_10", "textures/makkon_stone/stone_teal2_10"},
	{"textures/stone_teal2_11", "textures/makkon_stone/stone_teal2_11"},
	{"textures/stone_teal2_12", "textures/makkon_stone/stone_teal2_12"},
	{"textures/stone_teal2_13", "textures/makkon_stone/stone_teal2_13"},
	{"textures/stone_teal2_14", "textures/makkon_stone/stone_teal2_14"},
	{"textures/stone_teal2_15", "textures/makkon_stone/stone_teal2_15"},
	{"textures/stone_teal2_16", "textures/makkon_stone/stone_teal2_16"},
	{"textures/stone_teal2_17", "textures/makkon_stone/stone_teal2_17"},
	{"textures/stone_teal2_18", "textures/makkon_stone/stone_teal2_18"},
	{"textures/stone_teal2_19", "textures/makkon_stone/stone_teal2_19"},
	{"textures/stone_teal2_1_1", "textures/makkon_stone/stone_teal2_1_1"},
	{"textures/stone_teal2_1_2", "textures/makkon_stone/stone_teal2_1_2"},
	{"textures/stone_teal2_1_3", "textures/makkon_stone/stone_teal2_1_3"},
	{"textures/stone_teal2_1_4", "textures/makkon_stone/stone_teal2_1_4"},
	{"textures/stone_teal2_1_5", "textures/makkon_stone/stone_teal2_1_5"},
	{"textures/stone_teal2_2", "textures/makkon_stone/stone_teal2_2"},
	{"textures/stone_teal2_20", "textures/makkon_stone/stone_teal2_20"},
	{"textures/stone_teal2_21", "textures/makkon_stone/stone_teal2_21"},
	{"textures/stone_teal2_22", "textures/makkon_stone/stone_teal2_22"},
	{"textures/stone_teal2_23", "textures/makkon_stone/stone_teal2_23"},
	{"textures/stone_teal2_24", "textures/makkon_stone/stone_teal2_24"},
	{"textures/stone_teal2_25", "textures/makkon_stone/stone_teal2_25"},
	{"textures/stone_teal2_26", "textures/makkon_stone/stone_teal2_26"},
	{"textures/stone_teal2_3", "textures/makkon_stone/stone_teal2_3"},
	{"textures/stone_teal2_4", "textures/makkon_stone/stone_teal2_4"},
	{"textures/stone_teal2_5", "textures/makkon_stone/stone_teal2_5"},
	{"textures/stone_teal2_6", "textures/makkon_stone/stone_teal2_6"},
	{"textures/stone_teal2_7", "textures/makkon_stone/stone_teal2_7"},
	{"textures/stone_teal2_8", "textures/makkon_stone/stone_teal2_8"},
	{"textures/stone_teal2_9", "textures/makkon_stone/stone_teal2_9"},
	{"textures/stone_teal2_c1", "textures/makkon_stone/stone_teal2_c1"},
	{"textures/stone_teal2_c16", "textures/makkon_stone/stone_teal2_c16"},
	{"textures/stone_teal2_c17", "textures/makkon_stone/stone_teal2_c17"},
	{"textures/stone_teal2_c2", "textures/makkon_stone/stone_teal2_c2"},
	{"textures/stone_teal2_c3", "textures/makkon_stone/stone_teal2_c3"},
	{"textures/stone_teal2_c4", "textures/makkon_stone/stone_teal2_c4"},
	{"textures/stone_teal2_c5", "textures/makkon_stone/stone_teal2_c5"},
	{"textures/stone_teal2_t1", "textures/makkon_stone/stone_teal2_t1"},
	{"textures/stone_teal2_t10", "textures/makkon_stone/stone_teal2_t10"},
	{"textures/stone_teal2_t11", "textures/makkon_stone/stone_teal2_t11"},
	{"textures/stone_teal2_t12", "textures/makkon_stone/stone_teal2_t12"},
	{"textures/stone_teal2_t13", "textures/makkon_stone/stone_teal2_t13"},
	{"textures/stone_teal2_t14", "textures/makkon_stone/stone_teal2_t14"},
	{"textures/stone_teal2_t15", "textures/makkon_stone/stone_teal2_t15"},
	{"textures/stone_teal2_t16", "textures/makkon_stone/stone_teal2_t16"},
	{"textures/stone_teal2_t17", "textures/makkon_stone/stone_teal2_t17"},
	{"textures/stone_teal2_t2", "textures/makkon_stone/stone_teal2_t2"},
	{"textures/stone_teal2_t3", "textures/makkon_stone/stone_teal2_t3"},
	{"textures/stone_teal2_t4", "textures/makkon_stone/stone_teal2_t4"},
	{"textures/stone_teal2_t5", "textures/makkon_stone/stone_teal2_t5"},
	{"textures/stone_teal2_t6", "textures/makkon_stone/stone_teal2_t6"},
	{"textures/stone_teal2_t7", "textures/makkon_stone/stone_teal2_t7"},
	{"textures/stone_teal2_t8", "textures/makkon_stone/stone_teal2_t8"},
	{"textures/stone_teal_1", "textures/makkon_stone/stone_teal_1"},
	{"textures/stone_teal_10", "textures/makkon_stone/stone_teal_10"},
	{"textures/stone_teal_11", "textures/makkon_stone/stone_teal_11"},
	{"textures/stone_teal_12", "textures/makkon_stone/stone_teal_12"},
	{"textures/stone_teal_13", "textures/makkon_stone/stone_teal_13"},
	{"textures/stone_teal_14", "textures/makkon_stone/stone_teal_14"},
	{"textures/stone_teal_15", "textures/makkon_stone/stone_teal_15"},
	{"textures/stone_teal_16", "textures/makkon_stone/stone_teal_16"},
	{"textures/stone_teal_17", "textures/makkon_stone/stone_teal_17"},
	{"textures/stone_teal_18", "textures/makkon_stone/stone_teal_18"},
	{"textures/stone_teal_19", "textures/makkon_stone/stone_teal_19"},
	{"textures/stone_teal_1_1", "textures/makkon_stone/stone_teal_1_1"},
	{"textures/stone_teal_1_2", "textures/makkon_stone/stone_teal_1_2"},
	{"textures/stone_teal_1_3", "textures/makkon_stone/stone_teal_1_3"},
	{"textures/stone_teal_1_4", "textures/makkon_stone/stone_teal_1_4"},
	{"textures/stone_teal_2", "textures/makkon_stone/stone_teal_2"},
	{"textures/stone_teal_20", "textures/makkon_stone/stone_teal_20"},
	{"textures/stone_teal_21", "textures/makkon_stone/stone_teal_21"},
	{"textures/stone_teal_22", "textures/makkon_stone/stone_teal_22"},
	{"textures/stone_teal_23", "textures/makkon_stone/stone_teal_23"},
	{"textures/stone_teal_24", "textures/makkon_stone/stone_teal_24"},
	{"textures/stone_teal_25", "textures/makkon_stone/stone_teal_25"},
	{"textures/stone_teal_26", "textures/makkon_stone/stone_teal_26"},
	{"textures/stone_teal_3", "textures/makkon_stone/stone_teal_3"},
	{"textures/stone_teal_4", "textures/makkon_stone/stone_teal_4"},
	{"textures/stone_teal_5", "textures/makkon_stone/stone_teal_5"},
	{"textures/stone_teal_6", "textures/makkon_stone/stone_teal_6"},
	{"textures/stone_teal_7", "textures/makkon_stone/stone_teal_7"},
	{"textures/stone_teal_8", "textures/makkon_stone/stone_teal_8"},
	{"textures/stone_teal_9", "textures/makkon_stone/stone_teal_9"},
	{"textures/stone_teal_c1", "textures/makkon_stone/stone_teal_c1"},
	{"textures/stone_teal_c16", "textures/makkon_stone/stone_teal_c16"},
	{"textures/stone_teal_c17", "textures/makkon_stone/stone_teal_c17"},
	{"textures/stone_teal_c2", "textures/makkon_stone/stone_teal_c2"},
	{"textures/stone_teal_c3", "textures/makkon_stone/stone_teal_c3"},
	{"textures/stone_teal_c4", "textures/makkon_stone/stone_teal_c4"},
	{"textures/stone_teal_c5", "textures/makkon_stone/stone_teal_c5"},
	{"textures/stone_teal_t1", "textures/makkon_stone/stone_teal_t1"},
	{"textures/stone_teal_t10", "textures/makkon_stone/stone_teal_t10"},
	{"textures/stone_teal_t11", "textures/makkon_stone/stone_teal_t11"},
	{"textures/stone_teal_t12", "textures/makkon_stone/stone_teal_t12"},
	{"textures/stone_teal_t13", "textures/makkon_stone/stone_teal_t13"},
	{"textures/stone_teal_t14", "textures/makkon_stone/stone_teal_t14"},
	{"textures/stone_teal_t15", "textures/makkon_stone/stone_teal_t15"},
	{"textures/stone_teal_t16", "textures/makkon_stone/stone_teal_t16"},
	{"textures/stone_teal_t17", "textures/makkon_stone/stone_teal_t17"},
	{"textures/stone_teal_t2", "textures/makkon_stone/stone_teal_t2"},
	{"textures/stone_teal_t3", "textures/makkon_stone/stone_teal_t3"},
	{"textures/stone_teal_t4", "textures/makkon_stone/stone_teal_t4"},
	{"textures/stone_teal_t5", "textures/makkon_stone/stone_teal_t5"},
	{"textures/stone_teal_t6", "textures/makkon_stone/stone_teal_t6"},
	{"textures/stone_teal_t7", "textures/makkon_stone/stone_teal_t7"},
	{"textures/stone_teal_t8", "textures/makkon_stone/stone_teal_t8"},
	{"textures/stone_wht2_1", "textures/makkon_stone/stone_wht2_1"},
	{"textures/stone_wht2_10", "textures/makkon_stone/stone_wht2_10"},
	{"textures/stone_wht2_11", "textures/makkon_stone/stone_wht2_11"},
	{"textures/stone_wht2_12", "textures/makkon_stone/stone_wht2_12"},
	{"textures/stone_wht2_13", "textures/makkon_stone/stone_wht2_13"},
	{"textures/stone_wht2_14", "textures/makkon_stone/stone_wht2_14"},
	{"textures/stone_wht2_15", "textures/makkon_stone/stone_wht2_15"},
	{"textures/stone_wht2_16", "textures/makkon_stone/stone_wht2_16"},
	{"textures/stone_wht2_17", "textures/makkon_stone/stone_wht2_17"},
	{"textures/stone_wht2_18", "textures/makkon_stone/stone_wht2_18"},
	{"textures/stone_wht2_19", "textures/makkon_stone/stone_wht2_19"},
	{"textures/stone_wht2_1_1", "textures/makkon_stone/stone_wht2_1_1"},
	{"textures/stone_wht2_1_2", "textures/makkon_stone/stone_wht2_1_2"},
	{"textures/stone_wht2_1_3", "textures/makkon_stone/stone_wht2_1_3"},
	{"textures/stone_wht2_1_4", "textures/makkon_stone/stone_wht2_1_4"},
	{"textures/stone_wht2_1_5", "textures/makkon_stone/stone_wht2_1_5"},
	{"textures/stone_wht2_2", "textures/makkon_stone/stone_wht2_2"},
	{"textures/stone_wht2_20", "textures/makkon_stone/stone_wht2_20"},
	{"textures/stone_wht2_21", "textures/makkon_stone/stone_wht2_21"},
	{"textures/stone_wht2_22", "textures/makkon_stone/stone_wht2_22"},
	{"textures/stone_wht2_23", "textures/makkon_stone/stone_wht2_23"},
	{"textures/stone_wht2_24", "textures/makkon_stone/stone_wht2_24"},
	{"textures/stone_wht2_25", "textures/makkon_stone/stone_wht2_25"},
	{"textures/stone_wht2_26", "textures/makkon_stone/stone_wht2_26"},
	{"textures/stone_wht2_3", "textures/makkon_stone/stone_wht2_3"},
	{"textures/stone_wht2_4", "textures/makkon_stone/stone_wht2_4"},
	{"textures/stone_wht2_5", "textures/makkon_stone/stone_wht2_5"},
	{"textures/stone_wht2_6", "textures/makkon_stone/stone_wht2_6"},
	{"textures/stone_wht2_7", "textures/makkon_stone/stone_wht2_7"},
	{"textures/stone_wht2_8", "textures/makkon_stone/stone_wht2_8"},
	{"textures/stone_wht2_9", "textures/makkon_stone/stone_wht2_9"},
	{"textures/stone_wht2_c1", "textures/makkon_stone/stone_wht2_c1"},
	{"textures/stone_wht2_c16", "textures/makkon_stone/stone_wht2_c16"},
	{"textures/stone_wht2_c17", "textures/makkon_stone/stone_wht2_c17"},
	{"textures/stone_wht2_c2", "textures/makkon_stone/stone_wht2_c2"},
	{"textures/stone_wht2_c3", "textures/makkon_stone/stone_wht2_c3"},
	{"textures/stone_wht2_c4", "textures/makkon_stone/stone_wht2_c4"},
	{"textures/stone_wht2_c5", "textures/makkon_stone/stone_wht2_c5"},
	{"textures/stone_wht2_t1", "textures/makkon_stone/stone_wht2_t1"},
	{"textures/stone_wht2_t10", "textures/makkon_stone/stone_wht2_t10"},
	{"textures/stone_wht2_t11", "textures/makkon_stone/stone_wht2_t11"},
	{"textures/stone_wht2_t12", "textures/makkon_stone/stone_wht2_t12"},
	{"textures/stone_wht2_t13", "textures/makkon_stone/stone_wht2_t13"},
	{"textures/stone_wht2_t14", "textures/makkon_stone/stone_wht2_t14"},
	{"textures/stone_wht2_t15", "textures/makkon_stone/stone_wht2_t15"},
	{"textures/stone_wht2_t16", "textures/makkon_stone/stone_wht2_t16"},
	{"textures/stone_wht2_t17", "textures/makkon_stone/stone_wht2_t17"},
	{"textures/stone_wht2_t2", "textures/makkon_stone/stone_wht2_t2"},
	{"textures/stone_wht2_t3", "textures/makkon_stone/stone_wht2_t3"},
	{"textures/stone_wht2_t4", "textures/makkon_stone/stone_wht2_t4"},
	{"textures/stone_wht2_t5", "textures/makkon_stone/stone_wht2_t5"},
	{"textures/stone_wht2_t6", "textures/makkon_stone/stone_wht2_t6"},
	{"textures/stone_wht2_t7", "textures/makkon_stone/stone_wht2_t7"},
	{"textures/stone_wht2_t8", "textures/makkon_stone/stone_wht2_t8"},
	{"textures/stone_wht_1", "textures/makkon_stone/stone_wht_1"},
	{"textures/stone_wht_10", "textures/makkon_stone/stone_wht_10"},
	{"textures/stone_wht_11", "textures/makkon_stone/stone_wht_11"},
	{"textures/stone_wht_12", "textures/makkon_stone/stone_wht_12"},
	{"textures/stone_wht_13", "textures/makkon_stone/stone_wht_13"},
	{"textures/stone_wht_14", "textures/makkon_stone/stone_wht_14"},
	{"textures/stone_wht_15", "textures/makkon_stone/stone_wht_15"},
	{"textures/stone_wht_16", "textures/makkon_stone/stone_wht_16"},
	{"textures/stone_wht_17", "textures/makkon_stone/stone_wht_17"},
	{"textures/stone_wht_18", "textures/makkon_stone/stone_wht_18"},
	{"textures/stone_wht_19", "textures/makkon_stone/stone_wht_19"},
	{"textures/stone_wht_1_1", "textures/makkon_stone/stone_wht_1_1"},
	{"textures/stone_wht_1_2", "textures/makkon_stone/stone_wht_1_2"},
	{"textures/stone_wht_1_3", "textures/makkon_stone/stone_wht_1_3"},
	{"textures/stone_wht_1_4", "textures/makkon_stone/stone_wht_1_4"},
	{"textures/stone_wht_2", "textures/makkon_stone/stone_wht_2"},
	{"textures/stone_wht_20", "textures/makkon_stone/stone_wht_20"},
	{"textures/stone_wht_21", "textures/makkon_stone/stone_wht_21"},
	{"textures/stone_wht_22", "textures/makkon_stone/stone_wht_22"},
	{"textures/stone_wht_23", "textures/makkon_stone/stone_wht_23"},
	{"textures/stone_wht_24", "textures/makkon_stone/stone_wht_24"},
	{"textures/stone_wht_25", "textures/makkon_stone/stone_wht_25"},
	{"textures/stone_wht_26", "textures/makkon_stone/stone_wht_26"},
	{"textures/stone_wht_3", "textures/makkon_stone/stone_wht_3"},
	{"textures/stone_wht_4", "textures/makkon_stone/stone_wht_4"},
	{"textures/stone_wht_5", "textures/makkon_stone/stone_wht_5"},
	{"textures/stone_wht_6", "textures/makkon_stone/stone_wht_6"},
	{"textures/stone_wht_7", "textures/makkon_stone/stone_wht_7"},
	{"textures/stone_wht_8", "textures/makkon_stone/stone_wht_8"},
	{"textures/stone_wht_9", "textures/makkon_stone/stone_wht_9"},
	{"textures/stone_wht_c1", "textures/makkon_stone/stone_wht_c1"},
	{"textures/stone_wht_c16", "textures/makkon_stone/stone_wht_c16"},
	{"textures/stone_wht_c17", "textures/makkon_stone/stone_wht_c17"},
	{"textures/stone_wht_c2", "textures/makkon_stone/stone_wht_c2"},
	{"textures/stone_wht_c3", "textures/makkon_stone/stone_wht_c3"},
	{"textures/stone_wht_c4", "textures/makkon_stone/stone_wht_c4"},
	{"textures/stone_wht_c5", "textures/makkon_stone/stone_wht_c5"},
	{"textures/stone_wht_t1", "textures/makkon_stone/stone_wht_t1"},
	{"textures/stone_wht_t10", "textures/makkon_stone/stone_wht_t10"},
	{"textures/stone_wht_t11", "textures/makkon_stone/stone_wht_t11"},
	{"textures/stone_wht_t12", "textures/makkon_stone/stone_wht_t12"},
	{"textures/stone_wht_t13", "textures/makkon_stone/stone_wht_t13"},
	{"textures/stone_wht_t14", "textures/makkon_stone/stone_wht_t14"},
	{"textures/stone_wht_t15", "textures/makkon_stone/stone_wht_t15"},
	{"textures/stone_wht_t16", "textures/makkon_stone/stone_wht_t16"},
	{"textures/stone_wht_t17", "textures/makkon_stone/stone_wht_t17"},
	{"textures/stone_wht_t3", "textures/makkon_stone/stone_wht_t3"},
	{"textures/stone_wht_t4", "textures/makkon_stone/stone_wht_t4"},
	{"textures/stone_wht_t5", "textures/makkon_stone/stone_wht_t5"},
	{"textures/stone_wht_t6", "textures/makkon_stone/stone_wht_t6"},
	{"textures/stone_wht_t7", "textures/makkon_stone/stone_wht_t7"},
	{"textures/stone_wht_t8", "textures/makkon_stone/stone_wht_t8"},
	{"textures/_stone_blck2_f1", "textures/makkon_stone/_stone_blck2_f1"},
	{"textures/_stone_blck_f1", "textures/makkon_stone/_stone_blck_f1"},
	{"textures/_stone_blue2_f1", "textures/makkon_stone/_stone_blue2_f1"},
	{"textures/_stone_blue_f1", "textures/makkon_stone/_stone_blue_f1"},
	{"textures/_stone_brwn2_f1", "textures/makkon_stone/_stone_brwn2_f1"},
	{"textures/_stone_brwn_f1", "textures/makkon_stone/_stone_brwn_f1"},
	{"textures/_stone_grey2_f1", "textures/makkon_stone/_stone_grey2_f1"},
	{"textures/_stone_grey_f1", "textures/makkon_stone/_stone_grey_f1"},
	{"textures/_stone_pink2_f1", "textures/makkon_stone/_stone_pink2_f1"},
	{"textures/_stone_pink_f1", "textures/makkon_stone/_stone_pink_f1"},
	{"textures/_stone_prpl2_f1", "textures/makkon_stone/_stone_prpl2_f1"},
	{"textures/_stone_prpl_f1", "textures/makkon_stone/_stone_prpl_f1"},
	{"textures/_stone_red2_f1", "textures/makkon_stone/_stone_red2_f1"},
	{"textures/_stone_red_f1", "textures/makkon_stone/_stone_red_f1"},
	{"textures/_stone_tan2_f1", "textures/makkon_stone/_stone_tan2_f1"},
	{"textures/_stone_tan_f1", "textures/makkon_stone/_stone_tan_f1"},
	{"textures/_stone_teal2_f1", "textures/makkon_stone/_stone_teal2_f1"},
	{"textures/_stone_teal_f1", "textures/makkon_stone/_stone_teal_f1"},
	{"textures/_stone_wht2_f1", "textures/makkon_stone/_stone_wht2_f1"},
	{"textures/_stone_wht_f1", "textures/makkon_stone/_stone_wht_f1"},

	// Makkon tech 2022.8
	{"textures/__0techlight_1", "textures/makkon_tech/__0techlight_1"},
	{"textures/__0techlight_r1", "textures/makkon_tech/__0techlight_r1"},
	{"textures/__atechlight_1", "textures/makkon_tech/__atechlight_1"},
	{"textures/__atechlight_r1", "textures/makkon_tech/__atechlight_r1"},
	{"textures/0techlight_1", "textures/makkon_tech/0techlight_1"},
	{"textures/atechlight_1", "textures/makkon_tech/atechlight_1"},
	{"textures/atechlight_r1", "textures/makkon_tech/atechlight_r1"},
	{"textures/techcomp1_blk1", "textures/makkon_tech/techcomp1_blk1"},
	{"textures/techcomp1_blue1", "textures/makkon_tech/techcomp1_blue1"},
	{"textures/techcomp1_brwn1", "textures/makkon_tech/techcomp1_brwn1"},
	{"textures/techcomp1_brwn2", "textures/makkon_tech/techcomp1_brwn2"},
	{"textures/techcomp1_grey1", "textures/makkon_tech/techcomp1_grey1"},
	{"textures/techcomp1_grey2", "textures/makkon_tech/techcomp1_grey2"},
	{"textures/techcomp1_orng1", "textures/makkon_tech/techcomp1_orng1"},
	{"textures/techcomp1_prpl1", "textures/makkon_tech/techcomp1_prpl1"},
	{"textures/techcomp1_red1", "textures/makkon_tech/techcomp1_red1"},
	{"textures/techcomp1_red2", "textures/makkon_tech/techcomp1_red2"},
	{"textures/techcomp1_rust1", "textures/makkon_tech/techcomp1_rust1"},
	{"textures/techcomp1_rust2", "textures/makkon_tech/techcomp1_rust2"},
	{"textures/techcomp1_rust3", "textures/makkon_tech/techcomp1_rust3"},
	{"textures/techcomp1_teal1", "textures/makkon_tech/techcomp1_teal1"},
	{"textures/techcomp1_ylw1", "textures/makkon_tech/techcomp1_ylw1"},
	{"textures/techcomp1_ylw2", "textures/makkon_tech/techcomp1_ylw2"},
	{"textures/techcomp2_blk1", "textures/makkon_tech/techcomp2_blk1"},
	{"textures/techcomp2_blue1", "textures/makkon_tech/techcomp2_blue1"},
	{"textures/techcomp2_brwn1", "textures/makkon_tech/techcomp2_brwn1"},
	{"textures/techcomp2_brwn2", "textures/makkon_tech/techcomp2_brwn2"},
	{"textures/techcomp2_grey1", "textures/makkon_tech/techcomp2_grey1"},
	{"textures/techcomp2_grey2", "textures/makkon_tech/techcomp2_grey2"},
	{"textures/techcomp2_orng1", "textures/makkon_tech/techcomp2_orng1"},
	{"textures/techcomp2_prpl1", "textures/makkon_tech/techcomp2_prpl1"},
	{"textures/techcomp2_red1", "textures/makkon_tech/techcomp2_red1"},
	{"textures/techcomp2_red2", "textures/makkon_tech/techcomp2_red2"},
	{"textures/techcomp2_rust1", "textures/makkon_tech/techcomp2_rust1"},
	{"textures/techcomp2_rust2", "textures/makkon_tech/techcomp2_rust2"},
	{"textures/techcomp2_rust3", "textures/makkon_tech/techcomp2_rust3"},
	{"textures/techcomp2_teal1", "textures/makkon_tech/techcomp2_teal1"},
	{"textures/techcomp2_ylw1", "textures/makkon_tech/techcomp2_ylw1"},
	{"textures/techcomp2_ylw2", "textures/makkon_tech/techcomp2_ylw2"},
	{"textures/techcomp3_blk1", "textures/makkon_tech/techcomp3_blk1"},
	{"textures/techcomp3_blue1", "textures/makkon_tech/techcomp3_blue1"},
	{"textures/techcomp3_brwn1", "textures/makkon_tech/techcomp3_brwn1"},
	{"textures/techcomp3_brwn2", "textures/makkon_tech/techcomp3_brwn2"},
	{"textures/techcomp3_grey1", "textures/makkon_tech/techcomp3_grey1"},
	{"textures/techcomp3_grey2", "textures/makkon_tech/techcomp3_grey2"},
	{"textures/techcomp3_orng1", "textures/makkon_tech/techcomp3_orng1"},
	{"textures/techcomp3_prpl1", "textures/makkon_tech/techcomp3_prpl1"},
	{"textures/techcomp3_red1", "textures/makkon_tech/techcomp3_red1"},
	{"textures/techcomp3_red2", "textures/makkon_tech/techcomp3_red2"},
	{"textures/techcomp3_rust1", "textures/makkon_tech/techcomp3_rust1"},
	{"textures/techcomp3_rust2", "textures/makkon_tech/techcomp3_rust2"},
	{"textures/techcomp3_rust3", "textures/makkon_tech/techcomp3_rust3"},
	{"textures/techcomp3_teal1", "textures/makkon_tech/techcomp3_teal1"},
	{"textures/techcomp3_ylw1", "textures/makkon_tech/techcomp3_ylw1"},
	{"textures/techcomp3_ylw2", "textures/makkon_tech/techcomp3_ylw2"},
	{"textures/techcomp4_blk1", "textures/makkon_tech/techcomp4_blk1"},
	{"textures/techcomp4_blue1", "textures/makkon_tech/techcomp4_blue1"},
	{"textures/techcomp4_brwn1", "textures/makkon_tech/techcomp4_brwn1"},
	{"textures/techcomp4_brwn2", "textures/makkon_tech/techcomp4_brwn2"},
	{"textures/techcomp4_grey1", "textures/makkon_tech/techcomp4_grey1"},
	{"textures/techcomp4_grey2", "textures/makkon_tech/techcomp4_grey2"},
	{"textures/techcomp4_orng1", "textures/makkon_tech/techcomp4_orng1"},
	{"textures/techcomp4_prpl1", "textures/makkon_tech/techcomp4_prpl1"},
	{"textures/techcomp4_red1", "textures/makkon_tech/techcomp4_red1"},
	{"textures/techcomp4_red2", "textures/makkon_tech/techcomp4_red2"},
	{"textures/techcomp4_rust1", "textures/makkon_tech/techcomp4_rust1"},
	{"textures/techcomp4_rust2", "textures/makkon_tech/techcomp4_rust2"},
	{"textures/techcomp4_rust3", "textures/makkon_tech/techcomp4_rust3"},
	{"textures/techcomp4_teal1", "textures/makkon_tech/techcomp4_teal1"},
	{"textures/techcomp4_ylw1", "textures/makkon_tech/techcomp4_ylw1"},
	{"textures/techcomp4_ylw2", "textures/makkon_tech/techcomp4_ylw2"},
	{"textures/techcomp_t1blk1", "textures/makkon_tech/techcomp_t1blk1"},
	{"textures/techcomp_t1blk2", "textures/makkon_tech/techcomp_t1blk2"},
	{"textures/techcomp_t1blk3", "textures/makkon_tech/techcomp_t1blk3"},
	{"textures/techcomp_t1blk4", "textures/makkon_tech/techcomp_t1blk4"},
	{"textures/techcomp_t1blk5", "textures/makkon_tech/techcomp_t1blk5"},
	{"textures/techcomp_t1blk6", "textures/makkon_tech/techcomp_t1blk6"},
	{"textures/techcomp_t1blu1", "textures/makkon_tech/techcomp_t1blu1"},
	{"textures/techcomp_t1blu2", "textures/makkon_tech/techcomp_t1blu2"},
	{"textures/techcomp_t1blu3", "textures/makkon_tech/techcomp_t1blu3"},
	{"textures/techcomp_t1blu4", "textures/makkon_tech/techcomp_t1blu4"},
	{"textures/techcomp_t1blu5", "textures/makkon_tech/techcomp_t1blu5"},
	{"textures/techcomp_t1blu6", "textures/makkon_tech/techcomp_t1blu6"},
	{"textures/techcomp_t1brn1", "textures/makkon_tech/techcomp_t1brn1"},
	{"textures/techcomp_t1brn2", "textures/makkon_tech/techcomp_t1brn2"},
	{"textures/techcomp_t1brn3", "textures/makkon_tech/techcomp_t1brn3"},
	{"textures/techcomp_t1brn4", "textures/makkon_tech/techcomp_t1brn4"},
	{"textures/techcomp_t1brn5", "textures/makkon_tech/techcomp_t1brn5"},
	{"textures/techcomp_t1brn6", "textures/makkon_tech/techcomp_t1brn6"},
	{"textures/techcomp_t1brn7", "textures/makkon_tech/techcomp_t1brn7"},
	{"textures/techcomp_t1brn8", "textures/makkon_tech/techcomp_t1brn8"},
	{"textures/techcomp_t1prp1", "textures/makkon_tech/techcomp_t1prp1"},
	{"textures/techcomp_t1prp2", "textures/makkon_tech/techcomp_t1prp2"},
	{"textures/techcomp_t1prp3", "textures/makkon_tech/techcomp_t1prp3"},
	{"textures/techcomp_t1prp4", "textures/makkon_tech/techcomp_t1prp4"},
	{"textures/techcomp_t1red1", "textures/makkon_tech/techcomp_t1red1"},
	{"textures/techcomp_t1red2", "textures/makkon_tech/techcomp_t1red2"},
	{"textures/techcomp_t1rst1", "textures/makkon_tech/techcomp_t1rst1"},
	{"textures/techcomp_t1rst2", "textures/makkon_tech/techcomp_t1rst2"},
	{"textures/techcomp_t1rst3", "textures/makkon_tech/techcomp_t1rst3"},
	{"textures/techcomp_t1rst4", "textures/makkon_tech/techcomp_t1rst4"},
	{"textures/techcomp_t1rst5", "textures/makkon_tech/techcomp_t1rst5"},
	{"textures/techcomp_t1rst6", "textures/makkon_tech/techcomp_t1rst6"},
	{"textures/techcomp_t1tel1", "textures/makkon_tech/techcomp_t1tel1"},
	{"textures/techcomp_t1tel2", "textures/makkon_tech/techcomp_t1tel2"},
	{"textures/techcomp_t1tel3", "textures/makkon_tech/techcomp_t1tel3"},
	{"textures/techcomp_t1tel4", "textures/makkon_tech/techcomp_t1tel4"},
	{"textures/techcont1_blk1", "textures/makkon_tech/techcont1_blk1"},
	{"textures/techcont1_blu1", "textures/makkon_tech/techcont1_blu1"},
	{"textures/techcont1_brwn1", "textures/makkon_tech/techcont1_brwn1"},
	{"textures/techcont1_brwn2", "textures/makkon_tech/techcont1_brwn2"},
	{"textures/techcont1_grey1", "textures/makkon_tech/techcont1_grey1"},
	{"textures/techcont1_grey2", "textures/makkon_tech/techcont1_grey2"},
	{"textures/techcont1_orng1", "textures/makkon_tech/techcont1_orng1"},
	{"textures/techcont1_prpl1", "textures/makkon_tech/techcont1_prpl1"},
	{"textures/techcont1_red1", "textures/makkon_tech/techcont1_red1"},
	{"textures/techcont1_red2", "textures/makkon_tech/techcont1_red2"},
	{"textures/techcont1_rust1", "textures/makkon_tech/techcont1_rust1"},
	{"textures/techcont1_rust2", "textures/makkon_tech/techcont1_rust2"},
	{"textures/techcont1_rust3", "textures/makkon_tech/techcont1_rust3"},
	{"textures/techcont1_teal1", "textures/makkon_tech/techcont1_teal1"},
	{"textures/techcont1_ylw1", "textures/makkon_tech/techcont1_ylw1"},
	{"textures/techcont1_ylw2", "textures/makkon_tech/techcont1_ylw2"},
	{"textures/techcont2_blu1", "textures/makkon_tech/techcont2_blu1"},
	{"textures/techcont2_brwn1", "textures/makkon_tech/techcont2_brwn1"},
	{"textures/techcont2_brwn2", "textures/makkon_tech/techcont2_brwn2"},
	{"textures/techcont2_grey1", "textures/makkon_tech/techcont2_grey1"},
	{"textures/techcont2_grey2", "textures/makkon_tech/techcont2_grey2"},
	{"textures/techcont2_orng1", "textures/makkon_tech/techcont2_orng1"},
	{"textures/techcont2_prpl1", "textures/makkon_tech/techcont2_prpl1"},
	{"textures/techcont2_red1", "textures/makkon_tech/techcont2_red1"},
	{"textures/techcont2_red2", "textures/makkon_tech/techcont2_red2"},
	{"textures/techcont2_rust1", "textures/makkon_tech/techcont2_rust1"},
	{"textures/techcont2_rust2", "textures/makkon_tech/techcont2_rust2"},
	{"textures/techcont2_rust3", "textures/makkon_tech/techcont2_rust3"},
	{"textures/techcont2_teal1", "textures/makkon_tech/techcont2_teal1"},
	{"textures/techcont2_ylw1", "textures/makkon_tech/techcont2_ylw1"},
	{"textures/techcont2_ylw2", "textures/makkon_tech/techcont2_ylw2"},
	{"textures/techcont3_blu1", "textures/makkon_tech/techcont3_blu1"},
	{"textures/techcont3_brwn1", "textures/makkon_tech/techcont3_brwn1"},
	{"textures/techcont3_brwn2", "textures/makkon_tech/techcont3_brwn2"},
	{"textures/techcont3_grey1", "textures/makkon_tech/techcont3_grey1"},
	{"textures/techcont3_grey2", "textures/makkon_tech/techcont3_grey2"},
	{"textures/techcont3_orng1", "textures/makkon_tech/techcont3_orng1"},
	{"textures/techcont3_prpl1", "textures/makkon_tech/techcont3_prpl1"},
	{"textures/techcont3_red1", "textures/makkon_tech/techcont3_red1"},
	{"textures/techcont3_red2", "textures/makkon_tech/techcont3_red2"},
	{"textures/techcont3_rust3", "textures/makkon_tech/techcont3_rust3"},
	{"textures/techcont3_teal1", "textures/makkon_tech/techcont3_teal1"},
	{"textures/techcont3_ylw1", "textures/makkon_tech/techcont3_ylw1"},
	{"textures/techcont3_ylw2", "textures/makkon_tech/techcont3_ylw2"},
	{"textures/techdoor_blk1", "textures/makkon_tech/techdoor_blk1"},
	{"textures/techdoor_blk1b", "textures/makkon_tech/techdoor_blk1b"},
	{"textures/techdoor_blk1r", "textures/makkon_tech/techdoor_blk1r"},
	{"textures/techdoor_blk1y", "textures/makkon_tech/techdoor_blk1y"},
	{"textures/techdoor_blu1", "textures/makkon_tech/techdoor_blu1"},
	{"textures/techdoor_brwn1", "textures/makkon_tech/techdoor_brwn1"},
	{"textures/techdoor_brwn2", "textures/makkon_tech/techdoor_brwn2"},
	{"textures/techdoor_grey1", "textures/makkon_tech/techdoor_grey1"},
	{"textures/techdoor_grey1b", "textures/makkon_tech/techdoor_grey1b"},
	{"textures/techdoor_grey1r", "textures/makkon_tech/techdoor_grey1r"},
	{"textures/techdoor_grey1y", "textures/makkon_tech/techdoor_grey1y"},
	{"textures/techdoor_grey2", "textures/makkon_tech/techdoor_grey2"},
	{"textures/techdoor_grey2b", "textures/makkon_tech/techdoor_grey2b"},
	{"textures/techdoor_grey2r", "textures/makkon_tech/techdoor_grey2r"},
	{"textures/techdoor_grey2y", "textures/makkon_tech/techdoor_grey2y"},
	{"textures/techdoor_orng1", "textures/makkon_tech/techdoor_orng1"},
	{"textures/techdoor_prpl1", "textures/makkon_tech/techdoor_prpl1"},
	{"textures/techdoor_red1", "textures/makkon_tech/techdoor_red1"},
	{"textures/techdoor_red2", "textures/makkon_tech/techdoor_red2"},
	{"textures/techdoor_rst1", "textures/makkon_tech/techdoor_rst1"},
	{"textures/techdoor_rst2", "textures/makkon_tech/techdoor_rst2"},
	{"textures/techdoor_rst3", "textures/makkon_tech/techdoor_rst3"},
	{"textures/techdoor_teal1", "textures/makkon_tech/techdoor_teal1"},
	{"textures/techdoor_ylw1", "textures/makkon_tech/techdoor_ylw1"},
	{"textures/techdoor_ylw2", "textures/makkon_tech/techdoor_ylw2"},
	{"textures/techflat1_blk1", "textures/makkon_tech/techflat1_blk1"},
	{"textures/techflat1_blu1", "textures/makkon_tech/techflat1_blu1"},
	{"textures/techflat1_brwn1", "textures/makkon_tech/techflat1_brwn1"},
	{"textures/techflat1_brwn2", "textures/makkon_tech/techflat1_brwn2"},
	{"textures/techflat1_grey1", "textures/makkon_tech/techflat1_grey1"},
	{"textures/techflat1_grey2", "textures/makkon_tech/techflat1_grey2"},
	{"textures/techflat1_orng1", "textures/makkon_tech/techflat1_orng1"},
	{"textures/techflat1_prpl1", "textures/makkon_tech/techflat1_prpl1"},
	{"textures/techflat1_red1", "textures/makkon_tech/techflat1_red1"},
	{"textures/techflat1_red2", "textures/makkon_tech/techflat1_red2"},
	{"textures/techflat1_teal1", "textures/makkon_tech/techflat1_teal1"},
	{"textures/techflat1_ylw1", "textures/makkon_tech/techflat1_ylw1"},
	{"textures/techflat1_ylw2", "textures/makkon_tech/techflat1_ylw2"},
	{"textures/techflat2_blk1", "textures/makkon_tech/techflat2_blk1"},
	{"textures/techflat2_blu1", "textures/makkon_tech/techflat2_blu1"},
	{"textures/techflat2_brwn1", "textures/makkon_tech/techflat2_brwn1"},
	{"textures/techflat2_brwn2", "textures/makkon_tech/techflat2_brwn2"},
	{"textures/techflat2_grey1", "textures/makkon_tech/techflat2_grey1"},
	{"textures/techflat2_grey2", "textures/makkon_tech/techflat2_grey2"},
	{"textures/techflat2_orng1", "textures/makkon_tech/techflat2_orng1"},
	{"textures/techflat2_prpl1", "textures/makkon_tech/techflat2_prpl1"},
	{"textures/techflat2_red1", "textures/makkon_tech/techflat2_red1"},
	{"textures/techflat2_red2", "textures/makkon_tech/techflat2_red2"},
	{"textures/techflat2_teal1", "textures/makkon_tech/techflat2_teal1"},
	{"textures/techflat2_ylw1", "textures/makkon_tech/techflat2_ylw1"},
	{"textures/techflat2_ylw2", "textures/makkon_tech/techflat2_ylw2"},
	{"textures/techfloor1_blk1", "textures/makkon_tech/techfloor1_blk1"},
	{"textures/techfloor1_blu1", "textures/makkon_tech/techfloor1_blu1"},
	{"textures/techfloor1_brwn1", "textures/makkon_tech/techfloor1_brwn1"},
	{"textures/techfloor1_brwn2", "textures/makkon_tech/techfloor1_brwn2"},
	{"textures/techfloor1_grey1", "textures/makkon_tech/techfloor1_grey1"},
	{"textures/techfloor1_grey2", "textures/makkon_tech/techfloor1_grey2"},
	{"textures/techfloor1_prpl1", "textures/makkon_tech/techfloor1_prpl1"},
	{"textures/techfloor1_red1", "textures/makkon_tech/techfloor1_red1"},
	{"textures/techfloor1_teal1", "textures/makkon_tech/techfloor1_teal1"},
	{"textures/techfloor1_ylw1", "textures/makkon_tech/techfloor1_ylw1"},
	{"textures/techfloor1_ylw2", "textures/makkon_tech/techfloor1_ylw2"},
	{"textures/techflr1_blk1", "textures/makkon_tech/techflr1_blk1"},
	{"textures/techflr1_blu1", "textures/makkon_tech/techflr1_blu1"},
	{"textures/techflr1_brwn1", "textures/makkon_tech/techflr1_brwn1"},
	{"textures/techflr1_brwn2", "textures/makkon_tech/techflr1_brwn2"},
	{"textures/techflr1_grey1", "textures/makkon_tech/techflr1_grey1"},
	{"textures/techflr1_grey2", "textures/makkon_tech/techflr1_grey2"},
	{"textures/techflr1_orng1", "textures/makkon_tech/techflr1_orng1"},
	{"textures/techflr1_prpl1", "textures/makkon_tech/techflr1_prpl1"},
	{"textures/techflr1_red1", "textures/makkon_tech/techflr1_red1"},
	{"textures/techflr1_red2", "textures/makkon_tech/techflr1_red2"},
	{"textures/techflr1_teal1", "textures/makkon_tech/techflr1_teal1"},
	{"textures/techflr1_ylw1", "textures/makkon_tech/techflr1_ylw1"},
	{"textures/techflr1_ylw2", "textures/makkon_tech/techflr1_ylw2"},
	{"textures/techflr2_blk1", "textures/makkon_tech/techflr2_blk1"},
	{"textures/techflr2_blu1", "textures/makkon_tech/techflr2_blu1"},
	{"textures/techflr2_brwn1", "textures/makkon_tech/techflr2_brwn1"},
	{"textures/techflr2_brwn2", "textures/makkon_tech/techflr2_brwn2"},
	{"textures/techflr2_grey1", "textures/makkon_tech/techflr2_grey1"},
	{"textures/techflr2_grey2", "textures/makkon_tech/techflr2_grey2"},
	{"textures/techflr2_orng1", "textures/makkon_tech/techflr2_orng1"},
	{"textures/techflr2_prpl1", "textures/makkon_tech/techflr2_prpl1"},
	{"textures/techflr2_red1", "textures/makkon_tech/techflr2_red1"},
	{"textures/techflr2_red2", "textures/makkon_tech/techflr2_red2"},
	{"textures/techflr2_teal1", "textures/makkon_tech/techflr2_teal1"},
	{"textures/techflr2_ylw1", "textures/makkon_tech/techflr2_ylw1"},
	{"textures/techflr2_ylw2", "textures/makkon_tech/techflr2_ylw2"},
	{"textures/techflr3_blk1", "textures/makkon_tech/techflr3_blk1"},
	{"textures/techflr3_blu1", "textures/makkon_tech/techflr3_blu1"},
	{"textures/techflr3_brwn1", "textures/makkon_tech/techflr3_brwn1"},
	{"textures/techflr3_brwn2", "textures/makkon_tech/techflr3_brwn2"},
	{"textures/techflr3_grey1", "textures/makkon_tech/techflr3_grey1"},
	{"textures/techflr3_grey2", "textures/makkon_tech/techflr3_grey2"},
	{"textures/techflr3_orng1", "textures/makkon_tech/techflr3_orng1"},
	{"textures/techflr3_prpl1", "textures/makkon_tech/techflr3_prpl1"},
	{"textures/techflr3_red1", "textures/makkon_tech/techflr3_red1"},
	{"textures/techflr3_red2", "textures/makkon_tech/techflr3_red2"},
	{"textures/techflr3_teal1", "textures/makkon_tech/techflr3_teal1"},
	{"textures/techflr3_ylw1", "textures/makkon_tech/techflr3_ylw1"},
	{"textures/techflr3_ylw2", "textures/makkon_tech/techflr3_ylw2"},
	{"textures/techflr4_blk1", "textures/makkon_tech/techflr4_blk1"},
	{"textures/techflr4_blu1", "textures/makkon_tech/techflr4_blu1"},
	{"textures/techflr4_brwn1", "textures/makkon_tech/techflr4_brwn1"},
	{"textures/techflr4_brwn2", "textures/makkon_tech/techflr4_brwn2"},
	{"textures/techflr4_grey1", "textures/makkon_tech/techflr4_grey1"},
	{"textures/techflr4_grey2", "textures/makkon_tech/techflr4_grey2"},
	{"textures/techflr4_orng1", "textures/makkon_tech/techflr4_orng1"},
	{"textures/techflr4_prpl1", "textures/makkon_tech/techflr4_prpl1"},
	{"textures/techflr4_red1", "textures/makkon_tech/techflr4_red1"},
	{"textures/techflr4_red2", "textures/makkon_tech/techflr4_red2"},
	{"textures/techflr4_teal1", "textures/makkon_tech/techflr4_teal1"},
	{"textures/techflr4_ylw1", "textures/makkon_tech/techflr4_ylw1"},
	{"textures/techflr4_ylw2", "textures/makkon_tech/techflr4_ylw2"},
	{"textures/techflr5_blk1", "textures/makkon_tech/techflr5_blk1"},
	{"textures/techflr5_blu1", "textures/makkon_tech/techflr5_blu1"},
	{"textures/techflr5_brwn1", "textures/makkon_tech/techflr5_brwn1"},
	{"textures/techflr5_brwn2", "textures/makkon_tech/techflr5_brwn2"},
	{"textures/techflr5_grey1", "textures/makkon_tech/techflr5_grey1"},
	{"textures/techflr5_grey2", "textures/makkon_tech/techflr5_grey2"},
	{"textures/techflr5_orng1", "textures/makkon_tech/techflr5_orng1"},
	{"textures/techflr5_prpl1", "textures/makkon_tech/techflr5_prpl1"},
	{"textures/techflr5_red1", "textures/makkon_tech/techflr5_red1"},
	{"textures/techflr5_red2", "textures/makkon_tech/techflr5_red2"},
	{"textures/techflr5_teal1", "textures/makkon_tech/techflr5_teal1"},
	{"textures/techflr5_ylw1", "textures/makkon_tech/techflr5_ylw1"},
	{"textures/techflr5_ylw2", "textures/makkon_tech/techflr5_ylw2"},
	{"textures/techlight_1", "textures/makkon_tech/techlight_1"},
	{"textures/techlight_r1", "textures/makkon_tech/techlight_r1"},
	{"textures/techlzr_blu1", "textures/makkon_tech/techlzr_blu1"},
	{"textures/techlzr_blu2", "textures/makkon_tech/techlzr_blu2"},
	{"textures/techlzr_blu3", "textures/makkon_tech/techlzr_blu3"},
	{"textures/techlzr_grey1", "textures/makkon_tech/techlzr_grey1"},
	{"textures/techlzr_grn1", "textures/makkon_tech/techlzr_grn1"},
	{"textures/techlzr_orng1", "textures/makkon_tech/techlzr_orng1"},
	{"textures/techlzr_orng2", "textures/makkon_tech/techlzr_orng2"},
	{"textures/techlzr_pink1", "textures/makkon_tech/techlzr_pink1"},
	{"textures/techlzr_red1", "textures/makkon_tech/techlzr_red1"},
	{"textures/techlzr_red2", "textures/makkon_tech/techlzr_red2"},
	{"textures/techlzr_ylw1", "textures/makkon_tech/techlzr_ylw1"},
	{"textures/techlzr_ylw2", "textures/makkon_tech/techlzr_ylw2"},
	{"textures/techpipe1_blk1", "textures/makkon_tech/techpipe1_blk1"},
	{"textures/techpipe1_blu1", "textures/makkon_tech/techpipe1_blu1"},
	{"textures/techpipe1_brwn1", "textures/makkon_tech/techpipe1_brwn1"},
	{"textures/techpipe1_brwn2", "textures/makkon_tech/techpipe1_brwn2"},
	{"textures/techpipe1_grey1", "textures/makkon_tech/techpipe1_grey1"},
	{"textures/techpipe1_grey2", "textures/makkon_tech/techpipe1_grey2"},
	{"textures/techpipe1_orng1", "textures/makkon_tech/techpipe1_orng1"},
	{"textures/techpipe1_prpl1", "textures/makkon_tech/techpipe1_prpl1"},
	{"textures/techpipe1_red1", "textures/makkon_tech/techpipe1_red1"},
	{"textures/techpipe1_red2", "textures/makkon_tech/techpipe1_red2"},
	{"textures/techpipe1_rst1", "textures/makkon_tech/techpipe1_rst1"},
	{"textures/techpipe1_rst3", "textures/makkon_tech/techpipe1_rst3"},
	{"textures/techpipe1_teal1", "textures/makkon_tech/techpipe1_teal1"},
	{"textures/techpipe1_ylw1", "textures/makkon_tech/techpipe1_ylw1"},
	{"textures/techpipe1_ylw2", "textures/makkon_tech/techpipe1_ylw2"},
	{"textures/techpipe2_blu1", "textures/makkon_tech/techpipe2_blu1"},
	{"textures/techpipe2_brwn1", "textures/makkon_tech/techpipe2_brwn1"},
	{"textures/techpipe2_brwn2", "textures/makkon_tech/techpipe2_brwn2"},
	{"textures/techpipe2_grey1", "textures/makkon_tech/techpipe2_grey1"},
	{"textures/techpipe2_grey2", "textures/makkon_tech/techpipe2_grey2"},
	{"textures/techpipe2_orng1", "textures/makkon_tech/techpipe2_orng1"},
	{"textures/techpipe2_prpl1", "textures/makkon_tech/techpipe2_prpl1"},
	{"textures/techpipe2_red1", "textures/makkon_tech/techpipe2_red1"},
	{"textures/techpipe2_red2", "textures/makkon_tech/techpipe2_red2"},
	{"textures/techpipe2_rst1", "textures/makkon_tech/techpipe2_rst1"},
	{"textures/techpipe2_rst3", "textures/makkon_tech/techpipe2_rst3"},
	{"textures/techpipe2_teal1", "textures/makkon_tech/techpipe2_teal1"},
	{"textures/techpipe2_ylw1", "textures/makkon_tech/techpipe2_ylw1"},
	{"textures/techpipe2_ylw2", "textures/makkon_tech/techpipe2_ylw2"},
	{"textures/techpipe3_rst1", "textures/makkon_tech/techpipe3_rst1"},
	{"textures/techpipe3_rst2", "textures/makkon_tech/techpipe3_rst2"},
	{"textures/techpipe3_rst3", "textures/makkon_tech/techpipe3_rst3"},
	{"textures/techpnl_blk1", "textures/makkon_tech/techpnl_blk1"},
	{"textures/techpnl_blu1", "textures/makkon_tech/techpnl_blu1"},
	{"textures/techpnl_brwn1", "textures/makkon_tech/techpnl_brwn1"},
	{"textures/techpnl_brwn2", "textures/makkon_tech/techpnl_brwn2"},
	{"textures/techpnl_grey1", "textures/makkon_tech/techpnl_grey1"},
	{"textures/techpnl_grey2", "textures/makkon_tech/techpnl_grey2"},
	{"textures/techpnl_orng1", "textures/makkon_tech/techpnl_orng1"},
	{"textures/techpnl_prpl1", "textures/makkon_tech/techpnl_prpl1"},
	{"textures/techpnl_red1", "textures/makkon_tech/techpnl_red1"},
	{"textures/techpnl_red2", "textures/makkon_tech/techpnl_red2"},
	{"textures/techpnl_rst1", "textures/makkon_tech/techpnl_rst1"},
	{"textures/techpnl_rst2", "textures/makkon_tech/techpnl_rst2"},
	{"textures/techpnl_rst3", "textures/makkon_tech/techpnl_rst3"},
	{"textures/techpnl_teal1", "textures/makkon_tech/techpnl_teal1"},
	{"textures/techpnl_ylw1", "textures/makkon_tech/techpnl_ylw1"},
	{"textures/techpnl_ylw2", "textures/makkon_tech/techpnl_ylw2"},
	{"textures/techtrim1_blk1", "textures/makkon_tech/techtrim1_blk1"},
	{"textures/techtrim1_blu1", "textures/makkon_tech/techtrim1_blu1"},
	{"textures/techtrim1_brwn1", "textures/makkon_tech/techtrim1_brwn1"},
	{"textures/techtrim1_brwn2", "textures/makkon_tech/techtrim1_brwn2"},
	{"textures/techtrim1_grey1", "textures/makkon_tech/techtrim1_grey1"},
	{"textures/techtrim1_grey2", "textures/makkon_tech/techtrim1_grey2"},
	{"textures/techtrim1_orng1", "textures/makkon_tech/techtrim1_orng1"},
	{"textures/techtrim1_prpl1", "textures/makkon_tech/techtrim1_prpl1"},
	{"textures/techtrim1_red1", "textures/makkon_tech/techtrim1_red1"},
	{"textures/techtrim1_red2", "textures/makkon_tech/techtrim1_red2"},
	{"textures/techtrim1_rst1", "textures/makkon_tech/techtrim1_rst1"},
	{"textures/techtrim1_rst2", "textures/makkon_tech/techtrim1_rst2"},
	{"textures/techtrim1_rst3", "textures/makkon_tech/techtrim1_rst3"},
	{"textures/techtrim1_teal1", "textures/makkon_tech/techtrim1_teal1"},
	{"textures/techtrim1_ylw1", "textures/makkon_tech/techtrim1_ylw1"},
	{"textures/techtrim1_ylw2", "textures/makkon_tech/techtrim1_ylw2"},
	{"textures/techtrim2_blk1", "textures/makkon_tech/techtrim2_blk1"},
	{"textures/techtrim2_blu1", "textures/makkon_tech/techtrim2_blu1"},
	{"textures/techtrim2_brwn1", "textures/makkon_tech/techtrim2_brwn1"},
	{"textures/techtrim2_brwn2", "textures/makkon_tech/techtrim2_brwn2"},
	{"textures/techtrim2_grey1", "textures/makkon_tech/techtrim2_grey1"},
	{"textures/techtrim2_grey2", "textures/makkon_tech/techtrim2_grey2"},
	{"textures/techtrim2_orng1", "textures/makkon_tech/techtrim2_orng1"},
	{"textures/techtrim2_prpl1", "textures/makkon_tech/techtrim2_prpl1"},
	{"textures/techtrim2_red1", "textures/makkon_tech/techtrim2_red1"},
	{"textures/techtrim2_red2", "textures/makkon_tech/techtrim2_red2"},
	{"textures/techtrim2_teal1", "textures/makkon_tech/techtrim2_teal1"},
	{"textures/techtrim2_ylw1", "textures/makkon_tech/techtrim2_ylw1"},
	{"textures/techtrim2_ylw2", "textures/makkon_tech/techtrim2_ylw2"},
	{"textures/techtrim3_blk1", "textures/makkon_tech/techtrim3_blk1"},
	{"textures/techtrim3_blu1", "textures/makkon_tech/techtrim3_blu1"},
	{"textures/techtrim3_brwn1", "textures/makkon_tech/techtrim3_brwn1"},
	{"textures/techtrim3_brwn2", "textures/makkon_tech/techtrim3_brwn2"},
	{"textures/techtrim3_grey1", "textures/makkon_tech/techtrim3_grey1"},
	{"textures/techtrim3_grey2", "textures/makkon_tech/techtrim3_grey2"},
	{"textures/techtrim3_orng1", "textures/makkon_tech/techtrim3_orng1"},
	{"textures/techtrim3_prpl1", "textures/makkon_tech/techtrim3_prpl1"},
	{"textures/techtrim3_red1", "textures/makkon_tech/techtrim3_red1"},
	{"textures/techtrim3_red2", "textures/makkon_tech/techtrim3_red2"},
	{"textures/techtrim3_rst1", "textures/makkon_tech/techtrim3_rst1"},
	{"textures/techtrim3_rst2", "textures/makkon_tech/techtrim3_rst2"},
	{"textures/techtrim3_rst3", "textures/makkon_tech/techtrim3_rst3"},
	{"textures/techtrim3_teal1", "textures/makkon_tech/techtrim3_teal1"},
	{"textures/techtrim3_ylw1", "textures/makkon_tech/techtrim3_ylw1"},
	{"textures/techtrim3_ylw2", "textures/makkon_tech/techtrim3_ylw2"},
	{"textures/techvnt1_blk1", "textures/makkon_tech/techvnt1_blk1"},
	{"textures/techvnt1_blu1", "textures/makkon_tech/techvnt1_blu1"},
	{"textures/techvnt1_brwn1", "textures/makkon_tech/techvnt1_brwn1"},
	{"textures/techvnt1_brwn2", "textures/makkon_tech/techvnt1_brwn2"},
	{"textures/techvnt1_grey1", "textures/makkon_tech/techvnt1_grey1"},
	{"textures/techvnt1_grey2", "textures/makkon_tech/techvnt1_grey2"},
	{"textures/techvnt1_orng1", "textures/makkon_tech/techvnt1_orng1"},
	{"textures/techvnt1_prpl1", "textures/makkon_tech/techvnt1_prpl1"},
	{"textures/techvnt1_red1", "textures/makkon_tech/techvnt1_red1"},
	{"textures/techvnt1_red2", "textures/makkon_tech/techvnt1_red2"},
	{"textures/techvnt1_rust1", "textures/makkon_tech/techvnt1_rust1"},
	{"textures/techvnt1_rust2", "textures/makkon_tech/techvnt1_rust2"},
	{"textures/techvnt1_rust2_1", "textures/makkon_tech/techvnt1_rust2_1"},
	{"textures/techvnt1_treal1", "textures/makkon_tech/techvnt1_treal1"},
	{"textures/techvnt1_ylw1", "textures/makkon_tech/techvnt1_ylw1"},
	{"textures/techvnt1_ylw2", "textures/makkon_tech/techvnt1_ylw2"},
	{"textures/techvnt2_blk1", "textures/makkon_tech/techvnt2_blk1"},
	{"textures/techvnt2_blue1", "textures/makkon_tech/techvnt2_blue1"},
	{"textures/techvnt2_brwn1", "textures/makkon_tech/techvnt2_brwn1"},
	{"textures/techvnt2_brwn2", "textures/makkon_tech/techvnt2_brwn2"},
	{"textures/techvnt2_grey1", "textures/makkon_tech/techvnt2_grey1"},
	{"textures/techvnt2_grey2", "textures/makkon_tech/techvnt2_grey2"},
	{"textures/techvnt2_prpl1", "textures/makkon_tech/techvnt2_prpl1"},
	{"textures/techvnt2_red1", "textures/makkon_tech/techvnt2_red1"},
	{"textures/techvnt2_rust1", "textures/makkon_tech/techvnt2_rust1"},
	{"textures/techvnt2_rust2", "textures/makkon_tech/techvnt2_rust2"},
	{"textures/techvnt2_rust3", "textures/makkon_tech/techvnt2_rust3"},
	{"textures/techvnt2_teal1", "textures/makkon_tech/techvnt2_teal1"},
	{"textures/techvnt2_ylw1", "textures/makkon_tech/techvnt2_ylw1"},
	{"textures/techvnt2_ylw2", "textures/makkon_tech/techvnt2_ylw2"},
	{"textures/techvnt3_blk1", "textures/makkon_tech/techvnt3_blk1"},
	{"textures/techvnt3_blue1", "textures/makkon_tech/techvnt3_blue1"},
	{"textures/techvnt3_brwn1", "textures/makkon_tech/techvnt3_brwn1"},
	{"textures/techvnt3_brwn2", "textures/makkon_tech/techvnt3_brwn2"},
	{"textures/techvnt3_grey1", "textures/makkon_tech/techvnt3_grey1"},
	{"textures/techvnt3_grey2", "textures/makkon_tech/techvnt3_grey2"},
	{"textures/techvnt3_prpl1", "textures/makkon_tech/techvnt3_prpl1"},
	{"textures/techvnt3_red1", "textures/makkon_tech/techvnt3_red1"},
	{"textures/techvnt3_rust1", "textures/makkon_tech/techvnt3_rust1"},
	{"textures/techvnt3_rust3", "textures/makkon_tech/techvnt3_rust3"},
	{"textures/techvnt3_teal1", "textures/makkon_tech/techvnt3_teal1"},
	{"textures/techvnt3_ylw1", "textures/makkon_tech/techvnt3_ylw1"},
	{"textures/techvnt3_ylw2", "textures/makkon_tech/techvnt3_ylw2"},
	{"textures/techwal5_blk1", "textures/makkon_tech/techwal5_blk1"},
	{"textures/techwal5_blue1", "textures/makkon_tech/techwal5_blue1"},
	{"textures/techwal5_brwn1", "textures/makkon_tech/techwal5_brwn1"},
	{"textures/techwal5_brwn2", "textures/makkon_tech/techwal5_brwn2"},
	{"textures/techwal5_grey1", "textures/makkon_tech/techwal5_grey1"},
	{"textures/techwal5_grey2", "textures/makkon_tech/techwal5_grey2"},
	{"textures/techwal5_orng1", "textures/makkon_tech/techwal5_orng1"},
	{"textures/techwal5_prpl1", "textures/makkon_tech/techwal5_prpl1"},
	{"textures/techwal5_red1", "textures/makkon_tech/techwal5_red1"},
	{"textures/techwal5_red2", "textures/makkon_tech/techwal5_red2"},
	{"textures/techwal5_rust1", "textures/makkon_tech/techwal5_rust1"},
	{"textures/techwal5_rust2", "textures/makkon_tech/techwal5_rust2"},
	{"textures/techwal5_rust3", "textures/makkon_tech/techwal5_rust3"},
	{"textures/techwal5_teal1", "textures/makkon_tech/techwal5_teal1"},
	{"textures/techwal5_ylw1", "textures/makkon_tech/techwal5_ylw1"},
	{"textures/techwal5_ylw2", "textures/makkon_tech/techwal5_ylw2"},
	{"textures/techwal6_blk1", "textures/makkon_tech/techwal6_blk1"},
	{"textures/techwal6_blue1", "textures/makkon_tech/techwal6_blue1"},
	{"textures/techwal6_brwn1", "textures/makkon_tech/techwal6_brwn1"},
	{"textures/techwal6_brwn2", "textures/makkon_tech/techwal6_brwn2"},
	{"textures/techwal6_grey1", "textures/makkon_tech/techwal6_grey1"},
	{"textures/techwal6_grey2", "textures/makkon_tech/techwal6_grey2"},
	{"textures/techwal6_orng1", "textures/makkon_tech/techwal6_orng1"},
	{"textures/techwal6_prpl1", "textures/makkon_tech/techwal6_prpl1"},
	{"textures/techwal6_red1", "textures/makkon_tech/techwal6_red1"},
	{"textures/techwal6_red2", "textures/makkon_tech/techwal6_red2"},
	{"textures/techwal6_rust1", "textures/makkon_tech/techwal6_rust1"},
	{"textures/techwal6_rust2", "textures/makkon_tech/techwal6_rust2"},
	{"textures/techwal6_rust3", "textures/makkon_tech/techwal6_rust3"},
	{"textures/techwal6_teal1", "textures/makkon_tech/techwal6_teal1"},
	{"textures/techwal6_ylw1", "textures/makkon_tech/techwal6_ylw1"},
	{"textures/techwal6_ylw2", "textures/makkon_tech/techwal6_ylw2"},
	{"textures/techwal7_blue1", "textures/makkon_tech/techwal7_blue1"},
	{"textures/techwal7_brwn1", "textures/makkon_tech/techwal7_brwn1"},
	{"textures/techwal7_brwn2", "textures/makkon_tech/techwal7_brwn2"},
	{"textures/techwal7_grey1", "textures/makkon_tech/techwal7_grey1"},
	{"textures/techwal7_grey2", "textures/makkon_tech/techwal7_grey2"},
	{"textures/techwal7_orng1", "textures/makkon_tech/techwal7_orng1"},
	{"textures/techwal7_prpl1", "textures/makkon_tech/techwal7_prpl1"},
	{"textures/techwal7_red1", "textures/makkon_tech/techwal7_red1"},
	{"textures/techwal7_red2", "textures/makkon_tech/techwal7_red2"},
	{"textures/techwal7_rust1", "textures/makkon_tech/techwal7_rust1"},
	{"textures/techwal7_teal1", "textures/makkon_tech/techwal7_teal1"},
	{"textures/techwal7_ylw1", "textures/makkon_tech/techwal7_ylw1"},
	{"textures/techwal7_ylw2", "textures/makkon_tech/techwal7_ylw2"},
	{"textures/techwall1_blk1", "textures/makkon_tech/techwall1_blk1"},
	{"textures/techwall1_blu1", "textures/makkon_tech/techwall1_blu1"},
	{"textures/techwall1_brwn1", "textures/makkon_tech/techwall1_brwn1"},
	{"textures/techwall1_brwn2", "textures/makkon_tech/techwall1_brwn2"},
	{"textures/techwall1_grey1", "textures/makkon_tech/techwall1_grey1"},
	{"textures/techwall1_grey2", "textures/makkon_tech/techwall1_grey2"},
	{"textures/techwall1_orng1", "textures/makkon_tech/techwall1_orng1"},
	{"textures/techwall1_prpl1", "textures/makkon_tech/techwall1_prpl1"},
	{"textures/techwall1_red1", "textures/makkon_tech/techwall1_red1"},
	{"textures/techwall1_red2", "textures/makkon_tech/techwall1_red2"},
	{"textures/techwall1_rst1", "textures/makkon_tech/techwall1_rst1"},
	{"textures/techwall1_rst2", "textures/makkon_tech/techwall1_rst2"},
	{"textures/techwall1_rst3", "textures/makkon_tech/techwall1_rst3"},
	{"textures/techwall1_teal1", "textures/makkon_tech/techwall1_teal1"},
	{"textures/techwall1_ylw1", "textures/makkon_tech/techwall1_ylw1"},
	{"textures/techwall1_ylw2", "textures/makkon_tech/techwall1_ylw2"},
	{"textures/techwall2_blk1", "textures/makkon_tech/techwall2_blk1"},
	{"textures/techwall2_blu1", "textures/makkon_tech/techwall2_blu1"},
	{"textures/techwall2_brwn1", "textures/makkon_tech/techwall2_brwn1"},
	{"textures/techwall2_brwn2", "textures/makkon_tech/techwall2_brwn2"},
	{"textures/techwall2_grey1", "textures/makkon_tech/techwall2_grey1"},
	{"textures/techwall2_grey2", "textures/makkon_tech/techwall2_grey2"},
	{"textures/techwall2_orng1", "textures/makkon_tech/techwall2_orng1"},
	{"textures/techwall2_prpl1", "textures/makkon_tech/techwall2_prpl1"},
	{"textures/techwall2_red1", "textures/makkon_tech/techwall2_red1"},
	{"textures/techwall2_red2", "textures/makkon_tech/techwall2_red2"},
	{"textures/techwall2_rst1", "textures/makkon_tech/techwall2_rst1"},
	{"textures/techwall2_rst2", "textures/makkon_tech/techwall2_rst2"},
	{"textures/techwall2_rst3", "textures/makkon_tech/techwall2_rst3"},
	{"textures/techwall2_teal1", "textures/makkon_tech/techwall2_teal1"},
	{"textures/techwall2_ylw1", "textures/makkon_tech/techwall2_ylw1"},
	{"textures/techwall2_ylw2", "textures/makkon_tech/techwall2_ylw2"},
	{"textures/techwall3_blk1", "textures/makkon_tech/techwall3_blk1"},
	{"textures/techwall3_blu1", "textures/makkon_tech/techwall3_blu1"},
	{"textures/techwall3_brwn1", "textures/makkon_tech/techwall3_brwn1"},
	{"textures/techwall3_brwn2", "textures/makkon_tech/techwall3_brwn2"},
	{"textures/techwall3_grey1", "textures/makkon_tech/techwall3_grey1"},
	{"textures/techwall3_grey2", "textures/makkon_tech/techwall3_grey2"},
	{"textures/techwall3_orng1", "textures/makkon_tech/techwall3_orng1"},
	{"textures/techwall3_prpl1", "textures/makkon_tech/techwall3_prpl1"},
	{"textures/techwall3_red1", "textures/makkon_tech/techwall3_red1"},
	{"textures/techwall3_red2", "textures/makkon_tech/techwall3_red2"},
	{"textures/techwall3_rst1", "textures/makkon_tech/techwall3_rst1"},
	{"textures/techwall3_rst2", "textures/makkon_tech/techwall3_rst2"},
	{"textures/techwall3_rst3", "textures/makkon_tech/techwall3_rst3"},
	{"textures/techwall3_teal1", "textures/makkon_tech/techwall3_teal1"},
	{"textures/techwall3_ylw1", "textures/makkon_tech/techwall3_ylw1"},
	{"textures/techwall3_ylw2", "textures/makkon_tech/techwall3_ylw2"},
	{"textures/techwall4_blk1", "textures/makkon_tech/techwall4_blk1"},
	{"textures/techwall4_blu1", "textures/makkon_tech/techwall4_blu1"},
	{"textures/techwall4_brwn1", "textures/makkon_tech/techwall4_brwn1"},
	{"textures/techwall4_brwn2", "textures/makkon_tech/techwall4_brwn2"},
	{"textures/techwall4_grey1", "textures/makkon_tech/techwall4_grey1"},
	{"textures/techwall4_grey2", "textures/makkon_tech/techwall4_grey2"},
	{"textures/techwall4_orng1", "textures/makkon_tech/techwall4_orng1"},
	{"textures/techwall4_prpl1", "textures/makkon_tech/techwall4_prpl1"},
	{"textures/techwall4_red1", "textures/makkon_tech/techwall4_red1"},
	{"textures/techwall4_red2", "textures/makkon_tech/techwall4_red2"},
	{"textures/techwall4_rst1", "textures/makkon_tech/techwall4_rst1"},
	{"textures/techwall4_rst2", "textures/makkon_tech/techwall4_rst2"},
	{"textures/techwall4_rst3", "textures/makkon_tech/techwall4_rst3"},
	{"textures/techwall4_teal1", "textures/makkon_tech/techwall4_teal1"},
	{"textures/techwall4_ylw1", "textures/makkon_tech/techwall4_ylw1"},
	{"textures/techwall4_ylw2", "textures/makkon_tech/techwall4_ylw2"},
	{"textures/_techdc1_blk1", "textures/makkon_tech/_techdc1_blk1"},
	{"textures/_techdc1_blu1", "textures/makkon_tech/_techdc1_blu1"},
	{"textures/_techdc1_brwn1", "textures/makkon_tech/_techdc1_brwn1"},
	{"textures/_techdc1_brwn2", "textures/makkon_tech/_techdc1_brwn2"},
	{"textures/_techdc1_grey1", "textures/makkon_tech/_techdc1_grey1"},
	{"textures/_techdc1_grey2", "textures/makkon_tech/_techdc1_grey2"},
	{"textures/_techdc1_orn1", "textures/makkon_tech/_techdc1_orn1"},
	{"textures/_techdc1_prpl1", "textures/makkon_tech/_techdc1_prpl1"},
	{"textures/_techdc1_red1", "textures/makkon_tech/_techdc1_red1"},
	{"textures/_techdc1_red2", "textures/makkon_tech/_techdc1_red2"},
	{"textures/_techdc1_teal1", "textures/makkon_tech/_techdc1_teal1"},
	{"textures/_techdc1_ylw1", "textures/makkon_tech/_techdc1_ylw1"},
	{"textures/_techdc1_ylw2", "textures/makkon_tech/_techdc1_ylw2"},
	{"textures/_techdc2_blk1", "textures/makkon_tech/_techdc2_blk1"},
	{"textures/_techdc2_blu1", "textures/makkon_tech/_techdc2_blu1"},
	{"textures/_techdc2_brwn1", "textures/makkon_tech/_techdc2_brwn1"},
	{"textures/_techdc2_brwn2", "textures/makkon_tech/_techdc2_brwn2"},
	{"textures/_techdc2_grey1", "textures/makkon_tech/_techdc2_grey1"},
	{"textures/_techdc2_grey2", "textures/makkon_tech/_techdc2_grey2"},
	{"textures/_techdc2_orn1", "textures/makkon_tech/_techdc2_orn1"},
	{"textures/_techdc2_prpl1", "textures/makkon_tech/_techdc2_prpl1"},
	{"textures/_techdc2_red1", "textures/makkon_tech/_techdc2_red1"},
	{"textures/_techdc2_red2", "textures/makkon_tech/_techdc2_red2"},
	{"textures/_techdc2_teal1", "textures/makkon_tech/_techdc2_teal1"},
	{"textures/_techdc2_ylw1", "textures/makkon_tech/_techdc2_ylw1"},
	{"textures/_techdc2_ylw2", "textures/makkon_tech/_techdc2_ylw2"},
	{"textures/_techdc3_blk1", "textures/makkon_tech/_techdc3_blk1"},
	{"textures/_techdc3_blu1", "textures/makkon_tech/_techdc3_blu1"},
	{"textures/_techdc3_brwn1", "textures/makkon_tech/_techdc3_brwn1"},
	{"textures/_techdc3_brwn2", "textures/makkon_tech/_techdc3_brwn2"},
	{"textures/_techdc3_grey1", "textures/makkon_tech/_techdc3_grey1"},
	{"textures/_techdc3_grey2", "textures/makkon_tech/_techdc3_grey2"},
	{"textures/_techdc3_orn1", "textures/makkon_tech/_techdc3_orn1"},
	{"textures/_techdc3_prpl1", "textures/makkon_tech/_techdc3_prpl1"},
	{"textures/_techdc3_red1", "textures/makkon_tech/_techdc3_red1"},
	{"textures/_techdc3_red2", "textures/makkon_tech/_techdc3_red2"},
	{"textures/_techdc3_teal1", "textures/makkon_tech/_techdc3_teal1"},
	{"textures/_techdc3_ylw1", "textures/makkon_tech/_techdc3_ylw1"},
	{"textures/_techdc3_ylw2", "textures/makkon_tech/_techdc3_ylw2"},
	{"textures/_techdc4_blk1", "textures/makkon_tech/_techdc4_blk1"},
	{"textures/_techdc4_blu1", "textures/makkon_tech/_techdc4_blu1"},
	{"textures/_techdc4_brwn1", "textures/makkon_tech/_techdc4_brwn1"},
	{"textures/_techdc4_brwn2", "textures/makkon_tech/_techdc4_brwn2"},
	{"textures/_techdc4_grey1", "textures/makkon_tech/_techdc4_grey1"},
	{"textures/_techdc4_grey2", "textures/makkon_tech/_techdc4_grey2"},
	{"textures/_techdc4_orn1", "textures/makkon_tech/_techdc4_orn1"},
	{"textures/_techdc4_prpl1", "textures/makkon_tech/_techdc4_prpl1"},
	{"textures/_techdc4_red1", "textures/makkon_tech/_techdc4_red1"},
	{"textures/_techdc4_red2", "textures/makkon_tech/_techdc4_red2"},
	{"textures/_techdc4_teal1", "textures/makkon_tech/_techdc4_teal1"},
	{"textures/_techdc4_ylw1", "textures/makkon_tech/_techdc4_ylw1"},
	{"textures/_techdc4_ylw2", "textures/makkon_tech/_techdc4_ylw2"},
	{"textures/_techdc5_blk1", "textures/makkon_tech/_techdc5_blk1"},
	{"textures/_techdc5_blu1", "textures/makkon_tech/_techdc5_blu1"},
	{"textures/_techdc5_brwn1", "textures/makkon_tech/_techdc5_brwn1"},
	{"textures/_techdc5_brwn2", "textures/makkon_tech/_techdc5_brwn2"},
	{"textures/_techdc5_grey1", "textures/makkon_tech/_techdc5_grey1"},
	{"textures/_techdc5_grey2", "textures/makkon_tech/_techdc5_grey2"},
	{"textures/_techdc5_orn1", "textures/makkon_tech/_techdc5_orn1"},
	{"textures/_techdc5_prpl1", "textures/makkon_tech/_techdc5_prpl1"},
	{"textures/_techdc5_red1", "textures/makkon_tech/_techdc5_red1"},
	{"textures/_techdc5_red2", "textures/makkon_tech/_techdc5_red2"},
	{"textures/_techdc5_teal1", "textures/makkon_tech/_techdc5_teal1"},
	{"textures/_techdc5_ylw1", "textures/makkon_tech/_techdc5_ylw1"},
	{"textures/_techdc5_ylw2", "textures/makkon_tech/_techdc5_ylw2"},
	{"textures/_techgrt1_blk1", "textures/makkon_tech/_techgrt1_blk1"},
	{"textures/_techgrt1_blu1", "textures/makkon_tech/_techgrt1_blu1"},
	{"textures/_techgrt1_brwn1", "textures/makkon_tech/_techgrt1_brwn1"},
	{"textures/_techgrt1_brwn2", "textures/makkon_tech/_techgrt1_brwn2"},
	{"textures/_techgrt1_grey1", "textures/makkon_tech/_techgrt1_grey1"},
	{"textures/_techgrt1_grey2", "textures/makkon_tech/_techgrt1_grey2"},
	{"textures/_techgrt1_orng1", "textures/makkon_tech/_techgrt1_orng1"},
	{"textures/_techgrt1_prpl1", "textures/makkon_tech/_techgrt1_prpl1"},
	{"textures/_techgrt1_red1", "textures/makkon_tech/_techgrt1_red1"},
	{"textures/_techgrt1_red2", "textures/makkon_tech/_techgrt1_red2"},
	{"textures/_techgrt1_rst1", "textures/makkon_tech/_techgrt1_rst1"},
	{"textures/_techgrt1_rst2", "textures/makkon_tech/_techgrt1_rst2"},
	{"textures/_techgrt1_rst3", "textures/makkon_tech/_techgrt1_rst3"},
	{"textures/_techgrt1_teal1", "textures/makkon_tech/_techgrt1_teal1"},
	{"textures/_techgrt1_ylw1", "textures/makkon_tech/_techgrt1_ylw1"},
	{"textures/_techgrt1_ylw2", "textures/makkon_tech/_techgrt1_ylw2"},
	{"textures/_techgrt2_blu1", "textures/makkon_tech/_techgrt2_blu1"},
	{"textures/_techgrt2_brwn1", "textures/makkon_tech/_techgrt2_brwn1"},
	{"textures/_techgrt2_brwn2", "textures/makkon_tech/_techgrt2_brwn2"},
	{"textures/_techgrt2_grey1", "textures/makkon_tech/_techgrt2_grey1"},
	{"textures/_techgrt2_grey2", "textures/makkon_tech/_techgrt2_grey2"},
	{"textures/_techgrt2_orng1", "textures/makkon_tech/_techgrt2_orng1"},
	{"textures/_techgrt2_prpl1", "textures/makkon_tech/_techgrt2_prpl1"},
	{"textures/_techgrt2_red1", "textures/makkon_tech/_techgrt2_red1"},
	{"textures/_techgrt2_red2", "textures/makkon_tech/_techgrt2_red2"},
	{"textures/_techgrt2_rst1", "textures/makkon_tech/_techgrt2_rst1"},
	{"textures/_techgrt2_rst3", "textures/makkon_tech/_techgrt2_rst3"},
	{"textures/_techgrt2_teal1", "textures/makkon_tech/_techgrt2_teal1"},
	{"textures/_techgrt2_ylw1", "textures/makkon_tech/_techgrt2_ylw1"},
	{"textures/_techgrt2_ylw2", "textures/makkon_tech/_techgrt2_ylw2"},
	{"textures/_techgrt3_blk1", "textures/makkon_tech/_techgrt3_blk1"},
	{"textures/_techgrt3_blu1", "textures/makkon_tech/_techgrt3_blu1"},
	{"textures/_techgrt3_brwn1", "textures/makkon_tech/_techgrt3_brwn1"},
	{"textures/_techgrt3_brwn2", "textures/makkon_tech/_techgrt3_brwn2"},
	{"textures/_techgrt3_grey1", "textures/makkon_tech/_techgrt3_grey1"},
	{"textures/_techgrt3_grey2", "textures/makkon_tech/_techgrt3_grey2"},
	{"textures/_techgrt3_orng1", "textures/makkon_tech/_techgrt3_orng1"},
	{"textures/_techgrt3_prpl1", "textures/makkon_tech/_techgrt3_prpl1"},
	{"textures/_techgrt3_red1", "textures/makkon_tech/_techgrt3_red1"},
	{"textures/_techgrt3_red2", "textures/makkon_tech/_techgrt3_red2"},
	{"textures/_techgrt3_rst1", "textures/makkon_tech/_techgrt3_rst1"},
	{"textures/_techgrt3_rst2", "textures/makkon_tech/_techgrt3_rst2"},
	{"textures/_techgrt3_rst3", "textures/makkon_tech/_techgrt3_rst3"},
	{"textures/_techgrt3_teal1", "textures/makkon_tech/_techgrt3_teal1"},
	{"textures/_techgrt3_ylw1", "textures/makkon_tech/_techgrt3_ylw1"},
	{"textures/_techgrt3_ylw2", "textures/makkon_tech/_techgrt3_ylw2"},
	{"textures/_techgrt4_blu1", "textures/makkon_tech/_techgrt4_blu1"},
	{"textures/_techgrt4_brwn1", "textures/makkon_tech/_techgrt4_brwn1"},
	{"textures/_techgrt4_brwn2", "textures/makkon_tech/_techgrt4_brwn2"},
	{"textures/_techgrt4_grey1", "textures/makkon_tech/_techgrt4_grey1"},
	{"textures/_techgrt4_grey2", "textures/makkon_tech/_techgrt4_grey2"},
	{"textures/_techgrt4_orng1", "textures/makkon_tech/_techgrt4_orng1"},
	{"textures/_techgrt4_prpl1", "textures/makkon_tech/_techgrt4_prpl1"},
	{"textures/_techgrt4_red1", "textures/makkon_tech/_techgrt4_red1"},
	{"textures/_techgrt4_red2", "textures/makkon_tech/_techgrt4_red2"},
	{"textures/_techgrt4_rst1", "textures/makkon_tech/_techgrt4_rst1"},
	{"textures/_techgrt4_rst3", "textures/makkon_tech/_techgrt4_rst3"},
	{"textures/_techgrt4_teal1", "textures/makkon_tech/_techgrt4_teal1"},
	{"textures/_techgrt4_ylw1", "textures/makkon_tech/_techgrt4_ylw1"},
	{"textures/_techgrt4_ylw2", "textures/makkon_tech/_techgrt4_ylw2"},
	{"textures/_techtrs1_blk1", "textures/makkon_tech/_techtrs1_blk1"},
	{"textures/_techtrs1_blu1", "textures/makkon_tech/_techtrs1_blu1"},
	{"textures/_techtrs1_brwn1", "textures/makkon_tech/_techtrs1_brwn1"},
	{"textures/_techtrs1_brwn2", "textures/makkon_tech/_techtrs1_brwn2"},
	{"textures/_techtrs1_grey1", "textures/makkon_tech/_techtrs1_grey1"},
	{"textures/_techtrs1_grey2", "textures/makkon_tech/_techtrs1_grey2"},
	{"textures/_techtrs1_orng1", "textures/makkon_tech/_techtrs1_orng1"},
	{"textures/_techtrs1_prpl1", "textures/makkon_tech/_techtrs1_prpl1"},
	{"textures/_techtrs1_red1", "textures/makkon_tech/_techtrs1_red1"},
	{"textures/_techtrs1_red2", "textures/makkon_tech/_techtrs1_red2"},
	{"textures/_techtrs1_rst1", "textures/makkon_tech/_techtrs1_rst1"},
	{"textures/_techtrs1_rst2", "textures/makkon_tech/_techtrs1_rst2"},
	{"textures/_techtrs1_teal1", "textures/makkon_tech/_techtrs1_teal1"},
	{"textures/_techtrs1_ylw1", "textures/makkon_tech/_techtrs1_ylw1"},
	{"textures/_techtrs1_ylw2", "textures/makkon_tech/_techtrs1_ylw2"},
};

static const int numTextureConvertNames = sizeof( textureConvertNames ) / sizeof( textureConvertNames[0] );
